package de.aristaflow.adept2.core.timemanager;

import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.globals.TimeModelConstants.BaseGranularity;
import de.aristaflow.adept2.model.timemodel.Schedule;
import de.aristaflow.adept2.model.timemodel.ScheduleInterval;
import de.aristaflow.adept2.model.timemodel.ScheduleIntervalSet;

/**
 * A ScheduleEvaluation is used for evaluating the schedule of a
 * ScheduleRestrictedElement.
 */
public interface ScheduleEvaluation
{
  /* {src_lang=Java} */

  /**
   * Parses the Rule of the Schedule into a Schedule-Object which can be used
   * for more efficient evaluation of the Schedule.
   */
  public Schedule parseSchedule(String schedule) throws ScheduleResolutionException;

  public ScheduleIntervalSet evaluateSchedule(Schedule schedule, CalendarElement from,
      CalendarElement until) throws ScheduleResolutionException;

  public ScheduleIntervalSet evaluateSchedule(Schedule schedule, CalendarElement from,
      int intervalCount) throws ScheduleResolutionException;

//  public ScheduleIntervalSet evaluateSchedule(Schedule schedule,
//      BaseGranularity granularity, CalendarElement from, CalendarElement until);
//
//  public ScheduleIntervalSet evaluateSchedule(Schedule schedule,
//      BaseGranularity granularity, CalendarElement from, int intervalCount);

  public ScheduleInterval getNextInterval(Schedule schedule, CalendarElement from) throws ScheduleResolutionException;

  public boolean hasNextInterval(Schedule schedule, CalendarElement from) throws ScheduleResolutionException;

  boolean hasNextInterval(Schedule schedule, CalendarElement from, CalendarElement until) throws ScheduleResolutionException;

}