package de.aristaflow.adept2.core.timemanager;

import de.aristaflow.adept2.base.communication.annotations.ProxyAllow;
import de.aristaflow.adept2.base.service.ADEPT2Service;

@ProxyAllow
public interface TimeManager extends ADEPT2Service
{
  /* {src_lang=Java} */

  /**
   * Gets an implementation for the management of process template time model,
   * for instance, publishing time models for new templates.
   * 
   * @return An object allowing to manage runtime aspects of process template
   *         time models.
   * @uml.property  name="templateTimeManager"
   * @uml.associationEnd  readOnly="true" inverse="timeManager:de.aristaflow.adept2.core.timemanager.TemplateTimeManager"
   */
  public TemplateTimeManager getTemplateTimeManager();

  /**
   * Gets an implementation for the management of process instance time models which allows to retrieve the time model of a running instance to provide access for retrieving the current temporal information and the execution history.
   * @return  An object allowing to manage time models of running process instances.
   * @uml.property  name="instanceTimeManager"
   * @uml.associationEnd  readOnly="true" inverse="timeManager:de.aristaflow.adept2.core.timemanager.InstanceTimeManager"
   */
  public InstanceTimeManager getInstanceTimeManager();
  
  public TimeModellingService getTimeModellingService();
  
  public WorklistInteraction getWorklistInteraction();

  //public TimeAnalysis getTimeAnalysis();

  //public ScheduleEvaluation getScheduleEvaluation();

}