package de.aristaflow.adept2.core.timemanager.defaultimplementation;

import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.aristaflow.adept2.base.service.InternalServiceException;
import de.aristaflow.adept2.core.complexeventmanager.EventClassManager;
import de.aristaflow.adept2.core.complexeventmanager.EventFilter;
import de.aristaflow.adept2.core.complexeventmanager.EventListener;
import de.aristaflow.adept2.core.complexeventmanager.EventManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.ProcessExecutionManager;
import de.aristaflow.adept2.model.eventmodel.EventClass;
import de.aristaflow.adept2.model.eventmodel.EventReference;
import de.aristaflow.adept2.model.globals.ExecutionEventConstants;
import de.aristaflow.adept2.util.LoggerTools;

/**
 * @author  lanz
 */
public class DefaultTimeEventNotification implements EventListener, EventFilter
{
  protected Logger logger = LoggerTools.getLogger(this);
  
  private final Set<String> consumedEvents;

  /**
   * @uml.property  name="eventClassManager"
   * @uml.associationEnd  
   */
  private final EventClassManager eventClassManager;

  /**
   * @uml.property  name="processExecutionManager"
   * @uml.associationEnd  
   */
  private final ProcessExecutionManager processExecutionManager;

  public DefaultTimeEventNotification(ProcessExecutionManager processExecutionManager,
      EventManager eventManager, String... consumedEvents)
  {
    this.processExecutionManager = processExecutionManager;
    this.consumedEvents = new HashSet<String>(Arrays.asList(consumedEvents));
    this.eventClassManager = eventManager.getEventClassManager();
  }

  public void eventOccurred(EventReference event)
  {
    //this.logger.log(Level.FINE, String.format("Event '%s' Received", event));
    
    if (getEventClass(event).getClassification().contains(ExecutionEventConstants.INSTANCE_EVENT_CLASS_ID))
    {
      try
      {
        this.processExecutionManager.processEvent(event);
      }
      catch (Exception e)
      {
        this.logger.log(Level.SEVERE, "Exception while processing event!", e);
      }
    }
    else
    {
      String msg = String.format("Received Event which is not a InstanceEventReference (%s)",event);
      this.logger.severe(msg);
      throw new InternalServiceException(msg);
    }
    //this.logger.log(Level.FINE, String.format("Finished Processing Event '%s'", event));
  }

  private EventClass getEventClass(EventReference event)
  {
    EventClass eventClass = this.eventClassManager.getEventClass(event.getEventClassID());
    return eventClass;
  }

  public boolean filter(EventReference event)
  {
    return (getEventClass(event).getClassification().contains(ExecutionEventConstants.INSTANCE_EVENT_CLASS_ID));
  }

  public Set<String> getConsumedEventClasses()
  {
    return Collections.unmodifiableSet(new HashSet<String>(this.consumedEvents));
  }
  

}
