package de.aristaflow.adept2.core.timemanager.defaultimplementation;

import de.aristaflow.adept2.core.timemanager.ScheduleEvaluation;
import de.aristaflow.adept2.core.timemanager.ScheduleManager;
import de.aristaflow.adept2.core.timemanager.TimeModellingService;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.schedules.DefaultScheduleEvaluation;
import de.aristaflow.adept2.core.timemanager.storage.TimeManagerStorage;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;

public class DefaultTimeModellingService implements TimeModellingService
{

  private final TimeModelFactory timeModelFactory;
  private DefaultScheduleEvaluation scheduleEvaluation;
  private TimeManagerStorage timeManagerStorage;

  public DefaultTimeModellingService(
      DefaultTimeManager defaultTimeManager, TimeModelFactory timeModelFactory,
      DefaultTemplateTimeManager templateTimeManager, TimeManagerStorage timeManagerStorage)
  {
    this.timeModelFactory = timeModelFactory;
    this.timeManagerStorage = timeManagerStorage;
  }
  
  @Override
  public TemplateTimeModel createTimeModelForTemplate(Template template)
  {
    TemporalTemplate temporalTemplate = timeModelFactory.createTemporalTemplate(template);
    return timeModelFactory.createTimeModelForTemplate(temporalTemplate);
  }

  @Override
  public TimeModelFactory getTimeModelFactory()
  {
    return timeModelFactory;
  }

  public ScheduleEvaluation getScheduleEvaluation()
  {
    if (scheduleEvaluation == null)
    {
      scheduleEvaluation = new DefaultScheduleEvaluation(timeManagerStorage);
    }
    return scheduleEvaluation;
  }
  
  public ScheduleManager getScheduleManager()
  {
    if (scheduleEvaluation == null)
    {
      scheduleEvaluation = new DefaultScheduleEvaluation(timeManagerStorage);
    }
    return scheduleEvaluation;
  }

}
