package de.aristaflow.adept2.core.timemanager.defaultimplementation;

import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.UUID;

import de.aristaflow.adept2.base.service.AbstractSubService;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.timemanager.WorklistInteraction;
import de.aristaflow.adept2.core.worklistmanager.WorklistNotification2;

public class TimeAwareWorklistInteraction extends AbstractSubService implements WorklistInteraction, WorklistAccess
{

  /** The registered worklist managers. */
  protected final Set<WorklistNotification2> worklistManagers = new HashSet<WorklistNotification2>();
  private DefaultTimeManager timeManager;
  
  
  /**
   * The constructor of this class.
   * 
   * @param executionManager The execution manager this module is a submodule
   *          of.
   */
  public TimeAwareWorklistInteraction(final DefaultTimeManager timeManager)
  {
    super(new String[0],
          new String[0], timeManager);
    this.timeManager = timeManager;
  }

  public void addWorklistManager(SessionToken session, WorklistNotification2 worklistManager)
  {
    synchronized (worklistManagers)
    {
      worklistManagers.add(worklistManager);
    }
  }

  public void removeWorklistManager(SessionToken session, WorklistNotification2 worklistManager)
  {
    synchronized (worklistManagers)
    {
      worklistManagers.remove(worklistManager);
    }
  }
  
  protected Set<WorklistNotification2> getWorklistManagers()
  {
    Set<WorklistNotification2> ret;
    synchronized (worklistManagers)
    {
      ret = new HashSet<WorklistNotification2>(worklistManagers);
    }    
    return Collections.unmodifiableSet(ret);
  }
  
  public void timeModelUpdated(SessionToken session,
      UUID instanceID)
  {
    System.out.println("Time Model Updated");
    super.sessionActive(session);
    try
    { // sessionActive
      
      // use a copy to prevent concurrent modifications; the worklist managers
      // are checked for removal before updated
      Set<WorklistNotification2> wlManagers = getWorklistManagers();
      for (final WorklistNotification2 worklistManager : wlManagers)
      {
          // check whether the worklist manager has already been removed
          // in this case skip the update 
          boolean worklistManagerPresent;
          synchronized (worklistManagers)
          {
            worklistManagerPresent = worklistManagers.contains(worklistManager);
          }
          if (worklistManagerPresent)
          {
            worklistManager.updateInstance(instanceID);
          }
      }
    }
    finally
    {
      super.sessionFinished(session);
    }
  }
  
//  protected List<EBPInstanceReference> getActiveActivities(SessionToken session, UUID instanceID)
//  {
//    List<EBPInstanceReference> result = new ArrayList<EBPInstanceReference>();
//    super.sessionActive(session);
//    try
//    { // sessionActive
//      
//      
//      NodeState[] nodeStates = new NodeState[]{ NodeState.NS_ACTIVATED,
//                                                NodeState.NS_SELECTED,
//                                                NodeState.NS_SUSPENDED};
//      
//      ExecutionManager executionManager = timeManager.getExecutionManager();
//      
//      
//      ProcessManager pm = timeManager.getProcessManager();
//      
//      InstanceManager instanceManager =
//        pm.getInstanceManager();
//      
//      
//      ExecutableInstance instance;
//      try
//      {
//        instance = instanceManager.getAndLockInstanceForExecution(session, instanceID);
//      }
//      catch (LockException e)
//      {
//        String warningMessage =
//            String.format("Cannot get a lock for the given instance '%s'!",
//                          instanceID);
//        
//        logger.log(Level.WARNING, warningMessage, e);
//        
//        throw new InternalServiceException(warningMessage, e);
//      }
//      
//      try
//      { // unlock instance execution
//        Template template = instance.getTemplate();
//        
//        for (NodeState nodeState : nodeStates)
//        {
//          int[] affectedNodes = instance.getNodesInState(nodeState);
//          for (int affectedNodeID : affectedNodes)
//          {
//            Node affectedNode = template.getNode(affectedNodeID);
//            
//            if ((affectedNode.getID() == template.getStartNode().getID()) ||
//                (affectedNode.getID() == template.getEndNode().getID()))
//            {
//              continue;
//            }
//            ExecutableBusinessProcess ebp = affectedNode
//              .getExecutableBusinessProcess();
//            
//            boolean isActivity = isActivity(ebp);//ActivityExecutionTools.isActivity(ebp);
//            EBPType ebpType = getEBPType(ebp);//ActivityExecutionTools.getEBPType(ebp);
//            int nodeIteration = instance.getNodeIteration(affectedNodeID);
//            
//            
//            
//            final EBPInstanceReference ebpIR = executionManager
//              .getProcessModelFactory()
//              .createEBPInstanceReference(ebpType, instanceID,
//                                          instance.getBaseTemplateID(),
//                                          affectedNodeID, nodeIteration,
//                                          isActivity, getURIs());
//            
//            result.add(ebpIR);
//          }
//        }
//      }
//      finally
//      {
//        try
//        {
//          instanceManager.unlockExecutableInstance(session, instance.getID());
//        }
//        catch (LockException le)
//        {
//          String msg =
//            String.format("removeActivitiesFromWorklist: Cannot unlock the instance"
//                          + " '%1$s' for execution. Another method may have "
//                          + "unlocked it.", instance.getID());
//          logger.log(Level.SEVERE, msg, le);
//          throw new InternalServiceException(msg, le);
//        }
//      }
//    }
//    finally
//    {
//      super.sessionFinished(session);
//    }
//  }
//
//  /**
//   * Helper method to check, whether an EBP is an activity or not.
//   * 
//   * @param ebp The EBP the is activity flag is needed from
//   * @return true, if the EBP is an activity, false otherwise
//   * @throws RuntimeException if an unknown EBP type is provided 
//   */
//  public static boolean isActivity(ExecutableBusinessProcess ebp)
//  {
//    boolean isActivity;
//    if (ebp == null || ebp instanceof Activity)
//    {
//      isActivity = true;
//    }
//    else if (ebp instanceof LightWeightProcess
//             || ebp instanceof VariableParallelismEBP)
//    {
//      isActivity = false;
//    }
//    else
//    {
//      throw new RuntimeException("Unknown EBP type: "
//                                 + ebp.getClass().getName());
//    }
//    return isActivity;
//  }
//  
//  /**
//   * Helper method to get the ebpType of an EBP.
//   * 
//   * @param ebp The EBP the type is needed from.
//   * @return The type of the EBP or {@link EBPType#NULL} if the EBP is null.
//   */
//  public static EBPType getEBPType(final ExecutableBusinessProcess ebp)
//  {
//    EBPType type;
//    
//    if (ebp == null)
//    {
//      type = EBPType.NULL;
//    }
//    else
//    {
//      type = ebp.getType();
//    }
//    
//    return type;
//  }

}
