package de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution;

import java.util.UUID;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;

/**
 * The {@link ActivityStateUpdateManager} handles any required actions related
 * to the life cycle of an activity instance. More precisely it 
 * updates the {@link de.aristaflow.adept2.model.timemodel.InstanceTimeContext}
 * and calls {@link ActivityStarting} and {@link ActivityTermination} for
 * handling the time model, error detection, error handling, etc.
 * 
 * 
 * @author Andreas Lanz
 * 
 */
public interface ActivityStateUpdateManager
{
  // public void nodeActivated(SessionToken session, long timestamp,
  // UUID instanceID, int nodeID, int iteration);

  // public void nodeSelected(SessionToken session, long timestamp,
  // UUID instanceID, int nodeID, int iteration);
  // public void nodeDeselected(SessionToken session, long timestamp,
  // UUID instanceID, int nodeID, int iteration);

  /**
   * Handles any actions related to the start of a new activity instance. That
   * is it updates the
   * {@link de.aristaflow.adept2.model.timemodel.InstanceTimeContext} and calls
   * {@link ActivityStarting} to update the 
   * {@link de.aristaflow.adept2.model.timemodel.InstanceTimeModel}.
   * 
   * @param session
   * @param instanceID The ID of the newly started instance
   * @param nodeID The unique ID of the respective node
   * @param iteration The current iteration of the respective node
   * @param timestamp The exact time point the instance was started.
   */
  public void activityStarted(SessionToken session, UUID instanceID,
      int nodeID, int iteration, long timestamp);

  // public void nodeSuspended(SessionToken session, long timestamp,
  // UUID instanceID, int nodeID, int iteration);

  // public void nodeResumed(SessionToken session, long timestamp,
  // UUID instanceID, int nodeID, int iteration);

  // public void nodeResetSoft(SessionToken session, long timestamp,
  // UUID instanceID, int nodeID, int iteration);

  // public void nodeResetHard(SessionToken session, long timestamp,
  // UUID instanceID, int nodeID, int iteration);

  // public void nodeSkipped(SessionToken session, long timestamp,
  // UUID instanceID, int nodeID, int iteration);

  
  /**
   * Handles any actions related to the completion of an activity instance. That
   * is it updates the
   * {@link de.aristaflow.adept2.model.timemodel.InstanceTimeContext} and calls
   * {@link ActivityTermination} to update the 
   * {@link de.aristaflow.adept2.model.timemodel.InstanceTimeModel}.
   * 
   * @param session
   * @param instanceID The ID of the newly started instance
   * @param nodeID The unique ID of the respective node
   * @param iteration The current iteration of the respective node
   * @param timestamp The exact time point the instance was started.
   */
  public void activityFinished(SessionToken session, UUID instanceID,
      int nodeID, int iteration, long timestamp);

  // public void nodeFailed(SessionToken session, long timestamp,
  // UUID instanceID, int nodeID, int iteration);
}
