package de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution;

import java.util.UUID;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;

public interface InstanceStarting
{

  void instanceStarted(SessionToken session, UUID instanceID, long timestamp);

  /**
   * 
   * <b>Erzeugen der Prozessinstanz (ProcessInstantiation-Event) </b> <div>
   * <ul>
   * <li>Zunaechst wird ueberprueft, ob die <u>Validity Period</u> auf dem
   * <u>Earliest</u> oder <u>Latest Start Date</u> des Prozesses erfuellt ist
   * (Wert mglw. aus Konfiguration lesen)<span> </span>und der Prozess somit
   * ueberhaupt gestartet werden darf.</li>
   * <li>Abschaetzen der Erfuellbarkeit der <u>Validity Period</u> auf dem
   * <u>Earliest</u> oder <u>Latest Completion Date</u> des Prozesses</span>
   * (Wert mglw. aus Konfiguration lesen) mithilfe des Instanziierungszeitpunkts
   * und der zur Build-time vorberechneten Prozessdauer.
   * <ul>
   * <li>Eventuell kann eine <u>Validity Period</u> auf <u>Earliest</u> oder
   * <u>Latest Completion Date</u> waehrend der build-time bzw. waehrend dem
   * deploy mithilfe der berechneten Prozessdauer schon auf eine entsprechende
   * <u>Validity Period</u> auf dem Start abgebildet werden.</li>
   * </ul>
   * </li>
   * <li>Nun wird das Zeitmodell instanziiert und das ProzessInstantiation-Event
   * auf den aktuellen Zeitpunkt gesetzt.
   * <ul>
   * <li>Verwendung von zur Build-Time vorberechnetem Zeitmodell als Basis</li>
   * <li>XOR-Entscheidungen die lediglich von Prozess-Eingabeparametern abhaengen
   * werden ausgewertet und die abgewaehlten Pfade im Zeitmodell deaktiviert bzw.
   * aus diesem entfernt.</li>
   * <li><u>Time-dependent Variability</u> so weit moeglich auswerten</li>
   * <li>Es werden alle Constraints aktualisiert, die das
   * ProzessInstantiation-Event als Update-Trigger haben (d.h. ihre Daten aus
   * den Prozesseingabe-Parametern bzw. einer Konfiguration beziehen)<br/>
   * <ul>
   * <li>Die Constraints und Ausfuehrungsintervalle werden im Zeitmodell
   * entsprechend angepasst</li>
   * </ul>
   * </li>
   * <li>Auf Basis der aktuellen Daten des Zeitmodell wird fuer jedes Schedule
   * Restricted Element der erste und letzte im entsprechenden
   * Ausfuehrungs-Intervall des Events moegliche Slot ermittelt und falls noetig
   * das Ausfuehrungs-Intervall entsprechend eingeschraenkt
   * <ul>
   * <li>Falls kein Slot verfuegbar ist wir der Start des Prozesses abgebrochen</li>
   * </ul>
   * </li>
   * <li>Nun findet eine erste Zeitrechnung auf dem vollstaendigen Zeitmodell
   * statt. Dadurch werden die Werte an die aktuelle Instanz angepasst.
   * <ul>
   * <li>Relativ effizient aufgrund der zur Build-Time vorberechneten Werte.
   * (nur Schedule Restricted Elements sind problematisch)
   * <ul>
   * <li>-&gt;Time-dependent Variability wird falls moeglich ausgewertet (-&gt;
   * wichtig fuer Validity Period von Aktivitaeten)</li>
   * </ul>
   * </li>
   * <li>Inkonsistenzen sollten dabei hauptsaechlich im Zusammenhang mit
   * folgenden Pattern auftreten (vorausgesetzt die Werte der Konfiguration sind
   * korrekt und es handelt sich um keinen Prozess mit Zeitdaten als
   * Eingabeparameter):
   * <ul>
   * <li><u>Schedule Restricted Elements</u>: Es ist kein gueltiger Slot
   * verfuegbar</li>
   * <li><u>Validity Period</u>: Eine der Aktivitaeten kann nicht innerhalb ihrer
   * <u>Validity Period</u> ausgefuehrt werden bzw. der Prozess kann nicht
   * innerhalb seiner <u>Validity Period</u> auf dem <u>Earliest</u> bzw.
   * <u>Latest Completion Date</u> beendet werden.</li>
   * </ul>
   * </li>
   * <li>Inkonsistenzen in Zusammenhang mit Schedule Restricted Elements und
   * Validity Period sowie Constraints, welche ihre Werte aus der Konfiguration
   * beziehen, sind generelle Probleme, die nicht behoben werden koennen. Treten
   * solche Inkonsistenzen auf wird der Start der Prozessinstanz abgebrochen</li>
   * <li>Inkonsistenzen im Zusammenhang mit Prozess-Eingabeparameter koennen
   * moeglicherweise durch Veraendern der Eingabeparameter behoben werden.
   * <ul>
   * <li>-&gt; Exception und Einbezug des Benutzer der den Prozess gestartet
   * hat.</li>
   * <li>Betroffen: TP1, TP2, TP3, TP4, TP6, TP9, TP10</li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * <li>Damit ist, falls kein Fehler auftritt, die Initialisierung des
   * Zeitmodells abgeschlossen</li>
   * </ul>
   * </li>
   * <li>Nun wird die Ausfuehrungskomponente des Zeitmanagers initialisiert
   * <ul>
   * <li>Zunaechst wird ueberprueft, ob der Prozess sofort gestartet werden kann,
   * d.h. das ProcessStart-Event sofort auftreten darf, oder aufgrund eines
   * Constraints noch gewartet werden muss
   * <ul>
   * <li>Falls der Prozess sofort gestartet werden kann wird der Prozess im
   * ExecutionManager zur Ausfuehrung freigegeben</li>
   * <li>Andernfalls wird entweder ein entsprechender Timer initialisiert, der
   * bei Ablauf die Ausfuehrung im Execution Manager freigibt (siehe unten), oder
   * auf das Auftreten eines entsprechenden (externen) Events gewartet. (Trigger
   * auf Event setzen)</li>
   * <li>Moegliche Ursachen fuer Verzoegerungen:
   * <ul>
   * <li><u>TP3</u> vom ProcessInstantiation-Event (sinnvoll?) oder von einem
   * externen Event zum ProcessStart-Event</li>
   * <li><u>Fixed Date</u> auf ProcessStart (Wert aus Prozess-Eingabeparameter
   * oder externen Quelle (--&gt; zusaetzlich vorgeschaltetes Update-Event))</li>
   * <li><u>Schedule Restricted Element</u> auf ProcessStart</li>
   * </ul>
   * </li>
   * <li>Nach Freigabe des Prozesses wird dieser damit als "Running" betrachtet
   * und kann vom ExecutionManager ausgefuehrt werden.</li>
   * <li>-&gt; Abstand zwischen ProcessInstantiation und ProcessStart ~0, falls
   * dies von keinen Constraint direkt beeinflusst wird</li>
   * </ul>
   * </li>
   * <li>Nun wird ueberprueft, ob eine der <u>Validity Periods</u> eine Gefahr fuer
   * die Prozessausfuehrung darstellt, d.h. ob ihr Wert an den berechneten Wert
   * des entsprechenden Events angrenzt bzw. diesen sogar beschraenkt.
   * <ul>
   * <li>Ist dies nicht der Fall wird die Validity Period im weiteren Verlauf
   * ignoriert und koennte aus dem Zeitmodell geloescht werden, falls dies die
   * Zeitrechnung beschleunigt (Achtung muesste ggf. beim auftreten/"korrigieren"
   * von Zeitfehlern ueberprueft / neu eingefuegt werden)</li>
   * </ul>
   * </li>
   * <li>Parallel findet die Initialisierung des Laufzeitmodells statt
   * <ul>
   * <li>Das Laufzeitmodell besteht aus dem im ersten Schritt berechneten
   * Zeitmodell und vier Event-Mengen:
   * <ul>
   * <li><i>RecentHistory</i>: Die entlang des Kontrollflusses /
   * Abhaengigkeitsgraphen zuletzt aufgetretenen Events. Diese beschreiben den
   * aktuellen zeitlichen Zustand der Prozess-Instanz.</li>
   * <li><i>Pending</i>: Die Events deren Kontrollfluss-Vorbedingungen erfuellt
   * sind, d.h. die Events die nach Kontrollfluss als naechstes auftreten
   * koennten. (-&gt; Unabhaengig von Zeit-Constraints)</li>
   * <li><i>ActiveEvents</i>: Die Menge der Events, auf die sich ein aktives
   * Constraint bezieht. Das Ausfuehrungsintervall dieser Events ist meist
   * bereits stark eingeschraenkt und besitzt weniger aus dem Kontrollfluss
   * resultierende Varianz.
   * <ul>
   * <li>Dient zur Abschaetzung der Konsistenz, da diese Events grossen Einfluss
   * hierauf haben.</li>
   * <li>Nicht notwendigerweise disjunkt zu den anderen Mengen</li>
   * </ul>
   * </li>
   * <li><i>Horizon</i>: In "naher" Zukunft anstehende Events, die aufgrund des
   * aktuellen Vorschau-Horizonts bestimmt werden und fuer die Vorhersage von
   * Zeitdaten relevant sind.</li>
   * </ul>
   * </li>
   * <li>Zunaechst wird das ProcessInstantiation-Event zusammen mit dem
   * Zeitstempel in die Event-Menge <i>RecentHistory</i> eingefuegt</li>
   * <li>Dann werden zur Ausfuehrung anstehende Events in der Menge
   * <i>Pending</i> gespeichert (d.h. ProcessStart-Event und eventuelle externe
   * ProcessIntermediate-Events die kein eingehendes Constraint haben bzw. nur
   * eines zum ProzessInstantiation-Event)</li>
   * <li>Saemtliche <u>Fixed Date Elements</u> (TP4) deren Wert bereits bekannt
   * ist werden der Menge <i>ActiveEvents</i> hinzugefuegt.
   * 
   * <ul>
   * <li>(Ausfuehrungsintervall des Ziel-Events bereits zuvor anhand des
   * Constraints angepasst)</li>
   * <li>Moeglich: Exception-Timer starten, da Zeitmodell nicht immer
   * aktualisiert wird und damit Konsistenz nicht vollstaendig garantiert werden
   * kann <span> </span>(siehe auch weiter unten) <span
   * style="font-size: 11px;">&lt;-- Nur Events in Pending koennen direkt
   * beeinflusst werden. Alle anderen Events koennen nur indirekt ueber Pending
   * beeinflusst werden.</span></li>
   * </ul>
   * </li>
   * <li>Fuer saemtliche vom ProcessInstantiation-Event ausgehenden Constraints
   * (<u>TP3</u>) wird das Ziel-Event der <i>ActiveEvents</i> Menge hinzugefuegt.
   * <ul>
   * <li>(Ausfuehrungsintervall des Ziel-Events bereits zuvor angepasst)</li>
   * <li>Moeglich: Exception-Timer starten</li>
   * </ul>
   * </li>
   * <li>Fuer jedes <u>Schedule Restricted Element</u> wird der naechste Zeitslot
   * berechnet, der innerhalb des "Ausfuehrungs"-Intervalls des Events liegt
   * (Wert bereits aus Instanziierung des Zeitmodells bekannt / kann gecached
   * werden). Fuer das Ende des Slots wird ein Update-Timer registriert.
   * Entsprechend wird das Event ebenfalls der <i>ActiveEvents</i> Menge
   * hinzugefuegt.
   * <ul>
   * <li>(Ausfuehrungsintervall des Ziel-Events bereits zuvor angepasst)</li>
   * <li>Hier sind Abstufungen fuer den Update-Timer moeglich, je nachdem wie gross
   * der Zwischenraum zum darauffolgenden Slot im Verhaeltnis zur Varianz des
   * Events ist (d.h. zur Laenge des "Ausfuehrungs"-Intervalls). Ist der
   * Zwischenraum klein im Verhaeltnis zur Varianz koennen weitere Slots
   * hinzugenommen werden bis ein grosser Zwischenraum auftritt oder eine
   * bestimmte (?) Zeitspanne ueberschritten ist und dann erst fuer diesen
   * Zeitpunkt ein Update Timer gesetzt werden.</li>
   * <li>Ebenfalls sind Abstufung bei der Entscheidung, ob das Event den
   * <i>ActiveEvents</i> hinzugefuegt wird, moeglich. Je nachdem wie viel Varianz
   * das Element hat und wie sich Slots und Zwischenraeume verhalten, kann das
   * Event zunaechst aus der <i>ActiveEvents</i> menge weggelassen werden.
   * <ul>
   * <li>Was triggert dann schliesslich das hinzufuegen des Events zu
   * <i>ActiveEvents</i>? (-&gt; Update-Trigger?)</li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * <li>[Moegliche Erweiterung: Auch alle Constraints, die von einem Pending
   * Event ausgehen werden ausgewertet, d.h. <i>ActiveEvents</i> hinzugefuegt.]</li>
   * <li>Auf Basis von <i>Pending</i>, <i>ActiveEvents</i> und des aktuellen
   * Horizonts wird die Event-Menge <i>Horizon</i> bestimmt (siehe Notizen zu
   * Sliding-Horizon`).
   * <ul>
   * <li>Verschiedene Strategien fuer den Horizont denkbar: Zeit, Anzahl,
   * Kombination (min/max)</li>
   * <li>Konsequenz: Desto mehr Constraint bereits betrachtet werden muessen
   * desto weniger Vorausschau ist moeglich
   * <ul>
   * <li>Zeitrechnung wird komplexer und ungenauer</li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * <li>Alle Events aus <i>ActiveEvents</i>, die in keiner der anderen Mengen
   * sind, werden mit ihren direkten Vorgaengern (nach Kontrollfluss) aus der
   * Vereinigungsmenge von <i>Pending</i>, <i>Horizon</i> und
   * <i>ActiveEvents</i> verbunden, in dem "neue" Constraints gebildet werden,
   * die den entsprechenden Abstand aus dem Zeitmodell entlang des
   * Kontrollflusses widerspiegeln.</li>
   * <li>Auf Basis der Event-Mengen <i>Horizon</i>, <i>ActiveEvents</i>,
   * <i>Pending</i> und <i>RecentHistory</i> wird nun eine erneute Zeitrechnung
   * durchgefuehrt.
   * 
   * <ul>
   * <li>d.h. es wird der von diesen Events bzw. Event-Instanzen aufgespannte
   * Sub-Graph des Constraint-Networks betrachtet.</li>
   * <li>Hierbei wird zusaetzlich ein Free/Bounded-Schedule berechnet (Bettini02)
   * der die Werte fuer die Prediction der Zeitdaten liefert</li>
   * </ul>
   * </li>
   * <li>Fuer alle Events aus <i>Pending</i> werden entsprechende Timer
   * gestartet:
   * <ul>
   * <li>Falls das ProcessStart-Event verzoegert werden muss wird es im
   * ExecutionManager blockiert und ein entsprechender Timer gestartet, der das
   * Event zum entsprechenden Zeitpunkt freigibt (siehe oben)</li>
   * <li>Falls das ProcessStart-Event bis zu einem gewissen Zeitpunkt
   * aufgetreten sein muss wird ein entsprechender Exception-Timer gestartet</li>
   * <li>Der frueheste Zeitpunkt fuer ProcessIntermediate-Events kann nicht
   * kontrolliert werden, es kann nur auf einen Fehler reagiert werden.</li>
   * <li>Falls ein ProcessIntermediate-Event bis zu einem gewissen Zeitpunkt
   * aufgetreten sein muss wird ein entsprechender Exception-Timer gestartet</li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * </ul>
   * </li>
   * <li>Start des Escalation-Manager und bewerten der Kritikalitaet der
   * Events/Aktivitaeten
   * <ul>
   * <li><font color="#E30000">==&gt; TODO</font></li>
   * </ul>
   * </li>
   * </ul>
   * </div>
   * 
   * @param session
   * @param templateID
   * @param instanceID
   * @param timestamp
   */
  void instanceCreated(SessionToken session, UUID templateID, UUID instanceID,
      long timestamp);

}
