package de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution;

import java.util.UUID;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;

/**
 * The {@link InstanceStateUpdateManager} handles any required actions related
 * to the life cycle of a process instance. More precisely it creates and
 * updates the {@link de.aristaflow.adept2.model.timemodel.InstanceTimeContext}
 * and calls {@link InstanceStarting} and {@link InstanceTermination} for
 * handling the time model, error detection, error handling, etc.
 * 
 * 
 * @author Andreas Lanz
 * 
 */
public interface InstanceStateUpdateManager
{
  /**
   * Handles any actions related to the creation of a new process instance. That
   * is it creates and initializes the
   * {@link de.aristaflow.adept2.model.timemodel.InstanceTimeContext} and calls
   * {@link InstanceStarting} to create and initialize the 
   * {@link de.aristaflow.adept2.model.timemodel.InstanceTimeModel}.
   * 
   * @param session
   * @param templateID The Template based on which the instance has been created
   * @param instanceID The ID of the newly created instance
   * @param timestamp The exact time point the instance was created.
   */
  public void instanceCreated(SessionToken session, UUID templateID,
      UUID instanceID, long timestamp);
  
  /**
   * Handles any actions related to the start of a new process instance. That
   * is it updates the
   * {@link de.aristaflow.adept2.model.timemodel.InstanceTimeContext} and calls
   * {@link InstanceStarting} to update the 
   * {@link de.aristaflow.adept2.model.timemodel.InstanceTimeModel}.
   * 
   * @param session
   * @param instanceID The ID of the newly started instance
   * @param timestamp The exact time point the instance was started.
   */
  public void instanceStarted(SessionToken session, UUID instanceID,
      long timestamp);

//  public void instanceSuspended(SessionToken session, long timestamp,
//      UUID instanceID);

//  public void instanceResumed(SessionToken session, long timestamp,
//      UUID instanceID);

  /**
   * Handles any actions related to the completion of a process instance. That
   * is it updates the
   * {@link de.aristaflow.adept2.model.timemodel.InstanceTimeContext} and calls
   * {@link InstanceTermination} to handle the 
   * {@link de.aristaflow.adept2.model.timemodel.InstanceTimeModel}.
   * 
   * @param session
   * @param instanceID The ID of the completed instance
   * @param timestamp The exact time point the instance was completed.
   */
  public void instanceFinished(SessionToken session, UUID instanceID,
      long timestamp);

//  public void instanceFailed(SessionToken session, long timestamp,
//      UUID instanceID);

//  public void instanceChanged(SessionToken session, long timestamp,
//      UUID instanceID);

}
