package de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation;

import java.util.UUID;
import java.util.logging.Level;

import de.aristaflow.adept2.base.configuration.AbortServiceException;
import de.aristaflow.adept2.base.service.AbstractSubService;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultInstanceTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.WorklistAccess;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.ActivityStarting;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.timemodelcontrol.TemporalExceptionHandling;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.timemodelcontrol.defaultimplementation.SysErrTemporalExceptionHandling;
import de.aristaflow.adept2.model.timemodel.ChangeableTimeModelStatus;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus.TimeModelUpdateStatus;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;
import de.aristaflow.adept2.util.DataSourceException;
import de.aristaflow.adept2.util.LockException;

public class DefaultActivityStarting extends AbstractSubService implements ActivityStarting
{

  protected final  DefaultInstanceTimeManager instanceTimeManager;
  protected final  DefaultTimeManager timeManager;
  protected  TemporalExceptionHandling temporalExceptionHandling;
  protected final TimeModelFactory timeModelFactory;



  public DefaultActivityStarting(
      DefaultTimeManager defaultTimeManager,
      TimeModelFactory timeModelFactory,
      DefaultInstanceTimeManager instanceTimeManager)
  {
    super(defaultTimeManager);
    this.timeManager = defaultTimeManager;
    this.timeModelFactory = timeModelFactory;
    this.instanceTimeManager = instanceTimeManager;
  }
  
  @Override
  public void init() throws AbortServiceException
  {
    super.init();
//    this.timeModelTools = timeModelFactory.getTimeModelOperations();
    this.temporalExceptionHandling = new SysErrTemporalExceptionHandling();
  }
  
  private SessionToken subSession(SessionToken parentSession)
  {
    return this.timeManager.subSession(parentSession);
  }
  
  
  /**
   * TOOD check
   */
  public void activityStarted(SessionToken session, UUID instanceID,
      int nodeID, int iteration, long timestamp)
  {
    System.out.println(String.format("Activity (%1$d, %2$d) Started @ %3$d-> Updating Time Model",
                                     Integer.valueOf(nodeID), Integer.valueOf(iteration), Long.valueOf(timestamp)));
    
    try
    {
      ChangeableTimeModelStatus status = this.instanceTimeManager
        .getAndLockTimeModelStatus(session, instanceID);
      try
      {
    
        // Check if the time model is disable because of an invalid time model
        if (status.getTimeModelUpdateStatus().ordinal() 
            < TimeModelUpdateStatus.DISABLED.ordinal())
        {
          SessionToken childSession = subSession(session);
          try
          {
            UpdateableInstanceTimeModel timeModel = this.instanceTimeManager
              .getAndLockTimeModel(childSession, instanceID);
      
            try
            {
              // Set Instance Start Time
              timeModel.setActivityInstanceStartTime(nodeID, iteration, timestamp);
              this.instanceTimeManager.updateTimeModel(childSession, timeModel);
              
              // Check if calculations are disable because of an invalid time model      
              if (status.getTimeModelUpdateStatus().ordinal() 
                  < TimeModelUpdateStatus.NO_CALCULATION.ordinal())
              {
                  boolean stateChanged = TimeManagerTools.updateTimeModel(childSession, timeModelFactory, timeModel, status, this.temporalExceptionHandling);
                  if (stateChanged)
                  {
                    this.instanceTimeManager.updateTimeModelStatus(childSession, status);
                    //AL: determine changes and update triggers!
                  }
                    
                
                  this.instanceTimeManager.updateTimeModel(childSession, timeModel);
              }
              handleValidityPeriodAtActivityStart(childSession, instanceID, timestamp, status);
            }
            finally
            {
              this.instanceTimeManager.unlockTimeModel(childSession, instanceID);
            }
                 
          }
          catch (LockException e)
          {
            logger.log(Level.SEVERE, "Lock Exception", e);
          }
          catch (DataSourceException e)
          {
            logger.log(Level.SEVERE, "Data Source Exception", e);
          }
        }
      }
      finally 
      {
        this.instanceTimeManager.unlockTimeModelStatus(session, instanceID);
      }

      if (status.getTimeModelUpdateStatus().ordinal() <= TimeModelUpdateStatus.NO_UPDATE.ordinal())
      {
        SessionToken childSession = subSession(session);
        handleActivityDuration(childSession, instanceID, timestamp);
      }
      
      WorklistAccess notification = timeManager.getWorklistAccess();
      
      notification.timeModelUpdated(session, instanceID);
    }
    catch (LockException e)
    {
      logger.log(Level.SEVERE, "Lock Exception", e);
    }
  }

  private void handleActivityDuration(SessionToken childSession, UUID instanceID, long timestamp)
  {
    // TODO Auto-generated method stub
    
  }

  private void handleValidityPeriodAtActivityStart(SessionToken childSession, UUID instanceID, long timestamp,
      ChangeableTimeModelStatus status)
  {
    // TODO Auto-generated method stub
    
  }



}
