package de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation;

import java.net.URI;
import java.util.List;
import java.util.UUID;
import java.util.logging.Level;

import de.aristaflow.adept2.base.configuration.AbortServiceException;
import de.aristaflow.adept2.base.service.AbstractSubService;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.processmanager.InstanceManager;
import de.aristaflow.adept2.core.processmanager.ProcessManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultInstanceTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.WorklistAccess;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.ActivityTermination;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.timemodelcontrol.TemporalExceptionHandling;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.timemodelcontrol.defaultimplementation.SysErrTemporalExceptionHandling;
import de.aristaflow.adept2.model.globals.ProcessConstants.ProcessElementIdentifier;
import de.aristaflow.adept2.model.globals.TimeModelConstants.ReferencePoint;
import de.aristaflow.adept2.model.processmodel.Instance;
import de.aristaflow.adept2.model.processmodel.TemporalNode;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.timemodel.FixedDate;
import de.aristaflow.adept2.model.processmodel.timemodel.FixedDateUpdateHandler;
import de.aristaflow.adept2.model.processmodel.timemodel.FixedDateUpdateProcedureData;
import de.aristaflow.adept2.model.processmodel.timemodel.defaultimplementation.DefaultFixedDateUpdateMeasures;
import de.aristaflow.adept2.model.processmodel.timemodel.updatehandlers.UpdateHandlerRegistry;
import de.aristaflow.adept2.model.processmodel.tools.ProcessElementIdentifierTools;
import de.aristaflow.adept2.model.timemodel.ChangeableTimeModelStatus;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus.TimeModelUpdateStatus;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.TimeModelTools;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;
import de.aristaflow.adept2.util.DataSourceException;
import de.aristaflow.adept2.util.LockException;

public class DefaultActivityTermination extends AbstractSubService implements ActivityTermination
{

  private DefaultTimeManager timeManager;
  private DefaultInstanceTimeManager instanceTimeManager;
  private TimeModelFactory timeModelFactory;
  private TemporalExceptionHandling temporalExceptionHandling;

  public DefaultActivityTermination(
      DefaultTimeManager defaultTimeManager,
      TimeModelFactory timeModelFactory,
      DefaultInstanceTimeManager instanceTimeManager)
  {
    super(defaultTimeManager);
    this.timeManager = defaultTimeManager;
    this.timeModelFactory = timeModelFactory;
    this.instanceTimeManager = instanceTimeManager;
  }
  
  @Override
  public void init() throws AbortServiceException
  {
    super.init();
//    this.timeModelTools = timeModelFactory.getTimeModelOperations();
    this.temporalExceptionHandling = new SysErrTemporalExceptionHandling();
  }
  
  private SessionToken subSession(SessionToken parentSession)
  {
    return this.timeManager.subSession(parentSession);
  }
  
  
  /**
   * TOOD check
   */
  public void activityFinished(SessionToken session, UUID instanceID,
      int nodeID, int iteration, long timestamp)
  {
    // TODO Auto-generated method stub
    System.out.println(String.format("Activity (%1$d, %2$d) Finished @ %3$d-> Updating Time Model", Integer.valueOf(nodeID), Integer.valueOf(iteration), Long.valueOf(timestamp)));
    
    try
    {
      ChangeableTimeModelStatus status = this.instanceTimeManager
        .getAndLockTimeModelStatus(session, instanceID);
      try
      {
    
        // Check if the time model is disable because of an invalid time model
        if (status.getTimeModelUpdateStatus().ordinal() 
            < TimeModelUpdateStatus.DISABLED.ordinal())
        {
          SessionToken childSession = subSession(session);
          updateTimeModel(childSession, instanceID, nodeID, iteration, timestamp, status);
        }
      }
      finally 
      {
        this.instanceTimeManager.unlockTimeModelStatus(session, instanceID);
      }
      // <code>status</code> is now read-only!

      if (status.getTimeModelUpdateStatus().ordinal() <= TimeModelUpdateStatus.NO_UPDATE.ordinal())
      {
        SessionToken childSession = subSession(session);
        handleActivityDuration(childSession, instanceID, timestamp);
      }
      
      if (status.getTimeModelUpdateStatus().ordinal() 
          < TimeModelUpdateStatus.DISABLED.ordinal())
      {
        SessionToken childSession = subSession(session);
        handleTriggerEvents(childSession, instanceID, nodeID);
      }
      
      WorklistAccess notification = timeManager.getWorklistAccess();
      
      notification.timeModelUpdated(session, instanceID);
    }
    catch (LockException e)
    {
      logger.log(Level.SEVERE, "Lock Exception", e);
    }
    
    
  }

  /**
   * 
   * @param session
   * @param instanceID
   * @param nodeID
   * @param iteration
   * @param timestamp
   * @param status MUST BE LOCKED BY <code>session</code>
   */
  private void updateTimeModel(SessionToken session, UUID instanceID, int nodeID, int iteration, long timestamp,
      ChangeableTimeModelStatus status)
  {
    
    try
    {
      UpdateableInstanceTimeModel timeModel = this.instanceTimeManager
        .getAndLockTimeModel(session, instanceID);
   
      try
      {
        // Set Instance Start Time
        timeModel.setActivityInstanceFinishTime(nodeID, iteration, timestamp);
        this.instanceTimeManager.updateTimeModel(session, timeModel);
        
        // Check if calculations are disable because of an invalid time model      
        if (status.getTimeModelUpdateStatus().ordinal() 
            < TimeModelUpdateStatus.NO_CALCULATION.ordinal())
        {
            boolean stateChanged = TimeManagerTools.updateTimeModel(session, timeModelFactory, timeModel, status, this.temporalExceptionHandling);
            if (stateChanged)
            {
              this.instanceTimeManager.updateTimeModelStatus(session, status);
              //AL: determine changes and update triggers!
            }
              
          
            this.instanceTimeManager.updateTimeModel(session, timeModel);
        }
        handleValidityPeriodAtActivityTermination(session, instanceID, timestamp, status);
      }
      finally
      {
        this.instanceTimeManager.unlockTimeModel(session, instanceID);
      }
      
    }
    catch (LockException e)
    {
      logger.log(Level.SEVERE, "Lock Exception", e);
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "Data Source Exception", e);
    }
  }

  private void handleTriggerEvents(SessionToken session, UUID instanceID, int nodeID)
  {
    try
    {
      InstanceTimeModel timeModel = this.instanceTimeManager
          .getInstanceTimeModelReadonly(session, instanceID);
      
      List<URI> updateTrigger = timeModel.getUpdateTrigger(nodeID);
      
      for (URI trigger : updateTrigger)
      {
        String elementType = ProcessElementIdentifierTools.getElementType(trigger);
        if (ProcessElementIdentifier.PROCESS_ELEMENT_NODE.equals(elementType))
        {
          String targetNodeIDString = ProcessElementIdentifierTools.getElementID(trigger);
          int targetNodeID = Integer.parseInt(targetNodeIDString);
          String field = ProcessElementIdentifierTools.getField(trigger);
          updateNodeField(session, instanceID, targetNodeID, field, trigger);
        }
        else
        {
          logger.log(Level.SEVERE, "Update trigger target is not a node: %s", trigger);
        }
        
      }
      
      
      
      
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "Data Source Exception", e);
    }
    
  }

  private void updateNodeField(SessionToken session, UUID instanceID, int targetNodeID, String field, URI trigger)
  {
    try
    {
      UpdateableInstanceTimeModel timeModel = this.instanceTimeManager
          .getAndLockTimeModel(session, instanceID);
      
      try
      {
   
        ProcessManager processManager = this.timeManager.getProcessManager();
        InstanceManager instanceManager = processManager.getInstanceManager();
        Instance instance = instanceManager.getInstance(session, instanceID);
        
        TemporalTemplate template = timeModelFactory.createTemporalTemplate(instance.getTemplate());
        
        TemporalNode node = template.getNode(targetNodeID);
        
        if (TimeModelTools.START_FIXED_DATE_FIELD.equals(field))
        {
          FixedDate fixedDate = node.getStartFixedDate();
          if (fixedDate != null)
          {
            updateFixedDate(session, targetNodeID, trigger, timeModel, instance, fixedDate, ReferencePoint.START);
          }
          
        }
        else if (TimeModelTools.COMPLETION_FIXED_DATE_FIELD.equals(field))
        {
          FixedDate fixedDate = node.getCompletionFixedDate();
          if (fixedDate != null)
          {
            updateFixedDate(session, targetNodeID, trigger, timeModel, instance, fixedDate, ReferencePoint.COMPLETION);
          }
        }
        else
        {
          logger.log(Level.SEVERE, "Unkown field to be updated!", field);
        }
          
        
      }
      finally
      {
        try
        {
          this.instanceTimeManager.unlockTimeModel(session, instanceID);
        }
        catch (LockException e)
        {
          // TODO Auto-generated catch block
          e.printStackTrace();
        }
      }
      
    }
    catch (DataSourceException e)
    {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }
  }

  private void updateFixedDate(SessionToken session, int targetNodeID, URI trigger,
      UpdateableInstanceTimeModel timeModel, Instance instance, FixedDate fixedDate, ReferencePoint referencePoint)
  {
    FixedDateUpdateProcedureData updateProcedureData = fixedDate.getFixedDateUpdateProcedure();
    if (updateProcedureData != null)
    {
      String procedureID = updateProcedureData.getProcedureID();
      FixedDateUpdateHandler updateHandler = UpdateHandlerRegistry.getFixedDateUpdateHandler(procedureID);
      if (updateHandler != null)
      {
  //            updateHandler.getUpdateTriggerList(template, node, configuration);
        int nodeIteration = instance.getNodeIteration(targetNodeID);
        
        DefaultFixedDateUpdateMeasures updateMeasures = new DefaultFixedDateUpdateMeasures(
                                                                                           timeManager
                                                                                             .getDependencyRegistry(),
                                                                                           timeManager
                                                                                             .getExecutionManager(),
                                                                                           instance, timeModel, 
                                                                                           targetNodeID, nodeIteration, 
                                                                                           fixedDate,
                                                                                           referencePoint);
        
        updateHandler.updateFixedDate(session, targetNodeID, nodeIteration, 
                                      trigger, updateProcedureData, updateMeasures);
      }
      else
      {
        // log severe error
      }
    }
  }

  private void handleActivityDuration(SessionToken session, UUID instanceID, long timestamp)
  {
    // TODO Auto-generated method stub
    
  }

  private void handleValidityPeriodAtActivityTermination(SessionToken session, UUID instanceID, long timestamp,
      ChangeableTimeModelStatus status)
  {
    // TODO Auto-generated method stub
    
  }

}
