package de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation;

import java.util.UUID;
import java.util.logging.Level;

import de.aristaflow.adept2.base.service.AbstractSubService;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultInstanceTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.DefaultTimeManager;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.InstanceTermination;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.InstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.util.DataSourceException;

/**
 * @author  Andreas Lanz
 */
public class DefaultInstanceTermination extends AbstractSubService implements
    InstanceTermination
{
  
  /**
   * @uml.property  name="instanceTimeManager"
   * @uml.associationEnd  
   */
  private final DefaultInstanceTimeManager instanceTimeManager;

  public DefaultInstanceTermination(DefaultTimeManager timeManager,
      DefaultInstanceTimeManager instanceTimeManager)
  {
    super(timeManager);
    this.instanceTimeManager = instanceTimeManager;
  }

  public void instanceFinished(SessionToken session, UUID instanceID,
      long timestamp)
  {
    System.out.println(String.format("Instance (%1$s) Finished @ %2$d-> Updating Time Model", instanceID, Long.valueOf(timestamp)));
    logger.entering(logger.getName(), "instanceFinished(SessionToken,UUID,long)");

    
    InstanceTimeModel timeModel;
    InstanceTimeContext timeContext;
    try
    {
      timeContext = this.instanceTimeManager.getInstanceTimeContext(session, instanceID);
      timeModel = this.instanceTimeManager.getInstanceTimeModelReadonly(session, instanceID);
    }
    catch (DataSourceException e)
    {
      logger.log(Level.SEVERE, "DataSourceException", e);
      throw new RuntimeException(e);
    }

    if (timeModel != null)
    {
      long instanceStart = timeContext.getInstanceStarted();
      Interval<TimeDistance> durationRestriction = timeModel.getInstanceDuration();
  
      if (durationRestriction != null)
      {
        long instanceDuration = timestamp - instanceStart;
        if (TimeManagerTools.compare(durationRestriction.getMaxValue(), instanceDuration) > 0)
        {
          logger.severe("Process missed Duration Restriction");
        }
      }
      
      long duration = (timestamp - instanceStart) / 1000;
      logger.log(Level.FINE, 
                 String.format("Instance '%1$s' finished! Duration %2$d seconds", 
                               instanceID,
                               Long.valueOf(duration)));
    }
    logger.exiting(logger.getName(), "instanceFinished(SessionToken,UUID,long)");
  }

}
