package de.aristaflow.adept2.core.timemanager.defaultimplementation.processexecution.defaultimplementation;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.timemanager.defaultimplementation.timemodelcontrol.TemporalExceptionHandling;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.ChangeableTimeModelStatus;
import de.aristaflow.adept2.model.timemodel.InconsistencyException;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.TimeModelSolver;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus.TimeModelState;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus.TimeModelUpdateStatus;

public abstract class TimeManagerTools
{

  /**
   * Returns a negative integer, zero, or a positive integer as the first
   * argument is less than, equal to, or greater than difference between the
   * second and the third.
   * <p>
   * 
   * @param timePoint
   * @param instanceStart
   * @param instanceFinish
   * @return
   */
  public static int compare(TimeDistance timePoint,
      long value)
  {
    // TODO Auto-generated method stub
    return 0;
  }
  
  protected static boolean updateStatus(ChangeableTimeModelStatus currentStatus, TimeModelState status,
      TimeModelUpdateStatus updateStatus, String msg)
  {
    boolean change = false;
    if (!currentStatus.getTimeModelUpdateStatus().equals(updateStatus))
    {
      currentStatus.setTimeModelUpdateStatus(updateStatus,
                                msg);
      change = true;
    }
    
    if (!currentStatus.getTimeModelStatus().equals(status.ordinal()))
    {
      currentStatus.setTimeModelStatus(status);
      change = true;
    }
    
    if (status == TimeModelState.INVALID)
    {
      change |= downgradeUpdateStatus(currentStatus, TimeModelUpdateStatus.NO_CALCULATION, msg);
    }
    return change;
  }
  
  protected static boolean downgradeUpdateStatus(ChangeableTimeModelStatus currentStatus, 
      TimeModelUpdateStatus updateStatus, String msg)
  {
    boolean change = false;
    if (currentStatus.getTimeModelUpdateStatus().ordinal() < updateStatus.ordinal())
    {
      currentStatus.setTimeModelUpdateStatus(updateStatus,
                                msg);
      change = true;
    }
    
    return change;
  }
  
//  protected static boolean downgradeStatus(ChangeableTimeModelStatus currentStatus, TimeModelState state, String msg)
//  {
//    boolean change = false;
//    if (currentStatus.getTimeModelStatus().ordinal() < state.ordinal())
//    {
//      currentStatus.setTimeModelStatus(state);
//      if (state == TimeModelState.INVALID && currentStatus.getTimeModelUpdateStatus() != TimeModelUpdateStatus.DISABLED)
//      {
//        downgradeUpdateStatus(currentStatus, TimeModelUpdateStatus.DISABLED, msg);
//      }
//      change = true;
//    }
//    return change;
//  }
  

  /**
   * 
   * @param session
   * @param timeModelFactory
   * @param timeModel
   * @param status
   * @param temporalExceptionHandling
   * @return whether or not the state of the time model (i.e., <code>status</code>) has changed.
   */
  protected static boolean updateTimeModel(SessionToken session, TimeModelFactory timeModelFactory,
      UpdateableInstanceTimeModel timeModel, ChangeableTimeModelStatus status,
      TemporalExceptionHandling temporalExceptionHandling)
  {
    if (!status.getTimeModelUpdateStatus().equals(TimeModelUpdateStatus.ENABLED))
    {
      // TODO log
      return false;
    }
    
    if (status.getTimeModelStatus().equals(TimeModelState.INVALID))
    {
      // log
      return false;
    }
    
    TimeModelSolver timeModelSolver = timeModelFactory.getTimeModelSolver(timeModel.getClass());
    
    try
    {
      timeModelSolver.solve(timeModel);
      System.out.println(String.format("Updating status of Instance (%1$s) to VALID", timeModel.getInstanceID()));
      return TimeManagerTools.updateStatus(status, TimeModelState.VALID, status.getTimeModelUpdateStatus(), null);
    }
    catch (InconsistencyException e)
    {
      temporalExceptionHandling.handleInconsistency(e);
      System.out.println(String.format("Updating status of Instance (%1$s) to INVALID", timeModel.getInstanceID()));
      return TimeManagerTools.updateStatus(status, TimeModelState.INVALID, TimeModelUpdateStatus.DISABLED, e.getMessage());
    }
    catch (InterruptedException e)
    {
      // Ignore
      System.out.println(String.format("Updating status of Instance (%1$s) to UNKOWN", timeModel.getInstanceID()));
      return TimeManagerTools.updateStatus(status, TimeModelState.UNKOWN, status.getTimeModelUpdateStatus(), e.getMessage());
    }
  }

}
