package de.aristaflow.adept2.core.timemanager.storage;

import java.util.UUID;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.model.timemodel.ChangeableTimeModelStatus;
import de.aristaflow.adept2.model.timemodel.InstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelStatus;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;
import de.aristaflow.adept2.util.DataSourceException;
import de.aristaflow.adept2.util.LockException;

// TODO AL this storage facility needs locking for concurrent access
public interface InstanceTimeModelStorage
{
  /* {src_lang=Java} */


  // /////////////////////////////////////////
  // Instance Time Model Handling
  // /////////////////////////////////////////

  public TimeModelStatus getTimeModelStatusReadonly(UUID instanceID)
      throws DataSourceException;

  public ChangeableTimeModelStatus getAndLockTimeModelStatus(SessionToken session,
      UUID instanceID) throws DataSourceException;

  public void updateTimeModelStatus(SessionToken session, TimeModelStatus status)
      throws DataSourceException, LockException;

  public void unlockTimeModelStatus(SessionToken session, UUID instanceID)
      throws LockException;
  
  public InstanceTimeModel getInstanceTimeModelReadonly(UUID instanceID)
  throws DataSourceException;

  public UpdateableInstanceTimeModel getAndLockInstanceTimeModel(SessionToken session,
    UUID instanceID) throws DataSourceException;
  
  public void updateInstanceTimeModel(SessionToken session, UUID instanceID,
    InstanceTimeModel timeModel) throws DataSourceException,
    LockException;
  
  public void lockInstanceTimeModel(SessionToken session, UUID instanceID);
  
  public void unlockInstanceTimeModel(SessionToken session, UUID instanceID)
    throws LockException;
  
  public boolean hasTimeModelForInstance(UUID instanceID) throws DataSourceException;

  // /////////////////////////////////////////
  // Instance Time Context Handling
  // /////////////////////////////////////////

  public UpdateableInstanceTimeContext createAndLockInstanceTimeContext(SessionToken session,
      UUID instanceID) throws DataSourceException;

  public InstanceTimeContext getInstanceTimeContextReadonly(UUID instanceID)
      throws DataSourceException;

  public UpdateableInstanceTimeContext getAndLockInstanceTimeContext(SessionToken session,
      UUID instanceID) throws DataSourceException;

  public void updateInstanceTimeContext(SessionToken session,
      InstanceTimeContext instanceTimeData) throws DataSourceException,
      LockException;

  public void unlockInstanceTimeContext(SessionToken session, UUID instanceID)
      throws LockException;

  public void lockInstanceTimeContext(SessionToken session, UUID instanceID);


}