package de.aristaflow.adept2.core.worklistmanager.timeaware;

import java.net.URI;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.base.configuration.AbortServiceException;
import de.aristaflow.adept2.base.configuration.ConfigurationException;
import de.aristaflow.adept2.base.service.Registry;
import de.aristaflow.adept2.base.service.ServiceNotKnownException;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.ComponentTypes;
import de.aristaflow.adept2.core.timemanager.InstanceTimeManager;
import de.aristaflow.adept2.core.timemanager.TimeManager;
import de.aristaflow.adept2.core.timemanager.WorklistInteraction;
import de.aristaflow.adept2.core.worklistmanager.WorklistNotification2;
import de.aristaflow.adept2.core.worklistmanager.defaultimplementation.TransientWorklistManager;

public class TimeAwareTransientWorklistManager extends TransientWorklistManager implements TimeAwareWorklistManager
{

  private String[] runtimeRequired;
  private TimeManager timeManager;

  public TimeAwareTransientWorklistManager(Configuration configuration, Registry registry)
      throws ConfigurationException
  {
    super(configuration, registry);
  }
  
  @Override
  public void init(URI[] myURIs) throws AbortServiceException
  {
    super.init(myURIs);
    // Override default worklist notification
    this.worklistNotification = new TimeAwareWorklistNotification(configuration, this);
  }

  @Override
  public void start() throws AbortServiceException
  {
    super.start();
    
    SessionToken startSession = getSessionFactory().getSessionToken(getURIs());

    
    // Register this Worklist Manager to the Execution Manager
    timeManager = getTimeManager();
    WorklistInteraction worklistInteraction = timeManager
      .getWorklistInteraction();
    // Request all currently active Activities from the Execution
    // Manager/Worklist Interaction
    worklistInteraction.addWorklistManager(startSession,
                                           (WorklistNotification2) this.worklistNotification);
  }

  
  @Override
  public String[] getRuntimeRequiredServices()
  {
    if (runtimeRequired == null)
    {
      Set<String> ret = new HashSet<String>();
      //Add the services directly required by this class:
      ret.addAll(Arrays.asList(super.getRuntimeRequiredServices()));
  
      //Add the services required by the sub modules of this execution manager:
      ret.addAll(Arrays.asList(ComponentTypes.COMPONENT_TYPE_TIME_MANAGER));
      runtimeRequired = ret.toArray(new String[ret.size()]);
    }
    return runtimeRequired;
  }

  public InstanceTimeManager getInstanceTimeManager()
  {
    return getTimeManager().getInstanceTimeManager();
  }

  public TimeManager getTimeManager()
  {
    if (timeManager == null)
    {
      SessionToken sessionToken = sessionFactory.getSessionToken(getURIs());
      try
      {
        this.timeManager =
          registry.getServiceOfType(sessionToken, ComponentTypes.COMPONENT_TYPE_TIME_MANAGER,
                                    TimeManager.class);
      }
      catch (ServiceNotKnownException e)
      {
        throw new RuntimeException(e);
      }
    }

    return timeManager;
  }


}
