package de.aristaflow.adept2.model.common.timedata;


/*
 * Checks for TimePoint
 */
/**
 * <p>
 * A TimeDistance describes, rather obviously, the distance between two points in time. 
 * </p>
 * <p>
 * A TimeDistance allows to describe a time distance using different precisions,
 * e.g., 18days2hours is in precision hours and 2hours3minutes4seconds is in
 * precision seconds. This allows to cover the inaccuracy of time in the human
 * understanding, e.g., 1day0hours0minutes0seconds is not considered the same as
 * 1day.
 * </p>
 * <p>
 * The granularities are ordered as follows: seconds < minutes < hours < days <
 * months<br />
 * If two granularities are specified all granularities between the two which
 * are not specified are considered to be zero.
 * </p>
 */
public interface TimeDistance // extends TimePoint 
{
  /* {src_lang=Java, java_import=de.aristaflow.adept2.model.common.viewer.Attribute, java_import=de.aristaflow.adept2.model.common.viewer.AttributeSetter}*/


  public static final long UNDEFINED_TIME_DISTANCE = Long.MIN_VALUE;
  public static final long POSITIVE_INFINITY = Long.MAX_VALUE;
  public static final long NEGATIVE_INFINITY = -Long.MAX_VALUE;
  
  /** 
   *  Month precision of this TimeDistance or UNDEFINED_TIME_DISTANCE if undefined.
   */
  public long getMonths();

  /** 
   *  Day precision of this TimeDistance or UNDEFINED_TIME_DISTANCE if undefined.
   */
  public long getDays();

  /** 
   *  Hour precision of this TimeDistance or UNDEFINED_TIME_DISTANCE if undefined.
   */
  public long getHours();

  /** 
   *  Minute precision of this TimeDistance or UNDEFINED_TIME_DISTANCE if undefined.
   */
  public long getMinutes();

  /** 
   *  Second precision of this TimeDistance or UNDEFINED_TIME_DISTANCE if undefined.
   */
  public long getSeconds();
  
  public long getMilliseconds();

  public TimeDistance neg();

  public int compareTo(TimeDistance compare);

  boolean isInfinity();

  public int sign();

  TimeDistance add(TimeDistance sum);
  public TimeDistance sub(TimeDistance timeDistance);

  public long toSeconds();
  
  
  long toMilliseconds();



}