package de.aristaflow.adept2.model.common.timedata;

import de.aristaflow.adept2.model.common.viewer.Attribute;
import de.aristaflow.adept2.model.common.viewer.AttributeSetter;

/*
 * Checks for TimePoint
 */
/**
 * <p>
 * A TimePoint describes, rather obviously, a point in time. It is always
 * relative to a zero point (for example but not necessarily the zero point of
 * the gregorian calendar).
 * </p>
 * <p>
 * A TimePoint allows to describe a point in time using different precisions,
 * e.g., 18days2hours is in precision hours and 2hours3minutes4seconds is in
 * precision seconds. This allows to cover the inaccuracy of time in the human
 * understanding, e.g., 1day0hours0minutes0seconds is not considered the same as
 * 1day.
 * </p>
 * <p>
 * All values of a TimePoint have to be positive value. A negative value
 * indicates that the respective granularity is not specified. (see
 * #UNDEFINED_TIME_POINT)
 * </p>
 * <p>
 * The granularities are ordered as follows: seconds < minutes < hours < days <
 * months<br />
 * If two granularities are specified all granularities between the two which
 * are not specified are considered to be zero.
 * </p>
 */
public interface TimePoint {
  /* {src_lang=Java, java_import=de.aristaflow.adept2.model.common.viewer.Attribute, java_import=de.aristaflow.adept2.model.common.viewer.AttributeSetter}*/


  public static final long UNDEFINED_TIME_POINT = Long.MIN_VALUE;
  
  /** 
   *  Month precision of this TimePoint or UNDEFINED_TIME_POINT if undefined.
   */
  @Attribute(name="Months", defaultValue=""+UNDEFINED_TIME_POINT)
  public long getMonths();

  /** 
   *  Day precision of this TimePoint or UNDEFINED_TIME_POINT if undefined.
   */
  @Attribute(name="Days", defaultValue=""+UNDEFINED_TIME_POINT)
  public long getDays();

  /** 
   *  Hour precision of this TimePoint or UNDEFINED_TIME_POINT if undefined.
   */
  @Attribute(name="Hours", defaultValue=""+UNDEFINED_TIME_POINT)
  public long getHours();

  /** 
   *  Minute precision of this TimePoint or UNDEFINED_TIME_POINT if undefined.
   */
  @Attribute(name="Minutes", defaultValue=""+UNDEFINED_TIME_POINT)
  public long getMinutes();

  /** 
   *  Second precision of this TimePoint or UNDEFINED_TIME_POINT if undefined.
   */
  @Attribute(name="Seconds", defaultValue=""+UNDEFINED_TIME_POINT)
  public long getSeconds();

  @AttributeSetter("Months")
  public void setMonths(long months);

  @AttributeSetter("Days")
  public void setDays(long days);

  @AttributeSetter("Hours")
  public void setHours(long hours);

  @AttributeSetter("Minutes")
  public void setMinutes(long minutes);

  @AttributeSetter("Seconds")
  public void setSeconds(long seconds);

}