package de.aristaflow.adept2.model.common.timedata.defaultimplementation;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.TimePoint;

public class DefaultTimePoint implements TimePoint
{

  private long seconds;
  private long minutes;
  private long hours;
  private long days;
  private long months;

  
  public DefaultTimePoint(long seconds)
  {
    // If one value is INFINITY all others are too with the same sign
    
    this(seconds, TimePoint.UNDEFINED_TIME_POINT, TimePoint.UNDEFINED_TIME_POINT, TimePoint.UNDEFINED_TIME_POINT, TimePoint.UNDEFINED_TIME_POINT);  
  }
  
  public DefaultTimePoint(long seconds, long minutes, long hours, long days, long months)
  {
    // If one value is INFINITY all others are too with the same sign
    
    assert seconds != TimeDistance.POSITIVE_INFINITY
           || minutes != TimeDistance.POSITIVE_INFINITY
           || hours != TimeDistance.POSITIVE_INFINITY
           || days != TimeDistance.POSITIVE_INFINITY
           || months != TimeDistance.POSITIVE_INFINITY
           || (seconds == minutes && seconds == hours && seconds == days && seconds == months);
    
    assert seconds != TimeDistance.NEGATIVE_INFINITY
        || minutes != TimeDistance.NEGATIVE_INFINITY
        || hours != TimeDistance.NEGATIVE_INFINITY
        || days != TimeDistance.NEGATIVE_INFINITY
        || months != TimeDistance.NEGATIVE_INFINITY
        || (seconds == minutes && seconds == hours && seconds == days && seconds == months);
    
    // Not all values are undefined
    assert seconds != TimeDistance.UNDEFINED_TIME_DISTANCE
        || minutes != TimeDistance.UNDEFINED_TIME_DISTANCE
        || hours != TimeDistance.UNDEFINED_TIME_DISTANCE
        || days != TimeDistance.UNDEFINED_TIME_DISTANCE
        || months != TimeDistance.UNDEFINED_TIME_DISTANCE;
    

    this.seconds = seconds;
    this.minutes = minutes;
    this.hours = hours;
    this.days = days;
    this.months = months;    
  }


  public long getSeconds()
  {
    return seconds;
  }


  
  public void setSeconds(long seconds)
  {
    this.seconds = seconds;
  }

  public long getMinutes()
  {
    return minutes;
  }

  public void setMinutes(long minutes)
  {
    this.minutes = minutes;
  }

  public long getHours()
  {
    return hours;
  }

  public void setHours(long hours)
  {
    this.hours = hours;
  }

  public long getDays()
  {
    return days;
  }

  public void setDays(long days)
  {
    this.days = days;
  }

  public long getMonths()
  {
    return months;
  }

  public void setMonths(long months)
  {
    this.months = months;
  }
  
  @Override
  public String toString()
  {
    
    if (this.getMonths() == TimeDistance.POSITIVE_INFINITY)
    {
      return "infty";
    }
    
    if (this.getMonths() == TimeDistance.NEGATIVE_INFINITY)
    {
      return "-infty";
    }

    StringBuilder s = new StringBuilder();
    boolean notZero = false;
    
    
    if (this.getMonths() != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (this.getMonths() != 0)
      {
        s.append(this.getMonths());
        s.append("M");
        notZero = true;
      }
      
    }

    if (this.getDays() != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (this.getDays() != 0 || notZero)
      {
        s.append(this.getDays());
        s.append("d");
        notZero = true;
      }
    }

    if (this.getHours() != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (this.getHours() != 0 || notZero)
      {
        s.append(this.getHours());
        s.append("h");
        notZero = true;
      }
    }

    if (this.getMinutes() != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      if (this.getMinutes() != 0 || notZero)
      {
        s.append(this.getMinutes());
        s.append("m");
        notZero = true;
      }
    }

    if (this.getSeconds() != TimeDistance.UNDEFINED_TIME_DISTANCE)
    {
      s.append(this.getSeconds());
      s.append("s");
      notZero = true;
    }

    
    if (!notZero)
    {
      s.append("0s");
    }

    return s.toString();

  }


  @Override
  public int hashCode()
  {
    int hashCode;

    final int prime = 31;
    hashCode = 1;
    hashCode = prime * hashCode + (days != TimeDistance.UNDEFINED_TIME_DISTANCE ? (int) (days ^ (days >>> 32)) : 0);
    hashCode = prime * hashCode + (hours != TimeDistance.UNDEFINED_TIME_DISTANCE ?(int) (hours ^ (hours >>> 32)) : 0);
    hashCode = prime * hashCode + (minutes != TimeDistance.UNDEFINED_TIME_DISTANCE ?(int) (minutes ^ (minutes >>> 32)) : 0);
    hashCode = prime * hashCode + (months != TimeDistance.UNDEFINED_TIME_DISTANCE ?(int) (months ^ (months >>> 32)) : 0);
    hashCode = prime * hashCode + (seconds != TimeDistance.UNDEFINED_TIME_DISTANCE ?(int) (seconds ^ (seconds >>> 32)) : 0);
    
    return hashCode;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj == this)
      return true;
    
    if (obj instanceof TimePoint)
    {
      TimePoint other = (TimePoint) obj;
      // TODO reconsider whether to use convertToSeconds or compareValue...
//      return this.convertToSeconds(this) == convertToSeconds(other);
//      return compareValue(this.seconds, other.getSeconds())
//             && compareValue(this.minutes, other.getMinutes())
//             && compareValue(this.hours, other.getHours()) 
//             && compareValue(this.days, other.getDays())
//             && compareValue(this.months, other.getMonths());
      //return compareTo(other) == 0;
    }

    return super.equals(obj);
  }

}
