package de.aristaflow.adept2.model.processmodel.algorithms;

import java.util.logging.Logger;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.changeoperations.ExtractSubprocess;
import de.aristaflow.adept2.core.updatemanager.UpdateManager;
import de.aristaflow.adept2.model.processmodel.ChangeableTemplate;
import de.aristaflow.adept2.model.processmodel.ProcessModelFactory;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition.NestedSESE;
import de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition.SESERegion;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTimeModelFactory;
import de.aristaflow.adept2.util.LoggerTools;

public class TemplateHierarchization
{

  private ProcessModelFactory pmf;
  private Logger logger = LoggerTools.getLogger(this);

  public TemplateHierarchization(ProcessModelFactory pmf)
  {
    this.pmf = pmf;
  }

  public void hierachize(SessionToken session, ChangeableTemplate template, UpdateManager updateManager)
  {
    
    TimeModelFactory timeModelFactory = new CSTNUTimeModelFactory();
    TemporalTemplate temporalTemplate = timeModelFactory.createTemporalTemplate(template);
    
    NestedSESE seseDecomposition = new NestedSESE();
    SESERegion seseTree = seseDecomposition.transform(temporalTemplate);
    
    
    
    for (SESERegion childBlock : seseTree.getChildren())
    {
      hierachize(session, template, childBlock, updateManager);
    }
  }

  private void hierachize(SessionToken session, ChangeableTemplate template, SESERegion block, UpdateManager updateManager)
  {
    for (SESERegion childBlock : block.getChildren())
    {
      hierachize(session, template, childBlock, updateManager);
    }
    
    if (ExtractSubprocess.isPossible(template, block.getEntryNode(), block.getExitNode()))
    {
      ExtractSubprocess.performOperation(session, template, block.getEntryNode(), block.getExitNode(), updateManager);
    }
    else
    {
      logger.severe(String.format("Could not extract SESE-Block '%s'", block));
    }
    
    
  }

}
