package de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Deque;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition.UndirectedGraphTransformation.UndirectedGraphFactory;
import de.aristaflow.adept2.model.timemodel.tools.DefaultUndirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.UndirectedGraph;

/**
 * Computes the SESE blocks of an Undirected Graph or Template (which is transformed to an undirected graph first {@link UndirectedGraphTransformation}).
 * 
 * See <i>Ananian 1997 - The Static Single Information Form</i> for details.
 * 
 * Implemented with some help from an implementation by Scott Ananian under GNU GPL Licence (see
 * <link>http://bpmn2bpel.googlecode.com/svn/trunk/BPMN2BPEL/src/translation/analysis/sese/SESE.java</link>)
 * 
 * @author lanz
 *
 */
public class NestedSESE
{
  
  public SESERegion transform(TemporalTemplate template)
  {
    // First transform the template to an undirected graph as helper structure
    UndirectedGraphTransformation<NodeStruct, EdgeStruct> graphTrafo = new UndirectedGraphTransformation<NodeStruct, EdgeStruct>(new UndirectedGraphFactory<NodeStruct, EdgeStruct>()
    {

      public NodeStruct createVertex(Object vertex, boolean representative)
      {
        return new NodeStruct(vertex, representative);
      }

      public EdgeStruct createEdge(NodeStruct v1, NodeStruct v2, Object edge)
      {
        return new EdgeStruct(v1, v2);
      }

      public UndirectedGraph<NodeStruct, EdgeStruct> createGraph(List<NodeStruct> vertices)
      {
        return new DefaultUndirectedGraph<NodeStruct, EdgeStruct>(vertices);
      }
    });

    UndirectedGraph<NodeStruct, EdgeStruct> graph = graphTrafo.transform(template);
    NodeStruct sourceNode = graphTrafo.sourceNode;
    NodeStruct sinkNode = graphTrafo.sinkNode;
    
    // Now determine SESE regions of the undirected graph and canonicalize them
    NestedSESE sese = new NestedSESE();
    RegionStruct rootRegion = sese.transform(graph, sourceNode, sinkNode);
    
//    for (Iterator<RegionStruct> iterator = rootRegion.children.iterator(); iterator.hasNext();)
//    {
//      RegionStruct region = iterator.next();
//      if (region.exit.n1 == sourceNode)
//      {
//        if (region.children.size() == 0 && region.nodes.size() == 0)
//          iterator.remove();
//        else
//        {
////          for (RegionStruct region2 : rootRegion.children)
////          {
////            if (region2.exit == region.entry)
////            {
////              
////            }
////          }
//        }
//      }
//    }
    
    rootRegion = sese.canonicalize(rootRegion);
    
    // Transform the helper structure back to the SESE regions of the original graph.
    SESERegion regions = graphTrafo.transform(rootRegion);
    return regions;
  }
  
  public RegionStruct transform(UndirectedGraph<NodeStruct, EdgeStruct> graph, NodeStruct sourceNode, NodeStruct sinkNode)
  {
    
    
    
    // compute cycle equivalence.
    CycleEquivalence ceq = new CycleEquivalence();
    // use cycle equivalence classes to determine canonical sese regions.
    Map<CQClass, List<EdgeStruct>> cdClasses = ceq.transform(graph, sourceNode);
    
    Map<EdgeStruct, RegionStruct> entryRegion= new HashMap<EdgeStruct, RegionStruct>();
    Map<EdgeStruct, RegionStruct> exitRegion = new HashMap<EdgeStruct, RegionStruct>();
    
    System.out.println(cdClasses);
    
    for (List<EdgeStruct> l : cdClasses.values())
    {
      EdgeStruct prev = null;
      for (EdgeStruct o : l)
      {
        if (prev != null)
        {
          RegionStruct r = new RegionStruct(prev, o);
          if (entryRegion.put(prev, r) != null)
            assert false;
          if (exitRegion.put(o, r) != null)
            assert false;
        }
        prev = o;
      }
    }
    
    // now do DFS traversal of CFG to determine nesting of regions.
    
    // state for DFS
    Set<Object> visited = new HashSet<Object>();
    visited.add(sourceNode);
    
//    List<EdgeStruct> outgoingEdges = getOutgoingEdges(graph, sourceNode);
//    for (EdgeStruct e : outgoingEdges)
//    {
//      if (e.n2 == sinkNode)
//      {
//        visited.add(e);
//      }
//        
//    }
    
    Deque<Object> nodeS = new LinkedList<Object>();
    nodeS.push(sourceNode);
    
    RegionStruct topLevel = new RegionStruct(null, null);
    Deque<RegionStruct> regionS = new LinkedList<RegionStruct>();
    regionS.push(topLevel);
    
   
    
    Map<NodeStruct, RegionStruct> smallestSESE = new HashMap<NodeStruct, RegionStruct>();

    while (!nodeS.isEmpty())
    {
      assert nodeS.size() == regionS.size();
      Object o = nodeS.pop(); // either an NodeStruct or an EdgeStruct
      RegionStruct cR = regionS.pop(); // currentRegion.
      // deal with region entry/exit.
      if (o instanceof NodeStruct)
      {
        // cR is smallest enclosing canonical SESE of o.
        smallestSESE.put((NodeStruct) o, cR);
        cR.nodes.add((NodeStruct) o);
      }
      else
      { // update current region.
        // if this is found in the cycle-equivalency stuff, push/pop.
        RegionStruct r1 = entryRegion.get(o);
        RegionStruct r2 = exitRegion.get(o);
        // pop up one level if we are leaving the region.
        if (cR.equals(r1))
        {
          cR = cR.parent;
          r1 = null;
        }
        if (cR.equals(r2))
        {
          cR = cR.parent;
          r2 = null;
        }
        // push a level if we are entering the region.
        if (r1 != null)
        {
//          if (r1.parent != cR) // LOW something here is not right yet...
//          {
          if (!isParentOf(r1, cR))
          {
            RegionStruct.link(cR, r1);
          }
//            
//          }
//          else
//          {
            
//            assert false : "Ooops";
//          }
          cR = r1;
          
        }
        else if (r2 != null)
        {
//          if (r2.parent != cR) // LOW something here is not right yet...
//          {
          if (!isParentOf(r2, cR))
          {
            RegionStruct.link(cR, r2);
          }
            
//          }
//          else
//          {
//            assert false : "Ooops";
//          }
          cR = r2;
          
        }
        else
        {
          System.out.println("empty region: "+o);
          
        }
      }
      // continue DFS traversal.
      if (o instanceof NodeStruct) // Node. Push edges.
      {
        List<EdgeStruct> succ = getOutgoingEdges(graph, ((NodeStruct) o));
        for (EdgeStruct edge : succ)
        {
          if (!visited.contains(edge))
          {
            nodeS.push(edge);
            regionS.push(cR);
            visited.add(edge);
          }
        }
      }
      else if (o instanceof EdgeStruct) // Edge. Push node.
      {
         NodeStruct node = ((EdgeStruct) o).n2 ;
        
         if (!visited.contains(node))
         {
           nodeS.push(node);
           regionS.push(cR);
           visited.add(node);
         }
      }
      
      
    } // END while.
    assert nodeS.isEmpty() && regionS.isEmpty();


    return topLevel;
  }

  private boolean isParentOf(RegionStruct potentialParent, RegionStruct region)
  {
    RegionStruct p = region;
    while (p.parent != null)
    {
      if (p.parent == potentialParent)
        return true;
      p = p.parent;
    }
    return false;
  }

  private List<EdgeStruct> getOutgoingEdges(UndirectedGraph<NodeStruct, EdgeStruct> graph, NodeStruct node)
  {
    Collection<NodeStruct> vertices = graph.getVertices();
    List<EdgeStruct> edges = new ArrayList<EdgeStruct>(vertices.size());
    for (NodeStruct v : vertices)
    {
      EdgeStruct edge = graph.getEdge(node, v);
      if (edge != null)
        edges.add(edge);
    }
    return edges;
  }

  public Map<NodeStruct, RegionStruct> canonicalize(Map<NodeStruct, RegionStruct> smallestSESE)
  {
    Map<NodeStruct, RegionStruct> seseBlocks = new HashMap<NodeStruct, RegionStruct>(smallestSESE);
    for (Entry<NodeStruct, RegionStruct> entry : seseBlocks.entrySet())
    {
      RegionStruct region = entry.getValue();
      
      if (region.parent != null) // Not the root region
      {
        
        boolean change = true;
        while (change)
        {
          change = false;
          for (RegionStruct region2 : region.parent.children)
          {
            if (region.exit == region2.entry)
            {
              region = RegionStruct.merge(region, region2);
              entry.setValue(region);
              change = true;
            }
            else if (region2.exit == region.entry)
            {
              region = RegionStruct.merge(region2, region);
              entry.setValue(region);
              change = true;
            }
          }
        }
      }
      
    }
    
    return seseBlocks;
  }
  
  public RegionStruct canonicalize(RegionStruct topLevel)
  {
    boolean change = true;
    while (change)
    {
      change = false;
      outer: for (Iterator<RegionStruct> iterator = topLevel.children.iterator(); iterator.hasNext();)
      {
        RegionStruct region1 = iterator.next();
//        region1 = canonicalize(region1);
        
        for (RegionStruct region2 : topLevel.children)
        {
          if (region1 == region2)
            continue;
            
          
          if (region1.exit == region2.entry)
          {
            System.out.print("Merging "+region1+" and "+region2+" to ");
            
            iterator.remove();
            region2.children.addAll(region1.children);
            region2.nodes.addAll(region1.nodes);
            region2.entry = region1.entry;
            
            System.out.println(region2);
            change = true;
            continue outer;
          }
          else if (region2.exit == region1.entry)
          {
            iterator.remove();
            region2.children.addAll(region1.children);
            region2.nodes.addAll(region1.nodes);
            region2.exit = region1.exit;
            change = true;
            continue outer;
          }
        }
      }
    }
    return topLevel;
  }
  
}
