package de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition;

import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * Helper class to compute the Cycle Equivalence classes of an undirected graph (see {@link CycleEquivalence}).
 * 
 * This class represents a Node in the undirected graph (see {@link node}) with necessary additional information.
 * 
 * @author lanz
 *
 */
class NodeStruct implements Comparable<NodeStruct>
{
  public Object node;
  public boolean representative;
  
  /** Number(n:node): integer : DFS preorder number of node */
  public int number = -1;
  
  /** BList(n:node): BracketList : List of brackets of node */
  public LinkedList<EdgeStruct> blist;
  /**  Hi(n:node): integer : Highest destination node of any edge originating from a descendant of node n */
  public NodeStruct hi;
  
  public boolean mark = false;

  public List<EdgeStruct> capping = new ArrayList<EdgeStruct>();
  public List<EdgeStruct> children = new ArrayList<EdgeStruct>();

  protected List<EdgeStruct> backedges = new ArrayList<EdgeStruct>();

  protected EdgeStruct parent;

  private int id;
  private static int ID = 0;

  public NodeStruct(Object nodeRep, boolean representative)
  {
    this.id = ID++;
    this.node = nodeRep;
    //this.number = topologicalID;
    this.representative = representative;
  }
  
  @Override
  public String toString()
  {
    StringBuilder stringBuilder = new StringBuilder();
    
    stringBuilder.append(String.valueOf(number));
    stringBuilder.append('@');
    stringBuilder.append('(');
    stringBuilder.append(id);
    stringBuilder.append(": ");
    if (representative)
      stringBuilder.append(String.valueOf(node));
    else
      stringBuilder.append("x");
    stringBuilder.append(')');
    return stringBuilder.toString();
  }

  public int compareTo(NodeStruct o)
  {
    return this.number - o.number;
  }
}