package de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition;

import java.util.ArrayList;
import java.util.List;

import de.aristaflow.adept2.model.processmodel.Node;

public class SESERegion 
{

  private final Node entryNode;
  private final Node exitNode;
  private final List<Node> nodes = new ArrayList<Node>();
  private final List<SESERegion> children = new ArrayList<SESERegion>();
  private SESERegion parent;

  public SESERegion(Node entryNode, Node exitNode)
  {
    this.entryNode = entryNode;
    this.exitNode = exitNode;
    this.nodes.add(entryNode);
    this.nodes.add(exitNode);
  }

  public Node getEntryNode()
  {
    return entryNode;
  }

  public Node getExitNode()
  {
    return exitNode;
  }

  @Override
  public String toString()
  {
    StringBuilder stringBuilder = new StringBuilder();
    stringBuilder.append("\n[");
    stringBuilder.append(entryNode);
    stringBuilder.append("--");
    stringBuilder.append(exitNode);
    stringBuilder.append("]");
    stringBuilder.append("{");
    String sep = "";
    for (Node n : nodes)
    {
      stringBuilder.append(sep);
      stringBuilder.append(n);
      sep = ", ";
      
    }
    stringBuilder.append("}");
    stringBuilder.append("{\n");
    sep = "";
    for (SESERegion child : getChildren())
    {
      stringBuilder.append(sep);
      stringBuilder.append(child);
      sep = ", ";
    }
    stringBuilder.append("\n}\n");
    return stringBuilder.toString();
  }

  void addNode(Node node)
  {
    this.nodes.add(node);
  }

  void addChildRegion(SESERegion region)
  {
    this.getChildren().add(region);
    region.parent = this;
    
  }

  public List<Node> getNodes()
  {
    return nodes;
  }

  public List<SESERegion> getChildren()
  {
    return children;
  }
  
  public SESERegion getParent()
  {
    return parent;
  }
  
}
