package de.aristaflow.adept2.model.processmodel.defaultimplementation;

import java.util.HashSet;
import java.util.Set;

import de.aristaflow.adept2.model.globals.ProcessConstants.EdgeType;
import de.aristaflow.adept2.model.globals.TimeModelConstants.RelationType;
import de.aristaflow.adept2.model.processmodel.AbstractWrappingTemplate;
import de.aristaflow.adept2.model.processmodel.ConstraintEdge;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.TemporalNode;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.timemodel.FixedDate;
import de.aristaflow.adept2.model.processmodel.timemodel.ProcessDuration;
import de.aristaflow.adept2.model.processmodel.timemodel.ScheduleRestrictedElement;
import de.aristaflow.adept2.model.processmodel.timemodel.TimeLag;
import de.aristaflow.adept2.model.processmodel.timemodel.TimeModelTools;
import de.aristaflow.adept2.model.processmodel.timemodel.ValidityPeriod;
import de.aristaflow.adept2.util.NullArgumentException;

public class DefaultTemporalTemplate extends AbstractWrappingTemplate implements TemporalTemplate
{

  /**
   * 
   */
  private static final long serialVersionUID = -3495014167149036655L;

  public DefaultTemporalTemplate(Template template)
  {
    super(template);
    if (template == null)
      throw new NullArgumentException();
  }

  public Template getBaseTemplate()
  {
    return this.template;
  }

  public ProcessDuration getProcessDuration()
  {
    return TimeModelTools.getProcessDuration(getBaseTemplate());
  }

  public ValidityPeriod getStartValidityPeriod()
  {
    return TimeModelTools.getStartValidityPeriod(getBaseTemplate());
  }

  public ValidityPeriod getCompletionValidityPeriod()
  {
    return TimeModelTools.getCompletionValidityPeriod(getBaseTemplate());
  }
  
  public FixedDate getStartFixedDate()
  {
    return TimeModelTools.getStartFixedDate(getBaseTemplate());
  }

  public FixedDate getCompletionFixedDate()
  {
    return TimeModelTools.getCompletionFixedDate(getBaseTemplate());
  }

  public ScheduleRestrictedElement getStartScheduleRestrictedElement()
  {
    return TimeModelTools.getStartSchedule(getBaseTemplate());
  }

  public ScheduleRestrictedElement getCompletionScheduleRestrictedElement()
  {
    return TimeModelTools.getCompletionSchedule(getBaseTemplate());
  } 
  
  public TimeLag getTimeLagConstraint(int srcNodeID, int destNodeID, RelationType relationType)
  {
    ConstraintEdge constraintEdge = getConstraintEdge(srcNodeID, destNodeID);
    if (constraintEdge != null)
    {
      return TimeModelTools.getTimeLag(constraintEdge, relationType);
    }
    else
    {
      if (RelationType.END_START.equals(relationType)
          && (getEdge(srcNodeID, destNodeID, EdgeType.ET_CONTROL) != null
              || getEdge(srcNodeID, destNodeID, EdgeType.ET_SYNC) != null))
      {
        return null; // TODO return a default time lag
      }
      return null;
    }
  }
  
  public TimeLag[] getTimeLagsConstraint(int srcNodeID, int destNodeID)
  {
    ConstraintEdge constraintEdge = getConstraintEdge(srcNodeID, destNodeID);
    return TimeModelTools.getTimeLags(constraintEdge);
  }  
  
  @Override
  public TemporalNode getNode(int nodeID)
  {
    Node node = super.getNode(nodeID);
    if (node != null)
    {
      return new DefaultTemporalNode(node);
    }
    return null;
  }

  @Override
  public String toString()
  {
    return getBaseTemplate().getName();
  }

  public Set<TemporalNode> getTemporalNodes()
  {
    Set<Node> nodes = super.getNodes();
    
    Set<TemporalNode> temporalNodes = new HashSet<TemporalNode>();
    
    for (Node node : nodes)
    {
      temporalNodes.add(new DefaultTemporalNode(node));
    }
    
    return temporalNodes;
  }



  
}
