package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.model.eventmodel.EventReference;

/*
 * Current Duration
 * 
 * 
 * Data Elements
 */
/** 
 *  <p>This interface supports user-defined update procedures for duration constraints. This allows to update activity duration constraints during runtime. This is for example necessary if the duration of an activity significantly depends on the context of the current instance. <br />
 *  An implementing class is notified when one out of a predefined set of triggers is raised. Based on the activity instance reference, the trigger, and the access to  <code>ActivityDurationUpdateMeasures</code> arbitrary source can be used to determine the new activity duration constraint.</p>
 *  <p>
 *  Each activity duration update handler is identified by a (unique) ID. This ID is also found in the activity duration.</p>
 */
public interface ActivityDurationUpdateHandler {
  /* {src_lang=Java, author=Andreas Lanz}*/


  public void updateDuration(ActivityInstanceReference activity, EventReference trigger, ActivityDurationUpdateProcedureData configuration, ActivityDurationUpdateMeasures updateMeasures, SessionToken session);

  /** 
   *  <p>Gets the (unique) ID of this activity duration update handling procedure to identify it as specified by the activity duration.</p>
   *  
   *  @returns The (unique) ID of this activity duration update handler procedure.
   */
  public String getID();

}