package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.model.common.viewer.Attribute;
import de.aristaflow.adept2.model.common.viewer.AttributeSetter;
import de.aristaflow.adept2.model.globals.TimeModelConstants.ReferencePoint;

/*
 * Checks for FixedDate
 * 
 */
/**
 * <p> A fixed date represents a restriction to the possible earliest or latest execution time of an activity in a process as described by time pattern 4.<br /> A particular element has to be executed at a particular date. Fixed Date Elements often determine the latest or earliest start / completion time of preceding / succeeding activities as well. If the deadline is missed the activity or process may even become obsolete. </p> <p> Whether the fixed date restricts the execution of the start or end event of the activity is decided by the ReferencePoint. </p> <p> Since execution times cannot be fixed at process template level the respective value needs to be provided during runtime. Therefore the <code>FixedDateUpdateProcedure</code> provides means to determine and update the respective activity duration constraint during runtime. </p> <p> Use <blockquote> <pre> PluginData pluginData = node.getPluginData(FixedDate.ID); FixedDate fixedDate = null; if (pluginData != null) { fixedDate = ModelViewer.createView(pluginData.getPluginData(), FixedDate.class); } </pre> </blockquote> or  {@link TimeModelTools#getFixedDate(de.aristaflow.adept2.model.processmodel.Node)}  to get an instance of this interface from a  {@link de.aristaflow.adept2.model.processmodel.Node}  . </p>
 */
@SuppressWarnings("deprecation")
public interface FixedDate {
  /* {src_lang=Java, java_import=de.aristaflow.adept2.model.common.viewer.Attribute, java_import=de.aristaflow.adept2.model.common.viewer.AttributeSetter}*/


  /**
   * The reference point determines whether the fixed date applies to the start or end of the respective activity.
   * @returns   whether this Fixed Date applies to the start or end of the respective activity.
   * @uml.property  name="referencePoint"
   * @uml.associationEnd  
   * @deprecated
   */
  @Attribute(name="ReferencePoint")
  public ReferencePoint getReferencePoint();

  /**
   * <p>The Configuration of the FixedDateUpdateProcedure which allows to update the execution date of this FixedDate during process execution.</p>
   * @returns   the Configuration of the FixedDateUpdateProcedure which will be used to update the execution date of this fixed date during runtime.
   * @uml.property  name="fixedDateUpdateProcedure"
   * @uml.associationEnd  
   */
  @Attribute(name="FixedDateUpdateProcedureData")
  public FixedDateUpdateProcedureData getFixedDateUpdateProcedure();

  /**
   * @param  data
   * @uml.property  name="fixedDateUpdateProcedure"
   */
  @AttributeSetter("FixedDateUpdateProcedureData")
  public void setFixedDateUpdateProcedure(FixedDateUpdateProcedureData data);

}