package de.aristaflow.adept2.model.processmodel.timemodel;

import java.net.URI;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.model.common.PluginData;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.processmodel.Template;

/*
 */
/**
 * <p>
 * This interface supports user-defined update procedures for fixed date
 * elements. This allows to update fixed date elements during runtime. This is
 * necessary since the execution date of an activity cannot be determined at
 * process template level. <br />
 * An implementing class is notified when one out of a predefined set of
 * triggers is raised. Based on the activity instance reference, the trigger,
 * and the access to <code>FixedDateUpdateMeasures</code> arbitrary source can
 * be used to determine the new execution date of the activity.
 * </p>
 * <p>
 * Each fixed date update handler is identified by a (unique) ID. This ID is
 * also found in the fixed date element.
 * </p>
 */
public interface FixedDateUpdateHandler
{
  /* {src_lang=Java} */

  // TODO reconsider return type
  public PluginData[] getUpdateTriggerList(Template template, Node node, FixedDateUpdateProcedureData configuration);


  /**
   * <p>
   * Notifies that on of the triggers specified by
   * {@link FixedDateUpdateHandlerData#getUpdateTriggerList()} has been raised.
   * The <code>FixedDateUpdateMeasures</code> allow to access the time manager,
   * for instance, to determine the current duration restriction of the
   * respective activity, or to update the duration restriction with a new value
   * and to set a new set of trigger events.
   * </p>
   * @param session The session token that must be used for accessing the given
   *          {@link FixedDateUpdateMeasures} object and only for it. When e.g.
   *          other services are accessed through the registry, a new session
   *          token must be used.
   * @param trigger The trigger that has been raised; The trigger corresponds to
   *          an event within the process
   * @param updateMeasures Update measures that can be used to update the
   *          execution time restrictions of the activity, e.g. set a new
   *          earlist start date restriction.
   * @param activity A reference to the activity for which the execution time
   *          restriction should be updated.
   * @param timeManager The Time Manager
   */
  public void updateFixedDate(SessionToken session, int nodeID, int nodeIteration,
      URI trigger, FixedDateUpdateProcedureData updateProcedureData,
      FixedDateUpdateMeasures updateMeasures);
//  public void updateFixedDate(EventDescription activity, EventReference trigger,
//      FixedDateUpdateProcedureData configuration, FixedDateUpdateMeasures updateMeasures, SessionToken session);

  /**
   * <p>
   * Gets the (unique) ID of this fixed date update handling procedure to
   * identify it as specified by the fixed date element.
   * </p>
   * 
   * @returns The (unique) ID of this fixed date update handler procedure.
   */
  public String getID();



}