package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.common.viewer.Attribute;
import de.aristaflow.adept2.model.common.viewer.AttributeSetter;

/*
 * TP2
 */
/**
 * <p> An process duration represents the duration constraints of a process as described by time pattern 2. </p> <p> The respective process has a certain duration restriction. Durations result from both waiting and processing times. Durations are often determined by external benchmarks (e.g., regulations, policies, QoS agreements). </p> <p> Process Durations may vary within certain boundaries. More precisely the minimum duration is by default the lower bound (i.e., LowerMinimumDuration) but may be restricted up to the upper bound (i.e., UpperMinimumDuration). For the maximum duration the default value is the upper bound (i.e., UpperMaximumDuration) and it may be restricted down to the lower bound (i.e., LowerMaximumDuration). Ultimately the minimum duration must still be lower then the respective maximum duration although UpperMinimumDuration may be higher than LowerMaximumDuration. <br /> LowerMinimumDuration and UpperMaximumDuration may be null. If LowerMinimumDuration is null UpperMinimumDuration also has to be null. Likewise if UpperMaximumDuration is null LowerMaximumDuration also has to be null. </p> <p> Use <blockquote> <pre> PluginData pluginData = template.getPluginData(ProcessDuration.ID); ProcessDuration processDuration = null; if (pluginData != null) { processDuration = ModelViewer.createView(pluginData.getPluginData(), ProcessDuration.class); } </pre> </blockquote> or  {@link TimeModelTools#getProcessDuration(de.aristaflow.adept2.model.processmodel.Template)}  to get an instance of this interface from a  {@link de.aristaflow.adept2.model.processmodel.Template}  . </p>
 */
public interface ProcessDuration {
  /* {src_lang=Java, java_import=de.aristaflow.adept2.model.common.viewer.Attribute, java_import=de.aristaflow.adept2.model.common.viewer.AttributeSetter}*/


  
  //public static final String ID = "TimeData.ProcessDuration";

  /**
   * @uml.property  name="lowerMinimumDuration"
   * @uml.associationEnd  
   */
  @Attribute(name="LowerMinimumDuration")
  public TimePoint getMinimumDuration();

  /**
   * @uml.property  name="upperMinimumDuration"
   * @uml.associationEnd  
   */
  @Attribute(name="UpperMinimumDuration")
  public TimePoint getGuardMinimumDuration();

  /**
   * @uml.property  name="lowerMaximumDuration"
   * @uml.associationEnd  
   */
  @Attribute(name="LowerMaximumDuration")
  public TimePoint getGuardMaximumDuration();

  /**
   * @uml.property  name="upperMaximumDuration"
   * @uml.associationEnd  
   */
  @Attribute(name="UpperMaximumDuration")
  public TimePoint getMaximumDuration();

  @Attribute(name="Contingency", defaultValue="0")
  public TimePoint getContingency();
  
  /**
   * @uml.property  name="expectedDuration"
   * @uml.associationEnd  
   */
  @Attribute(name="ExpectedDuration")
  public TimePoint getExpectedDuration();


  
  /**
   * @param  duration
   * @uml.property  name="lowerMinimumDuration"
   */
  @AttributeSetter("LowerMinimumDuration")
  public void setFlexibleMinimumDuration(TimePoint duration);

  /**
   * @param  duration
   * @uml.property  name="upperMinimumDuration"
   */
  @AttributeSetter("UpperMinimumDuration")
  public void setGuardMinimumDuration(TimePoint duration);

  /**
   * @param  duration
   * @uml.property  name="lowerMaximumDuration"
   */
  @AttributeSetter("LowerMaximumDuration")
  public void setGuardMaximumDuration(TimePoint duration);

  /**
   * @param  duration
   * @uml.property  name="upperMaximumDuration"
   */
  @AttributeSetter("UpperMaximumDuration")
  public void setFlexibleMaximumDuration(TimePoint duration);

  @AttributeSetter("Contingency")
  public void setContingency(TimePoint contingency);
  
  /**
   * @param  duration
   * @uml.property  name="expectedDuration"
   */
  @AttributeSetter("ExpectedDuration")
  public void setExpectedDuration(TimePoint duration);

}