package de.aristaflow.adept2.model.processmodel.timemodel;

import java.util.HashMap;
import java.util.Map;

import de.aristaflow.adept2.model.common.PluginData;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.common.viewer.ModelViewer;
import de.aristaflow.adept2.model.globals.TimeModelConstants.ReferencePoint;
import de.aristaflow.adept2.model.globals.TimeModelConstants.RelationType;
import de.aristaflow.adept2.model.processmodel.ConstraintEdge;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.processmodel.Template;

@SuppressWarnings("deprecation")
public class TimeModelTools
{
  protected static TemplateTemporalInformation getTemplateTemporalInformation(Template template)
  {
    PluginData pluginData = template.getPluginData(TemplateTemporalInformation.ID);
    TemplateTemporalInformation info = null;
    if (pluginData != null)
    {
      Map<String, String> map = new HashMap<String, String>(pluginData.getPluginData());
      info = ModelViewer.createView(map, TemplateTemporalInformation.class);
    }
    return info;
  }
  

  public static ProcessDuration getProcessDuration(Template template)
  {
    TemplateTemporalInformation info = getTemplateTemporalInformation(template);
    
    ProcessDuration processDuration = null;
    if (info != null)
    {
      processDuration = info.getProcessDuration();
    }
    else
    {
      // Downward compatibility
      PluginData pluginData = template.getPluginData("TimeData.ProcessDuration");
      
      if (pluginData != null)
      {
        processDuration = ModelViewer.createView(pluginData.getPluginData(), 
                                                 ProcessDuration.class);
      }
    }
    return processDuration;
  }
  
  public static ActivityDuration getActivityDuration(Node node)
  {
    ActivityDuration activityDuration = null;
    
    NodeTemporalInformation nodeTemporalInformation = getNodeTemporalInformation(node);
    if (nodeTemporalInformation != null)
    {
      activityDuration = nodeTemporalInformation.getActivityDuration();
    }
    else
    {
      // Downward compatibility
      PluginData pluginData = node.getPluginData("TimeData.ActivityDuration");
      
      if (pluginData != null)
      {
        activityDuration = ModelViewer.createView(new HashMap<String, String>(pluginData.getPluginData()), 
                                                               ActivityDuration.class);
      }
    }
    return activityDuration;
  }
  
  protected static ConstraintEdgeTemporalInformation getEdgeTemporalInformation(ConstraintEdge constraintEdge)
  {
    PluginData pluginData = constraintEdge.getPluginData(ConstraintEdgeTemporalInformation.ID);
    ConstraintEdgeTemporalInformation info = null;
    if (pluginData != null)
    {
      Map<String, String> map = new HashMap<String, String>(pluginData.getPluginData());
      info = ModelViewer.createView(map, ConstraintEdgeTemporalInformation.class);
    }
    return info;
  }
  
  public static TimeLag[] getTimeLags(ConstraintEdge constraintEdge)
  {
    ConstraintEdgeTemporalInformation temporalInformation = getEdgeTemporalInformation(constraintEdge);
    TimeLag[] timeLags = new TimeLag[0];
    if (temporalInformation != null)
    {
      timeLags = temporalInformation.getTimeLags();
    }
    else
    {
      // Downward compatibility
      PluginData pluginData = constraintEdge.getPluginData("TimeData.TimeLag");

      if (pluginData != null)
      {
        timeLags = new TimeLag[1];
        timeLags[0] = ModelViewer.createView(new HashMap<String, String>(pluginData.getPluginData()), 
                                                          TimeLag.class);      
      }
    }
    return timeLags;
  }

  public static TimeLag getTimeLag(ConstraintEdge constraintEdge, RelationType relationType)
  {
    ConstraintEdgeTemporalInformation temporalInformation = getEdgeTemporalInformation(constraintEdge);
    
    TimeLag timeLag = null;
    
    if (temporalInformation != null)
    {
      TimeLag[] timeLags = temporalInformation.getTimeLags();
      if (timeLags != null)
      {
        for (TimeLag t : timeLags)
        {
          if (relationType.equals(t.getRelationType()))
          {
            timeLag = t;
            break;
          }
        }
      }
    }
    else
    {
      // Downward compatibility
      PluginData pluginData = constraintEdge.getPluginData("TimeData.TimeLag");
      
      if (pluginData != null)
      {
        timeLag = ModelViewer.createView(pluginData.getPluginData(), 
                                                          TimeLag.class); 
        if (!timeLag.getRelationType().equals(relationType))
          timeLag = null;
      }
    }
    return timeLag;
  }
  
  
  public static FixedDate getStartFixedDate(Node node)
  {
    FixedDate fixedDate = null;
    
    
    NodeTemporalInformation nodeTemporalInformation = getNodeTemporalInformation(node);
    if (nodeTemporalInformation != null)
    {
      fixedDate = nodeTemporalInformation.getStartFixedDate();
    }
    else
    {
      PluginData pluginData = node.getPluginData("TimeData.FixedDate");
      
      
      if (pluginData != null)
      {
        
        fixedDate = ModelViewer.createView(pluginData.getPluginData(), 
                                                          FixedDate.class);
        if (fixedDate.getReferencePoint() != ReferencePoint.START)
          fixedDate = null;
      }
    }
    return fixedDate;
  }
  
  public static FixedDate getCompletionFixedDate(Node node)
  {
    FixedDate fixedDate = null;
    
    
    NodeTemporalInformation nodeTemporalInformation = getNodeTemporalInformation(node);
    if (nodeTemporalInformation != null)
    {
      fixedDate = nodeTemporalInformation.getCompletionFixedDate();
    }
    else
    {
      PluginData pluginData = node.getPluginData("TimeData.FixedDate");
      
      
      if (pluginData != null)
      {
        
        fixedDate = ModelViewer.createView(pluginData.getPluginData(), 
                                                          FixedDate.class);
        if (fixedDate.getReferencePoint() != ReferencePoint.COMPLETION)
          fixedDate = null;
      }
    }
    return fixedDate;
  }

//  public static ScheduleRestrictedElement getScheduleRestrictedElement(Node node)
//  {
//    PluginData pluginData = node.getPluginData(ScheduleRestrictedElement.ID);
//
//    ScheduleRestrictedElement scheduleRestrictedElement = null;
//    if (pluginData != null)
//    {
//      scheduleRestrictedElement = ModelViewer
//        .createView(pluginData.getPluginData(), ScheduleRestrictedElement.class);
//    }
//
//    return scheduleRestrictedElement;
//  }
  
  public static ScheduleRestrictedElement getStartSchedule(Node node)
  {
    ScheduleRestrictedElement scheduleRestrictedElement = null;
    
    
    NodeTemporalInformation nodeTemporalInformation = getNodeTemporalInformation(node);
    if (nodeTemporalInformation != null)
    {
      scheduleRestrictedElement = nodeTemporalInformation.getStartSchedule();
    }
    else
    {
      PluginData pluginData = node.getPluginData("TimeData.ScheduleRestrictedElement");
      
      
      if (pluginData != null)
      {
        
        scheduleRestrictedElement = ModelViewer.createView(pluginData.getPluginData(), 
                                                           ScheduleRestrictedElement.class);
        if (scheduleRestrictedElement.getReferencePoint() != ReferencePoint.START)
          scheduleRestrictedElement = null;
      }
    }
    return scheduleRestrictedElement;
  }
  
  public static ScheduleRestrictedElement getCompletionSchedule(Node node)
  {
    ScheduleRestrictedElement scheduleRestrictedElement = null;
    
    
    NodeTemporalInformation nodeTemporalInformation = getNodeTemporalInformation(node);
    if (nodeTemporalInformation != null)
    {
      scheduleRestrictedElement = nodeTemporalInformation.getCompletionSchedule();
    }
    else
    {
      PluginData pluginData = node.getPluginData("TimeData.ScheduleRestrictedElement");
      
      
      if (pluginData != null)
      {
        
        scheduleRestrictedElement = ModelViewer.createView(pluginData.getPluginData(), 
                                                           ScheduleRestrictedElement.class);
        if (scheduleRestrictedElement.getReferencePoint() != ReferencePoint.COMPLETION)
          scheduleRestrictedElement = null;
      }
    }
    return scheduleRestrictedElement;
  }
  
  public static ValidityPeriod getStartValidityPeriod(Node node)
  {
    ValidityPeriod validityPeriod = null;
    
    
    NodeTemporalInformation nodeTemporalInformation = getNodeTemporalInformation(node);
    if (nodeTemporalInformation != null)
    {
      validityPeriod = nodeTemporalInformation.getStartValidityPeriod();
    }
    else
    {
      PluginData pluginData = node.getPluginData("TimeData.ValidityPeriod");
      
      
      if (pluginData != null)
      {
        
        validityPeriod = ModelViewer.createView(pluginData.getPluginData(), 
                                                ValidityPeriod.class);
        if (validityPeriod.getReferencePoint() != ReferencePoint.START)
          validityPeriod = null;
      }
    }
    return validityPeriod;
  }

  public static ValidityPeriod getCompletionValidityPeriod(Node node)
  {
    ValidityPeriod validityPeriod = null;
    
    
    NodeTemporalInformation nodeTemporalInformation = getNodeTemporalInformation(node);
    if (nodeTemporalInformation != null)
    {
      validityPeriod = nodeTemporalInformation.getCompletionValidityPeriod();
    }
    else
    {
      PluginData pluginData = node.getPluginData("TimeData.ValidityPeriod");
      
      
      if (pluginData != null)
      {
        
        validityPeriod = ModelViewer.createView(pluginData.getPluginData(), 
                                                ValidityPeriod.class);
        if (validityPeriod.getReferencePoint() != ReferencePoint.COMPLETION)
          validityPeriod = null;
      }
    }
    return validityPeriod;
  }

  protected static NodeTemporalInformation getNodeTemporalInformation(Node node)
  {
    PluginData pluginData = node.getPluginData(NodeTemporalInformation.ID);
    NodeTemporalInformation nodeTemporalInformation = null;
    if (pluginData != null)
    {
      HashMap<String, String> map = new HashMap<String, String>(pluginData
        .getPluginData());

      nodeTemporalInformation = ModelViewer
        .createView(map, NodeTemporalInformation.class);
    }
    return nodeTemporalInformation;
  }

  public static ValidityPeriod getStartValidityPeriod(Template template)
  {
    ValidityPeriod validityPeriod = null;
    
    
    TemplateTemporalInformation temporalInformation = getTemplateTemporalInformation(template);
    if (temporalInformation != null)
    {
      validityPeriod = temporalInformation.getStartValidityPeriod();
    }
    else
    {
      PluginData pluginData = template.getPluginData("TimeData.ValidityPeriod");
      
      
      if (pluginData != null)
      {
        
        validityPeriod = ModelViewer.createView(pluginData.getPluginData(), 
                                                ValidityPeriod.class);
        if (validityPeriod.getReferencePoint() != ReferencePoint.START)
          validityPeriod = null;
      }
    }
    return validityPeriod;
  }

  public static ValidityPeriod getCompletionValidityPeriod(Template template)
  {
    ValidityPeriod validityPeriod = null;
    
    
    TemplateTemporalInformation temporalInformation = getTemplateTemporalInformation(template);
    if (temporalInformation != null)
    {
      validityPeriod = temporalInformation.getCompletionValidityPeriod();
    }
    else
    {
      PluginData pluginData = template.getPluginData("TimeData.ValidityPeriod");
      
      
      if (pluginData != null)
      {
        
        validityPeriod = ModelViewer.createView(pluginData.getPluginData(), 
                                                ValidityPeriod.class);
        if (validityPeriod.getReferencePoint() != ReferencePoint.COMPLETION)
          validityPeriod = null;
      }
    }
    return validityPeriod;
  }
  
  public static FixedDate getStartFixedDate(Template template)
  {
    FixedDate fixedDate = null;
    
    
    TemplateTemporalInformation temporalInformation = getTemplateTemporalInformation(template);
    if (temporalInformation != null)
    {
      fixedDate = temporalInformation.getStartFixedDate();
    }
    else
    {
      PluginData pluginData = template.getPluginData("TimeData.FixedDate");
      
      
      if (pluginData != null)
      {
        
        fixedDate = ModelViewer.createView(pluginData.getPluginData(), 
                                           FixedDate.class);
        if (fixedDate.getReferencePoint() != ReferencePoint.START)
          fixedDate = null;
      }
    }
    return fixedDate;
  }


  public static FixedDate getCompletionFixedDate(Template template)
  {
    FixedDate fixedDate = null;
    
    
    TemplateTemporalInformation temporalInformation = getTemplateTemporalInformation(template);
    if (temporalInformation != null)
    {
      fixedDate = temporalInformation.getCompletionFixedDate();
    }
    else
    {
      PluginData pluginData = template.getPluginData("TimeData.FixedDate");
      
      
      if (pluginData != null)
      {
        
        fixedDate = ModelViewer.createView(pluginData.getPluginData(), 
                                           FixedDate.class);
        if (fixedDate.getReferencePoint() != ReferencePoint.COMPLETION)
          fixedDate = null;
      }
    }
    return fixedDate;
  }


  public static ScheduleRestrictedElement getStartSchedule(Template template)
  {
    ScheduleRestrictedElement scheduleRestrictedElement = null;
    
    
    TemplateTemporalInformation temporalInformation = getTemplateTemporalInformation(template);
    if (temporalInformation != null)
    {
      scheduleRestrictedElement = temporalInformation.getStartScheduleRestrictedElement();
    }
    else
    {
      PluginData pluginData = template.getPluginData("TimeData.ScheduleRestrictedElement");
      
      
      if (pluginData != null)
      {
        
        scheduleRestrictedElement = ModelViewer.createView(pluginData.getPluginData(), 
                                                           ScheduleRestrictedElement.class);
        if (scheduleRestrictedElement.getReferencePoint() != ReferencePoint.START)
          scheduleRestrictedElement = null;
      }
    }
    return scheduleRestrictedElement;
  }


  public static ScheduleRestrictedElement getCompletionSchedule(
      Template template)
  {
    ScheduleRestrictedElement scheduleRestrictedElement = null;
    
    
    TemplateTemporalInformation temporalInformation = getTemplateTemporalInformation(template);
    if (temporalInformation != null)
    {
      scheduleRestrictedElement = temporalInformation.getCompletionScheduleRestrictedElement();
    }
    else
    {
      PluginData pluginData = template.getPluginData("TimeData.ScheduleRestrictedElement");
      
      
      if (pluginData != null)
      {
        
        scheduleRestrictedElement = ModelViewer.createView(pluginData.getPluginData(), 
                                                           ScheduleRestrictedElement.class);
        if (scheduleRestrictedElement.getReferencePoint() != ReferencePoint.COMPLETION)
          scheduleRestrictedElement = null;
      }
    }
    return scheduleRestrictedElement;
  }


  public static String toString(ActivityDuration activityDuration)
  {
    String flexibleMin = null;
    String guardMin = null;
    String guardMax = null; 
    String flexibleMax = null;
    
    
    TimePoint duration = activityDuration
        .getMinimumDuration();
    if (duration != null)
    {
      flexibleMin = convertToString(duration);
    }
    else
    {
      flexibleMin = "1s";
    }
    
    duration = activityDuration
        .getGuardMinimumDuration();
      
    if (duration != null)
    {
      guardMin = convertToString(duration);
      
    }
    
    duration = activityDuration
        .getGuardMaximumDuration();
    
    if (duration != null)
    {
      guardMax = convertToString(duration);
      
    }
    
    duration = activityDuration
        .getMaximumDuration();
    
    if (duration != null)
    {
      flexibleMax = convertToString(duration);
      
    }
    
    
    StringBuilder s = new StringBuilder();
    s.append("[[");
    s.append(flexibleMin);
    s.append(", ");
    
    if (guardMin != null)
      s.append(guardMin);
    else
      s.append(flexibleMax);
    
    s.append("] [");
    
    if (guardMax != null)
      s.append(guardMax);
    else 
      s.append(flexibleMin);
    s.append(", ");
    s.append(flexibleMax);
    
    s.append("]");

    return s.toString();
  }


  public static String toString(TimeLag timeLag)
  {
    String lowerMin = null;
    String upperMin = null;
    String lowerMax = null; 
    String upperMax = null;
    
    
    TimePoint timeDist = timeLag
        .getLowerMinimumTimeLag();
    if (timeDist != null)
    {
      lowerMin = convertToString(timeDist);
    }
    
    timeDist = timeLag
        .getUpperMinimumTimeLag();
      
    if (timeDist != null)
    {
      upperMin = convertToString(timeDist);
      
    }
    
    timeDist = timeLag
        .getLowerMaximumTimeLag();
    
    if (timeDist != null)
    {
      lowerMax = convertToString(timeDist);
      
    }
    
    timeDist = timeLag
        .getUpperMaximumTimeLag();
    
    if (timeDist != null)
    {
      upperMax = convertToString(timeDist);
      
    }
    
    
    StringBuilder s = new StringBuilder();
    s.append("[");
    if (lowerMin != null)
      s.append(lowerMin);
    else if (upperMax != null)
      s.append("-Inf");
    
    if (upperMin != null || lowerMax != null)
    {
      s.append(" [");
      
      if (upperMin != null)
        s.append(upperMin);
      
      if (upperMin != null || lowerMax != null)
        s.append(", ");
      
      if (lowerMax != null)
        s.append(lowerMax);
      
      s.append("] ");
    }
    else
    {
      s.append(", ");
    }
    
    if (upperMax != null)
      s.append(upperMax);
    else if (lowerMin != null)
      s.append("Inf");
    
    s.append("]");
    return s.toString();
  }

  public static String convertToString(TimePoint timeLag)
  {

      StringBuilder s = new StringBuilder();
      
      if (timeLag.getMonths() != TimePoint.UNDEFINED_TIME_POINT)
      {
        s.append(timeLag.getMonths());
        s.append("M");
      }
      
      if (timeLag.getDays() != TimePoint.UNDEFINED_TIME_POINT)
      {
        s.append(timeLag.getDays());
        s.append("d");
      }
      
      if (timeLag.getHours() != TimePoint.UNDEFINED_TIME_POINT)
      {
        s.append(timeLag.getHours());
        s.append("h");
      }
      
      if (timeLag.getMinutes() != TimePoint.UNDEFINED_TIME_POINT)
      {
        s.append(timeLag.getMinutes());
        s.append("m");
      }
      
      if (timeLag.getSeconds() != TimePoint.UNDEFINED_TIME_POINT)
      {
        s.append(timeLag.getSeconds());
        s.append("s");
      }

      return s.toString();
  }


  public static String toString(FixedDate fixedDate)
  {
    
    String description = fixedDate.getFixedDateUpdateProcedure().getDescription();
    if (description != null)
    {
      return description;
    }
    return "Fixed Date";
  }


  public static String toString(ScheduleRestrictedElement schedule)
  {
    return schedule.getSchedule();
  }










  
}
