package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.common.viewer.Attribute;
import de.aristaflow.adept2.model.common.viewer.AttributeSetter;
import de.aristaflow.adept2.model.globals.TimeModelConstants.ReferencePoint;

/*
 * Checks for Validity Period
 * 
 * 
 */
/**
 * <p> A validity period represents a restriction on the lifetem of a activity, i.e., a process step, as described by time pattern 7.<br /> The respective process step may only be instantiated within this validity period. In general, different versions of a process element may exist, but only one is valid at a specific point in time. Validity dates are especially relevant in the context of process evolution to restrict the remaining lifetime of an obsolete process implementation and to schedule rollout of the new process. </p> <p> Whether the validity period restricts the execution of the start or end event of the activity is decided by the ReferencePoint of the <code>ValidityPeriod</code>. </p> <p> The validity period may be provided by an external source, e.g. a data element or the configuration, , as referenced by of #getValidityPeriodSource(). </p> <p> Either #getValidFrom(), #getValidUntil() or #getValidityPeriodSource() needs to be provided. </p> <p> Use <blockquote> <pre> PluginData pluginData = template.getPluginData(ValidityPeriod.ID); ValidityPeriod validityPeriod = null; if (pluginData != null) { validityPeriod = ModelViewer.createView(pluginData.getPluginData(), ValidityPeriod.class); } </pre> </blockquote> or    {@link TimeModelTools#getValidityPeriod(de.aristaflow.adept2.model.processmodel.Template)}    to get an instance of this interface from a    {@link de.aristaflow.adept2.model.processmodel.Template}     or use <blockquote> <pre> PluginData pluginData = node.getPluginData(ValidityPeriod.ID); ValidityPeriod validityPeriod = null; if (pluginData != null) { validityPeriod = ModelViewer.createView(pluginData.getPluginData(), ValidityPeriod.class); } </pre> </blockquote> or    {@link TimeModelTools#getValidityPeriod(de.aristaflow.adept2.model.processmodel.Node)     )}     to get an instance of this interface from a    {@link de.aristaflow.adept2.model.processmodel.Node}    . </p>
 * @uml.dependency   supplier="de.aristaflow.adept2.model.processmodel.timemodel.ValidityPeriodSystemBehaviour"
 */
@SuppressWarnings("deprecation")
public interface ValidityPeriod {
  /* {src_lang=Java, java_import=de.aristaflow.adept2.model.common.viewer.Attribute, java_import=de.aristaflow.adept2.model.common.viewer.AttributeSetter}*/


    
  //public static final String ID = "TimeData.ValidityPeriod";

  /**
   * <p> The start of the validity period of the respective activity.</p>
   * @returns     The start of the validity period of the respective activity.
   * @uml.property   name="validFrom"
   * @uml.associationEnd   
   * @returns    The start of the validity period of the respective activity.
   */
  @Attribute(name="ValidFrom")
  public CalendarElement getValidFrom();

  /**
   * <p> The end of the validity period of the respective activity.</p>
   * @returns     The end of the validity period of the respective activity.
   * @uml.property   name="validUntil"
   * @uml.associationEnd   
   * @returns    The end of the validity period of the respective activity.
   */
  @Attribute(name="ValidUntil")
  public CalendarElement getValidUntil();

  /**
   * The reference point determines whether the validity period applies to the start or end of the respective activity.
   * @returns     whether this validity period applies to the start or end of the respective activity.
   * @uml.property   name="referencePoint"
   * @uml.associationEnd   
   * @returns    whether this validity period applies to the start or end of the respective activity.
   * @deprecated
   */
  @Attribute(name="ReferencePoint")
  public ReferencePoint getReferencePoint();

  /**
   * <p> TODO AL</p> <p>Either #getPeriod() or #getValidityPeriodSource() needs to be provided but not both at the same time</p>
   * @uml.property  name="validityPeriodSource"
   */
  @Attribute(name="ValidityPeriodSource")
  public String getValidityPeriodSource();

  /**
   * @param  from
   * @uml.property  name="validFrom"
   */
  @AttributeSetter("ValidFrom")
  public void setValidFrom(CalendarElement from);

  /**
   * @param  until
   * @uml.property  name="validUntil"
   */
  @AttributeSetter("ValidUntil")
  public void setValidUntil(CalendarElement until);

  /**
   * @param  referencePoint
   * @uml.property  name="referencePoint"
   */
  @AttributeSetter("ReferencePoint")
  public void setReferencePoint(ReferencePoint referencePoint);

  /**
   * @param  source
   * @uml.property  name="validityPeriodSource"
   */
  @AttributeSetter("ValidityPeriodSource")
  public void setValidityPeriodSource(String source);

  @Attribute(name="SystemBehaviour")
  public String getSystemBehaviourID();
}