package de.aristaflow.adept2.model.processmodel.timemodel.defaultimplementation;

import java.net.URI;

import de.aristaflow.adept2.base.service.Registry;
import de.aristaflow.adept2.base.service.ServiceNotKnownException;
import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.core.datamanager.DataManager;
import de.aristaflow.adept2.core.datamanager.ProcessAwareAccess;
import de.aristaflow.adept2.core.executionmanager.ExecutionManager;
import de.aristaflow.adept2.core.executionmanager.defaultimplementation.ParameterRefResolver;
import de.aristaflow.adept2.model.common.plugindata.LongDateSystemDataFormatter;
import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.datamanagement.InputDataContainer;
import de.aristaflow.adept2.model.eventmodel.EventDescription;
import de.aristaflow.adept2.model.globals.TimeModelConstants.ReferencePoint;
import de.aristaflow.adept2.model.processmodel.Instance;
import de.aristaflow.adept2.model.processmodel.timemodel.FixedDate;
import de.aristaflow.adept2.model.processmodel.timemodel.FixedDateUpdateMeasures;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;

public class DefaultFixedDateUpdateMeasures implements FixedDateUpdateMeasures
{

//  private ExecutionManager executionManager;
  private Registry registry;
  private Instance instance;
  private ExecutionManager executionManager;
  private UpdateableInstanceTimeModel timeModel;
  private FixedDate fixedDate;
  private ReferencePoint referencePoint;
  private int targetNodeID;
  private int nodeIteration;

  public DefaultFixedDateUpdateMeasures(Registry registry, ExecutionManager executionManager, Instance instance,
      UpdateableInstanceTimeModel timeModel, int targetNodeID, int nodeIteration, 
      FixedDate fixedDate, ReferencePoint referencePoint)
  {
    this.registry = registry;
    this.executionManager = executionManager;
    this.instance = instance;
    this.timeModel = timeModel;
    this.targetNodeID = targetNodeID;
    this.nodeIteration = nodeIteration;
    this.fixedDate = fixedDate;
    this.referencePoint = referencePoint;

  }

  @Override
  public boolean updateEarliestDate(SessionToken session, int nodeID, int iteration, CalendarElement date,
      boolean ignoreInconsistency)
  {
    switch (referencePoint)
    {
      case START:
        Interval<CalendarElement> startTimeframe = timeModel.getActivityInstanceStartTimeframe(nodeID, iteration);
        
        if (compare(startTimeframe.getMaxValue(), date) > 0 || ignoreInconsistency)
        {
          timeModel.setEarliestStartDate(session, nodeID, iteration, date);
          return true;
        }
        return false;
      case COMPLETION:
        Interval<CalendarElement> completionTimeframe = timeModel.getActivityInstanceCompletionTimeframe(nodeID, iteration);
        
        if (compare(completionTimeframe.getMaxValue(), date) > 0 || ignoreInconsistency)
        {
          timeModel.setEarliestCompletionDate(session, nodeID, iteration, date);
          return true;
        }
        return false;
      default:
        return false;
        
    }
    
  }

  @Override
  public boolean updateLatestDate(SessionToken session, int nodeID, int iteration, CalendarElement date,
      boolean ignoreInconsistency)
  {
    switch (referencePoint)
    {
      case START:
        Interval<CalendarElement> startTimeframe = timeModel.getActivityInstanceStartTimeframe(nodeID, iteration);
        
        if (compare(startTimeframe.getMinValue(), date) < 0 || ignoreInconsistency)
        {
          timeModel.setLatestStartDate(session, nodeID, iteration, date);
          return true;
        }
        return false;
      case COMPLETION:
        Interval<CalendarElement> completionTimeframe = timeModel.getActivityInstanceCompletionTimeframe(nodeID, iteration);
        
        if (compare(completionTimeframe.getMinValue(), date) < 0 || ignoreInconsistency)
        {
          timeModel.setLatestCompletionDate(session, nodeID, iteration, date);
          return true;
        }
        return false;
      default:
        return false;
        
    }
  }

  private int compare(CalendarElement minValue, CalendarElement date)
  {
    long diff = date.getTimeInMillis() - minValue.getTimeInMillis();
    return diff == 0 ? 0 : (diff > 0 ? 1 : -1);
  }


  @Override
  public void addUpdateTrigger(SessionToken session, EventDescription activity, EventDescription trigger)
  {
    // TODO Auto-generated method stub

  }

  @Override
  public void removeUpdateTrigger(SessionToken session, EventDescription activity, EventDescription trigger)
  {
    // TODO Auto-generated method stub

  }

  @Override
  public void clearUpdateTriggerList(SessionToken session, EventDescription activity)
  {
    // TODO Auto-generated method stub

  }
  
  public String resolveParameterString(SessionToken session, int nodeID, String param)
  {
    
    try
    {      
      URI[] dataManagerURI = executionManager.getDataManager(session, instance.getID());
      DataManager dataManager = registry.getService(session, dataManagerURI, DataManager.class);
      
      // Create the corresponding ExecutableBusinessProcessInstance:
      ProcessAwareAccess paa = dataManager.getProcessAwareAccess();
//      ParameterDataContext dataContext = paa.getPublicDataContext(session, instance, nodeID);
      InputDataContainer nodeDataContainer = paa.getNodeDataContainer(session, instance, nodeID);
//      Node node = instance.getTemplate().getNode(nodeID);

      String string = ParameterRefResolver.getFormattedReplacedString(session, nodeDataContainer,
                                                                      LongDateSystemDataFormatter.class, param);

      return string;
    }
    catch (ServiceNotKnownException e)
    {

      e.printStackTrace();
      return null;
    }

    

  }

}
