package de.aristaflow.adept2.model.processmodel.timemodel.schedules.evaluation;

import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.BasicCalendar;

public class ComplexIntervalSet implements IntervalSet
{

  private final List<TimeFrame> timeFrames = new ArrayList<TimeFrame>();
  private final BasicCalendar granule;
  private final int offset;
  
  public ComplexIntervalSet(BasicCalendar granule)
  {
    this.granule = granule;
    this.offset = 0;
  }
  
  public void add(TimeFrame interval)
  {
    timeFrames.add(interval);
  }

  public int size()
  {
    return this.timeFrames.size();
  }

  public boolean isEmpty()
  {
    return size() == 0;
  }
  
  public void removeLast()
  {
    this.timeFrames.remove(this.timeFrames.size() - 1);
  }

  public Iterable<TimeFrame> getTimeFrames()
  {
    return new Iterable<TimeFrame>()
    {
      
      public Iterator<TimeFrame> iterator()
      {
        return Collections.unmodifiableList(timeFrames).iterator();
      }
    };
  }

  @Override
  public Calendar getStart()
  {    
    return timeFrames.get(0).getStart();
  }

  @Override
  public Calendar getEnd()
  {
    return timeFrames.get(timeFrames.size()-1).getEnd();
  }
  
  @Override
  public String toString()
  {
    StringBuilder s = new StringBuilder();
    s.append("[");
    String sep = "";
    for (TimeFrame timeFrame : timeFrames)
    {
      s.append(sep);
      sep = ", ";
      s.append(timeFrame);
    }
    s.append("]@"+offset);
    return s.toString();
  }

  public TimeFrame get(int index)
  {
     return timeFrames.get(index);
  }
  
  public TimeFrame getAbsolute(int index)
  {
    return timeFrames.get(index-offset);
  }

  public BasicCalendar getGranule()
  {
    return granule;
  }

  public int getOffset()
  {
    return offset;
  }

  public IntervalSet range(int fromIndex, int toIndex)
  {
    if (fromIndex < 0)
      throw new IndexOutOfBoundsException();
    
    if (toIndex > this.timeFrames.size())
      throw new IndexOutOfBoundsException();
    
    ComplexIntervalSet intervalSet = new ComplexIntervalSet(this.granule);
    
    for (int i = 0; i <= toIndex; i++)
    {
      intervalSet.add(this.timeFrames.get(i));
    }
    
    return intervalSet;
  }

  public IntervalSet range(Calendar start, Calendar end)
  {
    ComplexIntervalSet set = new ComplexIntervalSet(this.granule);
    for (TimeFrame tf1 : getTimeFrames())
    {
      if (tf1.getStart().compareTo(start) >= 0 && tf1.getEnd().compareTo(end) <= 0)
      {
        set.add(tf1);
      } 
//        else if (!strict && tf1.getStart().compareTo(tf2.getEnd()) <= 0 && tf1.getEnd().compareTo(tf2.getStart()) >= 0)
//        {
//          set.add(tf1);
//        }
    }
    return set;
  }



}
