package de.aristaflow.adept2.model.processmodel.timemodel.schedules.evaluation;

import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Iterator;

import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.BasicCalendar;

public class ContinuousIntervalSet implements IntervalSet
{
  
  private static final SimpleDateFormat format = new SimpleDateFormat("dd.MM.yy HH:mm:ss.S");

  private final class ContinuousIntervalSetIterator implements Iterator<TimeFrame>
  {
    private Calendar current;

    public ContinuousIntervalSetIterator(Calendar start)
    {
      current = (Calendar) start.clone();
    }

    public boolean hasNext()
    {
      return current.before(end);
    }

    public TimeFrame next()
    {
      Calendar start = (Calendar) current.clone();
      current.add(baseCalender.getCalendarConstant(), 1);
      Calendar end = (Calendar)current.clone();
      end.add(Calendar.MILLISECOND, -1);
      return new Interval(baseCalender, start, end);
    }

    public void remove()
    {
      throw new UnsupportedOperationException();
    }
  }


  private BasicCalendar baseCalender;
  private Calendar start;
  private Calendar end;
  private int size = -1;

  public ContinuousIntervalSet(BasicCalendar baseCalender, Calendar start, Calendar end)
  {
    this.baseCalender = baseCalender;
    this.start = start;
    this.end = end;
  }
  
  
  public ContinuousIntervalSet()
  {
    // TODO Auto-generated constructor stub
  }


  public Calendar getStart()
  {
    return start;
  }

  public Calendar getEnd()
  {
    return end;
  }

  public BasicCalendar getGranule()
  {
    return baseCalender;
  }
  
//  SimpleDateFormat format = new SimpleDateFormat("dd.MM.yy hh:mm:ss.S");
  
  public Interval get(int index)
  {
    if (index < 0)
      throw new IndexOutOfBoundsException();
    
    Calendar from = (Calendar) start.clone();
    from.add(baseCalender.getCalendarConstant(), index);
      
    Calendar to = (Calendar) from.clone();
    to.add(baseCalender.getCalendarConstant(), 1);
    to.add(Calendar.MILLISECOND, -1);
    
    if (to.after(end))
      throw new IndexOutOfBoundsException();
    
    return new Interval(this.baseCalender, from, to);
  }
  
  public ContinuousIntervalSet range(int fromIndex, int toIndex)
  {
    if (fromIndex < 0)
      throw new IndexOutOfBoundsException();
    
    if (toIndex < 0)
      throw new IndexOutOfBoundsException();
    
    if (toIndex < fromIndex)
      throw new IllegalArgumentException();
    
    Calendar from = (Calendar) start.clone();
    from.add(baseCalender.getCalendarConstant(), fromIndex);
    
    Calendar to = (Calendar) start.clone();
    to.add(baseCalender.getCalendarConstant(), toIndex+1);
    to.add(Calendar.MILLISECOND, -1);
    
    if (to.after(end))
      throw new IndexOutOfBoundsException();
    
    return new ContinuousIntervalSet(baseCalender, from, to);
  }


  public Iterable<TimeFrame> getTimeFrames()
  {
    return new Iterable<TimeFrame>()
    {
      
      public Iterator<TimeFrame> iterator()
      {
        return new ContinuousIntervalSetIterator(start);
      }
    };
  }


  public int size()
  {
    if (size < 0)
    {
      Iterator<TimeFrame> iter = new ContinuousIntervalSetIterator(start);
      int i = 0;
      while (iter.hasNext())
      {
        i++;
        iter.next();
      }
      size = i;
    }
    
    
    return size;
  }
  
  @Override
  public String toString()
  {
    StringBuilder s = new StringBuilder();
    s.append("[");
    s.append(format.format(start.getTime()));
    s.append("-");
    s.append(this.baseCalender);
    s.append("-");
    s.append(format.format(end.getTime()));
    
    
    
//    String sep = "";
//    for (TimeFrame timeFrame : getTimeFrames())
//    {
//      s.append(sep);
//      sep = ", ";
//      s.append(timeFrame);
//    }
    s.append("]");
    return s.toString();
  }


  public IntervalSet range(Calendar from, Calendar to)
  {
    Calendar start = (Calendar) this.start.clone();
    
    while (start.compareTo(from) < 0)
    {
      start.add(baseCalender.getCalendarConstant(), 1);
    }
    
    Calendar end = (Calendar) start.clone();
    
    while (end.compareTo(to) < 0)
    {
      end.add(baseCalender.getCalendarConstant(), 1);
    }
    
    end.add(Calendar.MILLISECOND, -1);
    
    if (end.compareTo(to) > 0)
    {
      end.add(baseCalender.getCalendarConstant(), -1);
    }
        
    
    return new ContinuousIntervalSet(baseCalender, start, end);
  }

  
  
}
