package de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang;

import org.antlr.v4.runtime.ANTLRInputStream;
import org.antlr.v4.runtime.BailErrorStrategy;
import org.antlr.v4.runtime.CharStream;
import org.antlr.v4.runtime.CommonTokenStream;
import org.antlr.v4.runtime.IntStream;
import org.antlr.v4.runtime.NoViableAltException;
import org.antlr.v4.runtime.ParserRuleContext;
import org.antlr.v4.runtime.RecognitionException;
import org.antlr.v4.runtime.Token;
import org.antlr.v4.runtime.TokenStream;
import org.antlr.v4.runtime.misc.ParseCancellationException;

import de.aristaflow.adept2.core.timemanager.ScheduleResolutionException;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.Callang6Parser.CalendarExpressionContext;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.Callang6Parser.CalendarScriptContext;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.Callang6Parser.ParseCalendarExpressionContext;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.CalendarExpression;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.script.CalendarScript;

public class ScheduleParser
{
    public static CalendarExpression parse(String rule) throws ScheduleResolutionException
    {
      Callang6Parser parser = getParser(rule);
      ParseCalendarExpressionContext c;
      try
      {
        c = parser.parseCalendarExpression();
      }
      catch (RecognitionException e)
      {
        throw new ScheduleResolutionException(e.getMessage(), e);
      }
      catch (ParseCancellationException e)
      {
        String message = e.getMessage();
        if (e.getCause() instanceof RecognitionException)
        {
          RecognitionException re = (RecognitionException) e.getCause();
          ParserRuleContext context = (ParserRuleContext) re.getCtx();
          Token offendingToken = re.getOffendingToken();
          
          int line = 0;
          int charPositionInLine = 0;
          if (offendingToken != null)
          {
            line = offendingToken.getLine();           
            charPositionInLine = offendingToken.getCharPositionInLine();
          }
          
          String text = null;
          if (context != null)
          {
            Token start = context.getStart();
            if (start != null)
            {
              CharStream inputStream = start.getInputStream();
              if (inputStream != null)
                text = inputStream.toString();
            }
          }
  
          String msg = "";
          if (re instanceof NoViableAltException)
          {
            msg = "no viable alternative";
          }
  
          message = "line " + line + ":" + charPositionInLine + " " + msg + (text != null ? " at input '" + text + "'" : "");
        }
        
        throw new ScheduleResolutionException(message, e);
      }
      catch (RuntimeException e)
      {
        e.printStackTrace();
        throw e;
      }
      CalendarExpression script = c.expression;
      return script;
    }
    
    public static CalendarScript parseScript(String rule) throws ScheduleResolutionException
    {
      Callang6Parser parser = getParser(rule);
      CalendarScriptContext c;
      try
      {
        c = parser.calendarScript();
      }
      catch (RecognitionException e)
      {
        throw new ScheduleResolutionException(e.getMessage(), e);
      }
      catch (ParseCancellationException e)
      {
        Throwable ex = e.getCause() != null ? e.getCause() : e;
        throw new ScheduleResolutionException(ex.getMessage(), ex);
      }
      CalendarScript script = c.c;
      return script;
    }

    private static Callang6Parser getParser(String rule)
    {
      
      CharStream stream = new ANTLRInputStream(rule);
      Callang6Lexer lexer = new Callang6Lexer(stream);
      lexer.removeErrorListeners();
      TokenStream tokenstream = new CommonTokenStream(lexer);
      Callang6Parser parser = new Callang6Parser(tokenstream);
      parser.removeErrorListeners();
      parser.setErrorHandler(new BailErrorStrategy());
      
      return parser;
    }
}
