package de.aristaflow.adept2.model.processmodel.timemodel.updatehandlers;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.aristaflow.adept2.model.processmodel.timemodel.FixedDateUpdateHandler;
import de.aristaflow.adept2.model.processmodel.timemodel.TimeLagUpdateHandler;
import de.aristaflow.adept2.util.LoggerTools;

//FIXME AL: This is quick and dirty. Replace by real Registry / use the AristaFlow Plugin Registry instead.
public class UpdateHandlerRegistry
{

  private static Logger logger = LoggerTools.getLogger(UpdateHandlerRegistry.class);
  
  /**
   * @author  lanz
   */
  public static class UpdateHandlerReference
  {
    private final String id;

    /**
     * @uml.property  name="name"
     */
    private final String name;

    /**
     * @uml.property  name="description"
     */
    private final String description;

    /**
     * @uml.property  name="clazz"
     */
    private final Class<?> clazz;

    public UpdateHandlerReference(String id, String name, String description,
        Class<?> clazz)
    {
      this.id = id;
      this.name = name;
      this.description = description;
      this.clazz = clazz;
    }

    public String getHandlerID()
    {
      return id;
    }

    /**
     * @return
     * @uml.property  name="name"
     */
    public String getName()
    {
      return name;
    }

    /**
     * @return
     * @uml.property  name="clazz"
     */
    public Class<?> getClazz()
    {
      return clazz;
    }

    /**
     * @return
     * @uml.property  name="description"
     */
    public String getDescription()
    {
      return description;
    }
  }

  private static Map<String, UpdateHandlerReference> fixedDateHandlers;

  private static Map<String, UpdateHandlerReference> timeLagHandlers;

  static
  {
    fixedDateHandlers = new HashMap<String, UpdateHandlerReference>();
    fixedDateHandlers
      .put(
           DataElementFixedDateUpdateHandler.ID,
           new UpdateHandlerReference(
                                      DataElementFixedDateUpdateHandler.ID,
                                      "Read from Data Element",
                                      "Reads the value of the Fixed Date from a Data Element of the Process.",
                                      DataElementFixedDateUpdateHandler.class));

    timeLagHandlers = new HashMap<String, UpdateHandlerReference>();
    timeLagHandlers
      .put(
           DataElementTimeLagUpdateHandler.ID,
           new UpdateHandlerReference(
                                      DataElementTimeLagUpdateHandler.ID,
                                      "Read form Data Element",
                                      "Reads the value of the Time Lag from a Data Element of the Process.",
                                      DataElementTimeLagUpdateHandler.class));

  }

  public static Collection<UpdateHandlerReference> getFixedDateUpdateHandlers()
  {
    return Collections.unmodifiableCollection(fixedDateHandlers.values());
  }
  
  public static Collection<UpdateHandlerReference> getTimeLagUpdateHandlers()
  {
    return Collections.unmodifiableCollection(timeLagHandlers.values());
  }



  public static FixedDateUpdateHandler getFixedDateUpdateHandler(String id)
  {
    UpdateHandlerReference reference = fixedDateHandlers.get(id);
    FixedDateUpdateHandler updateHandler = null;
    try
    {

      updateHandler = (FixedDateUpdateHandler) reference.getClazz()
        .newInstance();
    }
    catch (InstantiationException e)
    {
      logger.log(Level.SEVERE, "InstantiationException occcurred", e);
    }
    catch (IllegalAccessException e)
    {
      logger.log(Level.SEVERE, "IllegalAccessException occcurred", e);
    }

    return updateHandler;
  }
  
  public TimeLagUpdateHandler getTimeLagUpdateHandler(String id)
  {
    UpdateHandlerReference reference = timeLagHandlers.get(id);
    TimeLagUpdateHandler updateHandler = null;
    try
    {

      updateHandler = (TimeLagUpdateHandler) reference.getClazz()
        .newInstance();
    }
    catch (InstantiationException e)
    {
      logger.log(Level.SEVERE, "InstantiationException occcurred", e);
    }
    catch (IllegalAccessException e)
    {
      logger.log(Level.SEVERE, "IllegalAccessException occcurred", e);
    }

    return updateHandler;
  }
}
