package de.aristaflow.adept2.model.timemodel;

import java.util.UUID;

import de.aristaflow.adept2.model.eventmodel.EventDescription;
import de.aristaflow.adept2.model.eventmodel.EventReference;

/**
 * <p>
 * Holds all temporal information related to a process instance.
 * </p>
 * <p>
 * This interface only holds information about the execution time stamps of the
 * process interface and does not consider any time constraints!
 * </p>
 */
public interface InstanceTimeContext
{
  /**
   * Represents a time stamp which is not set.
   */
  public static final long TIME_STAMP_NOT_SET = Long.MIN_VALUE;

  /**
   * 
   * @return The ID of the Instance this {@link InstanceTimeContext} belongs to.
   */
  public UUID getInstanceID();

  /**
   * 
   * @return The time stamp (in millis) at which the corresponding instance was
   *         created or {@link #TIME_STAMP_NOT_SET} if not set
   */
  public long getInstanceCreated();

  /**
   * 
   * @return The time stamp (in millis) at which the corresponding instance was
   *         started or {@link #TIME_STAMP_NOT_SET} if not set
   */
  public long getInstanceStarted();

  /**
   * 
   * @return The time stamp (in millis) at which the corresponding instance was
   *         finished or {@link #TIME_STAMP_NOT_SET} if not set
   */
  public long getInstanceFinished();

  /**
   * 
   * @param nodeID
   * @param iteration
   * @return The time stamp (in millis) at which the corresponding activity was
   *         started or {@link #TIME_STAMP_NOT_SET} if not set
   */
  public long getActivityStarted(int nodeID, int iteration);

  /**
   * 
   * @param nodeID
   * @param iteration
   * @return The time stamp (in millis) at which the corresponding activity was
   *         finished or {@link #TIME_STAMP_NOT_SET} if not set
   */
  public long getActivityFinished(int nodeID, int iteration);

  public long getEventOccurrence(EventDescription event);

  public EventReference[] getEventOccurrences();
}
