package de.aristaflow.adept2.model.timemodel;

import java.util.UUID;

public interface TimeModelStatus
{

  /**
   * @author   Andreas Lanz
   */
  public enum TimeModelUpdateStatus
  {
    /**
     * Time Management is completely enabled.<br />
     * Everything is enabled 
     */
    ENABLED,
    
    /**
     * Time Management is enabled but no predictions are calculated.<br />
     * Time Context and Time Model is kept up to date but no calculations are done on the time model. Simple Escalation and Notifications are scheduled. 
     */
    NO_CALCULATION,
    
    /**
     * Time Management is disabled but kept up to date.<br />
     * Time Context and Time Model is kept up to date but no calculations are done on the time model and no Escalation and Notifications are scheduled.
     */
    NO_UPDATE,
    
    /**
     * Time Management is disabled.<br />
     * Time Context is kept up to date. But no time model is created or update. No Escalation and Notifications are scheduled.
     */
    DISABLED;
  }

  /**
   * @author  Andreas Lanz
   */
  public enum TimeModelState
  {
    /**
     * The state of the time model is unknown. This may have several reasons:
     * <ul>
     * <li>The time model has not been fully initialized yet</li>
     * <li>The last update of the time model has been interrupted (i.e., it has
     * not been completed). Thus the time model is in an unkown state.</li>
     * <li>The time model is out of date; i.e., it has been modified but not yet been updated.</li>
     * </ul>
     * May be combined with {@link TimeModelUpdateStatus}#&#42;
     */
    UNKOWN, // TODO introduce #UNKOWN where required <- no update / no calculation and out of date...
    /**
     * <p>The time model is up-to-date and valid.</p>
     * 
     * May be combined with {@link TimeModelUpdateStatus}#&#42;
     */
    VALID, 
    /**
     * <p>The time model is invalid.</p>
     * 
     * May only be combined with {@link TimeModelUpdateStatus#DISABLED}, {@link TimeModelUpdateStatus#NO_UPDATE} and {@link TimeModelUpdateStatus#NO_CALCULATION}
     */
    INVALID,    
  }

  TimeModelUpdateStatus getTimeModelUpdateStatus();
  
  TimeModelState getTimeModelStatus();

  UUID getInstanceID();

  String getTimeModelUpdateStatusComment();
}
