package de.aristaflow.adept2.model.timemodel.defaultimplementation;

import java.io.Serializable;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;

import de.aristaflow.adept2.model.eventmodel.EventDescription;
import de.aristaflow.adept2.model.eventmodel.EventReference;
import de.aristaflow.adept2.model.timemodel.InstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeContext;

public class DefaultInstanceTimeContext implements InstanceTimeContext,
    UpdateableInstanceTimeContext
{

  private final UUID instanceID;
  private long instanceStarted;
  private long instanceFinished;
  private long instanceCreated;
  
  
  protected Map<Integer, ActivityInformation> activityInfo = new HashMap<Integer, DefaultInstanceTimeContext.ActivityInformation>();
  
  public static class ActivityInstanceInformation implements Serializable
  {

    /**
     * 
     */
    private static final long serialVersionUID = -8691130061127491365L;
    
    public long activityStarted;
    
    public long activityFinished;
    
    public int iteration;
  }
  
  public static class ActivityInformation implements Serializable
  {
    
    /**
     * 
     */
    private static final long serialVersionUID = -3968100593483352153L;
    
    protected Map<Integer, ActivityInstanceInformation> activityInfo = new HashMap<Integer, DefaultInstanceTimeContext.ActivityInstanceInformation>();
    

    public ActivityInstanceInformation createActivityInstance(int iteration)
    {
      ActivityInstanceInformation info;
      Integer nodeInteger = Integer.valueOf(iteration);
      if (activityInfo.containsKey(nodeInteger))
      {
        info = activityInfo.get(nodeInteger);
      }
      else
      {
        info = new ActivityInstanceInformation();
        info.iteration = iteration;
        activityInfo.put(nodeInteger, info);
      }
      return info;
    }


    public ActivityInstanceInformation getActivityInstanceInformation(int iteration)
    {
      ActivityInstanceInformation info = null;
      Integer nodeInteger = Integer.valueOf(iteration);
      if (activityInfo.containsKey(nodeInteger))
      {
        info = activityInfo.get(nodeInteger);
      }
      return info;
    }
    
  }
  

  public DefaultInstanceTimeContext(UUID instanceID)
  {
    this.instanceID = instanceID;
  }


  public UUID getInstanceID()
  {
    return this.instanceID;
  }


  public long getActivityFinished(int nodeID, int iteration)
  {
    ActivityInformation activityInfo = getActivityInfo(nodeID);
    if (activityInfo != null)
    {
      ActivityInstanceInformation activityInstanceInformation = activityInfo.getActivityInstanceInformation(iteration);
      if (activityInstanceInformation != null)
        return activityInstanceInformation.activityFinished;
    }
    
    return Long.MIN_VALUE;
  }


  public long getActivityStarted(int nodeID, int iteration)
  {
    ActivityInformation activityInfo = getActivityInfo(nodeID);
    if (activityInfo != null)
    {
      ActivityInstanceInformation activityInstanceInformation = activityInfo.getActivityInstanceInformation(iteration);
      if (activityInstanceInformation != null)
        return activityInstanceInformation.activityStarted;
    }
    
    return Long.MIN_VALUE;
  }





  public long getEventOccurrence(EventDescription event)
  {
    // future use
    return -1;
  }


  public EventReference[] getEventOccurrences()
  {
    // future use
    return null;
  }
  
  public void addEventOccurrence(EventReference event)
  {
    // future use
    
  }



  public long getInstanceCreated()
  {
    return instanceCreated;
  }


  public long getInstanceFinished()
  {
    return instanceFinished;
  }


  public long getInstanceStarted()
  {
    return instanceStarted;
  }

  public void setActivityFinished(int nodeID, int iteration, long timestamp)
  {
    createActivityInfo(nodeID).createActivityInstance(iteration).activityFinished = timestamp;
  }


  public void setActivityStarted(int nodeID, int iteration, long timestamp)
  {
    createActivityInfo(nodeID).createActivityInstance(iteration).activityStarted = timestamp;
  }


  protected ActivityInformation createActivityInfo(int nodeID)
  {
    ActivityInformation info;
    Integer nodeInteger = Integer.valueOf(nodeID);
    if (activityInfo.containsKey(nodeInteger))
    {
      info = activityInfo.get(nodeInteger);
    }
    else
    {
      info = new ActivityInformation();
      activityInfo.put(nodeInteger, info);
    }
    return info;
  }
  
  private ActivityInformation getActivityInfo(int nodeID)
  {
    ActivityInformation info = null;
    Integer nodeInteger = Integer.valueOf(nodeID);
    if (activityInfo.containsKey(nodeInteger))
    {
      info = activityInfo.get(nodeInteger);
    }
    return info;
  }


  public void setInstanceCreated(long timestamp)
  {
    this.instanceCreated = timestamp;    
  }


  public void setInstanceFinished(long timestamp)
  {
    this.instanceFinished = timestamp;    
  }


  public void setInstanceStarted(long timestamp)
  {
    this.instanceStarted = timestamp;    
  }
}
