package de.aristaflow.adept2.model.timemodel.defaultimplementation;

import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.aristaflow.adept2.model.common.PluginData;
import de.aristaflow.adept2.model.common.viewer.ModelViewer;
import de.aristaflow.adept2.model.globals.ActivityConstants.AccessType;
import de.aristaflow.adept2.model.processmodel.TemporalNode;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.timemodel.FixedDate;
import de.aristaflow.adept2.model.processmodel.timemodel.FixedDateUpdateProcedureData;
import de.aristaflow.adept2.model.processmodel.timemodel.updatehandlers.DataElementUpdateTrigger;
import de.aristaflow.adept2.model.processmodel.tools.ProcessElementIdentifierTools;
import de.aristaflow.adept2.util.LoggerTools;

public class TimeModelTools
{
  public static final String COMPLETION_FIXED_DATE_FIELD = "completionFixedDate";
  public static final String START_FIXED_DATE_FIELD = "startFixedDate";
  protected static Logger logger = LoggerTools.getLogger(TimeModelTools.class); 
  
  public static Map<Integer, List<URI>> createUpdateTriggerMapping(TemporalTemplate template)
  {
    URI baseURI = ProcessElementIdentifierTools.getTemplateIdentifier(template, "localhost");
    
    Map<Integer, List<URI>> updateTriggers = new HashMap<Integer, List<URI>>();
    
    Set<TemporalNode> nodes = template.getTemporalNodes();
    
    for (TemporalNode node : nodes)
    {
      //List<Integer> triggerNodes = new ArrayList<Integer>();
      FixedDate fixedDate = node.getStartFixedDate();
      if (fixedDate != null)
      {
        List<Integer> triggerNodes = createUpdateTriggers(template, fixedDate);
        
        URI identifier = ProcessElementIdentifierTools.getNodeIdentifier(baseURI, node, START_FIXED_DATE_FIELD);
        
        addTriggers(updateTriggers, triggerNodes, identifier);
      }
      
      fixedDate = node.getCompletionFixedDate();
      if (fixedDate != null)
      {
        List<Integer> triggerNodes = createUpdateTriggers(template, fixedDate);
        
        URI identifier = ProcessElementIdentifierTools.getNodeIdentifier(baseURI, node, COMPLETION_FIXED_DATE_FIELD);
        
        addTriggers(updateTriggers, triggerNodes, identifier);
      }
      
    }
    return updateTriggers;
    
  }

  private static void addTriggers(Map<Integer, List<URI>> updateTriggers, List<Integer> triggerNodes, URI identifier)
  {
    for (Integer i : triggerNodes)
    {
      List<URI> list = updateTriggers.get(i);
      if (list == null)
      {
        list = new ArrayList<URI>();
        updateTriggers.put(i, list);
      }
      
      list.add(identifier);
    }
  }

  private static List<Integer> createUpdateTriggers(TemporalTemplate template, FixedDate fixedDate)
  {
    List<Integer> triggerNodes = new ArrayList<Integer>();
    
    FixedDateUpdateProcedureData updateProcedure = fixedDate.getFixedDateUpdateProcedure();
    if (updateProcedure != null)
    {
      PluginData[] triggerList = updateProcedure.getUpdateTriggerList();
      if (triggerList != null)
      {
        List<Integer> triggers = createUpdateTriggers(template, triggerList);
        triggerNodes.addAll(triggers);
      }
    }
    return triggerNodes;
  }

  private static List<Integer> createUpdateTriggers(TemporalTemplate template, PluginData[] triggerList)
  {
    List<Integer> triggerNodes = new ArrayList<Integer>();
    for (PluginData data : triggerList)
    {
      if (data.getPluginID().equals(DataElementUpdateTrigger.ID))
      {
        DataElementUpdateTrigger updateTrigger = ModelViewer.createView(data.getPluginData(),
          DataElementUpdateTrigger.class);
        int dataElementID = updateTrigger.getDataElementID();
        List<Integer> writers = getWriters(template, dataElementID);
        triggerNodes.addAll(writers);
      }
      else
      {
        logger.log(Level.SEVERE, "Unknown PluginID '%s'!", data.getPluginID());
      }
    }
    return triggerNodes;

  }

  private static List<Integer> getWriters(TemporalTemplate template, int dataElementID)
  {
    List<Integer> nodes = new ArrayList<Integer>();
    int[] nodeIDs = template.getAccessingNodeIDsForDataElement(dataElementID, AccessType.WRITE);
    for (int id : nodeIDs)
    {
      nodes.add(Integer.valueOf(id));
    }
    
    nodeIDs = template.getAccessingNodeIDsForDataElement(dataElementID, AccessType.SYSTEM_WRITE);
    for (int id : nodeIDs)
    {
      nodes.add(Integer.valueOf(id));
    }
    return nodes;
  }
}
