package de.aristaflow.adept2.model.timemodel.tcn;

import java.util.Collection;

public interface ChangeableTCN<N extends TCNTimePoint, E extends TCNEdge<N, ?>> extends TCN<N, E>
{
  /**
   * 
   * @return set containing all changed edges, @
   *         java.util.Collections#emptySet()} indicates no changes,
   *         <code>null</code> indicates that changes are unknown (i.e., all edges are considered to be changed).
   */
  public Collection<E> getChanges();
  
  /**
   * Updates a Edge with the given value.
   * 
   * @param edge The edge to be updated to the new value
   * @return The updated edge. This value shall be used for any further
   *         operations on the <code>edge</code> as the internal representation
   *         of the edge may differ from the external and thus both may
   *         otherwise get out of sync.
   * @throws IllegalArgumentException in case the edge cannot be updated.
   * @throws IllegalStateException possibly throws {@link IllegalStateException}
   *           in case the cstnu becomes inconsistent after the update. Note
   *           that this property is not guaranteed, but may depend on the
   *           implementation (in current implementations this only happens if
   *           assertions enabled)
   */
  public E updateConstraint(E edge) throws IllegalArgumentException;
 
  
  public void setConsistent();
  
  public N insertTimePoint(N timePoint);
  
  public void removeTimePoint(N timePoint);
}
