package de.aristaflow.adept2.model.timemodel.tcn;

import java.util.Arrays;

/**
 * This object is immutable
 * 
 * @author Andreas
 *
 */
public class IterationChain implements Comparable<IterationChain>
{

  /**
   * This object is immutable
   * 
   * @author Andreas
   *
   */
  public static class Iteration
  {

    private final int blockNodeID;
    private final int iteration;

    public Iteration(int blockNodeID, int iteration)
    {
      this.blockNodeID = blockNodeID;
      this.iteration = iteration;
    }

    public int getNodeID()
    {
      return blockNodeID;
    }

    public int getIteration()
    {
      return iteration;
    }

    @Override
    public int hashCode()
    {
      return blockNodeID << 32 ^ iteration;
    }

    @Override
    public boolean equals(Object obj)
    {
      if (obj == this)
        return true;
      
      if (obj instanceof Iteration)
      {
        Iteration other = (Iteration) obj;
        if (this.blockNodeID == other.blockNodeID && this.iteration == other.iteration)
          return true;
        
        return false;
      }
      
      return super.equals(obj);
    }
    
    @Override
    public String toString()
    {
      return blockNodeID + "@" +iteration; 
    }
  }

  
  private final Iteration[] iterations;
  private int cachedHashCode = Integer.MIN_VALUE;

  public IterationChain(int startNodeID)
  {
    this.iterations = new Iteration[1];
    this.iterations[0] = new Iteration(startNodeID, 0);
  }
  
  @Deprecated // Should only be used by Import/Export and internally
  public IterationChain(Iteration[] iterations)
  {
    this.iterations = iterations;
  }

  public IterationChain subIteration(int startNodeID)
  {
    Iteration[] it = new Iteration[this.iterations.length+1];
    System.arraycopy(this.iterations, 0, it, 0, this.iterations.length);
    it[this.iterations.length] = new Iteration(startNodeID, 1); 
    return new IterationChain(it);
  }

  public IterationChain incrementIteration()
  {
    Iteration[] it = new Iteration[this.iterations.length];
    System.arraycopy(this.iterations, 0, it, 0, this.iterations.length);
    Iteration current = it[this.iterations.length-1];
    it[this.iterations.length-1] = new Iteration(current.getNodeID(), current.getIteration()+1);
    return new IterationChain(it);
  }

  public int getIteration()
  {
    int iteration = 1;
    for (int i = iterations.length-1; i>=1; i--) // DO NOT USE THE FIRST ITERATION AS IT IS THE PROCESS AND HAS ALWAYS ITERATION 0
    {
      iteration *= iterations[i].getIteration();
    }
    return iteration;
  }

  @Override
  public int hashCode()
  {
    if (cachedHashCode == Integer.MIN_VALUE)
    {
      cachedHashCode = Arrays.hashCode(this.iterations);
    }
    return cachedHashCode;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    
    if (obj instanceof IterationChain)
    {
      IterationChain other = (IterationChain) obj;
      boolean result = Arrays.equals(this.iterations, other.iterations);
      return result;
    }
    return super.equals(obj);
  }

  @Override
  public String toString()
  {
    return Arrays.toString(this.iterations);
  }
  
  public Iteration[] getIterations()
  {
    return this.iterations;
  }

  @Override
  public int compareTo(IterationChain other)
  {
    Iteration[] iter = this.iterations;
    Iteration[] anotherIter = other.iterations;
    
    if (iter.length > anotherIter.length)
      return 1;
    
    if (iter.length < anotherIter.length)
      return -1;
    
    int i = 0;
    while (i < iter.length && i < anotherIter.length)
    {
      if (iter[i] != anotherIter[i])
        break;
      i++;
    }
    
    if (i >= iter.length && i >= anotherIter.length)
      return 0;
    
    Iteration it = iter[i];
    Iteration anotherIt = anotherIter[i];
    
    int thisVal = it.blockNodeID;
    int anotherVal = anotherIt.blockNodeID;
    int compare = (thisVal < anotherVal ? -1 : (thisVal == anotherVal ? 0 : 1));
    
    if (compare != 0)
      return compare;
    
    thisVal = it.iteration;
    anotherVal = anotherIt.iteration;
    compare = (thisVal < anotherVal ? -1 : (thisVal == anotherVal ? 0 : 1));

    return compare;
  }
  

}
