package de.aristaflow.adept2.model.timemodel.tcn;

import java.lang.reflect.Array;
import java.util.Arrays;
import java.util.Collection;

public class JoinNodeMapping<N extends TCNTimePoint> extends NodeMapping<N>
{


  private final N[] prefixNodes;
  private final NodeInstance[] branches;

  public JoinNodeMapping(String label, NodeInstance nodeInstance, N[] prefixNodes, N startNode, N[] intermediateEvents, N endNode)
  {
    super(label, nodeInstance, startNode, intermediateEvents, endNode);
    this.prefixNodes = prefixNodes == null ? (N[])Array.newInstance(startNode.getClass(),0) : prefixNodes;
    this.branches = new NodeInstance[this.prefixNodes.length];
  }

  @Override
  public N getPrefixNode()
  {
    throw new IllegalAccessError();
  }
  
  public N[] getPrefixNodes()
  {
    return prefixNodes;
  }
  
  public NodeInstance[] getBranches()
  {
    return branches;
  }
  
  
  @Override
  public Collection<N> getNodes()
  {
    Collection<N> x = super.getNodes();
    x.addAll(Arrays.asList(prefixNodes));
    return x;
  }
  
  
  @Override
  public N getPrefixNode(int branchID)
  {
    if (this.branches.length > 0)
      return prefixNodes[branchID];
    return getStartNode();
  }

  
  @Override
  public void setBranchID(int branchID, NodeMapping<N> pred)
  {
    if (this.branches.length > 0)
      this.branches[branchID] = pred.getNodeInstance();
  }
  
  @Override
  public int getBranchID(N branchNode)
  {
    
    if (!this.getNodes().contains(branchNode))
      throw new IllegalArgumentException();
    
    if (branchNode == getStartNode() && prefixNodes.length == 0)
      return 0;
    
    for (int i = 0; i < prefixNodes.length; i++)
    {
      if (prefixNodes[i] == branchNode)
        return i;
    }

    
    return -1;
  }
  
  @Override
  public int getBranchID(NodeMapping<N> pred)
  {
    if (this.branches.length > 0)
    {
      NodeInstance nodeInstance = pred.getNodeInstance();
      for (int i = 0; i < branches.length; i++)
      {
        NodeInstance n = branches[i];
        
        if (n == nodeInstance)
          return i;
      }
    }
    else
    {
      return 0;
    }
    return -1;
  }
    
  @Override
  public boolean checkComplete()
  {
    //
    for (int i = 0; i < branches.length; i++)
    {
      assert branches[i] != null;
    }
    return super.checkComplete();
  }

}
