package de.aristaflow.adept2.model.timemodel.tcn;

public class NodeInstance implements Comparable<NodeInstance>
{
  private final int nodeID;
  private final IterationChain iteration;
  private final int topologicalID;

  public NodeInstance(int nodeID, IterationChain iteration, int topologicalID)
  {
    this.nodeID = nodeID;
    this.iteration = iteration;
    this.topologicalID = topologicalID;
  }

  public int getNodeID()
  {
    return nodeID;
  }

  public IterationChain getIteration()
  {
    return iteration;
  }

  @Override
  public int hashCode()
  {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((iteration == null) ? 0 : iteration.hashCode());
    result = prime * result + nodeID;
    return result;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    NodeInstance other = (NodeInstance) obj;
    if (iteration == null)
    {
      if (other.iteration != null)
        return false;
    }
    else if (!iteration.equals(other.iteration))
      return false;
    if (nodeID != other.nodeID)
      return false;
    return true;
  }
  
  @Override
  public String toString()
  {
    StringBuilder s = new StringBuilder();
    s.append(nodeID).append("^{").append(iteration).append("}");
    return s.toString();
  }

  public int getTopologicalID()
  {
    return topologicalID;
  }

  @Override
  public int compareTo(NodeInstance other)
  {
    int thisVal = this.nodeID;
    int anotherVal = other.nodeID;
    int compare = (thisVal < anotherVal ? -1 : (thisVal == anotherVal ? 0 : 1));
    
    if (compare != 0)
      return compare;
    
    return this.iteration.compareTo(other.getIteration());
  }
  
  
}
