package de.aristaflow.adept2.model.timemodel.tcn;

import java.lang.reflect.Array;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;


/**
 * This Object is immutable
 * 
 * @author Andreas
 *
 * @param <N>
 */
public class NodeMapping<N extends TCNTimePoint>
{

  private final String label;


  private final NodeInstance nodeInstance;
  
//  private final int nodeID;
//  private final IterationChain iterationChain;
  
  private final N startNode;
  private final N endNode;
  private final N[] intermediateEvents;


  
  

  @SuppressWarnings("unchecked")
  public NodeMapping(String label, NodeInstance nodeInstance, N startNode, N[] intermediateEvents, N endNode)
  {
    this.label = label;
    this.nodeInstance = nodeInstance;
    this.startNode = startNode;
    
    this.intermediateEvents = intermediateEvents == null ? (N[])Array.newInstance(startNode.getClass(), 0) : intermediateEvents;
    this.endNode = endNode;
  }
  
  public String getLabel()
  {
    return label;
  }

  public NodeInstance getNodeInstance()
  {
    return nodeInstance;
  }
  
  public N getStartNode()
  {
    return startNode;
  }
  


  public N getEndNode()
  {
    return endNode;
  }
  
  public Collection<N> getNodes()
  {
    List<N> asList = new ArrayList<N>();
    asList.add(startNode);
    asList.addAll(Arrays.asList(getIntermediateEvents()));
    asList.add(endNode);
    return asList;
  }

  
  @Override
  public String toString()
  {
    if (label != null && label.length() > 0)
      return label;
    else
      return "Node "+nodeInstance; 
  }

  public N[] getIntermediateEvents()
  {
    return intermediateEvents;
  }

  public N getPrefixNode()
  {
    return startNode;
  }

  public N getSuffixNode()
  {
    return endNode;
  }

  public void setBranchID(int branchID, NodeMapping<N> node)
  {
    //
  }
  
  public void setBranch(N branchNode, NodeMapping<N> node)
  {
    int branchID = getBranchID(branchNode);
    if (branchID != -1)
      setBranchID(branchID, node);
  }
  
  public int getBranchID(NodeMapping<N> node)
  {
    return 0;
  }
  
  public int getBranchID(N branchNode)
  {
    return 0;
  }
  
  public N getPrefixNode(int branchID)
  {
    return startNode;
  }
  
  public N getSuffixNode(int branchID)
  {
    return endNode;
  }

  public boolean checkComplete()
  {
    // 
    return true;
  }

  public N getSuffixNode(NodeMapping<N> succPoints)
  {
    int branchID = getBranchID(succPoints);
    return getSuffixNode(branchID);
  }

  public N getPrefixNode(NodeMapping<N> predPoints)
  {
    int branchID = getBranchID(predPoints);
    return getPrefixNode(branchID);
  }

}
