package de.aristaflow.adept2.model.timemodel.tcn;

import java.lang.reflect.Array;
import java.util.Arrays;
import java.util.Collection;


public class SplitNodeMapping<N extends TCNTimePoint> extends NodeMapping<N>
{

  private final N[] suffixNodes;
  private NodeInstance[] branches;


  public SplitNodeMapping(String label, NodeInstance nodeInstance, N startNode, N[] intermediateEvents, N endNode, N[] suffixNodes)
  {
    
    super(label, nodeInstance, startNode, intermediateEvents, endNode);
    this.suffixNodes = suffixNodes == null ? (N[])Array.newInstance(startNode.getClass(),0) : suffixNodes;
    this.branches = new NodeInstance[this.suffixNodes.length];
    
  }

  @Override
  public N getSuffixNode()
  {
    throw new IllegalAccessError();
  }
  
  public N[] getSuffixNodes()
  {
    return this.suffixNodes;
  }
  
  public NodeInstance[] getBranches()
  {
    return branches;
  }
  
  @Override
  public Collection<N> getNodes()
  {
    Collection<N> x = super.getNodes();
    x.addAll(Arrays.asList(suffixNodes));
    return x;
  }
  
  @Override
  public void setBranchID(int branchID, NodeMapping<N> pred)
  {
    if (branches.length > 0)
      this.branches[branchID] = pred.getNodeInstance();
  }
  
  @Override
  public int getBranchID(NodeMapping<N> succ)
  {
    if (branches.length > 0)
    {
      NodeInstance nodeInstance = succ.getNodeInstance();
      for (int i = 0; i < branches.length; i++)
      {
        NodeInstance n = branches[i];
        
        if (n == nodeInstance)
          return i;
      }
      
    }
    else
    {
      return 0;
    }
    return -1;
  }

  @Override
  public N getSuffixNode(int branchID)
  {
    if (suffixNodes.length > 0)
      return suffixNodes[branchID];
    return getEndNode();
  }
  
  @Override
  public int getBranchID(N branchNode)
  {
    if (!this.getNodes().contains(branchNode))
      throw new IllegalArgumentException();
    
    if (branchNode == getEndNode() && suffixNodes.length == 0)
      return 0;
    
    for (int i = 0; i < suffixNodes.length; i++)
    {
      if (suffixNodes[i] == branchNode)
        return i;
    }

    
    return -1;
  }
  
  @Override
  public boolean checkComplete()
  {
    //
    for (int i = 0; i < branches.length; i++)
    {
      assert branches[i] != null;
    }
    return super.checkComplete();
  }

}
