package de.aristaflow.adept2.model.timemodel.tcn;

/**
 * A TCNEdge represents an edge in a TCN graph between the nodes given by {@link #getSourceNode()} and 
 * {@link #getTargetNode()}. The particular edge is annotated with value {@link #getValue()}.
 *  
 * Note that a TCN is considered to be a simple graph. That is, between any pair of nodes at most one edge may exist.
 * In particular, if more than one edge shall exist between two nodes this has to be encoded in the value of a single 
 * edge (i.e., the value of an edge may be complex and contain the annotation of different edges).
 *  
 * @author lanz
 *
 * @param <N> The type of TCNNode connected by the edge
 * @param <V> The type of value the edge is annotated with.
 */
public interface TCNEdge<N extends TCNTimePoint, V extends ConstraintValue> extends Cloneable
{

  public enum FlowDirection 
  {
    FORWARD,
    BACKWARD,
    UNKOWN
  }
  
  
  public N getSourceNode();

  public N getTargetNode();

  public V getValue();

  public TCNEdge<N, V> clone();

  public FlowDirection getFlowDirection();

  public boolean isControlEdge();

  public boolean isImplicit();

}
