package de.aristaflow.adept2.model.timemodel.tcn.cstnImpl;


import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.UUID;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.CSTN;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.STN;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.Projection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.STNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.SimpleSTN;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractTCN.DirectedGraphImplementation;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;

public class CSTN2STNProjection implements STN<LabeledTimePoint, STNEdge>, Projection<LabeledTimePoint, LabeledEdge, STNEdge>, Cloneable
{
  protected CSTN cstn;

  public CSTN2STNProjection(CSTN cstn)
  {
    this.cstn = cstn;
  }

  @Override
  public STNEdge getEdge(LabeledTimePoint from, LabeledTimePoint to)
  {
    LabeledEdge labeledEdge = cstn.getEdge(from, to);
    
    TimeDistance resultEdgeValue = null;
    
    if (labeledEdge != null)
    {
      Collection<LabeledValue> values = labeledEdge.getValue().getLabeledValues();
      for (LabeledValue vx : values)
      {
        // Just a normal edge...
        resultEdgeValue = resultEdgeValue == null ? vx.getValue() : min(resultEdgeValue, vx.getValue());
      }
    }
    
    if (resultEdgeValue == null)
      return null;
    
    return new STNEdge(from, to, new DefaultConstraintValue(resultEdgeValue), 
                       labeledEdge != null ? labeledEdge.isControlEdge() : false,
                       labeledEdge != null ? labeledEdge.isImplicit() : false,                    
                       labeledEdge != null ? labeledEdge.getFlowDirection() : FlowDirection.UNKOWN);
  }
  
  public Collection<STNEdge> getIncomingEdges(LabeledTimePoint to)
  {
    Collection<LabeledTimePoint> timePoints = getTimePoints();
    List<STNEdge> edges = new ArrayList<STNEdge>(timePoints.size());
    for (LabeledTimePoint from : timePoints)
    {
      STNEdge edge = getEdge(from, to);
      if (edge != null)
      {
        edges.add(edge);
      }
    }
    return edges;
  }

  public Collection<STNEdge> getOutgoingEdges(LabeledTimePoint from)
  {
    Collection<LabeledTimePoint> timePoints = getTimePoints();
    List<STNEdge> edges = new ArrayList<STNEdge>(timePoints.size());
    for (LabeledTimePoint to : timePoints)
    {
      STNEdge edge = getEdge(from, to);
      if (edge != null)
      {
        edges.add(edge);
      }
    }
    return edges;
  }


  
  public TimeDistance min(TimeDistance v1, TimeDistance v2)
  {
    if (v1.compareTo(v2) <= 0)
      return v1;
    else
    return v2;
  }
  
  @Override
  public int hashCode()
  {
    return cstn.hashCode();
  }
  
  @Override
  public LabeledTimePoint getTimeBaseEvent()
  {
    return cstn.getTimeBaseEvent();
  }
  
  @Override
  public Collection<LabeledTimePoint> getTimePoints()
  {
    return cstn.getTimePoints();
  }

  public boolean contains(LabeledTimePoint timePoint)
  {
    return cstn.contains(timePoint);
  }
  
  @Override
  public Collection<STNEdge> getEdges()
  {
    Collection<STNEdge> edges = new ArrayList<STNEdge>();
    
    for (LabeledTimePoint n1 : cstn.getTimePoints())
    {
      for (LabeledTimePoint n2 : cstn.getTimePoints())
      {
        STNEdge edge = getEdge(n1, n2);
        if (edge != null)
          edges.add(edge);
      }
    }
    
    return edges;
  }

  @Override
  public boolean equals(Object obj)
  {
    return cstn.equals(obj);
  }

  @Override
  public void visit(GraphVisitor<LabeledTimePoint, STNEdge> visitor)
  {
    visitor.discoverGraph(this);
  
//  
//    for (NodeMapping<LabeledTimePoint> n : getNodeMapping())
//    {
//      
//      DirectedGraphImplementation<LabeledTimePoint, STNEdge> subGraph = new DirectedGraphImplementation<LabeledTimePoint, STNEdge>(n);
//      subGraph.visit(visitor);
//    }
//    
    for (LabeledTimePoint e : getTimePoints())
    {
      visitor.discoverVertex(e);
      visitor.finishVertex(e);
    }
    
    for (LabeledTimePoint ei : getTimePoints())
    {
      for (LabeledTimePoint ej : getTimePoints())
      {
        STNEdge e = getEdge(ei, ej);
        if (e != null)
        {
          visitor.discoverEdge(e.getSourceNode(), 
                               e.getTargetNode(), 
                               e);
          visitor.finishEdge(e.getSourceNode(), 
                             e.getTargetNode(), 
                             e);
        }
      }
    }
  
    visitor.finishGraph(this);
  }


  @Override
  public String toString()
  {
    return cstn.toString();
  }

  @Override
  public TCN<LabeledTimePoint, LabeledEdge> getBaseTCN()
  {
    return cstn;
  }

  @Override
  public STN<LabeledTimePoint, STNEdge> clone()
  {
    // Nodes are immutable
    Collection<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>(getTimePoints());
    Collection<STNEdge> edges = new ArrayList<STNEdge>();
    
    for (LabeledTimePoint n1 : nodes)
    {
      for (LabeledTimePoint n2 : nodes)
      {
        STNEdge edge = getEdge(n1, n2);
        if (edge != null)
          edges.add(edge.clone());
      }
    }
    
    return new SimpleSTN(nodes, edges);
  }

  @Override
  public boolean isConsistent()
  {
    return this.cstn.isConsistent();
  }


  public Collection<LabeledTimePoint> getVertices()
  {
    return getTimePoints();
  }
  
  
}
