package de.aristaflow.adept2.model.timemodel.tcn.cstnImpl;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.model.globals.ProcessConstants.NodeType;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.tcn.CSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTN;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.JoinNodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractLabeledTemplateTransformation;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultObservationTimePoint;

public class CSTNTemplateTransformation extends AbstractLabeledTemplateTransformation
{

  public CSTNTemplateTransformation(Configuration configuration, TemporalTemplate template, int lookAhead)
  {
    super(configuration, template, lookAhead);
  }

  public CSTNTemplateTransformation(Configuration configuration, TemporalTemplate template)
  {
    super(configuration, template);
  }
  
  public CSTNTemplateTransformation(TemporalTemplate template, int lookAhead)
  {
    this(null, template, lookAhead);
  }

  public CSTNTemplateTransformation(TemporalTemplate template)
  {
    this(null, template);
  }


  @Override
  protected ChangeableCSTN createTCN(Collection<LabeledTimePoint> nodes,
      Collection<LabeledEdge> edges,
      Collection<NodeMapping<LabeledTimePoint>> nodeMapping)
  {
    Set<ChoiceVariable> observations = new HashSet<ChoiceVariable>();
    
    for (LabeledTimePoint node : nodes)
    {
      if (node instanceof DefaultObservationTimePoint)
      {
        observations.add(((DefaultObservationTimePoint) node).getObservation());
      }
    }
    
    // FIXME
    ChangeableCSTN timeModel = new DefaultCSTN(nodes, edges, observations);
    return timeModel;
  }

  
  @Override
  public CSTN transform()
  {
    return (CSTN) super.transform();
  }
  
  @Override
  protected NodeMapping<LabeledTimePoint> transformActivity(int nodeID,
      IterationChain iterationChain, Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_NORMAL)
    {
      throw new IllegalArgumentException(""+template.getNodeType(nodeID));
    }
    
    int topologicalID = template.getNodeTopologicalID(nodeID);
    
    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);
    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);
    
    addNode(A_S);
    addNode(A_E);
    
    TCNActivityDuration duration = getActivityDuration(nodeID);
    

    LabeledEdge minEdge = new DefaultLabeledEdge(A_E, A_S, 
      new DefaultLabeledValueSet(new DefaultLabeledValue(branches, duration.minDuration.neg())), 
      true, false, FlowDirection.BACKWARD);
    minEdge = addEdge(minEdge);
    
    LabeledEdge maxEdge = new DefaultLabeledEdge(A_S, A_E, 
      new DefaultLabeledValueSet(new DefaultLabeledValue(branches, duration.maxDuration)), 
      true, false, FlowDirection.FORWARD);
    maxEdge = addEdge(maxEdge);

    NodeMapping<LabeledTimePoint> node = new NodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, 
        null, A_E);
    
    processActivityValidityPeriod(nodeID, node);
    
    return addNodeMapping(node);
  }
  
  @Override
  protected JoinNodeMapping<LabeledTimePoint> transformANDJoin(int nodeID, IterationChain iterationChain, Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_AND_JOIN)
    {
      throw new IllegalArgumentException("" + template.getNodeType(nodeID));
    }

    

    

    int topologicalID = template.getNodeTopologicalID(nodeID);

    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);

    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);
    addNode(A_S);
    addNode(A_E);


    createTCNEdge(A_S, A_E, branches, STRUCTURE_NODE_MIN_DURATION, STRUCTURE_NODE_MAX_DURATION, true);

//    int[] preds = template.getPredByEdgeType(nodeID, EdgeType.ET_CONTROL);
//    LabeledTimePoint[] nodes = new LabeledTimePoint[preds.length];
//    Arrays.fill(nodes, A_S);
    
    return addNodeMapping(new JoinNodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, null, A_S, null, A_E));
  }
  
  

}
