package de.aristaflow.adept2.model.timemodel.tcn.cstnImpl;

import java.net.URI;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.model.processmodel.Instance;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.defaultimplementation.DefaultTemporalTemplate;
import de.aristaflow.adept2.model.timemodel.InstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.TimeModelOperations;
import de.aristaflow.adept2.model.timemodel.TimeModelSolver;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.TimeModelTools;
import de.aristaflow.adept2.model.timemodel.tcn.CSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTN;

public class CSTNTimeModelFactory implements TimeModelFactory
{

  private final Configuration configuration;

  public CSTNTimeModelFactory()
  {
    this.configuration = null; // TODO
  }
  
  public CSTNTimeModelFactory(Configuration configuration)
  {
    this.configuration = configuration;    
  }
  
  public TemplateTimeModel createTimeModelForTemplate(TemporalTemplate template)
  {
    CSTNTemplateTransformation trafo = new CSTNTemplateTransformation(this.configuration, template);
    CSTN cstn = trafo.transform();
    
    Map<Integer, List<URI>> updateTriggers = TimeModelTools.createUpdateTriggerMapping(template);
    
    return new CSTNTemplateTimeModel(template, cstn, trafo.getNodeMapping(), trafo.getProcessCreationTimePoint(),
                                     trafo.getProcessStartTimePoint(), trafo.getProcessEndTimePoint(), updateTriggers);
  }

  

  public InstanceTimeModel createTimeModelForInstance(Instance instance, InstanceTimeContext timeContext,
      TemplateTimeModel templateTimeModel)
  {
    CSTNTemplateTimeModel timeModel = (CSTNTemplateTimeModel) templateTimeModel;

    UUID instanceID = instance != null ? instance.getID() : null;

    assert instance == null || instance.getTemplate().getID() == templateTimeModel.getTemplateID();

    return new CSTNInstanceTimeModel(instanceID, templateTimeModel.getTemplateID(),
                                     (ChangeableCSTN) timeModel.getBaseTimeModel(),
                                     timeModel.getNodeMapping(), timeModel.getProcessCreationTimePoint(),
                                     timeModel.getProcessStartTimePoint(), timeModel.getProcessEndTimePoint(),
                                     timeModel.getUpdateTriggers());
  }



  public TimeModelOperations getTimeModelOperations()
  {
    return new DefaultTimeModelOperations();
  }

  public void setCreateChangeable(boolean changeable)
  {
    // Not necessary, returned elements are always changeable....
    // throw new NotImplementedException();
  }

  public TemporalTemplate createTemporalTemplate(Template template)
  {
    return new DefaultTemporalTemplate(template);
  }

  public TimeModelSolver getTimeModelSolver(Class<? extends TimeModel> clazz)
  {
    return new DCCSTNTemplateTimeModelSolver();
  }

}
