package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;


import java.util.ArrayList;
import java.util.List;

import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.AbstractCSTNProjection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.STN;

public class CSTNU2CSTNProjection extends AbstractCSTNProjection implements STN<LabeledTimePoint, LabeledEdge>, Cloneable, ChangeableCSTN
{
  

  protected final boolean contingentValues;

  public CSTNU2CSTNProjection(ChangeableCSTNU cstnu, boolean propagate, boolean useContingentValues)
  {
    super(cstnu, propagate);
    this.contingentValues = useContingentValues;
  }


  @Override
  protected LabeledEdge transformEdge(LabeledTimePoint from, LabeledTimePoint to, LabeledEdge edge)
  {
    // if the edge contains a contingent constraint we need to process it.
    if (edge.isContingent()) 
    {
      edge = convertToNormalEdge(edge);
      return edge;
    }
    
    return edge;
  }



  private LabeledEdge convertToNormalEdge(LabeledEdge edge)
  {
    List<LabeledValue> values = new ArrayList<LabeledValue>();
    for (LabeledValue l : edge.getValue().getLabeledValues())
    {
      if (l instanceof ContingentLabeledValue)
      {
        ContingentLabeledValue cl = (ContingentLabeledValue) l;
        
        LabeledValue label;
        if (contingentValues)
        {
          if (cl.getTrigger() == edge.getTargetNode() || cl.getTrigger() == edge.getSourceNode())
          {
            // If it's a contingent duration use the "normal" value which has been initialized appropriately and has never been modified
            label = new DefaultLabeledValue(l.getLabel(), cl.getValue());
          }
          else
          {
            // Otherwise use the contingent value
            // LOW we could also use the min value of contingent and non-contingent here
            label = new DefaultLabeledValue(l.getLabel(), cl.getContingentValue());
          }
        }
        else
        {
          label = new DefaultLabeledValue(l.getLabel(), cl.getValue());
        }
        
        values.add(label);
      }
      else
      {
        values.add(l); // No processing required
      }            
    }
    LabeledEdge labeledEdge = new DefaultLabeledEdge(edge.getSourceNode(), edge.getTargetNode(), new DefaultLabeledValueSet(values), edge.isControlEdge(), edge.isImplicit(), edge.getFlowDirection());
    return labeledEdge;
  }

  @Override
  public List<ChoiceVariable> getObservations()
  {
    return getBaseTCN().getObservations();
  }


  public LabeledTimePoint insertTimePoint(LabeledTimePoint timePoint)
  {
    throw new UnsupportedOperationException();
  }


  public void removeTimePoint(LabeledTimePoint timePoint)
  {
    throw new UnsupportedOperationException();
  }



  
  
}
