package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;


import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.STN;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.Projection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.STNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.SimpleSTN;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;

public class CSTNU2STNProjection implements STN<LabeledTimePoint, STNEdge>, Projection<LabeledTimePoint, LabeledEdge, STNEdge>, Cloneable
{
  protected CSTNU cstnu;

  public CSTNU2STNProjection(CSTNU cstnu)
  {
    this.cstnu = cstnu;
  }

  @Override
  public STNEdge getEdge(LabeledTimePoint from, LabeledTimePoint to)
  {
    LabeledEdge labeledEdge = cstnu.getEdge(from, to);
    
    TimeDistance resultEdgeValue = null;
    
    if (labeledEdge != null)
    {
      Collection<LabeledValue> values = labeledEdge.getValue().getLabeledValues();
      for (LabeledValue vx : values)
      {
        if (vx instanceof ContingentLabeledValue)
        {
          ContingentLabeledValue v = (ContingentLabeledValue) vx;
          if (v.getTrigger() == to  || v.getTrigger() == from)
          {
            // If it is a contingent duration edge directions are reversed so use the other one.
            LabeledEdge opDir = cstnu.getEdge(to, from);
            
            if (opDir != null)
            {
              Collection<LabeledValue> opValues = opDir.getValue().getLabeledValues();
              
              for (LabeledValue ovx : opValues)
              {
                if (ovx instanceof ContingentLabeledValue)
                {
                  ContingentLabeledValue ov = (ContingentLabeledValue) ovx;
                  if (ov.getTrigger() == v.getTrigger())
                  {
                    resultEdgeValue = resultEdgeValue == null ? ov.getContingentValue().neg() : min(resultEdgeValue, ov.getContingentValue().neg());
                  }
                }
              }
            }
          } else
          {
            // If the edge has a trigger it may be started after either the trigger event or the duration has expired
            STNEdge triggerEdge = getEdge(from, v.getTrigger());
            
            TimeDistance min = triggerEdge == null ? v.getContingentValue() : min(v.getContingentValue(), triggerEdge.getValue().getValue());
            
            resultEdgeValue = resultEdgeValue == null ? min : min(resultEdgeValue, min);
          }
        }
        else
        {
          // Just a normal edge...
          resultEdgeValue = resultEdgeValue == null ? vx.getValue() : min(resultEdgeValue, vx.getValue());
        }
      }
    }
    else
    {
      LabeledEdge opDir = cstnu.getEdge(to, from);
      
      if (opDir != null)
      {
        Collection<LabeledValue> opValues = opDir.getValue().getLabeledValues();
        
        for (LabeledValue ovx : opValues)
        {
          if (ovx instanceof ContingentLabeledValue)
          {
            ContingentLabeledValue ov = (ContingentLabeledValue) ovx;
            resultEdgeValue = resultEdgeValue == null ? ov.getContingentValue().neg() : min(resultEdgeValue, ov.getContingentValue().neg());
          }
        }
      }
    }
    
    if (resultEdgeValue == null)
      return null;
    
    return new STNEdge(from, to, new DefaultConstraintValue(resultEdgeValue), 
                       labeledEdge != null ? labeledEdge.isControlEdge() : false,
                       labeledEdge != null ? labeledEdge.isImplicit() : false,                    
                       labeledEdge != null ? labeledEdge.getFlowDirection() : FlowDirection.UNKOWN);
  }
  
  public Collection<STNEdge> getIncomingEdges(LabeledTimePoint to)
  {
    Collection<LabeledTimePoint> timePoints = getTimePoints();
    List<STNEdge> edges = new ArrayList<STNEdge>(timePoints.size());
    for (LabeledTimePoint from : timePoints)
    {
      STNEdge edge = getEdge(from, to);
      if (edge != null)
      {
        edges.add(edge);
      }
    }
    return edges;
  }

  public Collection<STNEdge> getOutgoingEdges(LabeledTimePoint from)
  {
    Collection<LabeledTimePoint> timePoints = getTimePoints();
    List<STNEdge> edges = new ArrayList<STNEdge>(timePoints.size());
    for (LabeledTimePoint to : timePoints)
    {
      STNEdge edge = getEdge(from, to);
      if (edge != null)
      {
        edges.add(edge);
      }
    }
    return edges;
  }


  
  public TimeDistance min(TimeDistance v1, TimeDistance v2)
  {
    if (v1.compareTo(v2) <= 0)
      return v1;
    else
    return v2;
  }
  
  @Override
  public int hashCode()
  {
    return cstnu.hashCode();
  }
  
  @Override
  public LabeledTimePoint getTimeBaseEvent()
  {
    return cstnu.getTimeBaseEvent();
  }
  
  @Override
  public Collection<LabeledTimePoint> getTimePoints()
  {
    return cstnu.getTimePoints();
  }

  @Override
  public Collection<STNEdge> getEdges()
  {
    Collection<STNEdge> edges = new ArrayList<STNEdge>();
    
    for (LabeledTimePoint n1 : cstnu.getTimePoints())
    {
      for (LabeledTimePoint n2 : cstnu.getTimePoints())
      {
        STNEdge edge = getEdge(n1, n2);
        if (edge != null)
          edges.add(edge);
      }
    }
    
    return edges;
  }

  @Override
  public boolean equals(Object obj)
  {
    return cstnu.equals(obj);
  }

  @Override
  public void visit(GraphVisitor<LabeledTimePoint, STNEdge> visitor)
  {
    visitor.discoverGraph(this);
  
  
//    for (NodeMapping<LabeledTimePoint> n : getNodeMapping())
//    {
//      
//      DirectedGraphImplementation<LabeledTimePoint, STNEdge> subGraph = new DirectedGraphImplementation<LabeledTimePoint, STNEdge>(n);
//      subGraph.visit(visitor);
//    }
    
    for (LabeledTimePoint e : getTimePoints())
    {
      visitor.discoverVertex(e);
      visitor.finishVertex(e);
    }
    
    for (LabeledTimePoint ei : getTimePoints())
    {
      for (LabeledTimePoint ej : getTimePoints())
      {
        STNEdge e = getEdge(ei, ej);
        if (e != null)
        {
          visitor.discoverEdge(e.getSourceNode(), 
                               e.getTargetNode(), 
                               e);
          visitor.finishEdge(e.getSourceNode(), 
                             e.getTargetNode(), 
                             e);
        }
      }
    }
  
    visitor.finishGraph(this);
  }

//  @Override
//  public void setProcessBounds(LabeledNode processCreationNode, LabeledNode processStartNode, LabeledNode processEndNode)
//  {
//    cstnu.setProcessBounds(processCreationNode, processStartNode, processEndNode);
//  }

  @Override
  public String toString()
  {
    return cstnu.toString();
  }

  @Override
  public TCN<LabeledTimePoint, LabeledEdge> getBaseTCN()
  {
    return cstnu;
  }

  @Override
  public STN<LabeledTimePoint, STNEdge> clone()
  {
    // Nodes are immutable
    Collection<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>(getTimePoints());
    Collection<STNEdge> edges = new ArrayList<STNEdge>();
    
    for (LabeledTimePoint n1 : nodes)
    {
      for (LabeledTimePoint n2 : nodes)
      {
        STNEdge edge = getEdge(n1, n2);
        if (edge != null)
          edges.add(edge.clone());
      }
    }
    
    // NodeMappings are immutable
    return new SimpleSTN(nodes, edges);
  }

  @Override
  public boolean isConsistent()
  {
    return this.cstnu.isConsistent();
  }


  public Collection<LabeledTimePoint> getVertices()
  {
    return getTimePoints();
  }

  @Override
  public boolean contains(LabeledTimePoint timePoint)
  {
    return getTimePoints().contains(timePoint);
  }
  
  
}
