package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;
import java.util.Queue;
import java.util.logging.Level;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.ChangeSet;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Path;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Triangle;
import de.aristaflow.adept2.model.timemodel.tcn.CSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableTCN;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.SimpleConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.TCNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.AbstractCSTNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.LabelBasedPC2Solver;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.model.timemodel.tools.DuplicateFreeQueue;


/**
 * 
 * @author Andreas
 *
 */
/*
 * Use regular expression "Processing.*\r\n.*\r\n Ignoring.*" with "." not matching line break to filter unnecessary lines in log...
 */

public class CSTNUDCSolver extends AbstractCSTNSolver implements TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge>
{

  protected int updateCount;
  protected CSTNUAllMaxProjection allMax;
  protected Collection<LabeledEdge> changedEdges;
  
  public static boolean SIMPLIFY = true;
  
  
  public CSTNUDCSolver()
  {
    super();
  }
  
  public CSTNUDCSolver(Level debugLevel)
  {
    super(debugLevel);
  }
  
  @Override
  public <T extends ChangeableTCN<LabeledTimePoint, LabeledEdge>> T solve(T tcn) throws TCNInconsistencyException, InterruptedException
  {
    return solve(tcn, null);
  }
  
  @SuppressWarnings("unchecked")
  public <T extends ChangeableTCN<LabeledTimePoint, LabeledEdge>> T solve(
      T tcn,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet)
      throws TCNInconsistencyException, InterruptedException
  {
    if (tcn instanceof ChangeableCSTNU)
    {
      return (T)solve((ChangeableCSTNU)tcn, changeSet);
    }
    else
    {
      throw new IllegalArgumentException();
    }
  }
  
  public ChangeableCSTNU solve(ChangeableCSTNU cstnu) throws TCNInconsistencyException, InterruptedException
  {
    return solve(cstnu, null);
  }
  
  /**
   * Morris er al.
   * 
   * <p><code>
   * Boolean procedure determineDC()<br/>
   *  loop from 1 to Cutoff Bound do<br/>
   *    if AllMax projection inconsistent<br/>
   *      return false;<br/>
   *    Perform needed No-Case Reductions;<br/>
   *    Perform needed Upper-Case Reductions;<br/>
   *    Perform any Cross-Case Reductions;<br/>
   *    Perform any Lower-Case Reductions;<br/>
   *    Perform any Label Removal Reductions;<br/>
   *    if no reductions were found above<br/>
   *      return true;
   * </code></p>
   * 
   * @param cstnu
   * @throws TCNInconsistencyException 
   * @throws InterruptedException 
   */
  public ChangeableCSTNU solve(ChangeableCSTNU model,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet) throws TCNInconsistencyException, InterruptedException
  {   
    allMax = null;
    updateCount = 0;
    changedEdges = null;

    if (model.isControllable())
      return model;
    
    ChangeableCSTNU cstnu = preProcessing(model);
    
    internalSolveComponent(cstnu, changeSet);

    ChangeableCSTNU result = postProcessing(cstnu);
    
    result.setControllable();
    return result;
  }
  
  protected ChangeableCSTNU postProcessing(ChangeableCSTNU cstnu)
  {
    // Because we used a reduction projection the all max projection is no
    // longer valid (i.e., it's based of the reduced model)
//    allMax = null;
    
//    uncollapseRigidComponents(cstnu, changeSet, nodes, collapseRigidComponents);
    return cstnu;
  }

  protected ChangeableCSTNU preProcessing(ChangeableCSTNU model)
  {
    
//  Set<LabeledNode> nodes = new HashSet<LabeledNode>(model.getVertices());
//  
//  // Find any rigid components (if any) and collapse them
//  Collection<Set<LabeledNode>> rigidComponents = TCNAlgorithms.findRigidComponents(cstnu);
//  Map<LabeledNode, Set<LabeledNode>> collapseRigidComponents;
//  if (rigidComponents.size() > 0)
//  {
//    try
//    {
//      collapseRigidComponents = collapseRigidComponents(cstnu, changeSet, nodes, rigidComponents);
//    }
//    catch (IllegalStateException e)
//    {
////      debug(Level.SEVERE, "Inconsistent: %s -> %s: %s", newEdge.getSourceNode(), newEdge.getTargetNode(), newEdge);
////      throw new TCNInconsistencyException(String.format("Inconsistent: %s -> %s: %s vs. %s", newEdge.getSourceNode(),
////        newEdge.getTargetNode(), newEdge, cstnu.getEdge(newEdge.getTargetNode(), newEdge.getSourceNode())),
////        newEdge.getSourceNode(), newEdge.getTargetNode(), cstnu);
//      throw new TCNInconsistencyException("Inconsistency in rigid components", model);
//    }
//  }
//  else
//  {
//    collapseRigidComponents = Collections.emptyMap();
//  }
//  
//  //CSTN reducedModel = new ReductionProjection<LabeledEdge>(cstnu, nodes);
//  
//  ChangeableCSTNU reducedModel = new CSTNUReductionProjection(model, nodes);
//  
//  TCNTestTools.printTCN(reducedModel, this.getClass(), "ReducedModel");
    
    return model;
  }

  @Override
  protected void solveComponent(ChangeableCSTN model,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet)
      throws TCNInconsistencyException, InterruptedException
  {
    internalSolveComponent((ChangeableCSTNU)model, changeSet);
    
  }

  protected void internalSolveComponent(ChangeableCSTNU model,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet) throws TCNInconsistencyException, InterruptedException
  {
    Collection<LabeledTimePoint> nodes = model.getTimePoints();
    
    long cutoff;
    {
      // Determine the CutOff Bound
      BigInteger observationCount = BigInteger.valueOf(Math.max(model.getObservations().size(), 1));
      BigInteger nodeCount = BigInteger.valueOf(nodes.size());
      BigInteger contingentNodeCount = BigInteger.valueOf(nodes.size());
      //This is a bad approximation...
      BigInteger cutoffBound = observationCount.multiply(nodeCount.pow(2).add(nodeCount.multiply(contingentNodeCount)).add(contingentNodeCount)).multiply(nodeCount.pow(4));
      
      
      if (cutoffBound.compareTo(BigInteger.valueOf(Long.MAX_VALUE)) == 1)
      {
        // Overflow of arithmetic calculation!
        assert debug(Level.SEVERE,
              "Overflow detected when calculating cuttOff Bound."
                  + "Limiting cutOff Bound to Long.MAX_VALUE."
                  + "Note: The algorithm might not give accurate results in case cutoffBound is reached.");
        cutoff = Long.MAX_VALUE;
      }
      else
      {
        cutoff = cutoffBound.longValue();
      }
    }
    
    
 
    
    // Create the solver used for checking consistency of the allmax projection.
    LabelBasedPC2Solver cstnsolver = new LabelBasedPC2Solver();
    
    // Create the AllMax Projection of the CSTNU. This will be reused in each iteration. cf. updateContingentConstraint()
    allMax = new CSTNUAllMaxProjection(model);
       
    // Create the queue holding the edges to be updated
    Queue<Path<LabeledTimePoint>> currentQueue;
    int nodeCount = nodes.size();
    Queue<Path<LabeledTimePoint>> nextQueue = new DuplicateFreeQueue<Path<LabeledTimePoint>>(nodeCount*nodeCount*2);


    // Initially consider any edges which have been modified since the cstnu 
    // last has been marked as minimal
    changedEdges = model.getChanges();

    if (changedEdges == null) // null indicates we don't know about changed edges...
    {
      getInitialPaths(nodes, nextQueue);
    }
    else
    {

      for (LabeledEdge changedEdge : changedEdges)
      {
        nextQueue.add(new Path<LabeledTimePoint>(changedEdge.getSourceNode(), changedEdge.getTargetNode()));
//        addRelatedPaths(nodes, nextQueue, changedEdge.getSourceNode(), changedEdge.getTargetNode());
      }
    }
    
    // Repeat until cutoff bound reached
    long iter;
    for (iter = 0; iter < cutoff; iter++)
    {
      if (Thread.currentThread().isInterrupted())
        throw new InterruptedException();
      
      assert debug(Level.FINE, "Iteration %s", Long.valueOf(iter));
      updateCount = 0;
      
      // Check the allMax projection for consistency. Use the changed edges from
      // the last iteration as starting point! Only these have been changed
      // during the last run of DCSolver. Note that changedEdges may initially
      // be null (i.e., during the first run)
//      cstnsolver.solve(allMax, null);
      
      if (Thread.currentThread().isInterrupted())
        throw new InterruptedException();

      
      currentQueue = nextQueue;
      nextQueue = new DuplicateFreeQueue<Path<LabeledTimePoint>>(nodeCount*nodeCount*2);
      
      
      // Initialize Data Structure holding all changes during this run.
      changedEdges = new ArrayList<LabeledEdge>();
      
      // Variable determining if there are any changes
      boolean changes = false;
           
      // Repeat for each Triple
      while (!currentQueue.isEmpty())
      {
        if (Thread.currentThread().isInterrupted())
          throw new InterruptedException();
        
        final Path<LabeledTimePoint> triangle = currentQueue.remove();
        
        final LabeledTimePoint i = triangle.getFirst();
        final LabeledTimePoint j = triangle.getSecond();
//        final LabeledTimePoint k = triangle.getSecond();
        

        for (LabeledTimePoint k : nodes)
        {
          
          if (i.equals(k) || j.equals(k))
            continue;
        
          if (processPath(model, changeSet, new Path<LabeledTimePoint>(i, k), j))
          {
            changes = true;    
            // e_ik has been modified so re-evaluate all related paths 
            //addRelatedPaths(nodes, nextQueue, i, k, currentQueue);
            Path<LabeledTimePoint> e = new Path<LabeledTimePoint>(i, k);
            if (!currentQueue.contains(e))
            {
              nextQueue.add(e);
            }
          }
          
          if (processPath(model, changeSet, new Path<LabeledTimePoint>(k, j), i))
          {
            changes = true;    
            // e_ik has been modified so re-evaluate all related paths 
            //addRelatedPaths(nodes, nextQueue, i, k, currentQueue);
            Path<LabeledTimePoint> e = new Path<LabeledTimePoint>(k, j);
            if (!currentQueue.contains(e))
            {
              nextQueue.add(e);
            }
          }
        }
      }
      
      // If there are no more changes we are finished and the CSTNU is controllable
      if (!changes)
      {
        assert debug(Level.FINE, "Controllable!");
        
        assert nextQueue.isEmpty();
        
        break;
      }
      
      assert debug(Level.FINE, "%s Updates", Integer.valueOf(updateCount));
      
      // Otherwise restart the loop
    }
    

    // Check if we left the loop due to the Cutoff Bound 
    if (iter >= cutoff)
    {
      assert debug(Level.FINE, "Cutoff bound reached");
      throw new TCNInconsistencyException("CutoffBound reached", model);
    }
  }

  @Override
  protected boolean processPath(ChangeableCSTN model,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet,
      Path<LabeledTimePoint> p, final LabeledTimePoint via) throws TCNInconsistencyException
  {
    LabeledTimePoint from = p.getFrom();
    LabeledTimePoint to = p.getTo();
    //    // Get all three edges of the triple
    final LabeledEdge e_ij = model.getEdge(from, via);
    final LabeledEdge e_jk = model.getEdge(via, to);
//         
//    // Check if the triple has enough "real" edges, i.e., at least the two main edges must be non-null
//    if (e_ij == null || e_jk == null)
//      return false;
 
    
    
    final LabeledEdge e_ik = model.getEdge(from, to);
    assert debug(Level.FINE, "Processing (%s, %s) with %s ~ %s + %s", from, to, e_ik, e_ij, e_jk);
        
    // Apply Reduction Rules which generate new edges 
    Map<Label, LabeledValue> newValues = applyEdgeGenerationRules(model, from, via, to);
    assert debug(Level.FINE, "   Generation Rules generated %s", newValues);
    
    // Do we have any new edges?
    if (newValues == null || newValues.isEmpty())
    {
      // This can only happen in case either all constraints belong to different branches or at least
      // one of them is un-set (i.e., infinity)
      assert debug(Level.FINE, " Ignoring\t %s->%s (because no labels have been generated)", from, to);
      return false;
    }
      
      
    // IMPROVE at which point does it make sense to apply the simplification rule?
    // Apply simplification Rules to reduce number of edges
//    if (SIMPLIFY)
//      applySimplificationRules(newValues);
    
    // Get already existing edges
    LabeledValueSet ls_DA = null;
    if (e_ik != null)
    {
      ls_DA = e_ik.getValue();
    }
    
    // Merge existing edges with new edges
    try
    {
      newValues = intersect(ls_DA, newValues);
    }
    catch (AssertionError e)
    {
      System.err.println(String.format(" Error with %s->%s", from, to));
      throw e;
    }
    assert debug(Level.FINE, "   Intersection resulted in %s", newValues);
    
    // Apply label removal rules to remove any unnecessary edges
    newValues = applyLabelRemovalRules(model, from, via, to, newValues);
    applyPropositionModificationRules(model, from, via, to, newValues);
    assert debug(Level.FINE, "   Label Removal Rules resulted in %s", newValues);
    
    // Apply simplification Rules to reduce number of edges
    if (SIMPLIFY)
      applySimplificationRules(newValues);
    
    assert newValues != null;

    LabeledValueSet newValueSet = new DefaultLabeledValueSet(newValues.values());
    
    if (ls_DA != null && ls_DA.equals(newValueSet))
    {
      assert debug(Level.FINE, " Ignoring\t %s to %s (because no changes)", e_ik, newValueSet);

      return false;
    }
    else
    {
      // Create the updated edge
      LabeledEdge labeledEdge = createEdge(from,
                                           to,
                                           newValueSet,
                                           e_ik != null && e_ik.isControlEdge(),
                                           e_ik == null || e_ik.isImplicit(),
                                           e_ik != null ? e_ik.getFlowDirection() : FlowDirection.UNKOWN);
      
      assert debug(Level.FINE, " Updating\t %s to %s (because of %s and %s)", e_ik, labeledEdge, e_ij, e_jk);
  
      // Update the CSTNU etc.
      updateConstraint(changeSet, model, labeledEdge, new Triangle<LabeledTimePoint>(new Path<LabeledTimePoint>(from, to), via));
      
      return true;
    }
  }
    

  /**
   * This considers the triangle Q -> S -> T, where newLabelSet belongs to Q -> T
   * 
   * @param cstnu
   * @param Q
   * @param S
   * @param T
   * @param newLabel
   * @return
   * @throws TCNInconsistencyException
   */
  protected Map<Label, LabeledValue> applyLabelRemovalRules(CSTN cstnu, LabeledTimePoint Q, LabeledTimePoint S, LabeledTimePoint T,
      Map<Label, LabeledValue> newLabel) throws TCNInconsistencyException
  {
    // Basic STN label Removal Rule, cf. Combi13
    labelRemovalReduction(cstnu, Q, T, newLabel);
    return newLabel;
  }

  /**
   * 
   * S -R:v-> T becomes S -v-> T if v >= -x, where x is the lower bound for the contingent link from T to R
   * 
   * @param cstnu
   * @param ST
   * @param ST
   * @param labeledValues 
   * @param changeSet 
   * @return
   * @throws TCNInconsistencyException 
   */
  protected void labelRemovalReduction(CSTN cstnu, LabeledTimePoint S, LabeledTimePoint T,
      Map<Label, LabeledValue> labeledValues) throws TCNInconsistencyException
  {
    assert S != null;
    assert T != null;
    
    List<LabeledValue> newValues = new ArrayList<LabeledValue>();
    

    
    for (Iterator<LabeledValue> iterator = labeledValues.values().iterator(); iterator.hasNext();)
    {
      LabeledValue l_ST = iterator.next();
      // S -R:v-> T
      if (!(l_ST instanceof ContingentLabeledValue && ConstraintCase.UPPERCASE.equals(((ContingentLabeledValue)l_ST).getCase())))
      {
        continue;
      }
      
      ContingentLabeledValue cl_ST = (ContingentLabeledValue)l_ST;
      
      // Never apply this rule if this is a contingent duration constraint!!!
      if (cl_ST.getTrigger() == S || cl_ST.getTrigger() == T) 
      {
        continue;
      }
       
      ContingentLabeledValue cl_TR = getContingentDurationConstraint(cstnu, T, cl_ST.getTrigger());
      
      if (cl_TR == null)
      {
        continue;
      }
        
      
      assert cl_TR != null;
      
//      // T -b:x-> R
//      if (!(ConstraintCase.LOWERCASE.equals(cl_TR.getCase()))) // Can this happen?
//      {
//        DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, cl_ST);
//        continue;
//      }
      
      assert ConstraintCase.LOWERCASE.equals(cl_TR.getCase());
      
      TimeDistance v = cl_ST.getContingentValue();
      TimeDistance x = cl_TR.getContingentValue();
      
      if (v.compareTo(x.neg()) < 0) // z >= -x
      {
        //DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, cl_ST);
        continue;
      }
     
      //R <-r:x- T T <-R:z S adds T <-z- S if z>=-x
      //R <-r:x- T
//      assert cl_TR instanceof ContingentLabeledValue;
      assert cl_TR.getCase().equals(ConstraintCase.LOWERCASE);
      // T <-R:z S
      assert cl_ST.getCase().equals(ConstraintCase.UPPERCASE);
      // z >= -x
      
      
      
      // Calculate the new Label (same trigger and same case as CA)
      // S -z-> T
      // Label is downgraded to a non-contingent label so update is required
      
      LabeledValue newValue = new DefaultLabeledValue(cl_ST.getLabel(), v);
      
      assert debug(Level.FINER, "\tLabelRemoval:\t Replacing %s by %s", cl_ST, newValue);
      iterator.remove();
      newValues.add(newValue);
//      DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, newValue);
    }
          
    for (LabeledValue newValue : newValues)
    {
      addOrReplaceIfNecessary(labeledValues, newValue);
    }
    
//    return labeledValues;
  }

  /**
   * 
   * A <-x- C <-c:y- D adds A <-x+y- D if x<=0
   * @param cstnu
   * @param QS
   * @param ST
   * @param QT
   * @param changeSet 
   * 
   * @return
   * @throws TCNInconsistencyException 
   */
  protected LabeledValue lowerCaseReduction(ContingentLabeledValue cl_QS,
      LabeledValue l_ST, LabeledTimePoint S)
  {
    //S -x-> T
    if (l_ST instanceof ContingentLabeledValue)
      return null;
    
   // Q -c:y-> S
    if (!(ConstraintCase.LOWERCASE.equals(cl_QS.getCase())))
      return null;

          
          
    TimeDistance v = l_ST.getValue();
    long cv = v.sign();
    if (!((cv < 0) || (cv == 0))) // x <= 0
      return null;
    
    if (!satisfiable(cl_QS.getLabel(), l_ST.getLabel()))
      return null;
          
          
    //A <-x- C <-c:y- D adds A <-x+y- D if x<=0
    //A <-x- C
    assert !(l_ST instanceof ContingentLabeledValue);
    // C <-c:y- D
    assert ConstraintCase.LOWERCASE.equals(cl_QS.getCase());
    // x <= 0
    //assert cl_CA.getValue() <= 0;
    // "C = c"
    assert cl_QS.getTrigger() == S;
          
          
    
    
    // Calculate the new Label
    // A <-x+y- D
    TimeDistance value = l_ST.getValue().add(cl_QS.getContingentValue());
    Label proposition = mergeLabels(l_ST.getLabel(), cl_QS.getLabel());
    //CSTNULabel newLabel = new CSTNULabel(proposition, null, null, value);
    
    LabeledValue newLabel = new DefaultLabeledValue(proposition, value);
    
    assert debug(Level.FINER, "\tLowerCase:\t %s + %s -> %s", cl_QS, l_ST, newLabel);
    
    return newLabel;
  }

  /**
   * Applies all edge generation rules to the triangle Q -> S -> T, where the generated constraints 
   * belong to edge Q -> T
   * @param cstnu
   * @param QS
   * @param ST
   * @return
   */
  
  protected Map<Label, LabeledValue> applyEdgeGenerationRules(ChangeableCSTN cstnu, LabeledTimePoint Q, LabeledTimePoint S, LabeledTimePoint T)
  {
    // "S = S"
    
    LabeledEdge QS = cstnu.getEdge(Q, S);
    
    if (QS == null)
      return null;
    
    LabeledEdge SQ = cstnu.getEdge(S, Q);
    
    LabeledEdge ST = cstnu.getEdge(S, T);
    
    Map<Label, LabeledValue> labels = new HashMap<Label, LabeledValue>();
    
 
    if (ST != null)
    {  
      for (LabeledValue l_QS : QS.getValue().getLabeledValues())
      {
      
        ContingentLabeledValue cl_QS = null;
        if (l_QS instanceof ContingentLabeledValue)
        {
          cl_QS = (ContingentLabeledValue) l_QS;
        }
          
        for (LabeledValue l_ST : ST.getValue().getLabeledValues())
        {
          ContingentLabeledValue cl_ST = null;
          if (l_ST instanceof ContingentLabeledValue)
          {
            cl_ST = (ContingentLabeledValue) l_ST;
          }
  
          {
              LabeledValue noCase = noCaseReduction(l_QS, l_ST,
                                                         ST.getSourceNode());  
              if (noCase != null)
              {
                addOrReplaceIfNecessary(Q, T, labels, noCase);
              }
          }
          
          if (cl_QS != null && cl_ST != null) // Both constraints need to be contingent
          {
            LabeledValue crossCase = crossCaseReduction(cl_QS, cl_ST,
                                                             ST.getSourceNode());
  
            if (crossCase != null)
            {
              addOrReplaceIfNecessary(Q, T, labels, crossCase);
            }
          }
  
          if (cl_ST != null) // ST needs to be contingent
          {
            LabeledValue upperCase = upperCaseReduction(l_QS, cl_ST,
                                                             ST.getSourceNode());
  
            if (upperCase != null)
            {
              addOrReplaceIfNecessary(Q, T, labels, upperCase);
            }
          }
          
          if (cl_QS != null) // QS needs to be contingent
          {
            LabeledValue lowerCase = lowerCaseReduction(cl_QS, l_ST,
                                                             ST.getSourceNode());
  
            if (lowerCase != null)
            {
              addOrReplaceIfNecessary(Q, T, labels, lowerCase);
            }
          }
        }
      }      
    }
    
    if (SQ != null)
    {
      for (LabeledValue l_QS : QS.getValue().getLabeledValues())
      {
        for (LabeledValue l_SQ : SQ.getValue().getLabeledValues())
        {
          LabeledValue observationCase = observationCaseReduction(l_SQ, l_QS, T);
          
          if (observationCase != null)
          {
            addOrReplaceIfNecessary(Q, T, labels, observationCase);
          }
        }
      
      }
    }
  
    if (labels.size() > 0)
      return  labels;
    return null;
  }

  /** 
   * This is mainly to assert that the edge from <code>from</code> to <code>to</code> is indeed a contingent duration constraint.
   * @param cstnu
   * @param from
   * @param to
   * @return
   */
  protected ContingentLabeledValue getContingentDurationConstraint(CSTN cstnu,
      LabeledTimePoint from, LabeledTimePoint to)
  {
    LabeledEdge edge = cstnu.getEdge(from,
                                   to);
    
    ContingentLabeledValue res = null;
    
    if (edge != null)
    {
      res = getContingentLabeledValue(edge);     
      
      
      assert res != null || debug(Level.WARNING, "Case missing for contingent constraint '%s->%s'!", from, to);
      assert res == null || res.getTrigger() == edge.getSourceNode() || res.getTrigger() == edge.getTargetNode();
    }
    
    return res;
  }
  
  protected ContingentLabeledValue getContingentLabeledValue(LabeledEdge edge)
  {
    Collection<LabeledValue> labeledValues = edge.getValue().getLabeledValues();
    assert labeledValues.size() == 1;
    LabeledValue labeledValue = labeledValues.iterator().next();
//    assert labeledValue instanceof ContingentLabeledValue;
    if (labeledValue instanceof ContingentLabeledValue)
      return (ContingentLabeledValue) labeledValue;
    else
    {
      return null;
    }
  }
  
  protected LabeledValue crossCaseReduction(ContingentLabeledValue cl_QS,
      ContingentLabeledValue cl_ST, LabeledTimePoint S)
  {
    LabeledValue newLabel;
    //A <-B:x- C
    if (!(ConstraintCase.UPPERCASE.equals(cl_ST.getCase())))
      return null;
        
    // C <-c:y- D
    if (! (ConstraintCase.LOWERCASE.equals(cl_QS.getCase())) )
      return null;
          
    // x<=0, B != C
    if (cl_ST.getContingentValue().sign() > 0) 
      return null;
    
    // B != C
    
    if (cl_ST.getTrigger() == S)
      return null;
    
    if (!satisfiable(cl_QS.getLabel(), cl_ST.getLabel()))
      return null;
    
    
    //A <-B:x- C C <-c:y- D adds A<-B:(x+y)-D If x <= 0, B != C
    //A <-B:x- C
    assert cl_ST.getCase().equals(ConstraintCase.UPPERCASE);
    // C <-c:y- D
    assert cl_QS.getCase().equals(ConstraintCase.LOWERCASE);
    // x <= 0
    //assert cl_CA.getValue() <= 0;

    // B != C
    assert cl_ST.getTrigger() != S;
    // "C = c"
    assert cl_QS.getTrigger() == S;
    
    
          
    // Calculate the new Label (same trigger and same case as CA)
    // A<-B:(x+y)-D
    TimeDistance value = cl_ST.getContingentValue().add(cl_QS.getContingentValue());
    Label proposition = mergeLabels(cl_ST.getLabel(), cl_QS.getLabel());
    
    newLabel = createContingentConstraint(proposition, value, cl_ST.getTrigger());
    
    assert debug(Level.FINER, "\tCrossCase:\t %s + %s -> %s", cl_QS, cl_ST, newLabel);
    
    return newLabel;
  }

  /**
   * 
   * Q -u-> S -R:v-> T
   * ==> Q -R:u+v-> T
   * @param cstnu
   * @param QS
   * @param ST
   * @param QT
   * @param changeSet 
   * 
   * @return
   * @throws TCNInconsistencyException 
   */
  protected LabeledValue upperCaseReduction(LabeledValue l_QS,
      ContingentLabeledValue cl_ST, LabeledTimePoint S)
  {
    // Q -u-> S
//    if (l_QS instanceof ContingentLabeledValue) // QS might be contingent as well though we are only interested in the non contingent part
//      return null;
    
    //S -R:v-> T
    if (!(ConstraintCase.UPPERCASE.equals(cl_ST.getCase())))
      return null;
    
    if (!satisfiable(l_QS.getLabel(), cl_ST.getLabel()))
      return null;
    
    // Q -u-> S -R:v-> T
    
    // Q -u-> S
    //assert !(l_QS instanceof ContingentLabeledValue); // QS might be contingent as well though we are only interested in the non contingent part
    // S -R:v-> T
    assert cl_ST.getCase().equals(ConstraintCase.UPPERCASE);
    

    // Calculate the new Label (same trigger and same case as ST)
    // Q -R:u+v-> T
    TimeDistance value = l_QS.getValue().add(cl_ST.getContingentValue());
    

    
    Label proposition = mergeLabels(l_QS.getLabel(), cl_ST.getLabel());
    LabeledValue newLabel;
    LabeledTimePoint trigger = cl_ST.getTrigger();
    
    newLabel = createContingentConstraint(proposition, value, trigger);
    
    assert debug(Level.FINER, "\tUpperCase:\t %s + %s -> %s", l_QS, cl_ST, newLabel);
    return newLabel;
  }

  protected LabeledValue createContingentConstraint(Label proposition, TimeDistance value,
      LabeledTimePoint trigger)
  {
    LabeledValue newLabel = new DefaultContingentLabeledValue(proposition, 
                                                  ConstraintCase.UPPERCASE, 
                                                  DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, 
                                                  trigger, value);
    
    return newLabel;
  }

  /**
   * 
   * A <-x- C <-y- D adds A <-x+y- D
   * @param cstnu
   * @param QS
   * @param ST
   * @param QT 
   * @param changeSet 
   * 
   * @return
   * @throws TCNInconsistencyException 
   */
  protected LabeledValue noCaseReduction(LabeledValue l_QS,
      LabeledValue l_ST, LabeledTimePoint S)
  {   
    if (!satisfiable(l_QS.getLabel(), l_ST.getLabel()))
      return null;
    
    if (l_QS.getValue().isInfinity() || l_ST.getValue().isInfinity())
      return null;
        
    TimeDistance value = l_QS.getValue().add(l_ST.getValue());
    Label proposition = mergeLabels(l_QS.getLabel(), l_ST.getLabel());

    LabeledValue newLabel = new DefaultLabeledValue(proposition, value);
    
    assert debug(Level.FINER, "\tNoCase:\t\t %s + %s -> %s", l_QS, l_ST, newLabel);
   
    return newLabel;
  }


  protected void updateConstraint(
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet,
      ChangeableCSTN cstnu, LabeledEdge newEdge,
      Triangle<LabeledTimePoint> causedBy)
      throws TCNInconsistencyException
  {
    LabeledEdge oldEdge = cstnu.getEdge(newEdge.getSourceNode(), newEdge.getTargetNode());
    
    if (oldEdge != null && oldEdge.isContingentDuration())
    {
      ContingentLabeledValue oldValue = getContingentLabeledValue(oldEdge);

      if (!newEdge.isContingentDuration())
      {
        throw new TCNInconsistencyException(
                                            "Cannot update contingent constraints " + oldEdge + " to " + newEdge,
                                            new Path<LabeledTimePoint>(newEdge.getSourceNode(), newEdge.getTargetNode()),
                                            causedBy.getSecond(), cstnu);
      }

      ContingentLabeledValue newValue = getContingentLabeledValue(newEdge);

      if (!oldValue.getContingentValue().equals(newValue.getContingentValue()))
      {
        throw new TCNInconsistencyException(
                                            "Cannot update contingent constraints " + oldEdge + " to " + newEdge,
                                            new Path<LabeledTimePoint>(newEdge.getSourceNode(), newEdge.getTargetNode()),
                                            causedBy.getSecond(), cstnu);
      }
    }
    else
    { // Adding a contingent duration this way is not allowed
      if (newEdge.isContingentDuration())
        throw new IllegalArgumentException("Adding contingent duration constraints is not allowed!");
    }

    if (changeSet != null)
    {
      LabeledValueSet oldValue = oldEdge != null ? oldEdge.getValue() : null;
      logChange(changeSet, causedBy.getFirst(), causedBy.getSecond(), newEdge, oldValue);
    }
    
    checkConsistency(cstnu, newEdge, causedBy);

    try
    {
      cstnu.updateConstraint(newEdge);
    }
    catch (IllegalArgumentException e) // Should not happen because of previous checks...
    {
      debug(Level.SEVERE, "Inconsistent: %s -> %s: %s", newEdge.getSourceNode(), newEdge.getTargetNode(), newEdge);
      throw new TCNInconsistencyException(String.format("Inconsistent: %s -> %s: %s vs. %s", newEdge.getSourceNode(),
        newEdge.getTargetNode(), newEdge, oldEdge),
        new Path<LabeledTimePoint>(newEdge.getSourceNode(), newEdge.getTargetNode()), causedBy.getSecond(), cstnu);
    }
    
    if (changedEdges != null)
      changedEdges.add(newEdge);
    

    if (allMax != null)
      allMax.removeDelta(newEdge);
    updateCount++;
  }

  
  @Override
  protected boolean isConsistent(SimpleConstraintValue value1,
      SimpleConstraintValue value2)
  {
    boolean consistent = super.isConsistent(value1, value2);
    

    if (consistent && value1 instanceof ContingentLabeledValue)
    {
      ContingentLabeledValue contingent1 = (ContingentLabeledValue) value1;
      
      TimeDistance d1 = contingent1.getContingentValue().neg();
      TimeDistance d2 = value2.getValue();
      
      consistent &= d1.compareTo(d2) <= 0;
    }
    
    if (consistent && value2 instanceof ContingentLabeledValue)
    {
      ContingentLabeledValue contingent2 = (ContingentLabeledValue) value2;
      
      TimeDistance d1 = value1.getValue();
      TimeDistance d2 = contingent2.getContingentValue().neg();
      
      consistent &= d1.compareTo(d2) >= 0;
    }
    
    
    return consistent;
  }

  @Override
  protected LabeledValue updateLabel(LabeledValue labeledValue, Label label)
  {
    LabeledValue newLabeledValue;
    if (labeledValue instanceof ContingentLabeledValue)
    {
      ContingentLabeledValue cl_XY = (ContingentLabeledValue)labeledValue;
      newLabeledValue = new DefaultContingentLabeledValue(label, cl_XY.getCase(), cl_XY.getSTNValue(), cl_XY.getTrigger(), cl_XY.getContingentValue());
    }
    else
    {
      newLabeledValue = super.updateLabel(labeledValue, label);
    }
    return newLabeledValue;
  }
  
  /**
   * Redundant Label Elimination Rule 2
   * 
   * <p>
   * This is a variant of the RLE2 of Combi et al. (2013) pushing values down
   * instead of up. That is, for two labels \alpha p and \alpha -p instead of
   * removing any \alpha labeled value we push the value of either \alpha p or
   * \alpha -p down to \alpha (depending on which one is greater) and then
   * remove the labeled value which has been pushed down (or both if they have
   * the same value).<br/>
   * This ensures that all labels are as short (i.e., generic) as possible. 
   * As as side product calling this method (multiple times) will yield the same result
   * as when applying RLE 3.
   * 
   * 
   * <br/>
   * See Conrad et al. 2011 for details on when a labeled value is dominated by
   * another labeled value and thus redundant.
   * </p>
   * 
   * Complexity of this is roughly O(n^2) in the number of entries of
   * <code>labels</code>
   * 
   * @param labels
   * @return
   */
  protected boolean redundantLabelEliminationRule2(Map<Label, LabeledValue> labels)
  {
    boolean changes = false;
    
    Set<LabeledValue> markForAdd = new HashSet<LabeledValue>();
    Set<Label> markForRemoval = new HashSet<Label>();
    
    List<LabeledValue> list = new ArrayList<LabeledValue>(labels.values());
    
    // First find all choices for each choice variable in each context
    Map<Label, Map<ChoiceVariable, Set<LabeledValue>>> map = new HashMap<Label, Map<ChoiceVariable, Set<LabeledValue>>>();
    for (ListIterator<LabeledValue> iter1 = list.listIterator(); iter1.hasNext(); )
    {
      LabeledValue v1 = iter1.next();
      if (!iter1.hasNext())
        break;
      
      // Each pair needs to be considered only once as singlediff(l1, l2) = singlediff(l2, l1)
      for (ListIterator<LabeledValue> iter2 = list.listIterator(iter1.nextIndex()); iter2.hasNext(); )
      {
        LabeledValue v2 = iter2.next();
        
        ChoiceVariable diff = singlediff(v1.getLabel(), v2.getLabel());
        if (diff != null)
        {
          Label context = v1.getLabel().removeVariable(diff);
          
          Map<ChoiceVariable, Set<LabeledValue>> baseLabels = map.get(context);
          if (baseLabels == null)
          {
            baseLabels = new HashMap<Label.ChoiceVariable, Set<LabeledValue>>();
            map.put(context, baseLabels);
          }
          
          Set<LabeledValue> choices = baseLabels.get(diff);
          if (choices == null)
          {
            choices = new HashSet<LabeledValue>(diff.getOptions()*2);
            baseLabels.put(diff, choices);
          }
          
          choices.add(v1);
          choices.add(v2);
        }
      }
    }
    
    
    for (Entry<Label, Map<ChoiceVariable, Set<LabeledValue>>> m : map.entrySet())
    {
      // Go over all contexts
      Label context = m.getKey();
      Map<ChoiceVariable, Set<LabeledValue>> choices = m.getValue();
      // Check each choice variable if all choices are present.
      for (Entry<ChoiceVariable, Set<LabeledValue>> entry : choices.entrySet())
      {
        // check if for this choice variable all options are present
        if (entry.getKey().getOptions() == entry.getValue().size())
        {
          // if yes, find the max value
          
          TimeDistance max = DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE;
          TimeDistance maxCont = DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE;
          boolean allContingent = true;
          LabeledTimePoint trigger = null;
          for (LabeledValue l : entry.getValue())
          {
            TimeDistance value = l.getValue();
            int compare = value.compareTo(max);
            if (compare > 0)
            {
              max = value;
            }  
            
            if (allContingent && l instanceof ContingentLabeledValue)
            {
              ContingentLabeledValue contingentLabeledValue = (ContingentLabeledValue) l;
              
              if (trigger == null)
              {
                trigger = contingentLabeledValue.getTrigger();
              }
              else
              {
                if (trigger != contingentLabeledValue.getTrigger())
                {
                  allContingent = false;
                  continue;
                }
              }
              
              value = contingentLabeledValue.getContingentValue();
              compare = value.compareTo(maxCont);
              if (compare > 0)
              {
                maxCont = value;
              }  
              
              
            }
            else
            {
              allContingent = false;
            }
            
          }
          
          if (allContingent)
          {
            // TODO contingent should only exist in CSTNU
            markForAdd.add(new DefaultContingentLabeledValue(context, ConstraintCase.UPPERCASE, max, trigger, maxCont));
          }
          else
          {
            // propose to add the max value as new value for the context
            markForAdd.add(new DefaultLabeledValue(context, max));
          }
          
          // remove any value that is equal (or greater) to the max value as it is subsumed by the new value added
          // for the context
          for (LabeledValue l : entry.getValue())
          {
            TimeDistance value = l.getValue();
            int compare = value.compareTo(max);
            if (compare >= 0)
            {
              markForRemoval.add(l.getLabel());
            }
          }
          
        }
      }
    }
    
    for (Label lbl : markForRemoval)
    {
      changes |= labels.remove(lbl) != null;
    }
    
    for (LabeledValue lbl : markForAdd)
    {
      changes |= addOrReplaceIfNecessary(labels, lbl);
    }
    
    
    
    assert labels.size() > 0 : "MarkForAdd: "+markForAdd+ " | MarkForRemoval: "+markForRemoval;
    return changes;
  }
  

  @Override
  public LabeledValue merge(
      LabeledValue newValue, LabeledValue oldValue)
  {
    if (newValue instanceof ContingentLabeledValue)
    {
      if (oldValue instanceof ContingentLabeledValue)
      {
        // Both are contingent
        ContingentLabeledValue clNew = (ContingentLabeledValue) newValue;
        ContingentLabeledValue clOld = (ContingentLabeledValue) oldValue;
        
        
        int compareNonCont = clOld.getValue().compareTo(clNew.getValue());
        int compareCont = clOld.getContingentValue().compareTo(clNew.getContingentValue());
        
        // if one of the values has been decreased
        if (compareCont > 0 || compareNonCont > 0) //old > new
        {
          assert clNew.getCase() == clOld.getCase() : clOld + " vs. "+clNew;
          assert clNew.getTrigger() == clOld.getTrigger() : "Trying to merge contingent constraints with different triggers!";
          
          // Merge or replace
          if (compareCont >= 0 && compareNonCont >= 0) // old >= new
          {
            // Both have either been decreased or not changed so replace
            //labels.put(newValue.getLabel(), newValue);
            return newValue;
          }
          else // old < new
          {
            
            // Merge
            TimeDistance nonContValue = compareNonCont > 0 ? clNew.getValue() : clOld.getValue();
            TimeDistance contValue = compareCont > 0 ? clNew.getContingentValue() : clOld.getContingentValue();
            //labels.put(clNew.getLabel(), createNewConstraint(clNew, nonContValeu, contValue));
            return updateLabeledValue(oldValue, clNew, nonContValue, contValue);
          }
        }
        else
        {
          return oldValue;
        }
      }
      else
      {
        // new is contingent old is non-contingent
        // Merge
        
        ContingentLabeledValue clNew = (ContingentLabeledValue) newValue;
                 
        TimeDistance oldTD = oldValue.getValue();
        TimeDistance newTD = clNew.getValue();
        int compareNonCont = oldTD.compareTo(newTD);

        
        // if one of the values has been decreased
        if (compareNonCont >= 0) // old > new
        {
          return newValue;
        }
        else
        {
          // Merge
          TimeDistance contValue = clNew.getContingentValue();
          TimeDistance nonContValue = oldValue.getValue();
          //labels.put(clNew.getLabel(), createNewConstraint(clNew, nonContValue, contValue));
          return updateLabeledValue(oldValue, clNew, nonContValue, contValue);
        }
      }
    
    }
    else
    {
      if (oldValue instanceof ContingentLabeledValue)
      {
        // new is non-contingent old is contingent
        // Merge
        ContingentLabeledValue clOld = (ContingentLabeledValue) oldValue;
        
        int compareNonCont = clOld.getValue().compareTo(newValue.getValue());
        
        // if one of the values has been decreased
        if (compareNonCont > 0)
        {
          // Merge
          TimeDistance nonContValue = newValue.getValue();
          TimeDistance contValue = clOld.getContingentValue();
//          labels.put(clOld.getLabel(), createNewConstraint(clOld, nonContValue, contValue));
          return updateLabeledValue(oldValue, clOld, nonContValue, contValue);
        }
        else
        {
          return oldValue;
        }
      }
      else
      {
        // Both are non-contingent
        // Replace if necessary
        int compare = oldValue.getValue().compareTo(newValue.getValue());
        if (compare > 0) // old > new
        {
//          labels.put(newValue.getLabel(), newValue);
          return newValue;
        }
        else
        {
          return oldValue;
        }
      }
    }
    //return oldValue;
  }
  
  protected LabeledValue updateLabeledValue(LabeledValue oldValue, ContingentLabeledValue value,
      TimeDistance nonContValue, TimeDistance contValue)
  {
    LabeledValue newValue;
    if (nonContValue.compareTo(contValue) <= 0) // Non-contingent value is more
                                               // restrictive than contingent
    {
      newValue = new DefaultLabeledValue(value.getLabel(), nonContValue);
    }
    else
    {
      newValue = new DefaultContingentLabeledValue(value.getLabel(), value.getCase(),
                                    nonContValue, value.getTrigger(), contValue);
    }
    return newValue;
  }
  
}
