package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import java.net.URI;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import de.aristaflow.adept2.base.sessionmanagement.SessionToken;
import de.aristaflow.adept2.model.common.timedata.CalendarElement;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.model.timemodel.UpdateableInstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.STNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInstanceTimeModel;

public class CSTNUInstanceTimeModel extends TCNInstanceTimeModel<LabeledTimePoint, STNEdge> implements
    InstanceTimeModel, UpdateableInstanceTimeModel
{

  private ChangeableCSTNU cstnu;

  public CSTNUInstanceTimeModel(UUID instanceID,
      UUID templateID, ChangeableCSTNU cstnu, Collection<NodeMapping<LabeledTimePoint>> nodeMapping,
      LabeledTimePoint creationTimePoint, LabeledTimePoint startTimePoint, LabeledTimePoint endTimePoint, 
      Map<Integer, List<URI>> updateTriggers)
  {
    super(instanceID, templateID, new CSTNU2STNProjection(cstnu), nodeMapping, 
          creationTimePoint, startTimePoint, endTimePoint, updateTriggers);
    this.cstnu = cstnu;
  }
 
  public void setInstanceCreationTime(long timestamp)
  {
    LabeledTimePoint node = getProcessCreationTimePoint();
    setToTimeStamp(node, timestamp);
    
//    TCNTestTools.printTCN(cstnu, this.getClass(), ".creation");
    
    handleTriggerEvent(node);
  }

  public void setInstanceStartTime(long timestamp)
  {
//    NodeMapping<LabeledTimePoint> startNodeMapping = stn.getNodeMapping(getBaseTemplate().getStartNode().getID(), iterationChain);
    
    LabeledTimePoint processStart = getProcessStartTimePoint();

    setToTimeStamp(processStart, timestamp);   
    
//    TCNTestTools.printTCN(cstnu, this.getClass(), ".start");
    
    handleTriggerEvent(processStart);
  }

  private void setToTimeStamp(LabeledTimePoint event, long timestamp)
  {
    
    LabeledTimePoint timeBaseEvent = getTimeBaseEvent();

    TimeDistance timeDistance = timeModelOperations.millisecondsToTimeDistance(timestamp);
    
    System.out.println("Setting "+event+" to "+timeDistance);
    
    assert timeDistance.toMilliseconds() == timestamp;
    
    setToDistance(timeBaseEvent, event, timeDistance);
  }

  private void setToDistance(LabeledTimePoint from, LabeledTimePoint to,
      TimeDistance timeDistance)
  {
    LabeledEdge max = cstnu.getEdge(from, to);
    LabeledEdge min = cstnu.getEdge(to, from);
    
    DefaultLabeledValue maxValue = new DefaultLabeledValue(from.getLabel(), timeDistance);
    LabeledEdge maxEdge = new DefaultLabeledEdge(from, to, new DefaultLabeledValueSet(maxValue), false, false, FlowDirection.FORWARD);
    cstnu.updateContingentConstraint(maxEdge);
    
    DefaultLabeledValue minValue = new DefaultLabeledValue(from.getLabel(), timeDistance.neg());
    LabeledEdge minEdge = new DefaultLabeledEdge(to, from, new DefaultLabeledValueSet(minValue), false, false, FlowDirection.BACKWARD);
    cstnu.updateContingentConstraint(minEdge);
   
    
    System.out.println(String.format("Updating [%s, %s] to [%s, %s]", min == null ? "infty" : min.getValue(), 
                                                                      max == null ? "infty" : max.getValue(), 
                                                                      minEdge.getValue(), maxEdge.getValue()));
  }

  public void setInstanceFinishTime(long timestamp)
  {
//    NodeMapping<LabeledTimePoint> startNodeMapping = stn.getNodeMapping(getBaseTemplate().getEndNode().getID(), iterationChain);
    
    LabeledTimePoint processEnd = getProcessEndTimePoint();

    setToTimeStamp(processEnd, timestamp);   
    
    handleTriggerEvent(processEnd);
  }

  public void setNodeActivationTime(int nodeID, int iteration, long timestamp)
  {
    // Nothing to do
    
  }

  public void setActivityInstanceStartTime(int nodeID, int iteration,
      long timestamp)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getStartNode();

    setToTimeStamp(node, timestamp);
    
//    TCNTestTools.printTCN(cstnu, this.getClass(), ""+nodeID+"."+iteration+".start");
    
    handleTriggerEvent(node);
  }

  public void setActivityInstanceFinishTime(int nodeID, int iteration,
      long timestamp)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getEndNode();

    setToTimeStamp(node, timestamp);   
    
    Interval<CalendarElement> timePoint = getTimePoint(nodeMapping.getStartNode());
    
    long startTime = toTimeStamp(timePoint.getMinValue());
    
    long duration = timestamp - startTime;
    
    setToDistance(nodeMapping.getStartNode(), nodeMapping.getEndNode(), 
      timeModelOperations.millisecondsToTimeDistance(duration));
    
    // Any intermediate event is also finished by now, so handle triggers
    for (LabeledTimePoint n : nodeMapping.getIntermediateEvents())
    {
      handleTriggerEvent(n);
    }
    // Remove any contingent link which have the respective node as trigger event
    handleTriggerEvent(node);
    
//    setToTimeStamp(node, timestamp);
    
//    TCNTestTools.printTCN(cstnu, this.getClass(), ""+nodeID+"."+iteration+".finish");
  }

  private long toTimeStamp(CalendarElement value)
  {
    return value.getTimeInMillis();
  }

  private void handleTriggerEvent(LabeledTimePoint node)
  {
    Collection<LabeledEdge> edges = cstnu.getEdges();
    for (LabeledEdge edge : edges)
    {
      boolean needsUpdate = false;
      STNEdge triggerEdge = null;
      List<LabeledValue> newValues = new ArrayList<LabeledValue>();
      for (LabeledValue labeledValue : edge.getValue().getLabeledValues())
      {
        LabeledValue value = labeledValue;
        if (value instanceof ContingentLabeledValue && ((ContingentLabeledValue)value).getTrigger() == node)
        {
          needsUpdate = true;
          // Get the value of the trigger
          if (triggerEdge == null) // Variable will be initialized on first use
            triggerEdge = stn.getEdge(edge.getSourceNode(), node);
          if (triggerEdge != null)
          {
            // Replace edge with non contingent edge with value of the trigger Node
            newValues.add(new DefaultLabeledValue(value.getLabel(), triggerEdge.getValue().getValue()));
          }
          else
          {
            // Otherwise replace with non-contingent value
            newValues.add(new DefaultLabeledValue(value.getLabel(), value.getValue()));
          }
          
        }
        else
        {
          newValues.add(labeledValue);
        }
      }
      
      if (needsUpdate)
      {
        LabeledEdge newEdge = new DefaultLabeledEdge(edge.getSourceNode(), edge.getTargetNode(),
          new DefaultLabeledValueSet(newValues), edge.isControlEdge(), false, edge.getFlowDirection());
        cstnu.updateContingentConstraint(newEdge);
      }
      
    }
  }


  public void setActivityInstanceProgress(int nodeID, int iteration,
      float progress)
  {
    // Not implemented, future use
    
  }

  public void createNewIteration(List<Node> nodesInLoop, int iteration)
  {
    // TODO create new Iteration
    
  }

  public CSTNU getBaseTimeModel()
  {
    return cstnu;
  }
  
  @Override
  protected TimeDistance getValue(STNEdge edge)
  {
    if (edge != null)
    {
      return edge.getValue().getValue();
    }
    return null;
  }

  @Override
  public void setEarliestStartDate(SessionToken session, int nodeID, int iteration, CalendarElement date)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getStartNode();
    setEarliestDate(node, date);
    
//    TCNTestTools.printTCN(cstnu, this.getClass(), ""+nodeID+"."+iteration+".start");
  }

  private void setEarliestDate(LabeledTimePoint node, CalendarElement date)
  {
    LabeledTimePoint timeBaseEvent = getTimeBaseEvent();
    
    TimeDistance timeDistance = timeModelOperations.millisecondsToTimeDistance(date.getTimeInMillis());
    
    
    
//    LabeledEdge min = cstnu.getEdge(node, timeBaseEvent);

    DefaultLabeledValue minValue = new DefaultLabeledValue(timeBaseEvent.getLabel(), timeDistance.neg());
    LabeledEdge minEdge = new DefaultLabeledEdge(node, timeBaseEvent, new DefaultLabeledValueSet(minValue), false, false, FlowDirection.BACKWARD);
    cstnu.updateContingentConstraint(minEdge);
  }

  @Override
  public void setEarliestCompletionDate(SessionToken session, int nodeID, int iteration, CalendarElement date)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getEndNode();
    setEarliestDate(node, date);
    
//    TCNTestTools.printTCN(cstnu, this.getClass(), ""+nodeID+"."+iteration+".completion");
  }

  @Override
  public void setLatestStartDate(SessionToken session, int nodeID, int iteration, CalendarElement date)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getStartNode();

    setLatestDate(node, date);
        
    
//    TCNTestTools.printTCN(cstnu, this.getClass(), ""+nodeID+"."+iteration+".start");
  }

  private void setLatestDate(LabeledTimePoint node, CalendarElement date)
  {
    long timestamp = date.getTimeInMillis();
    LabeledTimePoint timeBaseEvent = getTimeBaseEvent();
    
    TimeDistance timeDistance = timeModelOperations.millisecondsToTimeDistance(timestamp);
    
    
//    LabeledEdge max = cstnu.getEdge(timeBaseEvent, node);
    
    DefaultLabeledValue maxValue = new DefaultLabeledValue(timeBaseEvent.getLabel(), timeDistance);
    LabeledEdge maxEdge = new DefaultLabeledEdge(timeBaseEvent, node, new DefaultLabeledValueSet(maxValue), false, false, FlowDirection.FORWARD);
    cstnu.updateContingentConstraint(maxEdge);
  }

  @Override
  public void setLatestCompletionDate(SessionToken session, int nodeID, int iteration, CalendarElement date)
  {
    NodeMapping<LabeledTimePoint> nodeMapping = getNodeMapping(nodeID, iterationChain);
    
    LabeledTimePoint node = nodeMapping.getEndNode();

    setLatestDate(node, date);
        
    
//    TCNTestTools.printTCN(cstnu, this.getClass(), ""+nodeID+"."+iteration+".completion");
  }


}
