package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.globals.ProcessConstants.EdgeType;
import de.aristaflow.adept2.model.globals.ProcessConstants.NodeType;
import de.aristaflow.adept2.model.processmodel.TemporalNode;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.timemodel.ActivityDuration;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.JoinNodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractLabeledTemplateTransformation;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultObservationTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.util.Tuple;

public class CSTNUTemplateTransformation extends
    AbstractLabeledTemplateTransformation
{

  public static final String CONF_ADVANCED_ACTIVITY_TRANSFORMATION = CONF_PREFIX + "AdvancedActivityTransformation";
  
  public static final int ACTIVITY_TRANSFORMATION_BASIC = 0;
  public static final int ACTIVITY_TRANSFORMATION_COOPIS = 1;
  
  protected int advancedActivityTransformation;

  public CSTNUTemplateTransformation(Configuration configuration, TemporalTemplate template)
  {
    this(configuration, template, 0);
  }

  public CSTNUTemplateTransformation(Configuration configuration, TemporalTemplate template, int lookAhead)
  {
    super(configuration, template, lookAhead);
//    this.advancedActivityTransformation = advancedActivityTransformation;
  }
  
  public CSTNUTemplateTransformation(TemporalTemplate template)
  {
    this(null, template, 0);
  }

  public CSTNUTemplateTransformation(TemporalTemplate template, int lookAhead)
  {
    this(null, template, lookAhead);
  }
  
  @Override
  protected void readConfiguration()
  {
    super.readConfiguration();
    this.advancedActivityTransformation = this.configuration.getInt(CONF_ADVANCED_ACTIVITY_TRANSFORMATION, ACTIVITY_TRANSFORMATION_BASIC);
  }
  
  @Override
  public CSTNU transform()
  {
    return (CSTNU) super.transform();
  }
  
  @Override
  protected ChangeableCSTNU createTCN(Collection<LabeledTimePoint> nodes,
      Collection<LabeledEdge> edges,
      Collection<NodeMapping<LabeledTimePoint>> nodeMapping)
  {
    Set<ChoiceVariable> observations = new HashSet<ChoiceVariable>();
    
    for (LabeledTimePoint node : nodes)
    {
      if (node instanceof DefaultObservationTimePoint)
      {
        observations.add(((DefaultObservationTimePoint) node).getObservation());
      }
    }
    
    // FIXME
    ChangeableCSTNU timeModel = new DefaultCSTNU(nodes, edges, observations);
    return timeModel;
  }
  
  protected static class ContingentActivityDuration extends TCNActivityDuration
  {

    public final TimeDistance minGuard;
    public final TimeDistance maxGuard;

    public ContingentActivityDuration(TimeDistance minDuration, TimeDistance maxDuration, 
        TimeDistance minGuard, TimeDistance maxGuard)
    {
      super(minDuration, maxDuration);
      this.minGuard = minGuard;
      this.maxGuard = maxGuard;
    }
    
  }

  @Override
  protected ContingentActivityDuration getActivityDuration(int nodeID)
  {
    TemporalNode node = template.getNode(nodeID);
    ActivityDuration activityDuration = node.getActivityDuration();

    ContingentActivityDuration duration;
    
    if (activityDuration != null)
    {
      
      TimeDistance flexibleMinDuration = NODE_MIN_DURATION;
      TimeDistance guardMinDuration;
      TimeDistance guardMaxDuration;
      TimeDistance flexibleMaxDuration = NODE_MAX_DURATION;
      
      TimePoint lowerMinimumDuration = activityDuration
        .getMinimumDuration();
      if (lowerMinimumDuration != null)
      {
        flexibleMinDuration = new DefaultTimeDistance(lowerMinimumDuration);
      }
      


      TimePoint flexibleMaximumDuration = activityDuration
        .getMaximumDuration();
      if (flexibleMaximumDuration != null)
      {
        flexibleMaxDuration = new DefaultTimeDistance(flexibleMaximumDuration);
      }
      
      TimePoint guardMinimumDuration = activityDuration.getGuardMinimumDuration();
      if (guardMinimumDuration != null)
      {
        guardMinDuration = new DefaultTimeDistance(guardMinimumDuration);
      }
      else 
      {
        guardMinDuration = flexibleMaxDuration;
      }
      
     
      
      TimePoint guardMaximumDuration = activityDuration
          .getGuardMaximumDuration();
      if (guardMaximumDuration != null)
      {
        guardMaxDuration = new DefaultTimeDistance(guardMaximumDuration);
      }
      else
      {
        guardMaxDuration = flexibleMinDuration; 
      }
      
      duration = new ContingentActivityDuration(flexibleMinDuration, flexibleMaxDuration, guardMinDuration, guardMaxDuration);
    }
    else
    {
      duration = new ContingentActivityDuration(NODE_MIN_DURATION, NODE_MAX_DURATION, NODE_MAX_DURATION, NODE_MIN_DURATION);
    }
    

    return duration;
  }
  
  @Override
  protected NodeMapping<LabeledTimePoint> transformActivity(int nodeID,
      IterationChain iterationChain, Label branch)
  {
    NodeMapping<LabeledTimePoint> node;
    switch (advancedActivityTransformation)
    {
      case 1:
        node = transformActivityAdvanced(nodeID, iterationChain, branch);
        
        break;
      case 2: 
        //node = transformActivityCSTNPSU(nodeID, iterationChain, branch);
        throw new RuntimeException("No longer supported; Use CSTNPSU Transformation");
        
      case 0:
      default:
        node = transformActivityBasic(nodeID, iterationChain, branch);
        break;
    }
    
    processActivityValidityPeriod(nodeID, node);
    return node;
  }
  
  protected NodeMapping<LabeledTimePoint> transformActivityBasic(int nodeID,
                                                                 IterationChain iterationChain, Label branches)
   {
     if (template.getNodeType(nodeID) != NodeType.NT_NORMAL)
     {
       throw new IllegalArgumentException(""+template.getNodeType(nodeID));
     }
     
     int topologicalID = template.getNodeTopologicalID(nodeID);
     
     NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);
     LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
     LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);
     
     addNode(A_S);
     addNode(A_E);
     
     ContingentActivityDuration contigentDuration = getActivityDuration(nodeID);     

     LabeledEdge minEdge = new DefaultLabeledEdge(A_S, A_E, new DefaultLabeledValueSet(
       new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, contigentDuration.maxDuration, 
         A_E, contigentDuration.minGuard)), true, false, FlowDirection.BACKWARD);
     addEdge(minEdge);
     LabeledEdge maxEdge = new DefaultLabeledEdge(A_E, A_S, new DefaultLabeledValueSet(
       new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, contigentDuration.minGuard.neg(), 
         A_E, contigentDuration.maxGuard.neg())), true, false, FlowDirection.FORWARD);
     addEdge(maxEdge);
     
     
     NodeMapping<LabeledTimePoint> node = new NodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, null, A_E);
     
    
     
     return addNodeMapping(node);
   }
  

//  protected NodeMapping<LabeledTimePoint> transformActivityBasic2(int nodeID,
//      IterationChain iterationChain, Label branches)
//  {
//    if (template.getNodeType(nodeID) != NodeType.NT_NORMAL)
//    {
//      throw new IllegalArgumentException(""+template.getNodeType(nodeID));
//    }
//    
//    int topologicalID = template.getNodeTopologicalID(nodeID);
//    
//    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);
//    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
//    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);
//    
//    addNode(A_S);
//    addNode(A_E);
//    
//    Tuple<Interval<TimeDistance>,Interval<TimeDistance>> duration = getActivityDuration(nodeID);
//    
//    Interval<TimeDistance> contigentDuration = duration.getFirst(); 
//    Interval<TimeDistance> flexibleDuration = duration.getSecond(); 
//    
//
//    LabeledEdge minEdge = new DefaultLabeledEdge(A_S, A_E, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, flexibleDuration.getMaxValue(), A_E, contigentDuration.getMinValue())), true, false, FlowDirection.BACKWARD);
//    addEdge(minEdge);
//    LabeledEdge maxEdge = new DefaultLabeledEdge(A_E, A_S, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, flexibleDuration.getMinValue().neg(), A_E, contigentDuration.getMaxValue().neg())), true, false, FlowDirection.FORWARD);
//    addEdge(maxEdge);
//    
//    
//    NodeMapping<LabeledTimePoint> node = new NodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, null, A_E);
//    
//    return addNodeMapping(node);
//  }
  
  
  

//  protected NodeMapping<LabeledTimePoint> transformActivityAdvanced_MOD(
//      int nodeID, IterationChain iterationChain, Label branches)
//  {
//    if (template.getNodeType(nodeID) != NodeType.NT_NORMAL)
//    {
//      throw new IllegalArgumentException(""+template.getNodeType(nodeID));
//    }
//    
//    int topologicalID = template.getNodeTopologicalID(nodeID);
//    
//    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);
//    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
//    LabeledTimePoint A_C = createTCNNode(nodeInstance, "C", branches);
//    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);
//    
//    addNode(A_S);
//    addNode(A_C);
//    addNode(A_E);
//    
//    Tuple<Interval<TimeDistance>,Interval<TimeDistance>> duration = getActivityDuration(nodeID);
//    
//    Interval<TimeDistance> contigentDuration = duration.getFirst();
//    Interval<TimeDistance> flexibleDuration = duration.getSecond();
//    
//    TimeDistance contingentMinValue = contigentDuration.getMinValue();
//    TimeDistance contingentMaxValue = contigentDuration.getMaxValue();
//    
//    if (contingentMinValue.isInfinity())
//    {
//      contingentMinValue = NODE_MIN_DURATION;
//    }
//    
//    if (contingentMaxValue.isInfinity())
//    {
//      contingentMaxValue = contingentMinValue;
//    }
//      
//    LabeledEdge minEdge = new DefaultLabeledEdge(A_C, A_E, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, contingentMaxValue, A_E, contingentMinValue)), true, false, FlowDirection.FORWARD);
//    minEdge = addEdge(minEdge);
//    
//    LabeledEdge maxEdge = new DefaultLabeledEdge(A_E, A_C, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, contingentMinValue.neg(), A_E, contingentMaxValue.neg())), true, false, FlowDirection.BACKWARD);
//    maxEdge = addEdge(maxEdge);
//    
//    
//    TimeDistance flexibleMaxValue = flexibleDuration.getMaxValue();
//    createTCNEdge(A_S, A_C, branches, DefaultTimeDistance.ZERO_TIME_DISTANCE, flexibleMaxValue.sub(contingentMinValue), true);
//  
//    createTCNEdge(A_C, A_S, branches, DefaultTimeDistance.ZERO_TIME_DISTANCE, flexibleDuration.getMaxValue(), true);
//    
//    
//    
//
//    NodeMapping<LabeledTimePoint> node = new NodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, new LabeledTimePoint[]{A_C}, A_E);
//
//    return addNodeMapping(node);
//  }

  protected NodeMapping<LabeledTimePoint> transformActivityAdvanced(int nodeID,
      IterationChain iterationChain, Label branches)
  {
   if (template.getNodeType(nodeID) != NodeType.NT_NORMAL)
   {
     throw new IllegalArgumentException(""+template.getNodeType(nodeID));
   }
   
   int topologicalID = template.getNodeTopologicalID(nodeID);
   
   NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);
   LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
   LabeledTimePoint A_C = createTCNNode(nodeInstance, "C", branches);
   LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);
   
   addNode(A_S);
   addNode(A_C);
   addNode(A_E);
   
   ContingentActivityDuration duration = getActivityDuration(nodeID);
   
   TimeDistance contingentMinValue = duration.minGuard;
   TimeDistance contingentMaxValue = duration.maxGuard;
   
   if (contingentMinValue.isInfinity())
   {
     contingentMinValue = NODE_MIN_DURATION;
   }
   
   if (contingentMaxValue.isInfinity())
   {
     contingentMaxValue = contingentMinValue;
   }
     
   LabeledEdge maxEdge = new DefaultLabeledEdge(A_S, A_C, new DefaultLabeledValueSet(
     new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, contingentMaxValue, A_C, contingentMinValue)), true, false, FlowDirection.FORWARD);
   maxEdge = addEdge(maxEdge);
   
   LabeledEdge minEdge = new DefaultLabeledEdge(A_C, A_S, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, contingentMinValue.neg(), A_C, contingentMaxValue.neg())), true, false, FlowDirection.BACKWARD);
   minEdge = addEdge(minEdge);
   
   createTCNEdge(A_C, A_E, branches, DefaultTimeDistance.ZERO_TIME_DISTANCE, duration.maxDuration.sub(contingentMinValue), true);
 
   createTCNEdge(A_S, A_E, branches, DefaultTimeDistance.ZERO_TIME_DISTANCE, duration.maxDuration, true);
   
   
   

   NodeMapping<LabeledTimePoint> node = new NodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, new LabeledTimePoint[]{A_C}, A_E);

   return addNodeMapping(node);
 }  
  

  
  @Override
  protected JoinNodeMapping<LabeledTimePoint> transformANDJoin(int nodeID, IterationChain iterationChain, Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_AND_JOIN)
    {
      throw new IllegalArgumentException("" + template.getNodeType(nodeID));
    }

    int[] preds = template.getPredByEdgeType(nodeID, EdgeType.ET_CONTROL);

    LabeledTimePoint[] nodes = new LabeledTimePoint[preds.length];

    int topologicalID = template.getNodeTopologicalID(nodeID);

    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);

    for (int i = 0; i < preds.length; i++)
    {
      LabeledTimePoint waitNode = createTCNNode(nodeInstance, "w_{" + i + "}", branches);
      addNode(waitNode);
      nodes[i] = waitNode;
    }

    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);

    addNode(A_S);
    addNode(A_E);

    for (int i = 0; i < preds.length; i++)
    {
      LabeledTimePoint waitNode = nodes[i];

      createTCNEdge(waitNode, A_S, branches, CONTROL_FLOW_MIN_TIME_LAG,
        CONTROL_FLOW_MAX_TIME_LAG, true);
    }

    createTCNEdge(A_S, A_E, branches, STRUCTURE_NODE_MIN_DURATION, STRUCTURE_NODE_MAX_DURATION, true);

    return addNodeMapping(new JoinNodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, nodes, A_S, null, A_E));
  }

  @Override
  protected LabeledEdge mergeEdges(LabeledEdge e1, LabeledEdge e2)
  {
    LabeledEdge mergedEdge = super.mergeEdges(e1, e2);
    
    assert mergedEdge.getValue().getLabeledValues().size() == 1;
    
    return mergedEdge;
  }
  
//  @Override
//  protected LabeledTimePoint transformInstanceCreationEvent(IterationChain iterationChain, Label branches)
//  {
//    return getTimeBaseNode();
//  }
//
//  @Override
//  protected NodeMapping<LabeledTimePoint> transformProcessStartNode(int nodeID, IterationChain iterationChain, Label branches)
//  {
//    if (template.getNodeType(nodeID) != NodeType.NT_STARTFLOW)
//    {
//      throw new IllegalArgumentException("" + template.getNodeType(nodeID));
//    }
//
//    int topologicalID = template.getNodeTopologicalID(nodeID);
//
//    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);
//    LabeledTimePoint A_S = getTimeBaseNode();
//    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);
//
//
//    addNode(A_E);
//
//    createTCNEdge(A_S, A_E, branches, STRUCTURE_NODE_MIN_DURATION, STRUCTURE_NODE_MAX_DURATION, true);
//
//    return addNodeMapping(new NodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, null, A_E));
//  }



  
}
