package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import java.net.URI;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.apache.commons.configuration.BaseConfiguration;
import org.apache.commons.configuration.Configuration;
import org.apache.commons.configuration.MapConfiguration;

import de.aristaflow.adept2.model.processmodel.Instance;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.defaultimplementation.DefaultTemporalTemplate;
import de.aristaflow.adept2.model.timemodel.InstanceTimeContext;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelFactory;
import de.aristaflow.adept2.model.timemodel.TimeModelOperations;
import de.aristaflow.adept2.model.timemodel.TimeModelSolver;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.TimeModelTools;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;

public class CSTNUTimeModelFactory implements TimeModelFactory
{
  
  protected final Configuration configuration;

  public CSTNUTimeModelFactory()
  {
    this.configuration =  new BaseConfiguration();
    configuration.setProperty(CSTNUTemplateTransformation.CONF_ADVANCED_ACTIVITY_TRANSFORMATION, CSTNUTemplateTransformation.ACTIVITY_TRANSFORMATION_BASIC);
  }
  
  public CSTNUTimeModelFactory(Configuration configuration)
  {
    this.configuration = configuration;
    
  }

  public TemplateTimeModel createTimeModelForTemplate(TemporalTemplate template)
  {
    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(this.configuration,
                                                                        template);
    CSTNU cstnu = trafo.transform();
    
    Map<Integer, List<URI>> updateTriggers = TimeModelTools.createUpdateTriggerMapping(template);
    
    return new CSTNUTemplateTimeModel(template, cstnu, trafo.getNodeMapping(), trafo.getProcessCreationTimePoint(),
                                      trafo.getProcessStartTimePoint(), trafo.getProcessEndTimePoint(), updateTriggers);
  }

  public InstanceTimeModel createTimeModelForInstance(Instance instance,
      InstanceTimeContext timeContext, TemplateTimeModel templateTimeModel)
  {
    CSTNUTemplateTimeModel timeModel = (CSTNUTemplateTimeModel) templateTimeModel;

    UUID instanceID = instance != null ? instance.getID() : null;
    
    assert instance == null || instance.getTemplate().getID() == templateTimeModel.getTemplateID();
    
    return new CSTNUInstanceTimeModel(instanceID, templateTimeModel.getTemplateID(),
                                      (ChangeableCSTNU) (timeModel.getBaseTimeModel()).clone(),
                                      timeModel.getNodeMapping(), timeModel.getProcessCreationTimePoint(),
                                      timeModel.getProcessStartTimePoint(), timeModel.getProcessEndTimePoint(),
                                      timeModel.getUpdateTriggers());
  }


  public TimeModelOperations getTimeModelOperations()
  {
    return new DefaultTimeModelOperations();
  }

  public void setCreateChangeable(boolean changeable)
  {
    // Not necessary, returned elements are always changeable....
  }

  public TemporalTemplate createTemporalTemplate(Template template)
  {
    return new DefaultTemporalTemplate(template);
  }

  public TimeModelSolver getTimeModelSolver(
      Class<? extends TimeModel> clazz)
  {
    return new DCCSTNUTemplateTimeModelSolver();
  }

}
