package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import de.aristaflow.adept2.model.timemodel.ChangeSet;
import de.aristaflow.adept2.model.timemodel.TimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelSolver;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Triangle;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;

public class DCCSTNUTemplateTimeModelSolver implements TimeModelSolver
{

  final CSTNUDCSolver dcSolver;
  

  public DCCSTNUTemplateTimeModelSolver()
  {
    dcSolver = initSolver();
  }

  protected CSTNUDCSolver initSolver()
  {
    return new CSTNUDCSolver();
  }
  
  public void solve(TimeModel timeModel) throws TCNInconsistencyException, InterruptedException
  {
    if (timeModel instanceof CSTNUTemplateTimeModel)
    {
      CSTNUTemplateTimeModel cstnuTimeModel = (CSTNUTemplateTimeModel) timeModel;
      solve(cstnuTimeModel);
      return;
    }
    if (timeModel instanceof CSTNUInstanceTimeModel)
    {
      CSTNUInstanceTimeModel cstnuTimeModel = (CSTNUInstanceTimeModel) timeModel;
      solve(cstnuTimeModel);
      return;
    }
    throw new IllegalArgumentException();
  }
  
  public void solve(CSTNUTemplateTimeModel timeModel) throws TCNInconsistencyException, InterruptedException
  {
    ChangeableCSTNU cstnu = (ChangeableCSTNU) timeModel.getBaseTimeModel();
    dcSolver.solve(cstnu);
  }
  

  public void solve(CSTNUInstanceTimeModel timeModel) throws TCNInconsistencyException, InterruptedException
  {
    solve(timeModel, null);
  }
  
  public void solve(
      CSTNUInstanceTimeModel timeModel,
      ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet) throws TCNInconsistencyException, InterruptedException
  {
    ChangeableCSTNU cstnu = (ChangeableCSTNU) timeModel.getBaseTimeModel();

    try
    {
      cstnu = dcSolver.solve(cstnu, changeSet);
    }
    catch (TCNInconsistencyException e1)
    {
      throw e1;
    }   
    

  }



}
