package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.CSTN;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.ObservationTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractTCN;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultObservationTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ObservationBinding;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.Proposition;
import de.aristaflow.adept2.util.NullArgumentException;

public class DefaultCSTNU extends AbstractTCN<LabeledTimePoint, LabeledValueSet, LabeledEdge> implements Cloneable, CSTN,
    CSTNU, ChangeableCSTNU
{
  
  public static boolean CONNECT_OBSERVATION_NODES = true;
  
  private final List<ChoiceVariable> observations;
  private boolean controllable;

  public DefaultCSTNU(Collection<LabeledTimePoint> nodes,
      Collection<LabeledEdge> edges,
      Collection<ChoiceVariable> labels)
  {
    super(nodes, edges);
    this.observations = new ArrayList<ChoiceVariable>(labels);
  }
  
  @Override
  public void init()
  {
    if (CONNECT_OBSERVATION_NODES)
    {
      Collection<LabeledTimePoint> nodes = getTimePoints();
      for (LabeledTimePoint node : nodes)
      {
        Label lbl = node.getLabel();
        
        Proposition proposition = (Proposition) lbl;
        for (ObservationBinding observation : proposition.getObservations())
        {
          DefaultObservationTimePoint observationNode = observation.getVariable().getDecisionTimePoint();
          LabeledEdge edge = getEdge(node, observationNode);
          if (edge == null)
          {
            updateConstraint(new DefaultLabeledEdge(node, observationNode, new DefaultLabeledValueSet(new DefaultLabeledValue(node.getLabel(), DefaultTimeDistance.EPSILON_TIME_DISTANCE.neg())), false, true, FlowDirection.BACKWARD));
          }
        }
        
      }
    }
  }
  
  @Override
  protected boolean checkConsistency(LabeledEdge tcnEdge1, LabeledEdge tcnEdge2)
  {
    if (tcnEdge1 != null && tcnEdge2 != null)
    {
      Collection<LabeledValue> lvs1 = tcnEdge1.getValue().getLabeledValues();
      Collection<LabeledValue> lvs2 = tcnEdge2.getValue().getLabeledValues();
      
      for (LabeledValue lv1 : lvs1)
      {
        checkConsistency(tcnEdge1, lv1);
        TimeDistance v1 = lv1.getValue();
      
      
        for (LabeledValue lv2 : lvs2)
        {
          checkConsistency(tcnEdge2, lv2);
          TimeDistance v2 = lv2.getValue();
          
          if (lv1.getLabel().satisfiable(lv2.getLabel()))
          {
            if (lv1 instanceof ContingentLabeledValue && lv2 instanceof ContingentLabeledValue)
            {
              ContingentLabeledValue clv1 = (ContingentLabeledValue) lv1;
              ContingentLabeledValue clv2 = (ContingentLabeledValue) lv2;
              // FIXME
//              if (clv1.getContingentValue().compareTo(clv2.getContingentValue().neg()) >  0)
//                throw new IllegalStateException(tcnEdge1 + " vs. "+tcnEdge2);
              
              
//              if (!clv1.getContingentValue().equals(clv2.getSTNValue().neg()))
//                throw new IllegalStateException("Not a CSTNU: "+clv1+ " vs. "+ clv2);
//              
//              if (!clv2.getContingentValue().equals(clv1.getSTNValue().neg()))
//                throw new IllegalStateException("Not a CSTNU: "+clv1+ " vs. "+ clv2);
            }
     
            // FIXME
//            if (v1.compareTo(v2.neg()) <  0)
//              throw new IllegalStateException(lv1 + " < "+ lv2);
          }
        }
      }
      return true;
    }  
    return true;
  }

  protected void checkConsistency(LabeledEdge tcnEdge1, LabeledValue lv1)
  {
    if (!lv1.getLabel().subsumes(tcnEdge1.getSourceNode().getLabel()))
      throw new IllegalStateException(String.format("!(%s => %s) for %s for %s", lv1.getLabel(), tcnEdge1.getSourceNode().getLabel(), tcnEdge1, tcnEdge1.getSourceNode()));
    if (!lv1.getLabel().subsumes(tcnEdge1.getTargetNode().getLabel()))
      throw new IllegalStateException(String.format("!(%s => %s) for %s for %s", lv1.getLabel(), tcnEdge1.getTargetNode().getLabel(), tcnEdge1, tcnEdge1.getTargetNode()));
  }
  
  @Override
  protected boolean doCheckConsistency(LabeledEdge edge1, LabeledEdge edge2)
  {
    
    LabeledValueSet dir = edge1.getValue();
    LabeledValueSet inverseDir = edge2.getValue();
    Collection<LabeledValue> lvs1 = dir.getLabeledValues();
    Collection<LabeledValue> lvs2 = inverseDir.getLabeledValues();
    
    for (LabeledValue lv1 : lvs1)
    {
      TimeDistance v1 = lv1.getValue();
    

      for (LabeledValue lv2 : lvs2)
      {
        TimeDistance v2 = lv2.getValue();
        
        if (lv1.getLabel().satisfiable(lv2.getLabel()))
        {
          if (v1.compareTo(v2.neg()) <  0)
            throw new IllegalStateException();
          
          if (lv1 instanceof ContingentLabeledValue && lv2 instanceof ContingentLabeledValue)
          {
            ContingentLabeledValue clv1 = (ContingentLabeledValue) lv1;
            ContingentLabeledValue clv2 = (ContingentLabeledValue) lv2;
            
            if (clv1.getContingentValue().compareTo(clv2.getContingentValue().neg()) >  0)
              throw new IllegalStateException();
          }

            
          
        }
      }
      
      // WD1
      if (!lv1.getLabel().satisfiable(lv1.getLabel()))
        throw new IllegalStateException();
      
      if (!lv1.getLabel().subsumes(edge1.getSourceNode().getLabel()))
        throw new IllegalStateException();
      if (!lv1.getLabel().subsumes(edge1.getTargetNode().getLabel()))
        throw new IllegalStateException();
      
      // WD3
      Proposition prop = (Proposition) lv1.getLabel();
      for (ObservationBinding observation : prop.getObservations())
      {
        DefaultObservationTimePoint observationNode = observation.getVariable().getDecisionTimePoint();
        if (!lv1.getLabel().subsumes(observationNode.getLabel()))
          throw new IllegalStateException();
      }
    }
    return true;
  }
  
  @Override
  protected boolean doCheckConsistency(LabeledTimePoint node)
  {
    // WD2

    Proposition proposition = (Proposition) node.getLabel();
    for (ObservationBinding observation : proposition.getObservations())
    {
      DefaultObservationTimePoint observationNode = observation.getVariable().getDecisionTimePoint();
      if (!proposition.subsumes(observationNode.getLabel()))
        throw new IllegalStateException();
      
      if (CONNECT_OBSERVATION_NODES)
      {
        LabeledEdge edge = getEdge(node, observationNode);
        
        if (edge == null)
          throw new IllegalStateException();
        else
        {
          for (LabeledValue value : edge.getValue().getLabeledValues())
          {
            if (value.getValue().compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0)
              throw new IllegalStateException();
          }
        }
      }
    }
    
    return true;
  }
  
  @Override
  public LabeledEdge updateConstraint(LabeledEdge edge)
  {
    if (edge == null)
      throw new NullArgumentException();

    
    assert edge.getValue().subsumes(edge.getSourceNode().getLabel());
    assert edge.getValue().subsumes(edge.getTargetNode().getLabel());
 
    
    int source = getNodeKey(edge.getSourceNode());
    int target = getNodeKey(edge.getTargetNode());
    
    LabeledEdge old = internalGetConstraint(source, target);
    
    if (old != null && old.isContingentDuration())
    {
      ContingentLabeledValue oldValue = getContingentLabeledValue(old);
        ContingentLabeledValue newValue = getContingentLabeledValue(edge);
        
        if (!oldValue.getContingentValue().equals(newValue.getContingentValue()))
        {        
          throw new IllegalArgumentException("Updating contingent duration constraints is not allowed!");
        }
      }
    else 
    { // Adding a contingent duration this way is not allowed
      if (edge.isContingentDuration())
        throw new IllegalArgumentException("Adding contingent duration constraints is not allowed!");
    }
    
    internalSetConstraint(source, target, edge);
    
    return edge;
  }

  public ContingentLabeledValue getContingentLabeledValue(LabeledEdge edge)
  {
    Collection<LabeledValue> labeledValues = edge.getValue().getLabeledValues();
    assert labeledValues.size() == 1;
    LabeledValue labeledValue = labeledValues.iterator().next();
    assert labeledValue instanceof ContingentLabeledValue;
    return (ContingentLabeledValue) labeledValue;
  }

  /* (non-Javadoc)
   * @see de.aristaflow.adept2.model.timemodel.tcn.cstnu.CSTNU#updateContingentConstraint(de.aristaflow.adept2.model.timemodel.tcn.cstnu.LabeledEdge)
   */
  @Override
  public LabeledEdge updateContingentConstraint(LabeledEdge edge)
  {
    int source = getNodeKey(edge.getSourceNode());
    int target = getNodeKey(edge.getTargetNode());
    
    internalSetConstraint(source, target, edge);
    
    return edge;
  }
  
//  /* (non-Javadoc)
//   * @see de.aristaflow.adept2.model.timemodel.tcn.cstnu.CSTNU#removeContingentConstraint(de.aristaflow.adept2.model.timemodel.tcn.cstnu.LabeledEdge)
//   */
//  @Override
//  public void removeContingentConstraint(LabeledEdge edge)
//  {
//    int source = getNodeKey(edge.getSourceNode());
//    int target = getNodeKey(edge.getTargetNode());
//    
//   
//    internalSetConstraint(source, target, null);
//  }

  @Override
  public ChangeableCSTNU clone()
  {
    // All elements are immutable!
    ChangeableCSTNU cstnu = new DefaultCSTNU(this.getTimePoints(),
                            this.getEdges(), 
                            this.getObservations());
    
    if (this.isConsistent())
      cstnu.setConsistent();
    
    if (this.isControllable())
      cstnu.setControllable();
    
//    cstnu.setProcessBounds(this.getProcessCreationTimePoint(), this.getProcessStartTimePoint(), this.getProcessEndTimePoint());
    
    return cstnu;
  }

  @Override
  public LabeledTimePoint getTimeBaseEvent()
  {
    return TIMEBASE_NODE;
  }


  /* (non-Javadoc)
   * @see de.aristaflow.adept2.model.timemodel.tcn.cstnu.CSTNU#getObservations()
   */
  @Override
  public List<ChoiceVariable> getObservations()
  {
    return observations;
  }

  public void setControllable()
  {
    this.controllable = true;
    setConsistent();
  }

  /* (non-Javadoc)
   * @see de.aristaflow.adept2.model.timemodel.tcn.cstnu.CSTNU#isControllable()
   */
  @Override
  public boolean isControllable()
  {
    return controllable;
  }

  @Override
  public void modified()
  {
    super.modified();
    controllable = false;
  }
  
  @Override
  public LabeledTimePoint insertTimePoint(LabeledTimePoint node)
  {
    LabeledTimePoint timePoint = super.insertTimePoint(node);
    
    if (timePoint instanceof ObservationTimePoint)
    {
      this.observations.add(((ObservationTimePoint) timePoint).getChoiceVariable());
    }
    
    return timePoint;
  }
  
  @Override
  public void removeTimePoint(LabeledTimePoint node) throws IllegalStateException
  {

    // TODO check observation is not used by any constraint
    
    if (node instanceof ObservationTimePoint)
    {
      this.observations.remove(((ObservationTimePoint) node).getChoiceVariable());
    }
    
    
    
    super.removeTimePoint(node);
  }


}
