package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ExecutableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariableBinding;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.ObservationTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.SimpleConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.TCNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.Proposition;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;

public class DefaultExecutableCSTNU implements ExecutableCSTNU
{
  
  private Label scenario;
  private ChangeableCSTNU coreCSTNU;
  
  
  /** map from contingent time points to corresponding activation time points */
  private Map<LabeledTimePoint, LabeledTimePoint> contingentTimePoints;
  private Set<LabeledTimePoint> executedTimePoints;
  private LabeledTimePoint origin;
  private Set<LabeledTimePoint> activeTimePoints;
  private TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge> solver;
  private Map<LabeledTimePoint, Set<LabeledTimePoint>> predecessors;
  private HashMap<LabeledTimePoint, Set<LabeledTimePoint>> successors;

  public DefaultExecutableCSTNU(ChangeableCSTNU coreCSTNU,
      TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge> solver, long timeStamp) throws TCNInconsistencyException
  {
    
    if (!coreCSTNU.isControllable())
      throw new IllegalArgumentException();
    
    this.scenario = Proposition.EMPTY_PROPOSITION;
    this.coreCSTNU = (ChangeableCSTNU) coreCSTNU.clone();
    this.solver = solver;
    

    this.executedTimePoints = new HashSet<LabeledTimePoint>();
    this.contingentTimePoints = getContingentTimePoints(this.coreCSTNU);
    this.origin = this.coreCSTNU.getTimeBaseEvent();
    this.executedTimePoints.add(origin);
    
    initializePartialOrder(this.coreCSTNU);
    
    assert !this.predecessors.containsKey(origin);
    
    TimeDistance now = toTimeDistance(timeStamp);
    updateNow(now);
    
    propagate();
    
    this.activeTimePoints = new HashSet<LabeledTimePoint>();
    this.activeTimePoints.addAll(updatePartialOrder(origin));
  }
  
  

  
  protected List<LabeledTimePoint> updatePartialOrder(LabeledTimePoint tp)
  {
    List<LabeledTimePoint> newRoot = new ArrayList<LabeledTimePoint>();
    
    for (LabeledTimePoint succ : successors.get(tp))
    {
      Set<LabeledTimePoint> set = this.predecessors.get(succ);
      set.remove(tp);
      if (set.isEmpty())
      {
        this.predecessors.remove(succ);
        newRoot.add(succ);
      }
    }
    
    return newRoot;
    
//    for (Iterator<Entry<LabeledTimePoint, Set<LabeledTimePoint>>> iterator = this.predecessors.entrySet().iterator(); iterator.hasNext();)
//    {
//      Entry<LabeledTimePoint,Set<LabeledTimePoint>> entry = iterator.next();
//      
//      entry.getValue().remove(tp);
//      
//      if (entry.getValue().isEmpty())
//        iterator.remove();
//      
//    }
    
  }




  private void initializePartialOrder(TCN<LabeledTimePoint, LabeledEdge> cstnu)
  {
    this.predecessors = new HashMap<LabeledTimePoint, Set<LabeledTimePoint>>();
    this.successors = new HashMap<LabeledTimePoint, Set<LabeledTimePoint>>();
    
    for (LabeledTimePoint tp : cstnu.getTimePoints())
    {
      for (LabeledTimePoint pred : cstnu.getTimePoints())
      {
        LabeledEdge edge = cstnu.getEdge(tp, pred);
        if (edge != null)
        {
          TimeDistance dist = minValue(edge);
          
          if (dist.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) < 0)
          {
            Set<LabeledTimePoint> set = predecessors.get(tp);
            if (set == null)
            {
              set = new HashSet<LabeledTimePoint>();
              predecessors.put(tp, set);
            }
            set.add(pred);
            
            set = successors.get(pred);
            if (set == null)
            {
              set = new HashSet<LabeledTimePoint>();
              successors.put(pred, set);
            }
            set.add(tp);
          }
        }
      }
    }
  }

  private TimeDistance minValue(LabeledEdge edge)
  {
    TimeDistance min = DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
    for (SimpleConstraintValue value : edge.getValue().getValues())
    {
      if (min.compareTo(value.getValue()) > 0)
      {
        min = value.getValue();
      }
    }
    return min;
  }

  protected boolean propagate() throws TCNInconsistencyException
  {
    try
    {
      solver.solve(this.coreCSTNU);
      return true;
    }
    catch (InterruptedException e)
    {
      return false;
    }    
  }


  /**
   * 
   * @param coreCSTNU
   * @return a map from contingent time points to corresponding activation time points
   */
  protected Map<LabeledTimePoint, LabeledTimePoint> getContingentTimePoints(CSTNU coreCSTNU)
  {
    Map<LabeledTimePoint, LabeledTimePoint> contingentTimePoints = new HashMap<LabeledTimePoint, LabeledTimePoint>();
    for (LabeledEdge edge : coreCSTNU.getEdges())
    {
      for (LabeledValue value : edge.getValue().getLabeledValues())
      {
        if (value instanceof ContingentLabeledValue)
        {
          ContingentLabeledValue contingentLabeledValue = (ContingentLabeledValue) value;
          LabeledTimePoint trigger = contingentLabeledValue.getTrigger();
          
          if (ConstraintCase.LOWERCASE.equals(contingentLabeledValue.getCase()))
          {
            contingentTimePoints.put(trigger, edge.getSourceNode());
          }
          else
          {
            contingentTimePoints.put(trigger, edge.getTargetNode());
          }
          
          
        }
      }
    }
    return contingentTimePoints;
  }

  /**
   * 
   * @return the next time points to be executed
   * 
   */
  protected Collection<LabeledTimePoint> getNextTimePoints()
  {
    List<LabeledTimePoint> tp = new ArrayList<LabeledTimePoint>(this.coreCSTNU.getTimePoints());
    tp.removeAll(this.predecessors.keySet());
    return tp;
  }
  
  protected TimeDistance toTimeDistance(long now)
  {
    return new DefaultTimeDistance(now, 0, 0, 0, 0, 0);
  }
  
  protected void updateNow(TimeDistance now)
  {
    // TODO check before updating
    LabeledValue labeledValue = new DefaultLabeledValue(Proposition.EMPTY_PROPOSITION, now.neg());
    
    for (LabeledTimePoint timePoint : this.coreCSTNU.getTimePoints())
    {      
      LabeledEdge edge = this.coreCSTNU.getEdge(timePoint, origin);
      
      if (edge == null)
      {
        this.coreCSTNU.updateConstraint(new DefaultLabeledEdge(timePoint, origin, new DefaultLabeledValueSet(labeledValue), false, true, FlowDirection.BACKWARD));
      }
      else
      {
        Map<Label, LabeledValue> labeledValues = DefaultLabeledValueSet.toMap(edge.getValue());
        DefaultLabeledValueSet.addOrReplaceIfNecessary(labeledValues, labeledValue);
        this.coreCSTNU.updateConstraint(replaceValues(edge, labeledValues.values()));
      }
    }
  }
  
  protected TimeDistance scenarioProjection(LabeledEdge edge)
  {
    TimeDistance min = DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
    for (LabeledValue value : edge.getValue().getLabeledValues())
    {
      if (value.getLabel().satisfiable(scenario))
      {
        
        if (value.getValue().compareTo(min) < 0)
        {
          min = value.getValue();
        }
      }
    }

    return min;
  }
  
  
  public Collection<LabeledTimePoint> getExecutableTimePoints(long now)
  {
    List<LabeledTimePoint> executable = new ArrayList<LabeledTimePoint>();
    
    TimeDistance currentTime = toTimeDistance(now).neg();
    
    for (LabeledTimePoint active : this.activeTimePoints)
    {
      LabeledEdge edge = this.coreCSTNU.getEdge(active, origin);
      if (edge != null)
      {
        TimeDistance minTime = scenarioProjection(edge);
        
        if (minTime.compareTo(currentTime) >= 0)
        {
          executable.add(active);
        }
        
      }
      else
      {
        executable.add(active);
      }
        
    }
    
    
    return executable;
  }

  
  public Collection<LabeledTimePoint> getActiveTimePoints()
  {
    return this.activeTimePoints;
  }

  public void executeTimePoint(LabeledTimePoint timePoint, long now) throws TCNInconsistencyException
  {
    if (timePoint instanceof ObservationTimePoint)
      throw new IllegalArgumentException();
    
    internallyExecuteTimePoint(timePoint, now);
    
    propagate();
    
    
    this.activeTimePoints.remove(timePoint);
    this.activeTimePoints.addAll(updatePartialOrder(timePoint));

  }




  protected void internallyExecuteTimePoint(LabeledTimePoint timePoint, long now) throws TCNInconsistencyException
  {
    this.executedTimePoints.add(timePoint);
    
    TimeDistance currentTime = toTimeDistance(now);
    
    // TODO check before updating!
    updateNow(currentTime);
    this.coreCSTNU.updateConstraint(new DefaultLabeledEdge(origin, timePoint, new DefaultLabeledValueSet(new DefaultLabeledValue(Proposition.EMPTY_PROPOSITION, currentTime))));
    this.coreCSTNU.updateConstraint(new DefaultLabeledEdge(timePoint, origin, new DefaultLabeledValueSet(new DefaultLabeledValue(Proposition.EMPTY_PROPOSITION, currentTime.neg()))));
    
    LabeledTimePoint activationTimePoint = contingentTimePoints.remove(timePoint);
    if (activationTimePoint != null)
    {
      // Replace all wait constraints of the time point by requirement constraints.
      // Wait constraints always end at the activation time point 
      for (LabeledTimePoint tp  : this.coreCSTNU.getTimePoints())
      {
        LabeledEdge edge = this.coreCSTNU.getEdge(tp, activationTimePoint);
        List<LabeledValue> values = new ArrayList<LabeledValue>();
        
        boolean changes = false;
        for (LabeledValue value : edge.getValue().getLabeledValues())
        {
          if (value instanceof ContingentLabeledValue)
          {
            assert ((ContingentLabeledValue) value).getTrigger() == timePoint;
            values.add(new DefaultLabeledValue(value.getLabel(), ((ContingentLabeledValue) value).getSTNValue()));
            
            changes = true;
            
          }
          else
          {
            values.add(value);
          }
        }

        if (changes)
        {
          LabeledEdge newEdge = new DefaultLabeledEdge(tp, activationTimePoint, new DefaultLabeledValueSet(values),
            edge.isControlEdge(), edge.isImplicit(), edge.getFlowDirection());
          this.coreCSTNU.updateContingentConstraint(newEdge);
        }
      }
    }
    

  }

  public void executeObservationTimePoint(ObservationTimePoint timePoint, long now, ChoiceVariableBinding observation) throws TCNInconsistencyException
  {
    
    scenario = scenario.addVariableBinding(observation);
    purgeUnsatisfiableLabels();

    internallyExecuteTimePoint(timePoint, now);
    
    propagate();
    
    
    this.activeTimePoints.remove(timePoint);
    this.activeTimePoints.addAll(updatePartialOrder(timePoint));
  }
  
  

/* CSTNU methods */
  
  
  private void purgeUnsatisfiableLabels()
  {
    Collection<LabeledTimePoint> timePoints = this.coreCSTNU.getTimePoints();
    for (LabeledTimePoint tp1 : timePoints)
    {
      // If the Label of the Time Point is no longer satisfiable remove it
      if (!tp1.getLabel().satisfiable(scenario))
      {
        this.coreCSTNU.removeTimePoint(tp1);
        continue;
      }
        
      for (LabeledTimePoint tp2 : this.coreCSTNU.getTimePoints())
      {
        // If the label of the Time Point is no longer satisfiable skip it. (Will be removed by outer loop)
        if (!tp2.getLabel().satisfiable(scenario))
          continue;
        
        LabeledEdge edge = this.coreCSTNU.getEdge(tp1, tp2);
        LabeledEdge newEdge = purgeUnsatisfiableLabels(edge);
        
        if (newEdge != edge)
          this.coreCSTNU.updateContingentConstraint(newEdge);
      }
    }
    
  }




  private LabeledEdge purgeUnsatisfiableLabels(LabeledEdge edge)
  {
    List<LabeledValue> values = new ArrayList<LabeledValue>(edge.getValue().getLabeledValues());
    boolean changed = false;
    
    for (Iterator<LabeledValue> iterator = values.iterator(); iterator.hasNext();)
    {
      LabeledValue value = iterator.next();
      if (!value.getLabel().satisfiable(scenario))
      {
        changed = true;
        iterator.remove();
      }
    }
    
    
    if (changed)
    {
      return replaceValues(edge, values);
    }

    return edge;
  }


  protected DefaultLabeledEdge replaceValues(LabeledEdge edge, Collection<LabeledValue> values)
  {
    return new DefaultLabeledEdge(edge.getSourceNode(), edge.getTargetNode(), new DefaultLabeledValueSet(values),
      edge.isControlEdge(), edge.isImplicit(), edge.getFlowDirection());
  }




  public ChangeableCSTNU getBaseTCN()
  {
    return this.coreCSTNU;
  }

  @Override
  public List<ChoiceVariable> getObservations()
  {
    return getBaseTCN().getObservations();
  }

  @Override
  public boolean isControllable()
  {
    return getBaseTCN().isControllable();
  }
  
  @Override
  public Collection<LabeledTimePoint> getTimePoints()
  {
    return Collections.unmodifiableCollection(this.coreCSTNU.getTimePoints());
  }
  
  protected LabeledEdge transformEdge(LabeledTimePoint from, LabeledTimePoint to, LabeledEdge originalEdge)
  {
    if (this.coreCSTNU.contains(from) && this.coreCSTNU.contains(to))
      return originalEdge;
    return null;
  }

  @Override
  public TCN<LabeledTimePoint, LabeledEdge> clone()
  {
    throw new UnsupportedOperationException();
  }

  @Override
 public LabeledEdge getEdge(LabeledTimePoint from, LabeledTimePoint to)
 {
    if (this.coreCSTNU.contains(from) && this.coreCSTNU.contains(to))
      return coreCSTNU.getEdge(from, to);
    return null;
 }


  public LabeledTimePoint getTimeBaseEvent()
  {
    return coreCSTNU.getTimeBaseEvent();
  }


  @Override
  public boolean equals(Object obj)
  {
    return coreCSTNU.equals(obj);
  }
  
  @Override
  public int hashCode()
  {
    return coreCSTNU.hashCode();
  }

  public void visit(GraphVisitor<LabeledTimePoint, LabeledEdge> visitor)
  {
    visitor.discoverGraph(this);
  
//  
//    for (NodeMapping<LabeledTimePoint> n : getNodeMapping())
//    {
//      
//      DirectedGraphImplementation<LabeledTimePoint, LabeledEdge> subGraph = new DirectedGraphImplementation<LabeledTimePoint, LabeledEdge>(n);
//      subGraph.visit(visitor);
//    }
    
    for (LabeledTimePoint e : getTimePoints())
    {
      visitor.discoverVertex(e);
      visitor.finishVertex(e);
    }
    
    for (LabeledTimePoint ei : getTimePoints())
    {
      for (LabeledTimePoint ej : getTimePoints())
      {
        LabeledEdge e = getEdge(ei, ej);
        if (e != null)
        {
          visitor.discoverEdge(e.getSourceNode(), 
                               e.getTargetNode(), 
                               e);
          visitor.finishEdge(e.getSourceNode(), 
                             e.getTargetNode(), 
                             e);
        }
      }
    }
  
    visitor.finishGraph(this);
  }

  public Collection<LabeledEdge> getEdges()
  {
    List<LabeledEdge> edges = new ArrayList<LabeledEdge>();
    
    for (LabeledTimePoint from : getTimePoints())
    {
      for (LabeledTimePoint to : getTimePoints())
      {
        LabeledEdge edge = getEdge(from, to);
        if (edge != null)
          edges.add(edge);
      }
    }
    
    return edges;
  }
  
  public Collection<LabeledEdge> getIncomingEdges(LabeledTimePoint to)
  {
    Collection<LabeledTimePoint> timePoints = getTimePoints();
    List<LabeledEdge> edges = new ArrayList<LabeledEdge>(timePoints.size());
    for (LabeledTimePoint from : timePoints)
    {
      LabeledEdge edge = getEdge(from, to);
      if (edge != null)
      {
        edges.add(edge);
      }
    }
    return edges;
  }

  public Collection<LabeledEdge> getOutgoingEdges(LabeledTimePoint from)
  {
    Collection<LabeledTimePoint> timePoints = getTimePoints();
    List<LabeledEdge> edges = new ArrayList<LabeledEdge>(timePoints.size());
    for (LabeledTimePoint to : timePoints)
    {
      LabeledEdge edge = getEdge(from, to);
      if (edge != null)
      {
        edges.add(edge);
      }
    }
    return edges;
  }


  @Override
  public String toString()
  {
    return coreCSTNU.toString();
  }


  public boolean isConsistent()
  {
    // FIXME
    return this.coreCSTNU.isConsistent();
  }
  
  public Collection<LabeledTimePoint> getVertices()
  {
    return getTimePoints();
  }

  @Override
  public boolean contains(LabeledTimePoint timePoint)
  {
    return this.coreCSTNU.contains(timePoint);
  }

}
