package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import de.aristaflow.adept2.model.timemodel.tcn.ChangeableTCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTimePoint;

public abstract class AbstractChangeableProjection<N extends TCNTimePoint, E extends TCNEdge<N, ?>, T extends TCNEdge<N, ?>>
    extends AbstractProjection<N, E, T> implements ChangeableTCN<N, T>
{

  protected final boolean propagate;

  private TCNEdge<?, ?>[][] edges;
  private Map<N, Integer> nodeKeys;

  private int nextKey;
  
  public AbstractChangeableProjection(ChangeableTCN<N, E> tcn, boolean propagate)
  {
    super(tcn);
    this.propagate = propagate;
    int nodeCount = tcn.getTimePoints().size()+1;
    this.edges = new TCNEdge[nodeCount][nodeCount];
    this.nodeKeys = new HashMap<N, Integer>();
    this.nextKey = 0;
  }

  
//  public void setProcessBounds(N processCreationNode, N processStartNode, N processEndNode)
//  {
//    getBaseTCN().setProcessBounds(getBaseNode(processCreationNode), getBaseNode(processStartNode), getBaseNode(processEndNode));
//  }
  

  @Override
  public ChangeableTCN<N, E> getBaseTCN()
  {
    return (ChangeableTCN<N, E>) super.getBaseTCN();
  }
  
  @Override
  public Collection<T> getChanges()
  {
    //Collection<E> changes = new ArrayList<E>(getBaseTCN().getChanges());
    
    return null;
  }

  public T updateConstraint(T edge)
  {
    if (!propagate)
    {
      internalSetEdge(edge);
      
      changed();
      
      return edge;
    }
    else
    {
      return propagateUpdate(edge);
    }
  }
  
  protected void internalSetEdge(T edge)
  {
    int fromID = getOrCreateNodeKey(edge.getSourceNode());
    int toID = getOrCreateNodeKey(edge.getTargetNode()); 
    edges[fromID][toID] = edge;
  }


  /**
   * Node keys will be initialized on the fly as needed. Once used always the same key needs to
   * be allocated to a particular node.
   * 
   * @param node
   * @return
   */
  protected int getOrCreateNodeKey(N node)
  {
    
    Integer integer = nodeKeys.get(node);
    if (integer == null)
    {
      integer = Integer.valueOf(nextKey++);
      nodeKeys.put(node, integer);
    }
    return integer.intValue();
  }
  
  protected int getNodeKey(N node)
  {
    
    Integer integer = nodeKeys.get(node);
    if (integer == null)
    {
      return nextKey; // Return any unused key..
    }
    return integer.intValue();
  }


  protected abstract T propagateUpdate(T edge);
  
  @SuppressWarnings("unchecked")
  protected T internalGetEdge(N from, N to)
  {
    if (edges != null)
    {
      int fromID = getNodeKey(from);
      int toID = getNodeKey(to);
      
      return (T) edges[fromID][toID];
    }
    else
    {
      return null;
    }
  }
  


  public void removeDelta(E edge)
  {
    if (edges != null)
    {
      int fromID = getNodeKey(edge.getSourceNode());
      int toID = getNodeKey(edge.getTargetNode());
      
      edges[fromID][toID] = null;
      
      changed();
    }
  }
  
  @Override
 public T getEdge(N from, N to)
 {
   T edge = internalGetEdge(from, to);
   if (edge != null)
     return edge;
   
   E oldEdge = getBaseTCN().getEdge(getBaseNode(from), getBaseNode(to));
   
   if (oldEdge == null)
     return null; 
   
   T transform = transformEdge(from, to, oldEdge);
   
   if (transform != null && transform != oldEdge)
     internalSetEdge(transform);
   
   return transform;
  
 }
  
  @Override
  public Collection<N> getTimePoints()
  {
    Collection<N> baseNodes = super.getTimePoints();
    List<N> nodes = new ArrayList<N>(baseNodes.size());
    
    for (N node : baseNodes)
    {
      N transform = transformNode(node);
      if (transform != null)
      {
        nodes.add(transform);
      }
    }
    
    return nodes;
  }
  


  protected  N transformNode(N node)
  {
    return node;
  }


  protected N getBaseNode(N node)
  {
    return node;
  }


  protected abstract T transformEdge(N from, N to, E originalEdge);

}
