package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Set;
import java.util.logging.Level;

import de.aristaflow.adept2.model.timemodel.defaultimplementation.Path;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Triangle;
import de.aristaflow.adept2.model.timemodel.tcn.ConstraintDependencyGraph;
import de.aristaflow.adept2.model.timemodel.tcn.ConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTimePoint;
import de.aristaflow.adept2.model.timemodel.tools.DirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;

public abstract class AbstractConstraintDependencyGraph<N extends TCNTimePoint, E extends TCNEdge<N, T>, T extends ConstraintValue> implements ConstraintDependencyGraph<E>
{


  protected static final class GraphWrapper<E extends TCNEdge<?, ?>> implements
      DirectedGraph<Object, Object>
  {
    private final ConstraintDependencyGraph<E> graph;
    
    public GraphWrapper(ConstraintDependencyGraph<E> graph)
    {
      this.graph = graph;
      
    }

    @Override
    public Collection<Object> getVertices()
    {
      return new ArrayList<Object>(graph.getVertices());
    }

    @Override
    public Collection<Object> getEdges()
    {
      return new ArrayList<Object>(graph.getEdges());
    }

    @Override
    public Object getEdge(Object from, Object to)
    {
      // TODO Auto-generated method stub
      return null;
    }

    @Override
    public void visit(
        GraphVisitor<Object, Object> visitor)
    {
      //
    }
  }



  protected Level debugLevel = Level.OFF;
  private Set<E> nodes = new HashSet<E>();
  private Set<Triangle<E>> edges = new HashSet<Triangle<E>>();

  public AbstractConstraintDependencyGraph(TCN<N, E> tcn)
  {
    super();
    
    if (!tcn.isConsistent())
      throw new IllegalArgumentException("TCN must be minimal in order to build a constraint dependency graph!");
    
    this.nodes.addAll(tcn.getEdges());
    
    initGraph(tcn);
  }



  private void initGraph(TCN<N, E> tcn)
  {
    for(N x : tcn.getTimePoints())
    {
      for (N z : tcn.getTimePoints())
      {
        if (x == z)
          continue;
        
        E c_xz = tcn.getEdge(x, z); 
        
        if (c_xz == null)
          continue;
        
        for (N y : tcn.getTimePoints())
        {
          if (x == y || z == y)
            continue;
          
          
          E c_xy = tcn.getEdge(x, y);
          E c_yz = tcn.getEdge(y, z);
          
          if (c_xy == null || c_yz == null)
            continue;
          
          
          T c_xyz = add(c_xy.getValue(), c_yz.getValue());
          
          if (equalBound(c_xz.getValue(), c_xyz))
          {
            this.edges.add(new Triangle<E>(new Path<E>(c_xy, c_yz), c_xz));
          }
        }
      }
    }
  }


  
  protected abstract boolean equalBound(T t, T t2);

  protected abstract T add(T t, T t2);



  /* (non-Javadoc)
   * @see de.aristaflow.adept2.model.timemodel.defaultimplementation.ConstraintDependencyGraph#getVertices()
   */
  @Override
  public Collection<E> getVertices()
  {
    return Collections.unmodifiableCollection(nodes);
  }

  /* (non-Javadoc)
   * @see de.aristaflow.adept2.model.timemodel.defaultimplementation.ConstraintDependencyGraph#getEdges()
   */
  @Override
  public Collection<Object> getEdges()
  {
    return new ArrayList<Object>(edges);
  }

  /* (non-Javadoc)
   * @see de.aristaflow.adept2.model.timemodel.defaultimplementation.ConstraintDependencyGraph#getEdge(E, E)
   */
  @Override
  public Object getEdge(E from, E to)
  {
    Path<E> o = new Path<E>(from, to);
    if (edges.contains(o))
      return o;
    
    return null;
  }

  protected void debug(Level logLevel, String message)
  {
    if (logLevel.intValue() >= debugLevel.intValue())
    {
      System.out.println(message);
      //logger.log(logLevel, message);    
    }
  }

  protected void debug(Level logLevel, String message, Object... params)
  {
    debug(logLevel, String.format(message, params));    
  }

  @Override
  public String toString()
  {
    return edges.toString();
  }



  @Override
  public void visit(
      GraphVisitor<Object, Object> visitor)
  {
    DirectedGraph<Object, Object> graph = new GraphWrapper<E>(this);
    visitor.discoverGraph(graph);

    for (E edge : nodes)
    {
      visitor.discoverVertex(edge);
      visitor.finishVertex(edge);
    }

    for (Triangle<E> edge : edges)
    {
      Path<E> jointVertex = edge.getFirst();
      visitor.discoverVertex(jointVertex);
      visitor.discoverEdge(edge.getFirst().getFirst(), jointVertex, edge.getFirst().getFirst());
      visitor.discoverEdge(edge.getFirst().getSecond(), jointVertex, edge.getFirst().getSecond());
      visitor.discoverEdge(jointVertex, edge.getSecond(), "");
     
      visitor.finishEdge(edge.getFirst().getFirst(), jointVertex, edge.getFirst().getFirst());
      visitor.finishEdge(edge.getFirst().getSecond(), jointVertex, edge.getFirst().getSecond());
      visitor.finishEdge(jointVertex, edge.getSecond(), "");
      visitor.finishVertex(jointVertex);
    }

    visitor.finishGraph(graph);
  }

}