package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.util.NullArgumentException;

public abstract class AbstractLabeledTemplateTransformation extends
    AbstractTemplateTransformation<LabeledEdge>
{

  public AbstractLabeledTemplateTransformation(Configuration configuration, TemporalTemplate template)
  {
    super(configuration, template);
  }

  public AbstractLabeledTemplateTransformation(Configuration configuration, TemporalTemplate template, int lookAhead)
  {
    super(configuration, template, lookAhead);
  }



  @Override
  protected Label createInitialBranch()
  {
    return Proposition.EMPTY_PROPOSITION;
  }

  

//  @Override
//  protected LabeledEdge[] createTCNEdge(LabeledNode from, LabeledNode to, BranchSet branches)
//  {
//    return createTCNEdge(from, to, branches,  null, null);
//  }

  @Override
  protected LabeledEdge[] createTCNEdge(LabeledTimePoint from, LabeledTimePoint to,
      Label branches, TimeDistance minValue, TimeDistance maxValue,
      boolean controlEdge)
  {

    if (minValue == null)
      throw new NullArgumentException();
    
    if (maxValue == null)
      throw new NullArgumentException();
      
    LabeledEdge minEdge;
    LabeledEdge maxEdge;


    minEdge = new DefaultLabeledEdge(to, from,
                              new DefaultLabeledValueSet(new DefaultLabeledValue(branches, minValue.neg())), controlEdge,
                              false, FlowDirection.BACKWARD);
    minEdge = addEdge(minEdge);

    maxEdge = new DefaultLabeledEdge(
                              from,
                              to,
                              new DefaultLabeledValueSet(new DefaultLabeledValue(branches, maxValue)),
                              controlEdge, false, FlowDirection.FORWARD);
    maxEdge = addEdge(maxEdge);


    return new LabeledEdge[] { minEdge, maxEdge };
  }
  
  @Override
  protected LabeledEdge mergeEdges(LabeledEdge e1, LabeledEdge e2)
  {   
    LabeledValueSet labeledValueSet = e1.getValue().merge(e2.getValue());
    return new DefaultLabeledEdge(e1.getSourceNode(), e2.getTargetNode(), labeledValueSet, e1.isControlEdge() || e2.isControlEdge(), e1.isImplicit() && e2.isImplicit(), e1.getFlowDirection() == e2.getFlowDirection() ? e1.getFlowDirection() : FlowDirection.UNKOWN);
  }


//  protected Label mergeLabels(Label branch1, Label branch2)
//  {
//    return branch1.merge(branch2);
//  }




}