package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTimePoint;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;

public abstract class AbstractProjection<N extends TCNTimePoint, E extends TCNEdge<N, ?>, T extends TCNEdge<N, ?>>
    implements TCN<N, T>, Projection<N, E, T>, Cloneable
{

  private final TCN<N, E> tcn;
  private boolean changes = false;

  public AbstractProjection(TCN<N, E> tcn)
  {
    super();
    this.tcn = tcn;
  }

  @Override
  public int hashCode()
  {
    return tcn.hashCode();
  }

  public N getTimeBaseEvent()
  {
    return tcn.getTimeBaseEvent();
  }

  protected void changed()
  {
    this.changes = false;
  }

  @Override
  public boolean equals(Object obj)
  {
    return tcn.equals(obj);
  }

  public void visit(GraphVisitor<N, T> visitor)
  {
    visitor.discoverGraph(this);
  
  
//    for (NodeMapping<N> n : getNodeMapping())
//    {
//      
//      DirectedGraphImplementation<N, T> subGraph = new DirectedGraphImplementation<N, T>(n);
//      subGraph.visit(visitor);
//    }
    
    for (N e : getTimePoints())
    {
      visitor.discoverVertex(e);
      visitor.finishVertex(e);
    }
    
    for (N ei : getTimePoints())
    {
      for (N ej : getTimePoints())
      {
        T e = getEdge(ei, ej);
        if (e != null)
        {
          visitor.discoverEdge(e.getSourceNode(), 
                               e.getTargetNode(), 
                               e);
          visitor.finishEdge(e.getSourceNode(), 
                             e.getTargetNode(), 
                             e);
        }
      }
    }
  
    visitor.finishGraph(this);
  }


//  public Collection<NodeMapping<N>> getNodeMapping()
//  {
//    return tcn.getNodeMapping();
//  }

  public T getEdge(N from, N to)
  {
    return getEdge(from, to);
  }

  public Collection<T> getEdges()
  {
    List<T> edges = new ArrayList<T>();
    
    for (N from : getTimePoints())
    {
      for (N to : getTimePoints())
      {
        T edge = getEdge(from, to);
        if (edge != null)
          edges.add(edge);
      }
    }
    
    return edges;
  }
  
  public Collection<T> getIncomingEdges(N to)
  {
    Collection<N> timePoints = getTimePoints();
    List<T> edges = new ArrayList<T>(timePoints.size());
    for (N from : timePoints)
    {
      T edge = getEdge(from, to);
      if (edge != null)
      {
        edges.add(edge);
      }
    }
    return edges;
  }

  public Collection<T> getOutgoingEdges(N from)
  {
    Collection<N> timePoints = getTimePoints();
    List<T> edges = new ArrayList<T>(timePoints.size());
    for (N to : timePoints)
    {
      T edge = getEdge(from, to);
      if (edge != null)
      {
        edges.add(edge);
      }
    }
    return edges;
  }

  public Collection<N> getTimePoints()
  {
    return tcn.getTimePoints();
  }
  
  public boolean contains(N timePoint)
  {
    return this.getTimePoints().contains(timePoint);
  }

//  public NodeMapping<N> getNodeMapping(int nodeId, IterationChain iterationChain)
//  {
//    return tcn.getNodeMapping(nodeId, iterationChain);
//  }

  @Override
  public String toString()
  {
    return tcn.toString();
  }

  public TCN<N, E> getBaseTCN()
  {
    return tcn;
  }

  public boolean isConsistent()
  {
    return !changes && this.tcn.isConsistent();
  }

  public void setConsistent()
  {
  //    if (propagate)
  //      this.cstnu.setMinimal();
      changes = false;
  }

//  public N getProcessCreationTimePoint()
//  {
//    return this.tcn.getProcessCreationTimePoint();
//  }
//
//  public N getProcessStartTimePoint()
//  {
//    return this.tcn.getProcessStartTimePoint();
//  }
//
//  public N getProcessEndTimePoint()
//  {
//    return this.tcn.getProcessEndTimePoint();
//  }
//
//  public UUID getTemplateID()
//  {
//    return tcn.getTemplateID();
//  }
  
  public Collection<N> getVertices()
  {
    return getTimePoints();
  }
  
  @Override
  public abstract TCN<N, T> clone();

  public N insertTimePoint(N timePoint)
  {
    throw new UnsupportedOperationException();
  }


  public void removeTimePoint(N timePoint)
  {
    throw new UnsupportedOperationException();
  }

}