package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.Collection;
import java.util.Queue;
import java.util.logging.Level;
import java.util.logging.Logger;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.ChangeSet;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Path;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Triangle;
import de.aristaflow.adept2.model.timemodel.tcn.ConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.SimpleConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.TCNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTimePoint;

public abstract class AbstractSTNSolver<N extends TCNTimePoint, V extends ConstraintValue, E extends TCNEdge<N, V>>
    implements TCNSolver<N, V, E>
{

  protected Logger logger = Logger.getLogger(this.getClass().getName());
  
  public AbstractSTNSolver()
  {
  }
  
  public AbstractSTNSolver(Level debugLevel)
  {
    this.logger.setLevel(debugLevel);
  }
//
//  protected abstract void checkConsistency(TCN<N, E> model, N from, N to,
//      V newValueSet) throws InconsistencyException;

//  protected void addRelatedPaths(Collection<N> nodes,
//      Queue<Triple<N, N, N>> queue, N source, N target)
//  {
//    addRelatedPaths(nodes, queue, source, target, null);
//  }
  
//  /**
//   * This method add any triple to the queue which contains the edge from <code>source</code> to <code>target</code>
//   * @param queue
//   * @param source
//   * @param target
//   * @param ignoreList 
//   * @param model
//   */
//  protected void addRelatedPaths(Collection<N> nodes,
//      Queue<Triple<N, N, N>> queue, N source, N target, Collection<Triple<N, N, N>> ignoreList)
//  {
//    for (N via : nodes)
//    {
//
//      if (source.equals(via) || target.equals(via))
//        continue;
//            
//      // Add all possible triples including from and to to the queue
//      Triple<N, N, N> triple = new IdentityTriple<N, N, N>(source, target, via);
//      if (ignoreList == null || !ignoreList.contains(triple))
//        queue.offer(triple);
//      
//      triple = new IdentityTriple<N, N, N>(via, source, target);
//      if (ignoreList == null || !ignoreList.contains(triple))
//        queue.offer(triple);
//      
//      //queue.offer(new Triple<N, N, N>(via, target, source));
//      //queue.offer(new Triple<N, N, N>(target, source, via));
//      //queue.offer(new Triple<N, N, N>(target, via, source));
//      //queue.offer(new Triple<N, N, N>(source, via, target));
//    }
//
//  }

  protected void getInitialPaths(Collection<N> nodes,
      Queue<Path<N>> queue)
  {
    // IMPROVE improve using iterator, each pair needs to be considered only once
    for (N from : nodes)
    {
      for (N to : nodes)
      {
        if (from.equals(to))
          continue;
       
        queue.add(new Path<N>(from, to));
        queue.add(new Path<N>(to, from));
//        addRelatedPaths(nodes, queue, from, to);
      }
    }
  }
  
  protected void logChange(ChangeSet<E, V, Triangle<N>> changeSet,
      Path<N> p, N k,  final E e_ik, V oldValue)
  {
    logChange(changeSet, p, k, e_ik, oldValue, this.getClass().getSimpleName());
  }
   
  protected void logChange(
      ChangeSet<E, V, Triangle<N>> changeSet,
      Path<N> p, N k, final E e_ik, V oldValue, String reason)
  {
    assert e_ik != null;
    assert debug(Level.FINEST, "Updating Constraint %s -> %s from %s to %s",
                                      e_ik.getSourceNode(), e_ik.getTargetNode(), oldValue, e_ik.getValue());
        
     if (changeSet != null)
     {
       changeSet.logChange(e_ik,
                           oldValue,
                           e_ik.getValue(), new Triangle<N>(p.getFirst(), p.getSecond(), k), reason);
     }
  }
  

  
  /**
   * Note that this method may return invalid labels e.g., A && -A
   * @param branch1
   * @param branch2
   * @return
   */
  protected Label mergeLabels(Label branch1, Label branch2)
  {
    return Proposition.EMPTY_PROPOSITION;
  }
  


  public boolean satisfiable(Label b1, Label b2)
  {
    return b1.satisfiable(b2);
  }
  
  protected TimeDistance getTimeDistance(E edge)
  {
    V value = edge.getValue();
    if (value instanceof LabeledValueSet)
    {
      LabeledValueSet valueSet = (LabeledValueSet)value;
      assert valueSet.getLabeledValues().size() == 1;
      LabeledValue labeledValue = valueSet.getLabeledValues().iterator().next();
      return labeledValue.getValue();
    }
    else if (value instanceof LabeledValue)
    {
      LabeledValue labeledValue = (LabeledValue) value;
      return labeledValue.getValue();
    } else if (value instanceof SimpleConstraintValue)
    {
      return ((SimpleConstraintValue) value).getValue();
    }
    
    throw new IllegalArgumentException();
  }

  /**
   * Cleaning logs: <br/>
   * * Processing(.+?\r\n){1,8}\h*Ignoring.+?\r\n <br/>
   * * Hidden Change Constraint.+?\r\n
   * 
   * 
   * @param logLevel
   * @param message
   * @return <code>true</code> (Only for use with assert)
   */
  protected boolean debug(Level logLevel, String message)
  {
    boolean assertOn = false;
    assert (assertOn = true) == true; // Tautologie to check if assertions are on
    if (!assertOn)
      throw new AssertionError("Debug should be on only if assertions is on");
    
    if (logger.isLoggable(logLevel))
    {
      System.out.println(message);
      logger.log(logLevel, message);    
    }
    return true;
  }
  
  protected boolean debug(Level logLevel, String message, Object... params)
  {
    if (logger.isLoggable(logLevel))
    {
      debug(logLevel, String.format(message, params));
    }
    return true;
  }

  protected void checkConsistency(TCN<N, E> model, E edge, Triangle<N> causedBy) throws TCNInconsistencyException
  {
    N from = edge.getSourceNode();
    N to = edge.getTargetNode();
    E oppositeEdge = model.getEdge(to, from);
    
    if (oppositeEdge != null)
    {
      V value1 = oppositeEdge.getValue();
      V value2 = edge.getValue();
      if (!checkConsistency(value1, value2))
      {
        assert debug(Level.SEVERE, "Inconsistent: %s -> %s: %s < %s", from, to, value1, value2);
        //         model.setConsistent(false);
       throw new TCNInconsistencyException(String.format("Distance cannot be empty! Inconsistent: %s -> %s: %s < %s", from, to, value1, value2), new Path<N>(from, to), causedBy.getSecond(), model);
      }
    }
    
  }
  
  protected abstract boolean checkConsistency(V value1, V value2);
  
  
  protected boolean isConsistent(SimpleConstraintValue value1,
      SimpleConstraintValue value2)
  {
    TimeDistance d1 = value1.getValue();
    TimeDistance d2 = value2.getValue();
    
    boolean consistent = d1.compareTo(d2.neg()) >= 0;
    
    // If it's not consistent it's already so because of the non-contingent values.
//    if (consistent && (value1 instanceof ContingentLabeledValue || value2 instanceof ContingentLabeledValue))
//    {
//      // consider contingent values      
//    }
    
    return consistent;
  }
}
