package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.HashMap;
import java.util.Map;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.tools.DirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.Graph;
import de.aristaflow.adept2.model.timemodel.tools.Operand;
import de.aristaflow.adept2.model.timemodel.tools.WeightedEdge;

public class BellmanFord<V>
{

  public BellmanFord()
  {

  }
  
  Operand ZERO = new Operand()
  {
    
    @Override
    public int compareTo(Operand o)
    {
      return -o.sign();
    }
    
    @Override
    public Operand add(Operand a)
    {
      return a;
    }

    @Override
    public int sign()
    {
      return 0;
    }
  };
  
  public Map<V, Operand> run(DirectedGraph<V, WeightedEdge<V, Operand>> graph, V start)
  {
    Map<V, Operand> distances = new HashMap<V, Operand>();
    distances.put(start, ZERO);
    for (int i = 0; i < graph.getVertices().size(); i++)
    {
      for (V v1 : graph.getVertices())
      {
        Operand distV1 = distances.get(v1);
        
        if (distV1 == null)
          continue;
        
        for (V v2 : graph.getVertices())
        {
          WeightedEdge<V, Operand> weight = graph.getEdge(v1, v2);
          
          if (weight == null)
            continue;
          
          Operand distV2 = distances.get(v2);
          
          
          
          Operand newDist = distV1.add(weight.getWeight());
          
          if (newDist != null && (distV2 == null || newDist.compareTo(distV2) < 0))
          {
            distances.put(v2, newDist);
          }
          
        }
      }
    }
    
    return distances;
    
  }

}
