package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;

/**
   * This object is immutable
   * 
   * @author Andreas
   *
   */
public class DefaultContingentLabeledValue extends DefaultLabeledValue implements ContingentLabeledValue
{

//  public static final char UP_ARROW = '\u2191';
//  public static final char DOWN_ARROW = '\u2193';
  public static final String DOWN_LEFT = "\\";//'\u230A';
  public static final String DOWN_RIGHT = "/";//'\u230B';
  public static final String UP_LEFT = "/";//'\u2308';
  public static final String UP_RIGHT = "\\";//'\u2309';
  
  private final LabeledTimePoint trigger;
  private final TimeDistance contingentValue;
  protected final ConstraintCase constraintCase;

  
  public DefaultContingentLabeledValue(Label proposition, ConstraintCase cstnuCase, TimeDistance value, LabeledTimePoint trigger, TimeDistance contingentValue)
  {
    super(proposition, value);
    

    assert trigger != null && cstnuCase != null; // If we have a trigger we need a case
    
    
    // Contingent Constraints only make sense if the stn value is smaller than the contingent value (note that signs
    // are inverted; x <= y <=> -x >= -y). Otherwise the target of the constraint cannot happen before the trigger time 
    // point and thus the contingent part of the constraint is unnecessary. Furthermore, the == case is only useful for 
    // contingent durations in case minDur == maxDur.
    //FIXME assert value.compareTo(contingentValue) >= 0; 
    
    
    this.trigger = trigger;
    this.contingentValue = contingentValue;
    this.constraintCase = cstnuCase;
  }

  
  
  @Override
  public String toString()
  {
    String caseLabel = ConstraintCase.LOWERCASE.equals(constraintCase) ? DOWN_LEFT+this.trigger.toString().toLowerCase()+DOWN_RIGHT : UP_LEFT+this.trigger.toString().toUpperCase()+UP_RIGHT; 
    StringBuilder stringBuilder = new StringBuilder();
    stringBuilder.append(LANGLE)
    .append(this.label.toString())
    .append(", ")
    .append(getSTNValue())
    .append(", ")
    .append(caseLabel)
    .append(", ")
    .append(getContingentValue())
    .append(RANGLE);
    return stringBuilder.toString(); 
  }

  public LabeledTimePoint getTrigger()
  {
    return trigger;
  }
  
  
  @Override
  public ConstraintCase getCase()
  {
    return constraintCase;
  }


  @Override
  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    DefaultContingentLabeledValue other = (DefaultContingentLabeledValue) obj;
    if (contingentValue == null)
    {
      if (other.contingentValue != null)
        return false;
    }
    else if (!contingentValue.equals(other.contingentValue))
      return false;
    if (constraintCase != other.constraintCase) 
      return false;
    if (label == null)
    {
      if (other.label != null)
        return false;
    }
    else if (!label.equals(other.label))
      return false;
    if (trigger == null)
    {
      if (other.trigger != null)
        return false;
    }
    else if (!trigger.equals(other.trigger))
      return false;
    if (value == null)
    {
      if (other.value != null)
        return false;
    }
    else if (!value.equals(other.value))
      return false;
    return true;
  }  

  @Override
  public TimeDistance getContingentValue()
  {
    return contingentValue;
  }

  @Override
  public TimeDistance getSTNValue()
  {
    return getValue();
  }
  
  @Override
  public boolean dominates(LabeledValue other)
  {
    if (other.getLabel().subsumes(this.label))
    {
      if (this.value.compareTo(other.getValue()) <= 0)
      {
        if (other instanceof ContingentLabeledValue) // Both are contingent
        {
            ContingentLabeledValue c1 = (ContingentLabeledValue) other;
            
            if (this.getContingentValue().compareTo(c1.getContingentValue()) <= 0)
              return true;
        }
        else
        {
          return true;
        }
      }
    }

    return false;
  }
  
}