package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import de.aristaflow.adept2.model.common.defaultimplementation.SerialisablePluginDataContainer;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTimePoint;

/**
 * This object is immutable!
 * 
 * @author Andreas
 * LOW because of PluginDataContainer, Nodes are no longer immutable, but this does not change anything atm...
 */
public class DefaultLabeledNode extends SerialisablePluginDataContainer implements TCNTimePoint, LabeledTimePoint, Comparable<LabeledTimePoint>
{

  /**
   * 
   */
  private static final long serialVersionUID = 8480685101570175759L;

  private final Label label;

  private final String nodeLabel;

  private final String eventID;

  private final NodeInstance nodeInstance;

  private int cachedHashCode = Integer.MIN_VALUE;

  /**
   * A TCN Node with a label representing a specific event of a node instance
   * 
   * @param nodeInstance the node instance (i.e., node and iteration) this labeled tcn node belongs to
   * @param eventID the event of the <code>nodeInstance</code> this labeled tcn node describes
   * @param labels the set of labels representing the process instance types this labeled node belongs to
   */
  public DefaultLabeledNode(NodeInstance nodeInstance, String eventID, Label labels)
  {
    this.nodeInstance = nodeInstance;
    this.nodeLabel = nodeInstance.getNodeID()+"_{"+eventID+"}";
    this.eventID = eventID;
    this.label = labels;
  }

  @Deprecated // Should only be used for TIME BASE EVENT
  DefaultLabeledNode(String nodeLabel, NodeInstance nodeInstance, Label label)
  {
    this.nodeLabel = nodeLabel;
    this.nodeInstance = nodeInstance;
    this.eventID = nodeLabel;
    this.label = label;
  }


  public Label getLabel()
  {
    return label;
  }

  @Override
  public String toString()
  {
    return "" + nodeLabel;
  }

  public NodeInstance getNodeInstance()
  {
    return nodeInstance;
  }


  @Override
  public int hashCode()
  {
    if (cachedHashCode == Integer.MIN_VALUE)
    {
      final int prime = 53;
      cachedHashCode = 1;
      cachedHashCode = prime * cachedHashCode + (getNodeInstance().hashCode());
      cachedHashCode = prime * cachedHashCode + getEventID().hashCode();
    }
    return cachedHashCode;
  }

  @Override
  public boolean equals(Object obj)
  {
    if (obj == this)
      return true;
    
    if (obj instanceof LabeledTimePoint)
    {
      LabeledTimePoint other = (LabeledTimePoint) obj;
      boolean result = (this.getNodeInstance().equals(other.getNodeInstance()));
          result &= (this.getEventID().equals(other.getEventID())); 
      return result;
    }
    return super.equals(obj);
  }

  public String getEventID()
  {
    return eventID;
  }

  @Override
  public int compareTo(LabeledTimePoint other)
  {
    int compare = this.nodeInstance.compareTo(other.getNodeInstance());
    if (compare != 0)
      return compare;

    return this.eventID.compareTo(other.getEventID());
  }

}
