package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.util.NullArgumentException;

/**
 * This object is immutable!
 * 
 * @author Andreas
 */
public class DefaultLabeledValue implements LabeledValue
{

  public static final char LANGLE = '<'; //'\u3008'; //'<'; //
  public static final char RANGLE = '>'; //'\u3009'; //'>'; //
  
  protected final Label label;
  protected final TimeDistance value;
    
  public DefaultLabeledValue(Label label, TimeDistance value)
  {
    
    if (value == null)
      throw new NullArgumentException();
    
    this.label = label;
    this.value = value;
  }

  @Override
  public Label getLabel()
  {
    return label;
  }

  @Override
  public TimeDistance getValue()
  {
    return value;
  }
  


  @Override
  public String toString()
  {
    StringBuilder stringBuilder = new StringBuilder();
    stringBuilder.append(LANGLE).append(this.label.toString()).append(", ").append(getValue())
      //.append(", ").append("#").append(", NaN")
      .append(RANGLE);
    return stringBuilder.toString();
  }
  
  @Override
  public boolean equals(Object obj)
  {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    DefaultLabeledValue other = (DefaultLabeledValue) obj;
    if (label == null)
    {
      if (other.label != null)
        return false;
    }
    else if (!label.equals(other.label))
      return false;
    if (value == null)
    {
      if (other.value != null)
        return false;
    }
    else if (!value.equals(other.value))
      return false;
    return true;
  }  
  
  @Override
  public int hashCode()
  {
    return 31 * label.hashCode() + ((value == null) ? 0 : value.hashCode());
  }

  @Override
  public boolean dominates(LabeledValue other)
  {
    if (other.getLabel().subsumes(this.label))
    {
      if (this.value.compareTo(other.getValue()) <= 0)
      {
        if (other instanceof ContingentLabeledValue)
        {
            ContingentLabeledValue c1 = (ContingentLabeledValue) other;
            
            if (this.getValue().compareTo(c1.getContingentValue()) <= 0)
              return true;
        }
        else
        {
          return true;
        }
      }
    }

    return false;
  }

}
