package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;

import de.aristaflow.adept2.model.timemodel.tcn.ChangeableTCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTimePoint;

public class ReductionProjection<N extends TCNTimePoint, E extends TCNEdge<N, ?>> extends AbstractChangeableProjection<N, E, E> implements
    ChangeableTCN<N, E>
{

  private final Collection<N> timePoints;

  public ReductionProjection(ChangeableTCN<N, E> tcn, Collection<N> timePoints)
  {
    super(tcn, true);
    this.timePoints = new HashSet<N>(timePoints);
  }
  
  @Override
  public Collection<N> getTimePoints()
  {
    return Collections.unmodifiableCollection(timePoints);
  }
  

  @Override
  protected E propagateUpdate(E edge)
  {
    return getBaseTCN().updateConstraint(edge);
  }

  @Override
  protected E transformEdge(N from, N to, E originalEdge)
  {
    if (timePoints.contains(from) && timePoints.contains(to))
      return originalEdge;
    return null;
  }

  @Override
  public TCN<N, E> clone()
  {
    throw new UnsupportedOperationException();
  }

  public void removeTimePoint(N timePoint)
  {
    this.timePoints.remove(timePoint);
  }

  public N insertTimePoint(N timePoint)
  {
    throw new UnsupportedOperationException();
  }

}
