package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.Collection;

import org.apache.commons.configuration.Configuration;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.globals.ProcessConstants.NodeType;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableSTN;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.JoinNodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.STN;
import de.aristaflow.adept2.model.timemodel.tcn.SimpleConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.util.NullArgumentException;

public class STNTemplateTransformation extends AbstractTemplateTransformation<STNEdge>
{

  public STNTemplateTransformation(Configuration configuration, TemporalTemplate template)
  {
    super(configuration, template);
  }
  
  public STNTemplateTransformation(Configuration configuration, TemporalTemplate template, int lookAhead)
  {
    super(configuration, template, lookAhead);
  }
  
  public STNTemplateTransformation(TemporalTemplate template)
  {
    this(null, template);
  }
  
  public STNTemplateTransformation(TemporalTemplate template, int lookAhead)
  {
    this(null, template, lookAhead);
  }

  @Override
  protected ChangeableSTN<LabeledTimePoint, STNEdge> createTCN(Collection<LabeledTimePoint> nodes, Collection<STNEdge> edges,
      Collection<NodeMapping<LabeledTimePoint>> nodeMapping)
  {
    // FIXME
    return new SimpleSTN(nodes, edges);
  }
  
  @Override
  public STN<LabeledTimePoint, STNEdge> transform()
  {
    return (STN<LabeledTimePoint, STNEdge>) super.transform();
  }

  @Override
  protected STNEdge[] createTCNEdge(LabeledTimePoint from, LabeledTimePoint to,
      Label branches, TimeDistance minValue, TimeDistance maxValue, boolean controlEdge)
  {
    if (minValue == null)
      throw new NullArgumentException();
    
    if (maxValue == null)
      throw new NullArgumentException();
      
    STNEdge minEdge;
    STNEdge maxEdge;


    minEdge = new STNEdge(to, from, new DefaultConstraintValue(minValue.neg()), controlEdge, false, FlowDirection.BACKWARD);
    minEdge = addEdge(minEdge);

    maxEdge = new STNEdge(from, to, new DefaultConstraintValue(maxValue), controlEdge, false, FlowDirection.FORWARD);
    maxEdge = addEdge(maxEdge);


    return new STNEdge[] { minEdge, maxEdge };
  }

  @Override
  protected NodeMapping<LabeledTimePoint> transformActivity(int nodeID, IterationChain iterationChain,
      Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_NORMAL)
    {
      throw new IllegalArgumentException(""+template.getNodeType(nodeID));
    }
    
    int topologicalID = template.getNodeTopologicalID(nodeID);
    
    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);
    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);
    
    addNode(A_S);
    addNode(A_E);
    
    TCNActivityDuration duration = getActivityDuration(nodeID);
    
    STNEdge minEdge = new STNEdge(A_E, A_S, new DefaultConstraintValue(duration.minDuration.neg()), true, false, FlowDirection.BACKWARD);
    minEdge = addEdge(minEdge);
    
    STNEdge maxEdge = new STNEdge(A_S, A_E, new DefaultConstraintValue(duration.maxDuration), true, false, FlowDirection.FORWARD);
    maxEdge = addEdge(maxEdge);

    NodeMapping<LabeledTimePoint> node = new NodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, A_S, null, A_E);
    
    processActivityValidityPeriod(nodeID, node);
    
    return addNodeMapping(node);
  }
  
  @Override
  protected JoinNodeMapping<LabeledTimePoint> transformANDJoin(int nodeID, IterationChain iterationChain, Label branches)
  {
    if (template.getNodeType(nodeID) != NodeType.NT_AND_JOIN)
    {
      throw new IllegalArgumentException("" + template.getNodeType(nodeID));
    }

    

    int topologicalID = template.getNodeTopologicalID(nodeID);

    NodeInstance nodeInstance = new NodeInstance(nodeID, iterationChain, topologicalID);

    LabeledTimePoint A_S = createTCNNode(nodeInstance, "S", branches);
    LabeledTimePoint A_E = createTCNNode(nodeInstance, "E", branches);
    
    

    addNode(A_S);
    addNode(A_E);

    createTCNEdge(A_S, A_E, branches, STRUCTURE_NODE_MIN_DURATION, STRUCTURE_NODE_MAX_DURATION, true);

//    int[] preds = template.getPredByEdgeType(nodeID, EdgeType.ET_CONTROL);
//    LabeledTimePoint[] nodes = new LabeledTimePoint[preds.length];
//    Arrays.fill(nodes, A_S);
    return addNodeMapping(new JoinNodeMapping<LabeledTimePoint>(getNodeName(nodeID), nodeInstance, null, A_S, null, A_E));
  }
  

  @Override
  protected Label createInitialBranch()
  {
    return Proposition.EMPTY_PROPOSITION;
  }

  @Override
  protected STNEdge mergeEdges(STNEdge e1, STNEdge e2)
  {
    SimpleConstraintValue value = e1.getValue().getValue().compareTo(e2.getValue().getValue()) <= 0 ? e1.getValue() : e2.getValue();
    return new STNEdge(e1.getSourceNode(), e2.getTargetNode(), value, 
                       e1.isControlEdge() || e2.isControlEdge(),
                       e1.isImplicit() && e2.isImplicit(),
                       e1.getFlowDirection() == e2.getFlowDirection() ? e1.getFlowDirection() : FlowDirection.UNKOWN);
  }

}
