package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTN;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.AbstractCSTNProjection;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;

public class ScenarioProjection extends AbstractCSTNProjection
{

  private final Label scenario;
  /**
   * Mapping from source to projection
   */
  private Map<LabeledTimePoint, LabeledTimePoint> nodeMappingPrj = new HashMap<LabeledTimePoint, LabeledTimePoint>();
  /**
   * Mapping from projection to source
   */
  private Map<LabeledTimePoint, LabeledTimePoint> nodeMappingSrc = new HashMap<LabeledTimePoint, LabeledTimePoint>();



  public ScenarioProjection(ChangeableCSTN cstn, Label scenario)
  {
    super(cstn, false);
    this.scenario = scenario;
  }
  
  

  @Override
  public List<ChoiceVariable> getObservations()
  {
    List<ChoiceVariable> baseObservations = super.getObservations();
    Set<ChoiceVariable> observations = new HashSet<ChoiceVariable>((int)(baseObservations.size()*1.4));
    
    for (ChoiceVariable var : baseObservations)
    {
      if (var.getDecisionTimePoint().getLabel().satisfiable(scenario))
      {
        observations.add(var);
      }
    }
    
    return new ArrayList<ChoiceVariable>(observations);
  }



//  @Override
//  public Collection<NodeMapping<LabeledTimePoint>> getNodeMapping()
//  {
//    // Not supported by this projection
//    return Collections.emptyList();
//  }

  @Override
  protected LabeledTimePoint transformNode(LabeledTimePoint node)
  {
    if (!node.getLabel().satisfiable(scenario))
    {
      return null;
    }
    
    LabeledTimePoint newNode = nodeMappingPrj.get(node);
    
    if (newNode != null)
      return newNode;
    

    
    if (node instanceof DefaultObservationTimePoint)
    {
      DefaultObservationTimePoint obsNode = (DefaultObservationTimePoint) node;
      if (scenario.containsVariable(obsNode.getObservation()))
      {
        // Observation has been made, so node is downgraded to a LabeledNode
        newNode = new DefaultLabeledNode(node.getNodeInstance(), node.getEventID(), scenarioLabel(node.getLabel()));
      }
      else
      {
        Map<Long, ObservationBinding> edgeCodeMapping = obsNode.getEdgeCodeMapping();
        long[] edgeCodes = new long[edgeCodeMapping.size()];
        int[] observationIDs = new int[edgeCodeMapping.size()];
        int i = 0;
        for (Entry<Long, ObservationBinding> entry : edgeCodeMapping.entrySet())
        {
          edgeCodes[i] = entry.getKey().longValue();
          observationIDs[i] = entry.getValue().getValue();
          i++;
        }
        
        newNode = new DefaultObservationTimePoint(obsNode.getNodeInstance(), obsNode.getEventID(), scenarioLabel(obsNode.getLabel()), 
                                   edgeCodes, observationIDs);
      }
    }
    else
    {
      if (node == getBaseTCN().getTimeBaseEvent())
        newNode = node; // Time Base Event is always the same...
      else
        newNode = new DefaultLabeledNode(node.getNodeInstance(), node.getEventID(), scenarioLabel(node.getLabel()));
    }
    
    nodeMappingPrj.put(node, newNode);
    nodeMappingSrc.put(newNode, node);
    
    return newNode;
  }
  
  



  @Override
  protected LabeledTimePoint getBaseNode(LabeledTimePoint node)
  {
    return nodeMappingSrc.get(node);
  }



  protected Label scenarioLabel(Label label)
  {
    return label.removeLabel(scenario);
  }



//  @Override
//  public NodeMapping<LabeledTimePoint> getNodeMapping(int nodeId, IterationChain iterationChain)
//  {
//    // Not supported by this projection
//    return null;
//  }
  
  



  @Override
  public LabeledEdge getEdge(LabeledTimePoint from, LabeledTimePoint to)
  {
    assert nodeMappingSrc.containsKey(from);
    assert nodeMappingSrc.containsKey(to);
    return super.getEdge(from, to);
  }



  @Override
  protected LabeledEdge transformEdge(LabeledTimePoint from, LabeledTimePoint to, LabeledEdge originalEdge)
  {
    // Nodes have been projected
    assert from.getLabel().satisfiable(scenario);
    assert to.getLabel().satisfiable(scenario);
    
    Collection<LabeledValue> originalValues = originalEdge.getValue().getLabeledValues();
    //List<LabeledValue> values = new ArrayList<LabeledValue>(originalValues.length);
    Map<Label, LabeledValue> labels = new HashMap<Label, LabeledValue>((int)(originalValues.size() * 1.4));
    for (LabeledValue value : originalValues)
    {
      if (value.getLabel().satisfiable(scenario))
      {
        LabeledValue newValue = updateLabel(value, scenario);
        if (newValue != null)
          DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, newValue);
      }
    }
    
    if (labels.size() == 0)
      return null; // No constraint remaining.
    
//    if (values.size() == originalValues.length)
//      return originalEdge; // No constraint removed... // should we consider updated labels?


    return new DefaultLabeledEdge(from, to, new DefaultLabeledValueSet(labels.values()), originalEdge.isControlEdge(),
                           originalEdge.isImplicit(), originalEdge.getFlowDirection());
  }



  protected LabeledValue updateLabel(LabeledValue value, Label scenario)
  {
    LabeledValue newLabeledValue;
    Label newLabel = scenarioLabel(value.getLabel());
    if (value instanceof ContingentLabeledValue)
    {
      ContingentLabeledValue cl_XY = (ContingentLabeledValue)value;
      newLabeledValue = new DefaultContingentLabeledValue(newLabel, cl_XY.getCase(), cl_XY.getSTNValue(), cl_XY.getTrigger(), cl_XY.getContingentValue());
    }
    else
    {
      newLabeledValue = new DefaultLabeledValue(newLabel, value.getValue());
    }
    return newLabeledValue;
  }
  
  public LabeledTimePoint insertTimePoint(LabeledTimePoint timePoint)
  {
    throw new UnsupportedOperationException();
  }


  public void removeTimePoint(LabeledTimePoint timePoint)
  {
    throw new UnsupportedOperationException();
  }

}
