package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.Collection;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.CompoundConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.SimpleConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTimePoint;
import de.aristaflow.adept2.model.timemodel.tools.DefaultWeightedEdge;
import de.aristaflow.adept2.model.timemodel.tools.DirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.Graph;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;
import de.aristaflow.adept2.model.timemodel.tools.Operand;
import de.aristaflow.adept2.model.timemodel.tools.WeightedEdge;

public class SemiReduciblePathsGraph<N extends TCNTimePoint>
    implements DirectedGraph<N, WeightedEdge<N, Operand>>
{

  public class SemiReducibleOperand implements Operand
  {

    public SemiReducibleOperand(LabeledValueSet labeledValueSet)
    {
      // TODO Auto-generated constructor stub
    }

    @Override
    public int compareTo(Operand o)
    {
      // TODO Auto-generated method stub
      return 0;
    }

    @Override
    public Operand add(Operand a)
    {
      // TODO Auto-generated method stub
      return null;
    }

    @Override
    public int sign()
    {
      // TODO Auto-generated method stub
      return 0;
    }

  }

  private TCN<N, ? extends LabeledEdge> tcn;

  public SemiReduciblePathsGraph(TCN<N, ? extends LabeledEdge> tcn)
  {
    this.tcn = tcn;
  }

  @Override
  public Collection<N> getVertices()
  {
    return tcn.getTimePoints();
  }

  @Override
  public Collection<WeightedEdge<N, Operand>> getEdges()
  {
    throw new UnsupportedOperationException();
  }

  @Override
  public WeightedEdge<N, Operand> getEdge(N v1, N v2)
  {
    LabeledEdge edge = tcn.getEdge(v1, v2);
    
    if (edge == null)
      return null;

    TimeDistance weight = getValue(edge, false);
    
    TimeDistance contValue = getContingentValue(edge);
    
    if (contValue != null)
      weight = contValue;
    
    if (weight.isInfinity())
      return null;

    return new DefaultWeightedEdge<N, Operand>(v1, v2, new SemiReducibleOperand(edge.getValue())) ;
  }

  public TimeDistance getValue(LabeledEdge edge, boolean maxValue)
  {
    int sgn = maxValue ? 1 : -1;

    TimeDistance timeDistance = null;

    SimpleConstraintValue[] values = edge.getValue().getValues();

    for (SimpleConstraintValue constraintValue : values)
    {
      if (timeDistance == null)
      {
        timeDistance = constraintValue.getValue();
      }
      else
      {
        if (constraintValue.getValue().compareTo(timeDistance) >= sgn)
        {
          timeDistance = constraintValue.getValue();
        }
      }

    }

    return timeDistance;
  }
  
  public TimeDistance getContingentValue(LabeledEdge edge)
  {
    TimeDistance contValue = null;
    for (SimpleConstraintValue v : edge.getValue().getValues())
    {
      if (v instanceof ContingentLabeledValue)
      {
        ContingentLabeledValue contingentLabeledValue = (ContingentLabeledValue) v;
        TimeDistance contingentValue = contingentLabeledValue.getContingentValue();
        if (contValue == null || contValue.compareTo(contingentValue) > 0)
        {
          contValue = contingentValue;
        }
        
      }
    }

    return contValue;
  }

  @Override
  public void visit(GraphVisitor<N, WeightedEdge<N, Operand>> visitor)
  {
    throw new UnsupportedOperationException();
  }

}
