package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.ArrayList;
import java.util.List;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTN;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.AbstractCSTNProjection;

public class SinkDijkstra extends AbstractCSTNProjection
{

  public SinkDijkstra(ChangeableCSTN cstn)
  {
    super(cstn, true);
  }

  @Override
  protected LabeledEdge transformEdge(LabeledTimePoint from, LabeledTimePoint to, LabeledEdge originalEdge)
  {
    List<LabeledValue> values = new ArrayList<LabeledValue>();
    for (LabeledValue value : originalEdge.getValue().getLabeledValues())
    {
      TimeDistance h_X = h(from, value.getLabel());
      TimeDistance h_Y = h(to, value.getLabel());
      
      TimeDistance u = h_X.add(value.getValue()).sub(h_Y);
      
      assert u.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0;
      
      LabeledValue newValue = new DefaultLabeledValue(value.getLabel(), u);
      values.add(newValue);
    }
    
    return new DefaultLabeledEdge(from, to, new DefaultLabeledValueSet(values));
  }
  
  private TimeDistance h(LabeledTimePoint X, Label label)
  {
    LabeledTimePoint Z = getBaseTCN().getTimeBaseEvent();
    
    LabeledEdge labeledEdge = getBaseTCN().getEdge(Z, X);
    
    return scenarioProjection(labeledEdge, label);
  }

  private TimeDistance scenarioProjection(LabeledEdge edge, Label scenario)
  {
    TimeDistance min = DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
    for (LabeledValue value : edge.getValue().getLabeledValues())
    {
      if (value.getLabel().satisfiable(scenario))
      {
        
        if (value.getValue().compareTo(min) < 0)
        {
          min = value.getValue();
        }
      }
    }

    return min;
  }

  @Override
  protected LabeledEdge propagateUpdate(LabeledEdge edge)
  {
    List<LabeledValue> values = new ArrayList<LabeledValue>();
    LabeledTimePoint from = edge.getSourceNode();
    LabeledTimePoint to = edge.getTargetNode();
    for (LabeledValue value : edge.getValue().getLabeledValues())
    {
      TimeDistance h_X = h(from, value.getLabel());
      
      TimeDistance u = h_X.sub(value.getValue());
      
//      assert u.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) >= 0;
      
      LabeledValue newValue = new DefaultLabeledValue(value.getLabel(), u);
      values.add(newValue);
    }
    

    return getBaseTCN().updateConstraint(new DefaultLabeledEdge(from, to, new DefaultLabeledValueSet(values)));
  }

}
