package de.aristaflow.adept2.model.timemodel.tcn.stnuImpl;

import de.aristaflow.adept2.model.timemodel.tcn.ChangeableSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.STN;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.DefaultCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractChangeableProjection;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.STNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.SimpleSTN;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;

public class STNUAllMaxProjection extends AbstractChangeableProjection<LabeledTimePoint, LabeledEdge, STNEdge> implements
    STN<LabeledTimePoint, STNEdge>, Cloneable, ChangeableSTN<LabeledTimePoint, STNEdge>
{

  private final boolean allMax;

  public STNUAllMaxProjection(ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> stnu)
  {
    this(stnu, true);
  }
  
  public STNUAllMaxProjection(ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> stnu, boolean allMax)
  {
    super(stnu, false);
    this.allMax = allMax;
  }

  @Override
  public DefaultCSTNU getBaseTCN()
  {
    return (DefaultCSTNU) super.getBaseTCN();
  }

   @Override
  public STNEdge getEdge(LabeledTimePoint from, LabeledTimePoint to)
  {
    STNEdge edge = internalGetEdge(from, to);
    if (edge != null)
      return edge;
    
    LabeledEdge labeledEdge = getBaseTCN().getEdge(from, to);
    
    if (labeledEdge == null)
      return null; 
    
    return transformEdge(from, to, labeledEdge);
  }

  @Override
  protected STNEdge transformEdge(LabeledTimePoint from, LabeledTimePoint to, LabeledEdge labeledEdge)
  {
    if (labeledEdge.isContingentDuration())
    {
      DefaultConstraintValue value = null;
      
      assert labeledEdge.getValue().getLabeledValues().size() == 1;
      
      ContingentLabeledValue cl = (ContingentLabeledValue) labeledEdge.getValue().getLabeledValues().iterator().next();

      if (allMax)
        value = toMax(from, to, labeledEdge, cl);
      else
        value = toMin(from, to, labeledEdge, cl);
      
      return new STNEdge(from, to, value, labeledEdge.isControlEdge(), labeledEdge.isImplicit(),
                         labeledEdge.getFlowDirection());
    }
    
    return new STNEdge(from, to, new DefaultConstraintValue(labeledEdge.getValue().getLabeledValues().iterator().next().getValue()),
                       labeledEdge.isControlEdge(), labeledEdge.isImplicit(),
                       labeledEdge.getFlowDirection());
  }

  private DefaultConstraintValue toMin(LabeledTimePoint from, LabeledTimePoint to, LabeledEdge labeledEdge,
      ContingentLabeledValue cl)
  {
    DefaultConstraintValue value;
    if (ConstraintCase.LOWERCASE.equals(cl.getCase()))
    {
      value = new DefaultConstraintValue(cl.getContingentValue());
    }
    else
    {
      LabeledEdge upperCaseEdge = getBaseTCN().getEdge(to, from);
      assert upperCaseEdge != null : "No opposite edge for "+labeledEdge;
      assert upperCaseEdge.getValue().getLabeledValues().size() == 1;
      LabeledValue l = upperCaseEdge.getValue().getLabeledValues().iterator().next();
      assert l instanceof ContingentLabeledValue;
      
      ContingentLabeledValue cu = (ContingentLabeledValue) l;
      assert ConstraintCase.LOWERCASE.equals(cu.getCase());
      
      value = new DefaultConstraintValue(cu.getContingentValue().neg());
    }
    return value;
  }

  private DefaultConstraintValue toMax(LabeledTimePoint from, LabeledTimePoint to, LabeledEdge labeledEdge,
      ContingentLabeledValue cl)
  {
    DefaultConstraintValue value;
    if (ConstraintCase.UPPERCASE.equals(cl.getCase()))
    {
      value = new DefaultConstraintValue(cl.getContingentValue());
    }
    else
    {
      LabeledEdge upperCaseEdge = getBaseTCN().getEdge(to, from);
      assert upperCaseEdge != null : "No opposite edge for "+labeledEdge;
      assert upperCaseEdge.getValue().getLabeledValues().size() == 1;
      LabeledValue l = upperCaseEdge.getValue().getLabeledValues().iterator().next();
      assert l instanceof ContingentLabeledValue;
      
      ContingentLabeledValue cu = (ContingentLabeledValue) l;
      assert ConstraintCase.UPPERCASE.equals(cu.getCase());
      
      value = new DefaultConstraintValue(cu.getContingentValue().neg());
    }
    return value;
  }

  @Override
  protected STNEdge propagateUpdate(STNEdge edge)
  {
    throw new IllegalStateException("Propagation is not supported by this projection!");
  }

  public STN<LabeledTimePoint, STNEdge> clone()
  {
    return new SimpleSTN(getTimePoints(), getEdges());
  }

}
