package de.aristaflow.adept2.model.timemodel.tcn.xml;

import java.util.Collection;
import java.util.Comparator;
import java.util.Map.Entry;
import java.util.SortedSet;
import java.util.TreeSet;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import de.aristaflow.adept2.model.common.XMLExportTools;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.CSTN;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain.Iteration;
import de.aristaflow.adept2.model.timemodel.tcn.JoinNodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.STNU;
import de.aristaflow.adept2.model.timemodel.tcn.SplitNodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValueWithContingency;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultObservationTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ObservationBinding;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.Proposition;

public class CSTNUXMLExportTools
{
  public static final String NS = null;

  public Element createCSTNElement(Document document, TCN<LabeledTimePoint, LabeledEdge> cstn)
  { 
    Element rootElement = document
        .createElementNS(NS, "cstnu");
    if (cstn instanceof CSTNU)
    {
      rootElement.setAttribute("type", "cstnu");
    }
    else if (cstn instanceof CSTN)
    {
      rootElement.setAttribute("type", "cstn");
    }
    else if (cstn instanceof STNU)
    {
      rootElement.setAttribute("type", "stnu");
    }
    else
    {
      throw new IllegalArgumentException();
    }
    
    rootElement.setAttributeNS(null, "version",
                                     "1.0");
    
//    rootElement.setAttributeNS(NS, "templateID", cstn.getTemplateID().toString());


    // list of nodes
    Element nodesElement = addElement(document, rootElement,
                                      "tcnNodes");

    
    fillNodesElement(document, nodesElement, cstn.getTimePoints());
    
    // list of nodes mapping
    //Element nodeMappingElement = 
        addElement(document, rootElement,
                                      "nodeMappings");

    
//    fillNodeMappingElement(document, nodeMappingElement, cstn.getNodeMapping());

    // list of edges
    Element edgesElement = addElement(document, rootElement,
                                        "tcnEdges");

    fillEdgesElement(document, edgesElement, cstn.getEdges());

    return rootElement;
  }

  private void fillEdgesElement(Document document, Element edgesElement,
      Collection<LabeledEdge> tcnEdges)
  {
    for (LabeledEdge edge : tcnEdges)
    {
      Element edgeElement = createEdgeElement(document, edge);
      edgesElement.appendChild(edgeElement);
    }
  }

  private Element createEdgeElement(Document document, LabeledEdge edge)
  {
    Element edgeElement = document
        .createElementNS(NS, "labeledEdge");
    
    Element sourceNode = addElement(document, edgeElement, "sourceNode");
    fillNodeReference(document, sourceNode, edge.getSourceNode());
    
    Element targetNode = addElement(document, edgeElement, "targetNode");
    fillNodeReference(document, targetNode, edge.getTargetNode());
    
    edgeElement.setAttributeNS(NS, "controlEdge", ""+edge.isControlEdge());
    edgeElement.setAttributeNS(NS, "implicit", ""+edge.isImplicit());
    edgeElement.setAttributeNS(NS, "flowDirection", edge.getFlowDirection().toString());
    
    Element valueNode = addElement(document, edgeElement, "value");
    fillLabeledValueSetElement(document, valueNode, edge.getValue());
    
   
    return edgeElement;
  }

  private void fillLabeledValueSetElement(Document document, Element valueNodeElement,
      LabeledValueSet values)
  {
    for (LabeledValue value : values.getLabeledValues())
    {
      Element valueElement = createLabeledValueElement(document, value);
      valueNodeElement.appendChild(valueElement);
    }
    
  }

  private Element createLabeledValueElement(Document document, 
      LabeledValue value)
  {
    Element valueElement;
    if (value instanceof ContingentLabeledValueWithContingency)
    {
      valueElement  = createContingentLabeledValueWithContingencyElement(document, (ContingentLabeledValueWithContingency) value);
    }
    else if (value instanceof ContingentLabeledValue)
    {
      valueElement = createContingentLabeledValueElement(document, (ContingentLabeledValue) value);
    }
    else
    {
      valueElement = document.createElementNS(NS, "labeledValue");
      fillLabeledValueElement(document, valueElement, value);
    }
  
    //
    
    return valueElement;
  }

  private Element createContingentLabeledValueWithContingencyElement(Document document, ContingentLabeledValueWithContingency value)
  {
    Element valueElement = createContingentLabeledValueElement(document, value);
    Element contingencyValueElement = addElement(document, valueElement, "contingencyValue");
    fillTimeDistanceElement(document, contingencyValueElement, value.getContingency());
    
    return valueElement;
  }

  private Element createContingentLabeledValueElement(Document document, ContingentLabeledValue value)
  {
    Element valueElement;
    valueElement = document.createElementNS(NS, "contingentLabeledValue");
    fillLabeledValueElement(document, valueElement, value);
    
    valueElement.setAttributeNS(NS, "constraintCase", value.getCase().toString());
    
    Element triggerElement = addElement(document, valueElement, "trigger");
    fillNodeReference(document, triggerElement, value.getTrigger());
    
    Element timeDistance = addElement(document, valueElement, "contingentValue");
    fillTimeDistanceElement(document, timeDistance, value.getContingentValue());
   
    return valueElement;
  }

  private void fillLabeledValueElement(Document document, Element valueElement,
      LabeledValue value)
  {
    valueElement.appendChild(createLabelElement(document, value.getLabel()));
    Element timeDistance = addElement(document, valueElement, "value");
    fillTimeDistanceElement(document, timeDistance, value.getValue());

  }

  private void fillTimeDistanceElement(Document document, Element timeDistanceElement,
      TimeDistance value)
  {
    timeDistanceElement.setAttributeNS(NS, "months", ""+value.getMonths());
    timeDistanceElement.setAttributeNS(NS, "days", ""+value.getDays());
    timeDistanceElement.setAttributeNS(NS, "hours", ""+value.getHours());
    timeDistanceElement.setAttributeNS(NS, "minutes", ""+value.getMinutes());
    timeDistanceElement.setAttributeNS(NS, "seconds", ""+value.getSeconds());
    timeDistanceElement.setAttributeNS(NS, "milliseconds", ""+value.getMilliseconds());   
  }

  private void fillNodeMappingElement(Document document,
      Element nodeMappingElement,
      Collection<NodeMapping<LabeledTimePoint>> nodeMapping)
  {
    for (NodeMapping<LabeledTimePoint> mapping : nodeMapping)
    {
      Element mappingElement;
      if (mapping.getClass().equals(SplitNodeMapping.class))
      {
        mappingElement = createSplitNodeMappingElement(document, (SplitNodeMapping<LabeledTimePoint>)mapping);
      }
      else if (mapping.getClass().equals(JoinNodeMapping.class))
      {
        mappingElement = createJoinNodeMappingElement(document, (JoinNodeMapping<LabeledTimePoint>)mapping);
      }
      else
      {
        mappingElement = createNodeMappingElement(document, mapping);
      }
      nodeMappingElement.appendChild(mappingElement);
    }
    
  }

  private Element createNodeMappingElement(Document document,
      NodeMapping<LabeledTimePoint> mapping)
  {
    Element nodeMappingElement = document.createElementNS(NS, "nodeMapping");
    fillNodeMappingElement(document, nodeMappingElement, mapping);

    return nodeMappingElement;
  }

  private void fillNodeMappingElement(Document document,
      Element nodeMappingElement, NodeMapping<LabeledTimePoint> mapping)
  {
    nodeMappingElement.setAttributeNS(NS, "label", mapping.getLabel());
    
    Element nodeInstanceElement = createNodeInstanceElement(document, mapping.getNodeInstance());
    nodeMappingElement.appendChild(nodeInstanceElement);
    
    Element startNodeElement = document.createElementNS(NS, "startNodeReference");
    fillNodeReference(document, startNodeElement, mapping.getStartNode());
    nodeMappingElement.appendChild(startNodeElement);

    Element intermendiateNodesElement = document.createElementNS(NS, "intermediateNodes");
    
    for (LabeledTimePoint n : mapping.getIntermediateEvents())
    {
      Element nodeElement = document.createElementNS(NS, "nodeReference");
      fillNodeReference(document, nodeElement, n);
      intermendiateNodesElement.appendChild(nodeElement);
    }
    
    nodeMappingElement.appendChild(intermendiateNodesElement);
    
    
    Element endNodeElement = document.createElementNS(NS, "endNodeReference");
    fillNodeReference(document, endNodeElement, mapping.getEndNode());
    nodeMappingElement.appendChild(endNodeElement);    
  }

  private void fillNodeReference(Document document, Element nodeReferenceElement,
      LabeledTimePoint node)
  {
    fillNodeInstanceElement(document, nodeReferenceElement, node.getNodeInstance());
    nodeReferenceElement.setAttributeNS(NS, "eventID", node.getEventID());
  }

  private Element createJoinNodeMappingElement(Document document,
      JoinNodeMapping<LabeledTimePoint> mapping)
  {
    Element nodeMappingElement = document.createElementNS(NS, "joinNodeMapping");
    fillNodeMappingElement(document, nodeMappingElement, mapping);
    
    Element prefixNodesElement = document.createElementNS(NS, "prefixNodes");
    
    LabeledTimePoint[] prefixNodes = mapping.getPrefixNodes();
    NodeInstance[] branches = mapping.getBranches();
    for (int i = 0; i < prefixNodes.length; i++)
    {
      LabeledTimePoint n = prefixNodes[i];
      NodeInstance branchCon = branches[i];
      Element nodeElement = document.createElementNS(NS, "nodeReference");
      fillNodeReference(document, nodeElement, n);
      prefixNodesElement.appendChild(nodeElement);
    }
    
    // FIXME export branches!
    
    nodeMappingElement.appendChild(prefixNodesElement);

    return nodeMappingElement;
  }

  private Element createSplitNodeMappingElement(Document document,
      SplitNodeMapping<LabeledTimePoint> mapping)
  {
    Element nodeMappingElement = document.createElementNS(NS, "splitNodeMapping");
    fillNodeMappingElement(document, nodeMappingElement, mapping);
    
    Element prefixNodesElement = document.createElementNS(NS, "suffixNodes");
    
    for (LabeledTimePoint n : mapping.getSuffixNodes())
    {
      Element nodeElement = document.createElementNS(NS, "nodeReference");
      fillNodeReference(document, nodeElement, n);
      prefixNodesElement.appendChild(nodeElement);
    }
    
    nodeMappingElement.appendChild(prefixNodesElement);

    return nodeMappingElement;
  }

  private void fillNodesElement(Document document, Element nodesElement,
      Collection<LabeledTimePoint> tcnNodes)
  {
    // Observation nodes generating a particular observation need to be stored before any node depending on this
    // observation.
    SortedSet<LabeledTimePoint> sortedSet = new TreeSet<LabeledTimePoint>(new Comparator<LabeledTimePoint>()
    {

      @Override
      public int compare(LabeledTimePoint node1, LabeledTimePoint node2)
      {
        if (node1.equals(node2))
          return 0;
        
        if (node1 instanceof DefaultObservationTimePoint || node2 instanceof DefaultObservationTimePoint)
        {
          if (node1 instanceof DefaultObservationTimePoint && node2 instanceof DefaultObservationTimePoint)
          {
            Proposition label1 = (Proposition) node1.getLabel();
            Proposition label2 = (Proposition) node2.getLabel();
            
            if (label1.satisfiable(label2))
            {
              // Both may occur in the same instance, i.e., they may be dependent on each other
              
              for (ObservationBinding ob1 : label1.getObservations())
              {
                if (ob1.getVariable().getDecisionTimePoint().equals(node2))
                {
                  // label1 contains observation of node2 -> node2 needs to come first
                  return +1;
                }
              }
              
              for (ObservationBinding ob2 : label2.getObservations())
              {
                if (ob2.getVariable().getDecisionTimePoint().equals(node1))
                {
                  // label2 contains observation of node1 -> node1 needs to come first
                  return -1;
                }
              }
            }
          }
          // Observation Nodes come first
          else if (node1 instanceof DefaultObservationTimePoint)
          {
            return -1;
          }
          else // if (o2 instanceof ObservationNode)
          {
            return +1;
          }
        }
        int sgn = node1.hashCode() - node2.hashCode() > 0 ? 1 : -1;
        return sgn; // Any order is fine...
      }
    });
    sortedSet.addAll(tcnNodes);
    
    assert sortedSet.size() == tcnNodes.size();
    
    for(LabeledTimePoint node : sortedSet)
    {
      Element nodeElement;
      if (node.getClass().equals(DefaultObservationTimePoint.class))
      {
        nodeElement = createObservationNodeElement(document, (DefaultObservationTimePoint)node);
      }
      else
      {
        nodeElement = createNodeElement(document, node);
      }
      nodesElement.appendChild(nodeElement);
    }
  }

  private Element createObservationNodeElement(Document document,
      DefaultObservationTimePoint node)
  {
    Element nodeElement = document.createElementNS(NS, "observationNode");
    fillNodeElement(document, node, nodeElement);
    
//    Observation observationLabel = node.getObservation();
//    Element observationLabelElement = document.createElementNS(NS, "observationLabel");
//    nodeElement.appendChild(observationLabelElement);
//    observationLabelElement.appendChild(createObservationElement(document, observationLabel));
    
    Element decisionLabelsElement = document.createElementNS(NS, "decisionLabels");    
    for (Entry<Long, ObservationBinding> ds : node.getEdgeCodeMapping().entrySet())
    {
      Element decisionLabel = document.createElementNS(NS, "decisionLabel");
      decisionLabelsElement.appendChild(decisionLabel);
      decisionLabel.setAttributeNS(NS, "edgeCode", String.valueOf(ds.getKey()));
      
      ObservationBinding observation = ds.getValue();
      decisionLabel.setAttributeNS(NS, "observationCode", String.valueOf(observation.getValue()));
    }
    nodeElement.appendChild(decisionLabelsElement);
    
    return nodeElement;
  }


  private Element createObservationBindingElement(Document document, ObservationBinding observation)
  {
    Element literalGroupElement = document.createElementNS(NS, "observation");

    
    Element splitNode = document.createElement("observationSource");
    fillNodeReference(document, splitNode, observation.getVariable().getDecisionTimePoint());
    literalGroupElement.appendChild(splitNode);
    
    literalGroupElement.setAttributeNS(NS, "observationCode", String.valueOf(observation.getValue()));
    literalGroupElement.setAttributeNS(NS, "edgeCode", String.valueOf(observation.getEdgeCode()));
    
    return literalGroupElement;
  }

  private Element createNodeElement(Document document, LabeledTimePoint node)
  {
    Element nodeElement = document.createElementNS(NS, "node");
    fillNodeElement(document, node, nodeElement);
    
    return nodeElement;
  }

  private void fillNodeElement(Document document, LabeledTimePoint node,
      Element nodeElement)
  {
    Element nodeInstanceElement = document.createElementNS(NS, "event");
    fillNodeInstanceElement(document, nodeInstanceElement, node.getNodeInstance());
    nodeInstanceElement.setAttributeNS(NS, "eventID", node.getEventID());
    nodeElement.appendChild(nodeInstanceElement);
    
    Element labelElement = createLabelElement(document, node.getLabel());
    nodeElement.appendChild(labelElement);
    
    Element pluginDataContainer = XMLExportTools.createPluginDataContainerElement(document, NS, node);
    nodeElement.appendChild(pluginDataContainer);
  }

  private Element createNodeInstanceElement(Document document,
      NodeInstance nodeInstance)
  {
    Element nodeInstanceElement = document.createElementNS(NS, "nodeInstance");
    fillNodeInstanceElement(document, nodeInstanceElement, nodeInstance);
    return nodeInstanceElement;
  }

  private void fillNodeInstanceElement(Document document, Element nodeInstanceElement,
      NodeInstance nodeInstance)
  {
    nodeInstanceElement.setAttributeNS(NS, "nodeID",
                               "n" + Integer.toString(nodeInstance.getNodeID()));
    
    nodeInstanceElement.setAttributeNS(NS, "topologicalID", Integer.toString(nodeInstance.getTopologicalID()));
    
    Element iterationChain = createIterationChainElement(document, nodeInstance.getIteration());
    nodeInstanceElement.appendChild(iterationChain);
  }

  private Element createLabelElement(Document document, Label label)
  {
    if (label instanceof Proposition)
      return createPropositionElement(document, (Proposition)label);
    throw new IllegalArgumentException("label must be a Proposition");
  }

  private Element createPropositionElement(Document document, Proposition label)
  {
    Element propositionElement = document.createElementNS(NS, "proposition");
    
    for (ObservationBinding literal : label.getObservations())
    {
      Element literalGroupElement = createObservationBindingElement(document, literal);
      propositionElement.appendChild(literalGroupElement);
    }
    
    return propositionElement;
  }

  private Element createIterationChainElement(Document document,
      IterationChain iterationChain)
  {
    Element iterationChainElement = document.createElementNS(NS, "iterationChain");

    for (Iteration iteration : iterationChain.getIterations())
    {
      Element iterationElement = createIterationElement(document, iteration);
      iterationChainElement.appendChild(iterationElement);
    }
    
    return iterationChainElement;
  }

  private Element createIterationElement(Document document, Iteration iteration)
  {
    Element iterationElement = document.createElementNS(NS, "iteration");
    
    iterationElement.setAttributeNS(NS, "nodeID", "n"+String.valueOf(iteration.getNodeID()));
    iterationElement.setAttributeNS(NS, "iteration", String.valueOf(iteration.getIteration()));
    
    return iterationElement;
  }

  /**
   * Create new child Element without data.
   *
   * @param document
   * @param parent Element to which the created Element is appended to as a child
   * @param type type of the created Element
   *
   * @return The created child element.
   */
  private static Element addElement(final Document document,
      final Element parent, final String type)
  {
    return addElement(document, NS, parent, type);
  }
  
  /**
   * Create new child Element without data.
   *
   * @param document
   * @param ns The namespace the element should be created in.
   * @param parent Element to which the created Element is appended to as a child
   * @param type type of the created Element
   *
   * @return The created child element.
   */
  protected static Element addElement(final Document document, final String ns,
      final Element parent, final String type)
  {
    Element element = document.createElementNS(ns, type);
    parent.appendChild(element);
    return element;
  }
}
