package de.aristaflow.adept2.model.timemodel.tcn.xml;

import static de.aristaflow.adept2.model.timemodel.tcn.xml.CSTNUXMLExportTools.NS;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.w3c.dom.Element;
import org.w3c.dom.Node;

import de.aristaflow.adept2.model.common.PluginData;
import de.aristaflow.adept2.model.common.defaultimplementation.DefaultPluginData;
import de.aristaflow.adept2.model.common.defaultimplementation.DefaultPluginDataContainer;
import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.globals.XMLConstants;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain.Iteration;
import de.aristaflow.adept2.model.timemodel.tcn.JoinNodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.SplitNodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.DefaultCSTN;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.DefaultCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractTCN;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultContingentLabeledValueWithContingency;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledNode;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultObservationTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ObservationBinding;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.Proposition;
import de.aristaflow.adept2.util.xml.XMLHelperTools;

public class CSTNUXMLImportTools
{

  
  protected static Element getElement(final Node node)
  {
    return XMLHelperTools.getElement(node);
  }

  public static TCN<LabeledTimePoint, LabeledEdge> getTCN(Element root)
  {
    if (!root.getTagName().equals("cstnu"))
      throw new IllegalArgumentException("Can only load cstnus");
    
    String type;
//    UUID templateID;
    Set<ChoiceVariable> labels; 
    List<LabeledTimePoint> tcnNodes;
//    List<NodeMapping<LabeledTimePoint>> nodeMappings;
    List<LabeledEdge> tcnEdges;
    
    type = root.getAttributeNS(NS, "type");
    if (type == null || type.trim().length() == 0)
      type = "cstnu";
    
//    String templateIDString = root.getAttributeNS(NS, "templateID");
//    templateID = UUID.fromString(templateIDString);
    
    Element element = getElement(root.getFirstChild());
    Map<NodeInstance, Map<String, LabeledTimePoint>> nodeMap;
    nodeMap = getLabeledNodes(element);
    
    //Map<NodeInstance, Map<String, LabeledNode>> nodeMap = new HashMap<NodeInstance, Map<String,LabeledNode>>();
    
//    for (LabeledNode node : tcnNodes)
//    {
//    
//      Map<String, LabeledNode> map = nodeMap.get(node.getNodeInstance());
//      if (map == null){
//        map = new HashMap<String, LabeledNode>();
//        nodeMap.put(node.getNodeInstance(), map);
//      }
//      
//      map.put(node.getEventID(), node);
//    }
    
    tcnNodes = new ArrayList<LabeledTimePoint>();
    for (Map<String, LabeledTimePoint> m : nodeMap.values())
    {
      for (LabeledTimePoint n : m.values())
      {
        tcnNodes.add(n);
      }
    }
    
    element = getElement(element.getNextSibling());
    if ("nodeMappings".equalsIgnoreCase(element.getNodeName()))
    {
//      nodeMappings = getNodeMappings(element, nodeMap);
      
      element = getElement(element.getNextSibling());
    }
    tcnEdges = getLabeledEdges(element, nodeMap);
    
    labels = new HashSet<ChoiceVariable>();
    
    for (LabeledTimePoint node : tcnNodes)
    {
      if (node instanceof DefaultObservationTimePoint)
      {
        labels.add(((DefaultObservationTimePoint) node).getObservation());
      }
    }
    
    if ("cstnu".equalsIgnoreCase(type))
    {
      return new DefaultCSTNU(tcnNodes, tcnEdges, labels);
    }
    else if ("cstn".equalsIgnoreCase(type))
    {
      return new DefaultCSTN(tcnNodes, tcnEdges, labels);
    }
//    else if ("stnu".equalsIgnoreCase(type))
//    {
//      return new DefaultSTNU(templateID, tcnNodes, tcnEdges, nodeMappings, labels);
//    }
    else
      throw new IllegalArgumentException();
  }

  private static List<LabeledEdge> getLabeledEdges(Element element,
      Map<NodeInstance, Map<String, LabeledTimePoint>> tcnNodes)
  {
    assert element.getNodeName().equals("tcnEdges");
    
    List<LabeledEdge> edges = new ArrayList<LabeledEdge>();
    
    Element edgeElement = getElement(element.getFirstChild());
    while (edgeElement != null)
    {
      LabeledEdge edge = getLabeledEdge(edgeElement,tcnNodes);
      edges.add(edge);
      
      edgeElement = getElement(edgeElement.getNextSibling());
    }

    return edges;
  }

  private static LabeledEdge getLabeledEdge(Element edgeElement,
      Map<NodeInstance, Map<String, LabeledTimePoint>> tcnNodes)
  {
    assert edgeElement.getNodeName().equals("labeledEdge");
    
    LabeledTimePoint sourceNode;
    LabeledTimePoint targetNode;
    boolean controlEdge;
    boolean implicit;
    FlowDirection flowDirection;
    
    Element element = getElement(edgeElement.getFirstChild());
    sourceNode = getNodeReference(element, tcnNodes);
    
    element = getElement(element.getNextSibling());
    targetNode = getNodeReference(element, tcnNodes);
    
    
    controlEdge = Boolean.parseBoolean(edgeElement.getAttributeNS(NS, "controlEdge"));
    implicit = Boolean.parseBoolean(edgeElement.getAttributeNS(NS, "implicit"));
    flowDirection = FlowDirection.valueOf(edgeElement.getAttributeNS(NS, "flowDirection"));
    
    element = getElement(element.getNextSibling());
    LabeledValueSet value = getLabeledValueSet(element, tcnNodes);
    
    
    return new DefaultLabeledEdge(sourceNode, targetNode, value, controlEdge, implicit, flowDirection);
  }

  private static LabeledValueSet getLabeledValueSet(Element element, Map<NodeInstance, Map<String, LabeledTimePoint>> tcnNodes)
  {
    Element valueElement = getElement(element.getFirstChild());
    List<LabeledValue> values = new ArrayList<LabeledValue>();
    
    while (valueElement != null)
    {
      LabeledValue value = getLabeledValue(valueElement, tcnNodes);
      values.add(value);
      
      valueElement = getElement(valueElement.getNextSibling());
    }
    
    
    return new DefaultLabeledValueSet(values);
  }

  private static LabeledValue getLabeledValue(Element valueElement, Map<NodeInstance, Map<String, LabeledTimePoint>> tcnNodes)
  {
    LabeledValue value;
    
    Element el = getElement(valueElement.getFirstChild());
    Label label = getLabel(el, tcnNodes);
    
    el = getElement(el.getNextSibling());
    TimeDistance timeDistance = getTimeDistance(el);
    
    if (valueElement.getNodeName().equals("contingentLabeledValue"))
    {
      ConstraintCase constraintCase = ConstraintCase.valueOf(valueElement.getAttributeNS(NS, "constraintCase"));
      
      el = getElement(el.getNextSibling());
      
      LabeledTimePoint trigger = getNodeReference(el, tcnNodes);
      
      el = getElement(el.getNextSibling());
      TimeDistance contingentValue = getTimeDistance(el); 
      
      el = getElement(el.getNextSibling());
      if (el != null && el.getTagName().equals("contingencyValue"))
      {
        TimeDistance contingencyValue = getTimeDistance(el);
        value = new DefaultContingentLabeledValueWithContingency(label, constraintCase, timeDistance, trigger, contingentValue, contingencyValue);
      }
      else
      {
      
        value = new DefaultContingentLabeledValue(label, constraintCase, timeDistance, trigger, contingentValue);
      }
    }
    else
    {
      value = new DefaultLabeledValue(label, timeDistance);
    }

    return value;
  }

  private static TimeDistance getTimeDistance(Element timeDistanceElement)
  {
    long millis = TimeDistance.UNDEFINED_TIME_DISTANCE;
    long seconds = TimeDistance.UNDEFINED_TIME_DISTANCE;
    long minutes = TimeDistance.UNDEFINED_TIME_DISTANCE;
    long hours = TimeDistance.UNDEFINED_TIME_DISTANCE;
    long days = TimeDistance.UNDEFINED_TIME_DISTANCE;
    long months = TimeDistance.UNDEFINED_TIME_DISTANCE;
    
    if (timeDistanceElement.hasAttributeNS(NS, "months"))
      months = Long.parseLong(timeDistanceElement.getAttributeNS(NS, "months"));
    if (timeDistanceElement.hasAttributeNS(NS, "days"))
      days = Long.parseLong(timeDistanceElement.getAttributeNS(NS, "days"));
    if (timeDistanceElement.hasAttributeNS(NS, "hours"))
      hours = Long.parseLong(timeDistanceElement.getAttributeNS(NS, "hours"));
    if (timeDistanceElement.hasAttributeNS(NS, "minutes"))
      minutes = Long.parseLong(timeDistanceElement.getAttributeNS(NS, "minutes"));
    if (timeDistanceElement.hasAttributeNS(NS, "seconds"))
      seconds = Long.parseLong(timeDistanceElement.getAttributeNS(NS, "seconds"));    
    if (timeDistanceElement.hasAttributeNS(NS, "milliseconds"))
      millis = Long.parseLong(timeDistanceElement.getAttributeNS(NS, "milliseconds"));    
    
    return new DefaultTimeDistance(millis, seconds, minutes, hours, days, months);
  }

  private static LabeledTimePoint getNodeReference(Element element,
      Map<NodeInstance, Map<String, LabeledTimePoint>> tcnNodes)
  {
    NodeInstance nodeInstance = getNodeInstance(element);
    String eventID = element.getAttributeNS(NS, "eventID");
    
    return tcnNodes.get(nodeInstance).get(eventID);
  }

  private static List<NodeMapping<LabeledTimePoint>> getNodeMappings(
      Element mappingElement, Map<NodeInstance, Map<String, LabeledTimePoint>> tcnNodes)
  {
    Element element = getElement(mappingElement.getFirstChild());
    
    List<NodeMapping<LabeledTimePoint>> mappings = new ArrayList<NodeMapping<LabeledTimePoint>>();
    
    while (element != null)
    {

      NodeMapping<LabeledTimePoint> mapping = getNodeMapping(element, tcnNodes);
      mappings.add(mapping);
      
      element = getElement(element.getNextSibling());
    }

    return mappings;
  }

  private static NodeMapping<LabeledTimePoint> getNodeMapping(Element mappingElement,
      Map<NodeInstance, Map<String, LabeledTimePoint>> tcnNodes)
  {
    NodeMapping<LabeledTimePoint> nodeMapping;
    
    String label;
    NodeInstance nodeInstance;
    LabeledTimePoint startNode;
    List<LabeledTimePoint> intermediateEvents;
    LabeledTimePoint endNode;
    
    
    label = mappingElement.getAttributeNS(NS, "label");
    
    Element element = getElement(mappingElement.getFirstChild());
    nodeInstance = getNodeInstance(element);
        
    element = getElement(element.getNextSibling());
    startNode = getNodeReference(element, tcnNodes);
    
    element = getElement(element.getNextSibling());
    
    Element child = getElement(element.getFirstChild());
    intermediateEvents = new ArrayList<LabeledTimePoint>();
    while (child != null)
    {
      LabeledTimePoint intermediateNode = getNodeReference(child, tcnNodes);
      intermediateEvents.add(intermediateNode);
      child = getElement(child.getNextSibling());
    }
    
    element = getElement(element.getNextSibling());
    endNode = getNodeReference(element, tcnNodes);
    
    
    if (mappingElement.getNodeName().equals("joinNodeMapping"))
    {
      element = getElement(element.getNextSibling());
      List<LabeledTimePoint> prefixNodes;
      
      Element pref = getElement(element.getFirstChild());
      prefixNodes = new ArrayList<LabeledTimePoint>();
      while (pref != null)
      {
        LabeledTimePoint intermediateNode = getNodeReference(pref, tcnNodes);
        prefixNodes.add(intermediateNode);
        pref = getElement(pref.getNextSibling());
      }

   // FIXME import branches!
      
      nodeMapping = new JoinNodeMapping<LabeledTimePoint>(label, nodeInstance, 
          prefixNodes.toArray(new LabeledTimePoint[prefixNodes.size()]), 
          startNode, 
          intermediateEvents.toArray(new LabeledTimePoint[intermediateEvents.size()]), 
          endNode);
    } 
    else if (mappingElement.getNodeName().equals("splitNodeMapping"))
    {
      element = getElement(element.getNextSibling());
      
      List<LabeledTimePoint> suffixNodes;
      
      Element pref = getElement(element.getFirstChild());
      suffixNodes = new ArrayList<LabeledTimePoint>();
      while (pref != null)
      {
        LabeledTimePoint intermediateNode = getNodeReference(pref, tcnNodes);
        suffixNodes.add(intermediateNode);
        pref = getElement(pref.getNextSibling());
      }

      nodeMapping = new SplitNodeMapping<LabeledTimePoint>(label, nodeInstance, 
          startNode, 
          intermediateEvents.toArray(new LabeledTimePoint[intermediateEvents.size()]), 
          endNode, suffixNodes.toArray(new LabeledTimePoint[suffixNodes.size()]));
    }
    else
    {

      nodeMapping = new NodeMapping<LabeledTimePoint>(label, nodeInstance, 
          startNode, 
          intermediateEvents.toArray(new LabeledTimePoint[intermediateEvents.size()]), 
          endNode);
    }
    
    return nodeMapping;
  }

  private static Map<NodeInstance, Map<String, LabeledTimePoint>> getLabeledNodes(Element nodesElement)
  {
    Map<NodeInstance, Map<String, LabeledTimePoint>> nodeMap = new HashMap<NodeInstance, Map<String,LabeledTimePoint>>();
    //List<LabeledNode> nodes = new ArrayList<LabeledNode>();
    Element element = getElement(nodesElement.getFirstChild());
    
    while (element != null)
    {
      LabeledTimePoint node;
      if (element.getTagName().equals("observationNode"))
      {
        node = getObservationNode(element, nodeMap);
      }
      else
      {
        node = getLabeledNode(element, nodeMap);
      }
      
      Map<String, LabeledTimePoint> map = nodeMap.get(node.getNodeInstance());
      if (map == null){
        map = new HashMap<String, LabeledTimePoint>();
        nodeMap.put(node.getNodeInstance(), map);
      }
      
      map.put(node.getEventID(), node);
      
      element = getElement(element.getNextSibling());
    }

    return nodeMap;
  }

  private static LabeledTimePoint getLabeledNode(Element nodeElement, Map<NodeInstance, Map<String, LabeledTimePoint>> tcnNodes)
  {
    Element pdcElement = null;
    Element element = getElement(nodeElement.getFirstChild());
    NodeInstance nodeInstance = getNodeInstance(element);
    
    String eventID = element.getAttributeNS(NS, "eventID");
    
    element = getElement(element.getNextSibling());
    Label label = getLabel(element, tcnNodes);
    
    element = getElement(element.getNextSibling());
    if (element != null && element.getTagName().equals(XMLConstants.pluginDataContainer))
    {
      pdcElement = element;
    }
    
    if (nodeInstance.getNodeID() == Integer.MIN_VALUE)
    {
      return AbstractTCN.TIMEBASE_NODE;//new DefaultLabeledNode("Origin", new NodeInstance(Integer.MIN_VALUE, new IterationChain(0), Integer.MIN_VALUE), Proposition.EMPTY_PROPOSITION);
    }
    else
    {
      DefaultLabeledNode labeledNode = new DefaultLabeledNode(nodeInstance, eventID, label);
      if (pdcElement != null)
      {
        getPluginDataContainer(labeledNode, pdcElement);
      }
      return labeledNode; 
    }
    
    
  }
  
  private static void getPluginDataContainer(
      DefaultPluginDataContainer container, final Element pdcElement)
  {
    Map<String, PluginData> pluginDatas = new HashMap<String, PluginData>();
    
    Element pdElement = getElement(pdcElement.getFirstChild());
    while (pdElement != null)
    {
      String extensionPoint = pdElement
        .getAttribute(XMLConstants.pluginDataExtensionPoint);
      String pluginID = pdElement.getAttribute(XMLConstants.pluginDataPluginID);

      Map<String, String> innerMap = new HashMap<String, String>();
      Element pdEntry = getElement(pdElement.getFirstChild());
      while (pdEntry != null)
      {
        innerMap.put(pdEntry.getAttribute(XMLConstants.pluginDataName), pdEntry
          .getTextContent());

        pdEntry = getElement(pdEntry.getNextSibling());
      }

      
      PluginData pluginData = new DefaultPluginData(pluginID, innerMap); 
      pluginDatas.put(pluginID, pluginData);

      container.addSupportedPlugin(extensionPoint, pluginID, pluginData);
      
      pdElement = getElement(pdElement.getNextSibling());
    }
  }

//  private static LabelSet getLabelSet(Element element, Map<NodeInstance, Map<String, LabeledNode>> tcnNodes)
//  {
//    assert element.getNodeName().equals("labelSet");
//
//    List<Label> labels = new ArrayList<Label>();
//    
//    Element lbl = getElement(element.getFirstChild());
//    
//    while (lbl != null)
//    {
//      Label label = getLabel(lbl, tcnNodes);
//      labels.add(label);
//      
//      lbl = getElement(lbl.getNextSibling());
//      
//    }
//    
//    return new DefaultLabelSet(labels.toArray(new Label[labels.size()]));
//  }

  private static NodeInstance getNodeInstance(Element nodeInstanceElement)
  {
    int nodeID;
    int topologicalID;
    IterationChain iterationChain;
    
    String nodeIDString = nodeInstanceElement.getAttributeNS(NS, "nodeID");
    
    assert nodeIDString != null && nodeIDString.length() > 1 : ((Element)nodeInstanceElement.getParentNode()).getTagName();
    
    nodeID = Integer.parseInt(nodeIDString.substring(1));
    
    String topoIDString = nodeInstanceElement.getAttributeNS(NS, "topologicalID");
    if (topoIDString != null && topoIDString.length() > 0)
      topologicalID = Integer.parseInt(topoIDString);
    else
      topologicalID = -1;
    
    Element element = getElement(nodeInstanceElement.getFirstChild());
    
    iterationChain = getIterationChain(element);
            
    return new NodeInstance(nodeID, iterationChain, topologicalID);
  }

  
  private static IterationChain getIterationChain(Element chainElement)
  {
    assert chainElement.getNodeName().equals("iterationChain");
    
    List<IterationChain.Iteration> iterations = new ArrayList<IterationChain.Iteration>();
    
    Element element = getElement(chainElement.getFirstChild());
    while (element != null)
    {
      IterationChain.Iteration iteration = getIteration(element);
      iterations.add(iteration);
      
      element = getElement(element.getNextSibling());      
    }
    
    @SuppressWarnings("deprecation")
    IterationChain iterationChain = new IterationChain(iterations.toArray(new Iteration[iterations.size()]));
    return iterationChain;
  }

  private static Iteration getIteration(Element element)
  {
    assert element.getNodeName().equals("iteration");
    
    String nodeIDString = element.getAttributeNS(NS, "nodeID");
    int nodeID = Integer.parseInt(nodeIDString.substring(1));
    
    String iterationString = element.getAttributeNS(NS, "iteration");
    int iteration = Integer.parseInt(iterationString);
    
    return new Iteration(nodeID, iteration);
  }

  private static LabeledTimePoint getObservationNode(Element nodeElement, Map<NodeInstance, Map<String, LabeledTimePoint>> tcnNodes)
  {
    assert nodeElement.getTagName().equals("observationNode");
    

    Element element = getElement(nodeElement.getFirstChild());
    NodeInstance nodeInstance = getNodeInstance(element);
    
    String eventID = element.getAttributeNS(NS, "eventID");
    
    element = getElement(element.getNextSibling());
    Label label = getLabel(element, tcnNodes);
    
    
//    element = getElement(element.getNextSibling());
//    Element observationElement = getElement(element.getFirstChild());
//    observationLabel = getObservation(observationElement, tcnNodes);    
    
    element = getElement(element.getNextSibling());
    Element pdcElement = null;
    if (element.getTagName().equals(XMLConstants.pluginDataContainer))
    {
      pdcElement = element;
      element = getElement(element.getNextSibling());
    }
    
    Map<Long, Integer> decisionLabels = new HashMap<Long, Integer>();
    
    Element ds = getElement(element.getFirstChild());
    while (ds != null)
    {
      
      long edgeCode = Long.parseLong(ds.getAttributeNS(NS, "edgeCode"));
      int observationID = Integer.parseInt(ds.getAttributeNS(NS, "observationCode"));
      
      decisionLabels.put(Long.valueOf(edgeCode), Integer.valueOf(observationID));
      
      ds = getElement(ds.getNextSibling());
    }
    
    long[] edgeCodes = new long[decisionLabels.size()];
    int[] observationIDs = new int[decisionLabels.size()];
    int i = 0;
    for (Entry<Long, Integer> x : decisionLabels.entrySet())
    {
      edgeCodes[i] = x.getKey().intValue();
      observationIDs[i++] = x.getValue().intValue();
    }  
    

    
    DefaultObservationTimePoint observationTimePoint = new DefaultObservationTimePoint(nodeInstance, eventID, label, edgeCodes, observationIDs);
    if (pdcElement != null)
    {
      getPluginDataContainer(observationTimePoint, pdcElement);
    }
    return observationTimePoint;
  }

  private static Label getLabel(Element element, Map<NodeInstance, Map<String, LabeledTimePoint>> tcnNodes)
  {
    assert element.getNodeName().equals("proposition");
    
    List<ObservationBinding> literals = new ArrayList<ObservationBinding>();
    
    Element lit = getElement(element.getFirstChild());
    while (lit != null)
    {
      ObservationBinding group = getObservationBinding(lit, tcnNodes);
      literals.add(group);
      
      lit = getElement(lit.getNextSibling()); 
    }
    
    
    return new Proposition(literals );
  }

  private static ObservationBinding getObservationBinding(Element element, Map<NodeInstance, Map<String, LabeledTimePoint>> tcnNodes)
  {
    assert element.getTagName().equals("observation");
    
    Element nv = getElement(element.getFirstChild());
    LabeledTimePoint splitNode = getNodeReference(nv, tcnNodes);
    assert splitNode instanceof DefaultObservationTimePoint;
    DefaultObservationTimePoint observationNode = (DefaultObservationTimePoint) splitNode;

    int value = Integer.parseInt(element.getAttributeNS(NS, "observationCode")) & 0xffff; // LOW legacy code...
    long edgeCode = Integer.parseInt(element.getAttributeNS(NS, "edgeCode"));
    
   
    return new ObservationBinding(observationNode.getObservation(), edgeCode, value);
  }

}
