package de.aristaflow.adept2.model.timemodel.tcn.xml;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.logging.Logger;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTimePoint;
import de.aristaflow.adept2.util.LoggerTools;
import de.aristaflow.adept2.util.xml.XMLFormatException;
import de.aristaflow.adept2.util.xml.XMLHelperTools;

public class TCNXMLExport
{
  private Logger logger = LoggerTools.getLogger(this);
  
  private CSTNUXMLExportTools cstnuExportTools = new CSTNUXMLExportTools();

  /**
   * Creates a DOM document from the tcn.
   *
   * @param tcn The source tcn
   *
   * @return A DOM document
   * @throws XMLFormatException Thrown if the validation of the document fails
   */
  public <N extends TCNTimePoint, E extends TCNEdge<N, ?>> Document getDocumentForTCN(TCN<N, E> tcn) throws XMLFormatException
  {
    Document document = XMLHelperTools.createDocument();

    // create the root element of the document, and set the ID as attribute
    Object bugFixTCN = tcn; // Bug in JavaC; see http://bugs.sun.com/view_bug.do?bug_id=6932571
    @SuppressWarnings("unchecked")
    Element root = cstnuExportTools.createCSTNElement(document, (TCN<LabeledTimePoint, LabeledEdge>)bugFixTCN);
    document.appendChild(root);
    
    try
    {
      XMLHelperTools.validateStream(XMLHelperTools.getInputStreamForDocument(document), TCNXMLImport.getSchema());
    }
    catch (IOException e)
    {
      throw new XMLFormatException(e);
    }
    
    return document;
  }
  
  /**
   * Returns The corresponding stream for a given tcn
   *
   * @param tcn The source tcn
   *
   * @return An input stream for the given tcn
   * @throws XMLFormatException Thrown if the validation of the document fails
   * @throws IOException Thrown if an unrecoverable error occurs during the
   *           course of the transformation 
   */
  public <N extends TCNTimePoint, E extends TCNEdge<N, ?>> InputStream getInputStreamForTCN(TCN<N, E> tcn)
      throws XMLFormatException, IOException
  {
    return XMLHelperTools.getInputStreamForDocument(getDocumentForTCN(tcn));
  }
  
  /**
   * Returns the xml code of a given tcn as string.
   * 
   * @param tcn The tcn which should be transformed.
   * 
   * @return The xml code of the given process as string.
   * @throws XMLFormatException Thrown if the validation of the document fails
   * @throws IOException Thrown if an unrecoverable error occurs during the
   *           course of the transformation
   */
  public <N extends TCNTimePoint, E extends TCNEdge<N, ?>> String getXMLStringForTCN(TCN<N, E> tcn)
      throws XMLFormatException, IOException
  {
    return XMLHelperTools.getXMLStringForDocument(getDocumentForTCN(tcn));
  }
  
  /**
   * Writes the given tcn to the given file.
   *
   * @param file The target file,
   * @param tcn The source tcn,
   *
   * @throws IOException Thrown if writing the XML to the file fails, e.g. if
   *           the provided file is read-only.
   * @throws XMLFormatException Thrown if the validation of the document fails
   */
  public <N extends TCNTimePoint, E extends TCNEdge<N, ?>> void writeTCNToFile(TCN<N, E> tcn, File file)
      throws IOException, XMLFormatException
  {
    Document doc = getDocumentForTCN(tcn);
    XMLHelperTools.writeDocumentToFile(doc, file, null);
  }
}
