package de.aristaflow.adept2.model.timemodel.tcn.xml;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;

import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.SAXException;

import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.util.xml.VersionException;
import de.aristaflow.adept2.util.xml.XMLFormatException;
import de.aristaflow.adept2.util.xml.XMLHelperTools;

public class TCNXMLImport
{

  private static Schema cstnuSchema;

  public TCN<LabeledTimePoint, LabeledEdge> getTCNFromFile(final File file)
      throws VersionException, XMLFormatException, IOException
  {
    // get the document from the file (without validation)
    Document document = XMLHelperTools.getDocumentFromFile(file, null, null);

    return getTCNFromDocument(document);
  }

  public TCN<LabeledTimePoint, LabeledEdge> getTCNFromStream(
      final InputStream inputStream) throws XMLFormatException, IOException
  {
    // get the document from the stream, and the process from the document
    Document document = XMLHelperTools.getDocumentFromStream(inputStream);

    return getTCNFromDocument(document);
  }

  public TCN<LabeledTimePoint, LabeledEdge> getTCNFromDocument(
      final Document document) throws XMLFormatException
  {
    
    XMLHelperTools.validateDoc(document, getSchema());
    
    // get the root element of the document
    Element root = document.getDocumentElement();
    return CSTNUXMLImportTools.getTCN(root);
  }

  public static Schema getSchema()
  {
    SchemaFactory schemaFactory = SchemaFactory.newInstance(javax.xml.XMLConstants.W3C_XML_SCHEMA_NS_URI);

    try
    {
      Schema schema = null;

      if (cstnuSchema == null)
      {
        cstnuSchema = schemaFactory.newSchema(TCNXMLImport.class
          .getResource("cstnu.xsd"));
      }
      schema = cstnuSchema;
      return schema;
    }
    catch (SAXException e)
    {
      throw new RuntimeException(e);
    }
  }
}
