package de.aristaflow.adept2.model.timemodel.tools;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

public abstract class AbstractGraph<V, E> implements Graph<V, E>
{

  protected int maxNodeID;


  protected Map<V, Integer> vertices;
  private Object[][] edges;

  public AbstractGraph(Collection<V> vertices)
  {
    this.maxNodeID = 0;
    this.vertices = new LinkedHashMap<V, Integer>();
    for (V n : vertices)
    {
      if (!this.vertices.containsKey(n))
        this.vertices.put(n, Integer.valueOf(maxNodeID++));
    }
    
    ensureCapacity(maxNodeID);
  }

  
  int getNodeKey(V sourceNode)
  {
    Integer key = this.vertices.get(sourceNode);
    
    if (key == null)
      throw new IllegalArgumentException(String.valueOf(key));
    
    assert key != null;
    
    return key.intValue();
  }
  
  @SuppressWarnings("unchecked")
  protected void ensureCapacity(int size)
  {
    if (edges == null)
    {
      int newSize = size;//(int)(size*1.1);
      this.edges = new Object[newSize][newSize];
    }
    else
    {
      // IMPROVE consider to perform a reorganization (i.e., removing or reusing unused rows and columns) (see removeTCNNode(N)); 
      if (this.edges.length <= size)
      {
        int newSize = (int)(size*1.1);
        Object[][] newArray = new Object[newSize][newSize];
        
        for (int i = 0; i < this.edges.length; i++)
        {
          System.arraycopy(this.edges[i], 0, newArray[i], 0, this.edges.length);
        }
        this.edges = newArray;
      }
    }
  }

  public Collection<E> getEdges()
  {
    List<E> eg = new ArrayList<E>(this.edges.length*2);
    for (int i = 0; i < maxNodeID; i++)
    {
      for (int j = 0; j < maxNodeID; j++)
      {
        E edge = internalGetEdge(i,j);
        if (edge != null)
          eg.add(edge);
      }
    }
    
    return Collections.unmodifiableList(eg);
  }

  public Collection<V> getVertices()
  {
    return Collections.unmodifiableCollection(this.vertices.keySet());
  }

  public abstract boolean addEdge(V v1, V v2, E edge);
  
  public abstract E getEdge(V v1, V v2);

  protected boolean internalSetEdge(int key1, int key2, E edge)
  {
    boolean exists = edges[key1][key2] != null;
    edges[key1][key2] = edge;
    return exists;
  }



  @SuppressWarnings("unchecked")
  E internalGetEdge(int key1, int key2)
  {
    return (E) edges[key1][key2];
  }



}