package de.aristaflow.adept2.model.timemodel.tools;

import java.util.Collection;

public class DefaultDirectedGraph<V, E> extends AbstractGraph<V, E> implements DirectedGraph<V, E>
{
  public DefaultDirectedGraph(Collection<V> vertices)
  {
    super(vertices);
  }

  @Override
  public boolean addEdge(V v1, V v2, E edge)
  {
    int key1 = getNodeKey(v1);
    int key2 = getNodeKey(v2);
    
    return internalSetEdge(key1, key2, edge);
  }
  
  @Override
  public E getEdge(V v1, V v2)
  {
    int key1 = getNodeKey(v1);
    int key2 = getNodeKey(v2);
    
    return internalGetEdge(key1, key2);
  }
  
  public void visit(GraphVisitor<V, E> visitor)
  {
    visitor.discoverGraph(this);
    
        
    @SuppressWarnings("unchecked")
    V[] vertices = (V[])getVertices().toArray();
    for (V e : vertices)
    {
      visitor.discoverVertex(e);
      visitor.finishVertex(e);
    }
    
    for (int i = 0; i < vertices.length; i++)
    {
      V ei = vertices[i];
      for (int j = 0; j < vertices.length; j++)
      {
        V ej = vertices[j];
        E e = getEdge(ei, ej);
        if (e != null)
        {
          visitor.discoverEdge(ei, 
                               ej, 
                               e);
          visitor.finishEdge(ei, 
                             ej, 
                             e);
        }
      }
    }
  
    visitor.finishGraph(this);
  }
}
