package de.aristaflow.adept2.model.timemodel.tools;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

public class DefaultUndirectedGraph<V, E> extends AbstractGraph<V, E> implements UndirectedGraph<V, E>
{

  public DefaultUndirectedGraph(Collection<V> vertices)
  {
    super(vertices);
  }

  @Override
  public boolean addEdge(V v1, V v2, E edge)
  {
    int keyV1 = getNodeKey(v1);
    int keyV2 = getNodeKey(v2);
    
    int key1 = Math.min(keyV1, keyV2);
    int key2 = Math.max(keyV1, keyV2);
    
    return internalSetEdge(key1, key2, edge);
  }
  
  @Override
  public E getEdge(V v1, V v2)
  {
    int keyV1 = getNodeKey(v1);
    int keyV2 = getNodeKey(v2);
    
    int key1 = Math.min(keyV1, keyV2);
    int key2 = Math.max(keyV1, keyV2);
    
    return internalGetEdge(key1, key2);
  }
  
  public Collection<V> getAdjacenceList(V v)
  {
    List<V> adjacenceList = new ArrayList<V>(); 
    
    for (V vertex : getVertices())
    {
      if (getEdge(v, vertex) != null)
      {
        adjacenceList.add(vertex);
      }
    }
    
    return adjacenceList;
  }
  
  public void visit(GraphVisitor<V, E> visitor)
  {
    visitor.discoverGraph(this);
    
        
    V[] vertices = (V[])getVertices().toArray();
    for (V e : vertices)
    {
      visitor.discoverVertex(e);
      visitor.finishVertex(e);
    }
    
    for (int i = 0; i < vertices.length; i++)
    {
      V ei = vertices[i];
      for (int j = i; j < vertices.length; j++)
      {
        V ej = vertices[j];
        E e = getEdge(ei, ej);
        if (e != null)
        {
          visitor.discoverEdge(ei, 
                               ej, 
                               e);
          visitor.finishEdge(ei, 
                             ej, 
                             e);
        }
      }
    }
  
    visitor.finishGraph(this);
  }
  


}
