package de.aristaflow.adept2.model.timemodel.tools;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

public class DepthFirstSearch<V, E>
{

  private Graph<V, E> graph;
  private boolean directed;

  public DepthFirstSearch(Graph<V, E> graph)
  {
    this.graph = graph;
    this.directed = graph instanceof DirectedGraph;
  }


  enum VertexState
  {
    White, Gray, Black
  }

  public void performDepthFirstSearch(GraphVisitor<V, E> visitor, V startNode)
  {
    Collection<V> vertices = graph.getVertices();
    int vertexCount = vertices.size();

    visitor.discoverGraph(graph);
    
    runDFS(visitor, startNode, new HashSet<V>((int)(vertexCount*1.5)), new HashSet<E>((int)(graph.getEdges().size()*1.5)));
    
    visitor.finishGraph(graph);
  }

  protected void runDFS(GraphVisitor<V, E> visitor, V u, Set<V> markedVertices, HashSet<E> markedEdges)
  {
    visitor.discoverVertex(u);
    markedVertices.add(u);

    
    for (V v : graph.getVertices())
    {
      E edge = graph.getEdge(u, v);
      if (edge != null)
      {
//        if (!markedEdges.contains(edge))
//        {
          markedEdges.add(edge);
          visitor.discoverEdge(u, v, edge);
          if (!markedVertices.contains(v))
            runDFS(visitor, v, markedVertices, markedEdges);
          visitor.finishEdge(u, v, edge);
//        }
      }
    }

    visitor.finishVertex(u);
  }

}
