package de.aristaflow.adept2.model.timemodel.tools;

/**
 * 
 * @author Andreas Lanz
 *
 * @param <V> Type of Vertices
 * @param <E> Type of Edges
 */
public interface GraphVisitor<V, E>
{
   /** Called at beginning of visiting a graph or subgraph. */
  public void discoverGraph (Graph<V, E> graph);
  
  /** Called at end of visiting a graph or subgraph. */
  public void finishGraph (Graph<V, E> graph);
  
  /**
   * Called when beginning visit to vertex. If the vertex is
   * the entry point for a subgraph, discoverGraph for that
   * subgraph must be called during the visit.
   */
  public void discoverVertex(V vertex);
  /**
   * Called at end of vertex visit. If the vertex is an exit
   * point for a subgraph, finishGraph for the subgraph must
   * be called during the visit.
   */
   public void finishVertex (V vertex);

   /** Called when initially visiting edge. */
   public void discoverEdge (V startVertex, V endVertex, E edge);

   /** Called at end of visit to edge. */
   public void finishEdge (V startVertex, V endVertex,E edge);
}