package de.aristaflow.adept2.core.checks.processmodel.timemodel;

import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.io.IOException;

import org.junit.AfterClass;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;

import de.aristaflow.adept2.TestSuiteADEPT2;
import de.aristaflow.adept2.base.security.AuthenticationException;
import de.aristaflow.adept2.base.service.Registry;
import de.aristaflow.adept2.base.service.ServiceNotKnownException;
import de.aristaflow.adept2.base.sessionmanagement.ClientSessionFactory;
import de.aristaflow.adept2.core.checks.processmodel.ProcessTemplateCheck;
import de.aristaflow.adept2.core.checks.processmodel.structure.TestStructuralChecks;
import de.aristaflow.adept2.core.client.ClientBootstrap;
import de.aristaflow.adept2.core.client.ClientService;
import de.aristaflow.adept2.core.client.ModellingClientService;
import de.aristaflow.adept2.model.globals.WorklistConstants;
import de.aristaflow.adept2.model.processmodel.ProcessModelFactory;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.tools.NodeRelations;
import de.aristaflow.adept2.model.processmodel.tools.ProcessElementIdentifierTools;
import de.aristaflow.adept2.model.processmodel.xml.ProcessModelXMLImport;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;
import de.aristaflow.adept2.util.CheckReport;
import de.aristaflow.adept2.util.CheckReport.ResultType;
import de.aristaflow.adept2.util.DataSourceException;
import de.aristaflow.adept2.util.xml.VersionException;
import de.aristaflow.adept2.util.xml.XMLFormatException;

public class BasicTimeModelCheckTest
{
  
  private static final String ADEPT2PLUGIN_DIR = "../de.aristaflow.adept2.adept2plugin/";
  private static final String ADEPT2TIME_DIR = "../de.aristaflow.adept2.adept2plugin.time/";
  /** The directory the test configuration could be found in. */
  private static final String CONFIG_DIR = "conf/test"; //$NON-NLS-1$
  /** The directory the test configuration could be found in. */
  private static final String EXAMPLES_DIR = "examples"; //$NON-NLS-1$

  /** The xml import used for the test. */
  static ProcessModelXMLImport xmlImport;

  private static ModellingClientService modellingClientService;

  private static Registry registry;

  private static ClientService service;
  
//  private static boolean cleanConfDir = false;

  /**
   * @throws ServiceNotKnownException
   * @throws DataSourceException
   * @throws AuthenticationException
   */
  @SuppressWarnings("deprecation")
  @BeforeClass
  public static void init() throws ServiceNotKnownException, AuthenticationException, DataSourceException
  {
    StringBuilder confDir;
//    if (System.getProperty("arflow.confdir") == null //$NON-NLS-1$
//        || System.getProperty("arflow.confdir").equals("")) //$NON-NLS-1$ //$NON-NLS-2$
//    {
      String adept2plugin = null;
      String adept2time = null;
      try
      {
        adept2plugin = new File(ADEPT2PLUGIN_DIR).getCanonicalPath();
      }
      catch (IOException e)
      {
        fail("Problems while setting the confdir!"); //$NON-NLS-1$
      }
      
      try
      {
        adept2time = new File(ADEPT2TIME_DIR).getCanonicalPath();
      }
      catch (IOException e)
      {
        fail("Problems while setting the confdir!"); //$NON-NLS-1$
      }
      confDir = new StringBuilder(adept2plugin);
      confDir.append(File.separatorChar);
      confDir.append(CONFIG_DIR);
      
      confDir.append(File.pathSeparatorChar);
      
      confDir.append(adept2time);
      confDir.append(File.separatorChar);
      confDir.append(CONFIG_DIR);
      
      StringBuilder dataDir = new StringBuilder(adept2plugin);
      dataDir.append(File.separatorChar);
      dataDir.append(EXAMPLES_DIR);
      
      // set confDir and dataDir
      System.setProperty("arflow.confdir", confDir.toString()); //$NON-NLS-1$
      System.setProperty("arflow.datadir", dataDir.toString()); //$NON-NLS-1$
//      cleanConfDir = true;
//    }
    
    ClientBootstrap clientBootstrap = TestSuiteADEPT2.initialiseRegistry();

    service = clientBootstrap.getClientService();

    ClientSessionFactory csf = service.getAuthentication().authenticate(WorklistConstants.SYSTEM_AGENT_NAME,
                                                                        WorklistConstants.SYSTEM_ORGPOSITION,
                                                                        WorklistConstants.SYSTEM_PASSWORD);
    service.setAuthenticatedAgents(new ClientSessionFactory[] { csf });

    
    
    registry = clientBootstrap.getRegistry();
    modellingClientService = (ModellingClientService) service;

    ProcessModelFactory pmf = modellingClientService.getProcessModelFactory();
    // pmf = new DefaultProcessModelFactory(null);
    xmlImport = new ProcessModelXMLImport(pmf);
  }

  /**
   * Log off the authenticated agent from the client service after the tests.
   */
  @AfterClass
  public static void logoff()
  {
    registry = null;
    modellingClientService = null;
    xmlImport = null;
    service.logoff();
    
//    if (cleanConfDir)
//    {
      // set confDir and dataDir
      System.setProperty("arflow.confdir", ""); //$NON-NLS-1$
      System.setProperty("arflow.datadir", ""); //$NON-NLS-1$
//    }
  }

  @Test
  public void testPerformCheck() throws VersionException, XMLFormatException, IOException
  {

    CheckReport checkReport;
    Template template;
    ProcessTemplateCheck check = new BasicTemporalProcessCheck(null, modellingClientService.getProcessChecks(),
                                                               registry);

    template = xmlImport.getTemplateFromStream(TestTemplates.class.getResourceAsStream("SimpleTimeTestContingent.template"));
    checkReport = new CheckReport(ProcessElementIdentifierTools.getTemplateIdentifier(template, "TestControlFlow1"));
    check.performCheck(template, new NodeRelations(template), checkReport);
    assertTrue(checkReport.getReportSummary(), checkReport.getCheckResult().ordinal() < ResultType.WARNING.ordinal());

  }

}
