package de.aristaflow.adept2.model.common.timedata.defaultimplementation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.TimePoint;

public class DefaultTimeDistanceTest
{

  @Test
  public void testDefaultTimeDistanceLongLongLongLongLong() throws Exception
  {
    TimeDistance timeDistance = new DefaultTimeDistance(5, 10, 15, 20, 25, 30);
    
    assertEquals((((((30l*30l)+25l)*24l+20l)*60l+15l)*60l+10l)*1000l+5l, timeDistance.toMilliseconds());
    
    TimeDistance infty = new DefaultTimeDistance(TimeDistance.POSITIVE_INFINITY, 0, 0, 0, 0, 0);
    assertTrue(infty.isInfinity());
    
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getMilliseconds());
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getSeconds());
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getMinutes());
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getHours());
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getDays());
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getMonths());
    
    infty = new DefaultTimeDistance(0, TimeDistance.POSITIVE_INFINITY, 0 , 0, 0, 0);
    assertTrue(infty.isInfinity());
    infty = new DefaultTimeDistance(0, 0, TimeDistance.POSITIVE_INFINITY, 0, 0, 0);
    assertTrue(infty.isInfinity());
    infty = new DefaultTimeDistance(0, 0, 0, TimeDistance.POSITIVE_INFINITY, 0, 0);
    assertTrue(infty.isInfinity());
    infty = new DefaultTimeDistance(0, 0, 0, 0, TimeDistance.POSITIVE_INFINITY, 0);
    assertTrue(infty.isInfinity());
    infty = new DefaultTimeDistance(0, 0, 0, 0, 0, TimeDistance.POSITIVE_INFINITY);
    assertTrue(infty.isInfinity());
    
    
    TimeDistance ninfty = new DefaultTimeDistance(TimeDistance.NEGATIVE_INFINITY, 0, 0 , 0, 0, 0);
    assertTrue(ninfty.isInfinity());
    
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getMilliseconds());
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getSeconds());
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getMinutes());
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getHours());
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getDays());
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getMonths());
    
    ninfty = new DefaultTimeDistance(0, TimeDistance.NEGATIVE_INFINITY, 0 , 0, 0, 0);
    assertTrue(ninfty.isInfinity());
    ninfty = new DefaultTimeDistance(0, 0, TimeDistance.NEGATIVE_INFINITY, 0, 0, 0);
    assertTrue(ninfty.isInfinity());
    ninfty = new DefaultTimeDistance(0, 0, 0, TimeDistance.NEGATIVE_INFINITY, 0, 0);
    assertTrue(ninfty.isInfinity());
    ninfty = new DefaultTimeDistance(0, 0, 0, 0, TimeDistance.NEGATIVE_INFINITY, 0);
    assertTrue(ninfty.isInfinity());
    ninfty = new DefaultTimeDistance(0, 0, 0, 0, 0, TimeDistance.NEGATIVE_INFINITY);
    assertTrue(ninfty.isInfinity());
    
  }

  @Test
  public void testDefaultTimeDistanceLongLongLongLongLongLong() throws Exception
  {
    TimeDistance timeDistance = new DefaultTimeDistance(10, 15, 20, 25, 30);
    
    assertEquals((((((30l*30l)+25l)*24l+20l)*60l+15l)*60l+10l)*1000l, timeDistance.toMilliseconds());
    
    TimeDistance infty = new DefaultTimeDistance(TimeDistance.POSITIVE_INFINITY, 0, 0, 0, 0);
    assertTrue(infty.isInfinity());
    
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getMilliseconds());
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getSeconds());
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getMinutes());
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getHours());
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getDays());
    assertEquals(TimeDistance.POSITIVE_INFINITY, infty.getMonths());
    
    infty = new DefaultTimeDistance(0, TimeDistance.POSITIVE_INFINITY, 0 , 0, 0);
    assertTrue(infty.isInfinity());
    infty = new DefaultTimeDistance(0, 0, TimeDistance.POSITIVE_INFINITY, 0, 0);
    assertTrue(infty.isInfinity());
    infty = new DefaultTimeDistance(0, 0, 0, TimeDistance.POSITIVE_INFINITY, 0);
    assertTrue(infty.isInfinity());
    infty = new DefaultTimeDistance(0, 0, 0, 0, TimeDistance.POSITIVE_INFINITY);
    assertTrue(infty.isInfinity());

    
    TimeDistance ninfty = new DefaultTimeDistance(TimeDistance.NEGATIVE_INFINITY, 0, 0 , 0, 0);
    assertTrue(ninfty.isInfinity());
    
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getMilliseconds());
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getSeconds());
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getMinutes());
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getHours());
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getDays());
    assertEquals(TimeDistance.NEGATIVE_INFINITY, ninfty.getMonths());
    
    ninfty = new DefaultTimeDistance(0, TimeDistance.NEGATIVE_INFINITY, 0 , 0, 0);
    assertTrue(ninfty.isInfinity());
    ninfty = new DefaultTimeDistance(0, 0, TimeDistance.NEGATIVE_INFINITY, 0, 0);
    assertTrue(ninfty.isInfinity());
    ninfty = new DefaultTimeDistance(0, 0, 0, TimeDistance.NEGATIVE_INFINITY, 0);
    assertTrue(ninfty.isInfinity());
    ninfty = new DefaultTimeDistance(0, 0, 0, 0, TimeDistance.NEGATIVE_INFINITY);
    assertTrue(ninfty.isInfinity());
  }

  @Test
  public void testDefaultTimeDistanceTimePoint() throws Exception
  {
    TimePoint timePoint = new DefaultTimePoint(10, 15, 20, 25, 30);
    TimeDistance timeDistance = new DefaultTimeDistance(timePoint);
    
    assertEquals((((((30l*30l)+25l)*24l+20l)*60l+15l)*60l+10l)*1000l, timeDistance.toMilliseconds());
  }

  @Test
  public void testGetSeconds() throws Exception
  {
    TimeDistance timeDistance = new DefaultTimeDistance(5, 10, 15, 20, 25, 30);
    assertEquals(10l, timeDistance.getSeconds());
  }

  @Test
  public void testGetMilliseconds() throws Exception
  {
    TimeDistance timeDistance = new DefaultTimeDistance(5, 10, 15, 20, 25, 30);
    assertEquals(5l, timeDistance.getMilliseconds());
  }

  @Test
  public void testGetMinutes() throws Exception
  {
    TimeDistance timeDistance = new DefaultTimeDistance(5, 10, 15, 20, 25, 30);
    assertEquals(15l, timeDistance.getMinutes());
  }

  @Test
  public void testGetHours() throws Exception
  {
    TimeDistance timeDistance = new DefaultTimeDistance(5, 10, 15, 20, 25, 30);
    assertEquals(20l, timeDistance.getHours());
  }

  @Test
  public void testGetDays() throws Exception
  {
    TimeDistance timeDistance = new DefaultTimeDistance(5, 10, 15, 20, 25, 30);
    assertEquals(30l*30l+25l, timeDistance.getDays());
  }

  @Test
  public void testGetMonths() throws Exception
  {
    TimeDistance timeDistance = new DefaultTimeDistance(5, 10, 15, 20, 25, 30);
    assertEquals(0l, timeDistance.getMonths());
  }

  @Test
  public void testToMilliseconds() throws Exception
  {
    TimeDistance timeDistance = new DefaultTimeDistance(10, 15, 20, 25, 30);
    
    assertEquals((((((30l*30l)+25l)*24l+20l)*60l+15l)*60l+10l)*1000l, timeDistance.toMilliseconds());
    
    TimeDistance timeDistance2 = new DefaultTimeDistance(5, 10, 15, 20, 25, 30);
    
    assertEquals((((((30l*30l)+25l)*24l+20l)*60l+15l)*60l+10l)*1000l+5l, timeDistance2.toMilliseconds());
  }

  @Test
  public void testToSeconds() throws Exception
  {
    TimeDistance timeDistance = new DefaultTimeDistance(10, 15, 20, 25, 30);
    
    assertEquals(((((30l*30l)+25l)*24l+20l)*60l+15l)*60l+10l, timeDistance.toSeconds());
    
    TimeDistance timeDistance2 = new DefaultTimeDistance(5, 10, 15, 20, 25, 30);
    
    assertEquals(((((30l*30l)+25l)*24l+20l)*60l+15l)*60l+10l, timeDistance2.toSeconds());
  }
  
  @BeforeClass
  public static void setUpBeforeClass() throws Exception
  {
    //
  }

  @Before
  public void setUp() throws Exception
  {
    //
  }

  @Test
  public void testNeg()
  {
    TimeDistance S60 = new DefaultTimeDistance(60, 0, 0, 0, 0);
    TimeDistance M1 = new DefaultTimeDistance(0, 1, 0, 0, 0);
    
    TimeDistance M2S2 = new DefaultTimeDistance(2, 2, 0, 0, 0);
    
    TimeDistance H24 = new DefaultTimeDistance(0, 0, 24, 0, 0);
    TimeDistance D1 = new DefaultTimeDistance(0, 0, 0, 1, 0);
    
    TimeDistance nS60 = new DefaultTimeDistance(-60, 0, 0, 0, 0);
    TimeDistance nM1 = new DefaultTimeDistance(0, -1, 0, 0, 0);
    
    TimeDistance nM2S2 = new DefaultTimeDistance(-2, -2, 0, 0, 0);
    
    TimeDistance nH24 = new DefaultTimeDistance(0, 0, -24, 0, 0);
    TimeDistance nD1 = new DefaultTimeDistance(0, 0, -24, 0, 0);
    
    assertEquals(S60.neg(), nS60);
    assertEquals(S60.neg().neg(), S60);
    assertEquals(M1.neg(), nM1);
    assertEquals(M1.neg(), nS60);
    
    assertEquals(M2S2.neg(), nM2S2);
    assertEquals(H24.neg(), nH24);
    assertEquals(D1.neg(), nD1);
  }

  @Test
  public void testEqualsObject()
  {
    TimeDistance timeDistance1 = new DefaultTimeDistance(0, 0, 2, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE);
    TimeDistance timeDistance2 = new DefaultTimeDistance(TimeDistance.UNDEFINED_TIME_DISTANCE, 0, 2, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE);
   
    
    assertTrue(timeDistance1.equals(timeDistance2));
    assertTrue(timeDistance2.equals(timeDistance1));
  }
  
  @Test
  public void testHashCode()
  {
    TimeDistance timeDistance1 = new DefaultTimeDistance(0, 0, 2, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE);
    TimeDistance timeDistance2 = new DefaultTimeDistance(TimeDistance.UNDEFINED_TIME_DISTANCE, 0, 2, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE);
   
    assertTrue(timeDistance1.hashCode() == timeDistance2.hashCode());
  }

  @Test
  public void testCompareTo()
  {
    TimeDistance S60 = new DefaultTimeDistance(60, 0, 0, 0, 0);
    TimeDistance M1 = new DefaultTimeDistance(0, 1, 0, 0, 0);
    
    TimeDistance M2S2 = new DefaultTimeDistance(2, 2, 0, 0, 0);
    TimeDistance S122 = new DefaultTimeDistance(122, 0, 0, 0, 0);
    
    TimeDistance M60 = new DefaultTimeDistance(0, 60, 0, 0, 0);
    TimeDistance H1 = new DefaultTimeDistance(0, 0, 1, 0, 0);
    TimeDistance H24 = new DefaultTimeDistance(0, 0, 24, 0, 0);
    TimeDistance D1 = new DefaultTimeDistance(0, 0, 0, 1, 0);
    
    
    assertTrue(DefaultTimeDistance.ONE_TIME_DISTANCE.compareTo(DefaultTimeDistance.ONE_TIME_DISTANCE) == 0);
    assertTrue(DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) == 0);
    assertTrue(DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE.compareTo(DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE) == 0);
    assertTrue(DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE.compareTo(DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE) == 0);
    assertTrue(M1.compareTo(M1) == 0);
    assertTrue(S60.compareTo(S60) == 0);
    
    assertTrue(M1.compareTo(S60) == 0);
    assertTrue(S60.compareTo(M1) == 0);
    
    assertTrue(M2S2.compareTo(S122) == 0);
    assertTrue(S122.compareTo(M2S2) == 0);
    
    assertTrue(M2S2.compareTo(M1) > 0);
    assertTrue(M1.compareTo(M2S2) < 0);
    
    assertTrue(M2S2.compareTo(S60) > 0);
    assertTrue(S60.compareTo(M2S2) < 0);
        
    assertTrue(H1.compareTo(M60) == 0);
    assertTrue(M60.compareTo(H1) == 0);
    
    assertTrue(D1.compareTo(H24) == 0);
    assertTrue(H24.compareTo(D1) == 0);
    
    assertTrue(DefaultTimeDistance.ONE_TIME_DISTANCE.compareTo(M1) < 0);
    assertTrue(DefaultTimeDistance.ONE_TIME_DISTANCE.compareTo(S60) < 0);
    assertTrue(M1.compareTo(DefaultTimeDistance.ONE_TIME_DISTANCE) > 0);
    assertTrue(S60.compareTo(DefaultTimeDistance.ONE_TIME_DISTANCE) > 0);
    
    assertTrue(DefaultTimeDistance.ONE_TIME_DISTANCE.compareTo(DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE) < 0);
    assertTrue(DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE.compareTo(DefaultTimeDistance.ONE_TIME_DISTANCE) > 0);
    
    assertTrue(DefaultTimeDistance.ONE_TIME_DISTANCE.compareTo(DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE) > 0);
    assertTrue(DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE.compareTo(DefaultTimeDistance.ONE_TIME_DISTANCE) < 0);
    
    assertTrue(DefaultTimeDistance.ZERO_TIME_DISTANCE.compareTo(DefaultTimeDistance.ONE_TIME_DISTANCE) < 0);
    assertTrue(DefaultTimeDistance.ONE_TIME_DISTANCE.compareTo(DefaultTimeDistance.ZERO_TIME_DISTANCE) > 0);
    
    assertTrue(DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE.compareTo(DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE) < 0);
    assertTrue(DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE.compareTo(DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE) > 0);
  }

  /**
   * 
   * @param v1
   * @param v2
   * @return the minimum of the two arguments. If both are equal the first argument (<code>v1</code>) is returned.
   */
  public TimeDistance min(TimeDistance v1, TimeDistance v2)
  {
    return v1.compareTo(v2) <= 0 ? v1 : v2;
  }

  @Test
  public void testIsInfinity()
  {
    TimeDistance S60 = new DefaultTimeDistance(60, 0, 0, 0, 0);
    TimeDistance M1 = new DefaultTimeDistance(0, 1, 0, 0, 0);
    
    assertFalse(S60.isInfinity());
    assertFalse(M1.isInfinity());
    assertFalse(DefaultTimeDistance.ONE_TIME_DISTANCE.isInfinity());
    assertFalse(DefaultTimeDistance.ZERO_TIME_DISTANCE.isInfinity());
    assertTrue(DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE.isInfinity());
    assertTrue(DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE.isInfinity());
  }

  @Test
  public void testSign()
  {
    TimeDistance S60 = new DefaultTimeDistance(60, 0, 0, 0, 0);
    TimeDistance M1 = new DefaultTimeDistance(0, 1, 0, 0, 0);
    
    TimeDistance S60n = new DefaultTimeDistance(-60, 0, 0, 0, 0);
    
    assertEquals(1, S60.sign());
    assertEquals(-1, S60n.sign());
    assertEquals(1, M1.sign());
    assertEquals(-1, M1.neg().sign());
    assertEquals(1, DefaultTimeDistance.ONE_TIME_DISTANCE.sign());
    assertEquals(0, DefaultTimeDistance.ZERO_TIME_DISTANCE.sign());
    assertEquals(1, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE.sign());
    assertEquals(-1, DefaultTimeDistance.NEGATIVE_INFINITE_TIME_DISTANCE.sign());
  }

  @Test
  public void testSub()
  {
    TimeDistance S60 = new DefaultTimeDistance(60, 0, 0, 0, 0);
    TimeDistance M1 = new DefaultTimeDistance(0, 1, 0, 0, 0);
    TimeDistance M2 = new DefaultTimeDistance(0, 2, 0, 0, 0);
    
    TimeDistance M2S2 = new DefaultTimeDistance(2, 2, 0, 0, 0);
    TimeDistance M3S2 = new DefaultTimeDistance(2, 3, 0, 0, 0);
//    TimeDistance S122 = new DefaultTimeDistance(122, 0, 0, 0, 0);
    
//    TimeDistance M60 = new DefaultTimeDistance(0, 60, 0, 0, 0);
//    TimeDistance H1 = new DefaultTimeDistance(0, 0, 1, 0, 0);
//    TimeDistance H24 = new DefaultTimeDistance(0, 0, 24, 0, 0);
//    TimeDistance D1 = new DefaultTimeDistance(0, 0, 24, 0, 0);
//    
//    TimeDistance M1D2H3M4S5 = new DefaultTimeDistance(5, 4, 3, 2, 1);
//    TimeDistance M2D4H6M8S0 = new DefaultTimeDistance(10, 8, 6, 4, 2);
    
    assertEquals(M1, M2.sub(S60));
    assertEquals(M1, M2.sub(S60));
    
    assertEquals(M2S2, M3S2.sub(S60));
    assertEquals(M2S2, M3S2.sub(S60));
  }

  @Test
  public void testAdd()
  {
    TimeDistance S60 = new DefaultTimeDistance(60, 0, 0, 0, 0);
    TimeDistance M1 = new DefaultTimeDistance(0, 1, 0, 0, 0);
    TimeDistance M2 = new DefaultTimeDistance(0, 2, 0, 0, 0);
    
    TimeDistance nM1 = new DefaultTimeDistance(0, -1, 0, 0, 0);
    TimeDistance nM2 = new DefaultTimeDistance(0, -2, 0, 0, 0);
    
    TimeDistance M2S2 = new DefaultTimeDistance(2, 2, 0, 0, 0);
    TimeDistance M3S2 = new DefaultTimeDistance(2, 3, 0, 0, 0);
    TimeDistance S122 = new DefaultTimeDistance(122, 0, 0, 0, 0);
    
//    TimeDistance M60 = new DefaultTimeDistance(0, 60, 0, 0, 0);
//    TimeDistance H1 = new DefaultTimeDistance(0, 0, 1, 0, 0);
//    TimeDistance H24 = new DefaultTimeDistance(0, 0, 24, 0, 0);
//    TimeDistance D1 = new DefaultTimeDistance(0, 0, 24, 0, 0);
    
    TimeDistance M1D2H3M4S5 = new DefaultTimeDistance(5, 4, 3, 2, 1);
    TimeDistance M2D4H6M8S0 = new DefaultTimeDistance(10, 8, 6, 4, 2);
    
    assertEquals(M2, S60.add(S60));
    assertEquals(M2, M1.add(S60));
    
    assertEquals(M3S2, M2S2.add(S60));
    assertEquals(M3S2, S122.add(S60));
    
    assertEquals(M2D4H6M8S0, M1D2H3M4S5.add(M1D2H3M4S5));
    
    assertEquals(nM2, nM1.add(nM1));
    assertEquals(M1, M2.add(nM1));
    assertEquals(nM1, nM2.add(M1));
    
    assertEquals(DefaultTimeDistance.ZERO_TIME_DISTANCE, M1.add(nM1));
  }

}
