package de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.util.Arrays;
import java.util.List;
import java.util.Map;

import org.junit.Test;

import de.aristaflow.adept2.model.timemodel.tools.DefaultUndirectedGraph;

public class CycleEquivalenceTest
{

  /**
   * This is the example from <i>Johnson et al. 94 - The Program Structure
   * Tree: Computing Control Regions in Linear Time</i> Fig. 3 (b)
   */
  @Test
  public void testTransform() throws Exception
  {
    int nodeID = 0;
    NodeStruct n0 = new NodeStruct(nodeID++, true);
    NodeStruct n1 = new NodeStruct(nodeID++, true);
    NodeStruct n2 = new NodeStruct(nodeID++, true);
    NodeStruct n3 = new NodeStruct(nodeID++, true);
    NodeStruct n4 = new NodeStruct(nodeID++, true);
    NodeStruct n5 = new NodeStruct(nodeID++, true);
    NodeStruct n6 = new NodeStruct(nodeID++, true);
    NodeStruct n7 = new NodeStruct(nodeID++, true);
    
    List<NodeStruct> vertices = Arrays.asList(new NodeStruct[]{n0, n1, n2, n3, n4, n5, n6, n7});
    DefaultUndirectedGraph<NodeStruct, EdgeStruct> graph = new DefaultUndirectedGraph<NodeStruct, EdgeStruct>(vertices);
    
    EdgeStruct e01 = new EdgeStruct(n0, n1);
    graph.addEdge(n0, n1, e01);
    EdgeStruct e12 = new EdgeStruct(n1, n2);
    graph.addEdge(n1, n2, e12);
    EdgeStruct e23 = new EdgeStruct(n2, n3);
    graph.addEdge(n2, n3, e23);
    EdgeStruct e34 = new EdgeStruct(n3, n4);
    graph.addEdge(n3, n4, e34);
    EdgeStruct e45 = new EdgeStruct(n4, n5);
    graph.addEdge(n4, n5, e45);
    EdgeStruct e56 = new EdgeStruct(n5, n6);
    graph.addEdge(n5, n6, e56);
    EdgeStruct e67 = new EdgeStruct(n6, n7);
    graph.addEdge(n6, n7, e67);
    
    EdgeStruct x = new EdgeStruct(n0, n7);
    graph.addEdge(n0, n7, x);
    EdgeStruct z = new EdgeStruct(n1, n5);
    graph.addEdge(n1, n5, z);
    EdgeStruct y = new EdgeStruct(n2, n6);
    graph.addEdge(n2, n6, y);
//    graph.addEdge(n3, n4, new EdgeStruct(n3, n4));
    
//    DefaultGraphVisitor<NodeStruct, EdgeStruct> visitor = new DefaultGraphVisitor<NodeStruct, EdgeStruct>("testTransform");    
//    graph.visit(visitor);
    
    CycleEquivalence cycleEquivalence = new CycleEquivalence();
    Map<CQClass, List<EdgeStruct>> equivalencyClasses = cycleEquivalence.transform(graph, n0);
    
//    graph.visit(visitor);
    assertNotNull(equivalencyClasses);
    assertEquals(5, equivalencyClasses.size());
    
    List<EdgeStruct> x1 = equivalencyClasses.get(new CQClass(x, 1));
    assertNotNull(x1);
    assertEquals(3, x1.size());
    assertTrue(x1.contains(e01));
    assertTrue(x1.contains(e67));
    assertTrue(x1.contains(x));
    
    List<EdgeStruct> z2 = equivalencyClasses.get(new CQClass(z, 2));
    assertNotNull(z2);
    assertEquals(1, z2.size());
    assertTrue(z2.contains(e12));
    
    List<EdgeStruct> z3 = equivalencyClasses.get(new CQClass(z, 3));
    assertNotNull(z3);
    assertEquals(3, z3.size());
    assertTrue(z3.contains(e23));
    assertTrue(z3.contains(e34));
    assertTrue(z3.contains(e45));
    
    
    List<EdgeStruct> y2 = equivalencyClasses.get(new CQClass(y, 2));
    assertNotNull(y2);
    assertEquals(2, y2.size());
    assertTrue(y2.contains(e56));
  }


}
