package de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.junit.Ignore;
import org.junit.Test;

import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition.UndirectedGraphTransformation.UndirectedGraphFactory;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;
import de.aristaflow.adept2.model.timemodel.tools.DefaultUndirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.UndirectedGraph;
import de.aristaflow.adept2.tools.dot.DefaultGraphVisitor;

public class NestedSESETest
{

  @Test @Ignore
  public void testSESE() throws Exception
  {
    TemporalTemplate template = TestTemplates.simpleTimeTestContingentTestTemplate;
    
    UndirectedGraphTransformation<NodeStruct, EdgeStruct> graphTrafo = new UndirectedGraphTransformation<NodeStruct, EdgeStruct>(new UndirectedGraphFactory<NodeStruct, EdgeStruct>()
    {

      public NodeStruct createVertex(Object vertex, boolean representative)
      {
        return new NodeStruct(vertex, representative);
      }

      public EdgeStruct createEdge(NodeStruct v1, NodeStruct v2, Object edge)
      {
        return new EdgeStruct(v1, v2);
      }

      public UndirectedGraph<NodeStruct, EdgeStruct> createGraph(List<NodeStruct> vertices)
      {
        return new DefaultUndirectedGraph<NodeStruct, EdgeStruct>(vertices);
      }
    });

    UndirectedGraph<NodeStruct, EdgeStruct> graph = graphTrafo.transform(template);
    NodeStruct sourceNode = graphTrafo.sourceNode;
    NodeStruct sinkNode = graphTrafo.sinkNode;
    

    
    NestedSESE sese = new NestedSESE();
    
    RegionStruct rootRegion = sese.transform(graph, sourceNode, sinkNode);
    
    DefaultGraphVisitor<NodeStruct, EdgeStruct> visitor = new DefaultGraphVisitor<NodeStruct, EdgeStruct>("testSESE");
    graph.visit(visitor);
//    System.out.println(rootRegion);
    rootRegion = sese.canonicalize(rootRegion);
    
//    System.out.println(rootRegion);
    
    SESERegion regions = graphTrafo.transform(rootRegion);
    System.out.println(regions);
    
    List<Node> nodes = new ArrayList<Node>();
    traverse(regions, nodes ); // Collects all the nodes in the SESERegions and asserts that no node is contained in more than one region
    
    assertEquals(template.getNodes().size(), nodes.size());
  }
  
  @Ignore @Test
  public void testSESE2() throws Exception
  {
    TemporalTemplate template = TestTemplates.coopIS13TestTemplate;
        
    NestedSESE sese = new NestedSESE();
    
    SESERegion regions = sese.transform(template);
    
    List<Node> nodes = new ArrayList<Node>();
    traverse(regions, nodes ); // Collects all the nodes in the SESERegions and asserts that no node is contained in more than one region
    
    assertEquals(nodes.size(), template.getNodes().size());
    
    System.out.println(regions);
  }

  private void traverse(SESERegion region, Collection<Node> nodes)
  {
    for (Node node : region.getNodes())
    {
      assertTrue(nodes.add(node)); // No duplicate nodes
    }
    for (SESERegion child : region.getChildren())
    {
      traverse(child, nodes);
    }
    
  }

}
