package de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition;

import java.util.Arrays;
import java.util.List;

import org.junit.Test;

import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.algorithms.sesedecomposition.UndirectedGraphTransformation.UndirectedGraphFactory;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;
import de.aristaflow.adept2.model.timemodel.tools.DefaultDirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.DefaultUndirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.DirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.UndirectedGraph;
import de.aristaflow.adept2.tools.dot.DefaultGraphVisitor;

public class UndirectedGraphTransformationTest
{
  
  @Test 
  public void testTransformDirectedGraphN()
  {
    NodeStruct sourceNode = new NodeStruct(null, false);
    DirectedGraph<NodeStruct, EdgeStruct> sampleGraph = getSampleGraph(sourceNode);
    
//    DefaultGraphVisitor<NodeStruct, EdgeStruct> visitor = new DefaultGraphVisitor<NodeStruct, EdgeStruct>("testTransformDirectedGraphN");
//    sampleGraph.visit(visitor);
    
    UndirectedGraphTransformation<NodeStruct, EdgeStruct> trafo = new UndirectedGraphTransformation<NodeStruct, EdgeStruct>(new UndirectedGraphFactory<NodeStruct, EdgeStruct>()
    {

      public NodeStruct createVertex(Object vertex, boolean representative)
      {
        return new NodeStruct(vertex, representative);
      }

      public EdgeStruct createEdge(NodeStruct n1, NodeStruct n2, Object edge)
      {
        return new EdgeStruct(n1, n2);
      }

      public UndirectedGraph<NodeStruct, EdgeStruct> createGraph(List<NodeStruct> vertices)
      {
        return new DefaultUndirectedGraph<NodeStruct, EdgeStruct>(vertices);
      }
    });
    
    
    UndirectedGraph<NodeStruct,EdgeStruct> graph = trafo.transform(sampleGraph, sourceNode);
    
//    graph.visit(visitor);
    
  }
  

  /**
   * @see Ananian 1997 - The Static Single Information Form pp. 28
   * @param sourceNode
   * @return
   */
  public static DirectedGraph<NodeStruct, EdgeStruct> getSampleGraph(NodeStruct sourceNode)
  {   
    int nodeID = 1;
    NodeStruct n1 = new NodeStruct(nodeID++, true);
    NodeStruct n2 = new NodeStruct(nodeID++, true);
    NodeStruct n3 = new NodeStruct(nodeID++, true);
    NodeStruct n4 = new NodeStruct(nodeID++, true);
    NodeStruct n5 = new NodeStruct(nodeID++, true);
    NodeStruct n6 = new NodeStruct(nodeID++, true);
    NodeStruct n7 = new NodeStruct(nodeID++, true);
    NodeStruct n8 = new NodeStruct(nodeID++, true);
    NodeStruct n9 = new NodeStruct(nodeID++, true);
    NodeStruct n10 = new NodeStruct(nodeID++, true);
    NodeStruct n11 = new NodeStruct(nodeID++, true);
    NodeStruct n12 = new NodeStruct(nodeID++, true);
    NodeStruct n13 = new NodeStruct(nodeID++, true);
    NodeStruct n14 = new NodeStruct(nodeID++, true);
    NodeStruct n15 = new NodeStruct(nodeID++, true);
    NodeStruct n16 = new NodeStruct(nodeID++, true);
    NodeStruct n17 = new NodeStruct(nodeID++, true);
    
    List<NodeStruct> vertices = Arrays.asList(new NodeStruct[]{sourceNode, n1, n2, n3, n4, n5, n6, n7, n8, n9, n10, n11, n12, n13, n14, n15, n16, n17});
    DefaultDirectedGraph<NodeStruct, EdgeStruct> graph = new DefaultDirectedGraph<NodeStruct, EdgeStruct>(vertices);
    
    graph.addEdge(sourceNode, n1, new EdgeStruct(sourceNode, n1));
    graph.addEdge(n1, n2, new EdgeStruct(n1, n2));
    graph.addEdge(n2, n3, new EdgeStruct(n2, n3));
    graph.addEdge(n3, n4, new EdgeStruct(n3, n4));
    graph.addEdge(n4, n5, new EdgeStruct(n4, n5));
    graph.addEdge(n4, n6, new EdgeStruct(n4, n6));
    graph.addEdge(n5, n7, new EdgeStruct(n5, n7));
    graph.addEdge(n6, n7, new EdgeStruct(n6, n7));
    graph.addEdge(n7, n8, new EdgeStruct(n7, n8));
    graph.addEdge(n8, n2, new EdgeStruct(n8, n2));
    graph.addEdge(n8, n16, new EdgeStruct(n8, n16));
    
    graph.addEdge(n1, n9, new EdgeStruct(n1, n9));
    graph.addEdge(n9, n10, new EdgeStruct(n9, n10));
    graph.addEdge(n10, n11, new EdgeStruct(n10, n11));
    graph.addEdge(n10, n12, new EdgeStruct(n10, n12));
    graph.addEdge(n11, n13, new EdgeStruct(n11, n13));
    graph.addEdge(n12, n13, new EdgeStruct(n12, n13));
    graph.addEdge(n13, n14, new EdgeStruct(n13, n14));
    graph.addEdge(n14, n12, new EdgeStruct(n14, n12));
    graph.addEdge(n14, n15, new EdgeStruct(n14, n15));
    graph.addEdge(n15, n16, new EdgeStruct(n15, n16));
    
    graph.addEdge(n16, n17, new EdgeStruct(n16, n17));

    return graph;
  }


  @Test
  public void testTransformTemporalTemplate() throws Exception
  {
    TemporalTemplate template = TestTemplates.simpleTimeTestContingentTestTemplate;
    
    
    UndirectedGraphTransformation<NodeStruct, EdgeStruct> trafo = new UndirectedGraphTransformation<NodeStruct, EdgeStruct>(new UndirectedGraphFactory<NodeStruct, EdgeStruct>()
    {

      public NodeStruct createVertex(Object vertex, boolean representative)
      {
        return new NodeStruct(vertex, representative);
      }

      public EdgeStruct createEdge(NodeStruct n1, NodeStruct n2, Object edge)
      {
        return new EdgeStruct(n1, n2);
      }

      public UndirectedGraph<NodeStruct, EdgeStruct> createGraph(List<NodeStruct> vertices)
      {
        return new DefaultUndirectedGraph<NodeStruct, EdgeStruct>(vertices);
      }
    });
    
    UndirectedGraph<NodeStruct,EdgeStruct> graph = trafo.transform(template);
    
    DefaultGraphVisitor<NodeStruct, EdgeStruct> visitor = new DefaultGraphVisitor<NodeStruct, EdgeStruct>("UndirectedGraphTransformationTest");
    graph.visit(visitor);
    
  }

}
