package de.aristaflow.adept2.model.processmodel.simulator;

import java.util.Arrays;

import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.globals.ProcessConstants.EdgeType;
import de.aristaflow.adept2.model.processmodel.TemporalNode;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.timemodel.ActivityDuration;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;

public class DefaultSimulationStrategy implements SimulationStrategy
{

  protected DefaultTimeModelOperations timeModelTools = new DefaultTimeModelOperations();
  private TemporalTemplate template;
  protected Distribution distribution;
  
  public DefaultSimulationStrategy(TemporalTemplate template,
      Distribution distribution)
  {
    this.template = template;
    this.distribution = distribution;
  }

  public long determineNodeDuration(TemporalNode node)
  {
    ActivityDuration activityDuration = node.getActivityDuration();
    
    long duration;
    
    if (activityDuration != null)
    {
      TimePoint minimumDuration = activityDuration.getMinimumDuration();
      TimePoint maximumDuration = activityDuration.getMaximumDuration();
      
      long minDuration = timeModelTools.convertToSeconds(minimumDuration);
      long maxDuration = timeModelTools.convertToSeconds(maximumDuration);
      
      
      duration = distribution.drawRandom(minDuration, maxDuration);
    }
    else
    {
      duration = 60;
    }
    return secondsToMillis(duration);  
  }


  protected long secondsToMillis(long duration)
  {
    return duration * 1000l;
  }

  public long getProcessInstanceCreationTime()
  {
    //return System.currentTimeMillis();
    return 2000;
  }

  public long determineNodeStartTime(TemporalNode node, long currentTime)
  {
    return currentTime + secondsToMillis(distribution.drawRandom(10, 5*60));
  }

  public int determineXORBranch(TemporalNode node)
  {
    int[] succs = template.getSuccByEdgeType(node.getID(), EdgeType.ET_CONTROL);
    
    double[] probs = new double[succs.length];
    
    Arrays.fill(probs, 1d/succs.length);
    
    int index = distribution.drawRandom(probs);
    return succs[index];
  }

  public boolean determineLoop(TemporalNode node)
  {
    double repeat = 0.5;
    return distribution.drawRandom(new double[]{repeat, 1-repeat}) == 0;
  }


  
}
