package de.aristaflow.adept2.model.processmodel.timemodel;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import de.aristaflow.adept2.model.common.PluginData;
import de.aristaflow.adept2.model.common.defaultimplementation.DefaultPluginData;
import de.aristaflow.adept2.model.common.plugindata.EscalationData;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.common.viewer.Attribute;
import de.aristaflow.adept2.model.common.viewer.AttributeSetter;
import de.aristaflow.adept2.model.common.viewer.ModelViewer;
import de.aristaflow.adept2.model.globals.ProcessConstants.EdgeType;
import de.aristaflow.adept2.model.globals.TimeModelConstants.RelationType;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.processmodel.StructuredEdge;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.defaultimplementation.DefaultConstraintEdge;
import de.aristaflow.adept2.model.processmodel.defaultimplementation.DefaultStructuredConstraintEdge;
import de.aristaflow.adept2.model.timemodel.tools.DirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;

public class TemporalTemplateDirectedGraph implements DirectedGraph<Node, StructuredEdge>
{
  
  private Template template;

  public TemporalTemplateDirectedGraph(Template template)
  {
    this.template = template;
  }

  public Collection<Node> getVertices()
  {
    return template.getNodes();
  }

  public Collection<StructuredEdge> getEdges()
  {
    Set<StructuredEdge> edgeStructure = new HashSet<StructuredEdge>(template.getEdgeStructure(EdgeType.ET_CONTROL, EdgeType.ET_LOOP, EdgeType.ET_SYNC));
    edgeStructure.addAll(template.getConstraintEdgeStructure());
    return edgeStructure;
  }

  public StructuredEdge getEdge(Node from, Node to)
  {
    // TODO Auto-generated method stub
    return null;
  }

  public void visit(
      GraphVisitor<Node, StructuredEdge> visitor)
  {
    visitor.discoverGraph(this);
    
    for (Node n : getVertices())
    {
      visitor.discoverVertex(n);
      visitor.finishVertex(n);
    }
    
    for (StructuredEdge e : getEdges())
    {
      
      Node startVertex = template.getNode(e.getSourceNodeID());
      Node endVertex = template.getNode(e.getDestinationNodeID());
      visitor.discoverEdge(startVertex, endVertex, e);
      
      visitor.finishEdge(startVertex, endVertex, e);
    }
    
    
    final ProcessDuration processDuration = TimeModelTools.getProcessDuration(template);
    
    if (processDuration != null)
    {
      int startNodeID = template.getStartNode().getID();
      int endNodeID = template.getEndNode().getID();
      Node startVertex = template.getNode(startNodeID);
      Node endVertex = template.getNode(endNodeID);
      DefaultConstraintEdge edge = new DefaultConstraintEdge();
      
      Map<String, String> temporalInformation = new HashMap<String, String>();
      
      ConstraintEdgeTemporalInformation edgeTemporalInformation = ModelViewer
          .createView(temporalInformation, ConstraintEdgeTemporalInformation.class);

       
        TimeLag[] timeLags = new TimeLag[]{new TimeLag(){

          @Attribute(name = "RelationType")
          public RelationType getRelationType()
          {
            return RelationType.END_START;
          }

          @Attribute(name = "LowerMinimumTimeLag")
          public TimePoint getLowerMinimumTimeLag()
          {
            return processDuration.getMinimumDuration();
          }

          @Attribute(name = "UpperMinimumTimeLag")
          public TimePoint getUpperMinimumTimeLag()
          {
            return processDuration.getMinimumDuration();
          }

          @Attribute(name = "LowerMaximumTimeLag")
          public TimePoint getLowerMaximumTimeLag()
          {
            return processDuration.getGuardMaximumDuration();
          }

          @Attribute(name = "UpperMaximumTimeLag")
          public TimePoint getUpperMaximumTimeLag()
          {
            return processDuration.getMaximumDuration();
          }

          @Attribute(name = "ExpectedTimeLag")
          public TimePoint getExpectedTimeLag()
          {
            return processDuration.getExpectedDuration();
          }

          @Attribute(name = "TimeLagUpdateHandler")
          public TimeLagUpdateProcedureData getTimeLagUpdateHandler()
          {
            return null;
          }

          @Attribute(name = "TemporalExceptionHandlingProcedure")
          public PluginData getTemporalExceptionHandlingProcedure()
          {
            return null;
          }

          @Attribute(name = "EscalationData")
          public EscalationData getEscalationData()
          {
            return null;
          }

          @AttributeSetter("RelationType")
          public void setRelationType(RelationType relationType)
          {
            //
          }

          @AttributeSetter("LowerMinimumTimeLag")
          public void setLowerMinimumTimeLag(TimePoint duration)
          {
            //
          }

          @AttributeSetter("UpperMinimumTimeLag")
          public void setUpperMinimumTimeLag(TimePoint duration)
          {
            //
          }

          @AttributeSetter("LowerMaximumTimeLag")
          public void setLowerMaximumTimeLag(TimePoint duration)
          {
            //
          }

          @AttributeSetter("UpperMaximumTimeLag")
          public void setUpperMaximumTimeLag(TimePoint duration)
          {
            //
          }

          @AttributeSetter("ExpectedTimeLag")
          public void setExpectedTimeLag(TimePoint duration)
          {
            //
          }
          
        }};

        edgeTemporalInformation.setTimeLags(timeLags);
      
      PluginData edgeTemporalInformationPluginData = new DefaultPluginData(ConstraintEdgeTemporalInformation.ID,
                                                                           temporalInformation);
      
      edge.addSupportedPlugin(ConstraintEdgeTemporalInformation.ID, ConstraintEdgeTemporalInformation.ID, edgeTemporalInformationPluginData);
      visitor.discoverEdge(startVertex, endVertex, new DefaultStructuredConstraintEdge(edge, startNodeID, endNodeID));
    }
    
    visitor.finishGraph(this);
  }

  public Template getTemplate()
  {
    return template;
  }

}
