package de.aristaflow.adept2.model.processmodel.timemodel;

import de.aristaflow.adept2.model.globals.ProcessConstants.EdgeType;
import de.aristaflow.adept2.model.globals.ProcessConstants.NodeType;
import de.aristaflow.adept2.model.processmodel.ConstraintEdge;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.processmodel.StructuredConstraintEdge;
import de.aristaflow.adept2.model.processmodel.StructuredEdge;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.tools.Graph;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;
import de.aristaflow.adept2.tools.dot.AbstractDotGraph;
import de.aristaflow.adept2.tools.dot.DefaultGraphVisitor;
import de.aristaflow.adept2.tools.dot.DotColor;
import de.aristaflow.adept2.tools.dot.DotEdge;
import de.aristaflow.adept2.tools.dot.DotEdge.EdgeStyle;
import de.aristaflow.adept2.tools.dot.DotNode;
import de.aristaflow.adept2.tools.dot.DotNode.NodeShape;
import de.aristaflow.adept2.tools.dot.DotTools;
import de.aristaflow.adept2.tools.dot.GraphFormatter;
import de.aristaflow.adept2.tools.dot.LabelProvider;

public class TemporalTemplateDotVisitor extends DefaultGraphVisitor<Node, StructuredEdge>
{

  public TemporalTemplateDotVisitor(String name, DotTools dotTools)
  {
    super(name, dotTools);
    
    setLabelProvider(new LabelProvider<Node, StructuredEdge>()
    {

      public String getVertexLabel(Node o)
      {
        String name = "";
        
        Graph<Node, StructuredEdge> graph = getGraph();
        
        if (graph instanceof TemporalTemplateDirectedGraph)
        {
          TemporalTemplateDirectedGraph ttg = (TemporalTemplateDirectedGraph) graph;
          Template template = ttg.getTemplate();
          NodeType nodeType = template.getNodeType(o.getID());
          
          switch (nodeType)
          {
            case NT_AND_JOIN:
            case NT_AND_SPLIT:
              name += "P "; 
              break;
            case NT_ENDLOOP:
            case NT_STARTLOOP:
              name += "O ";
              break;
            case NT_XOR_JOIN:
            case NT_XOR_SPLIT:
              name += "X ";
              break;
            case NT_ENDFLOW:
            case NT_NORMAL:
            case NT_STARTFLOW:
            default:
              break;
            
          }
          
        }
        
        
        name += o.getName() != null && o.getName().length() > 0 ? o.getName() : "#"+o.getID();
        
        
        ActivityDuration activityDuration = TimeModelTools.getActivityDuration(o);
        if (activityDuration != null)
        {
          String duration = TimeModelTools.toString(activityDuration);
          
          name += "\\n"+duration;
        }
        return name;
      }

      public String getEdgeLabel(StructuredEdge o)
      {
        if (EdgeType.ET_CONTROL.equals(o.getEdge().getType()))
          return "";
        
        if (EdgeType.ET_CONSTRAINT.equals(o.getEdge().getType()))
        {
          StructuredConstraintEdge e = (StructuredConstraintEdge) o;
          ConstraintEdge constraintEdge = e.getConstraintEdge();
          TimeLag[] timeLags = TimeModelTools.getTimeLags(constraintEdge);
          
          StringBuilder b = new StringBuilder();
          for (TimeLag t : timeLags)
          {
            b.append(TimeModelTools.toString(t));
          }
          
          return b.toString();
        }
        
        return "";
      }

      public String getGraphLabel(Graph<Node, StructuredEdge> g)
      {
        return "";
      }

      public void setGraphVisitor(GraphVisitor<Node, StructuredEdge> visitor)
      {
        //
      }

      @Override
      public String getEdgeTailLabel(StructuredEdge e)
      {
        return null;
      }
    });
    
    setGraphFormatter(new GraphFormatter<Node, StructuredEdge>()
    {

      public void formatVertex(Node vertex, DotNode node)
      {
        node.setShape(NodeShape.BOX);
      }

      public void formatEdge(StructuredEdge e, DotEdge edge)
      {
        if (EdgeType.ET_CONTROL.equals(e.getEdge().getType()))
        {
          edge.setStyle(EdgeStyle.BOLD);
        }
        if (EdgeType.ET_CONSTRAINT.equals(e.getEdge().getType()))
        {
          edge.setConstraint(false);
          edge.setColor(DotColor.valueOf("blue"));
        }
      }

      public boolean isVisible(StructuredEdge e)
      {
        return true;
      }

      public void formatGraph(Graph<Node, StructuredEdge> graph,
          AbstractDotGraph g)
      {
        // 
      }

      public void setGraphVisitor(GraphVisitor<Node, StructuredEdge> visitor)
      {
        // 
        
      }
    });
  }

  protected String getNodeID(Node vertex)
  {
    return "N"+vertex.getID();
  }
  
}
