package de.aristaflow.adept2.model.processmodel.timemodel;

import static org.junit.Assert.fail;

import java.io.File;
import java.io.IOException;
import java.net.URI;

import org.junit.BeforeClass;
import org.junit.Test;

import de.aristaflow.adept2.TestSuiteADEPT2;
import de.aristaflow.adept2.model.common.viewer.ViewCheck;
import de.aristaflow.adept2.util.CheckReport;

/**
 * Integrity check for all known time model viewer classes.
 * 
 * @author Andreas Lanz
 */
public class TestModelViewerClasses
{
  private static final String ADEPT2PLUGIN_DIR = "../de.aristaflow.adept2.adept2plugin/";
  /** The directory the test configuration could be found in. */
  private static final String CONFIG_DIR = "conf/test"; //$NON-NLS-1$
  /** The directory the test configuration could be found in. */
  private static final String EXAMPLES_DIR = "examples"; //$NON-NLS-1$

  /** Initialise the registry. */
  @BeforeClass
  public static void initialiseRegistry()
  {
    StringBuilder confDir;
    if (System.getProperty("arflow.confdir") == null //$NON-NLS-1$
        || System.getProperty("arflow.confdir").equals("")) //$NON-NLS-1$ //$NON-NLS-2$
    {
      String adept2plugin = null;
      try
      {
        adept2plugin = new File(ADEPT2PLUGIN_DIR).getCanonicalPath();
      }
      catch (IOException e)
      {
        fail("Problems while setting the confdir!"); //$NON-NLS-1$
      }
      confDir = new StringBuilder(adept2plugin);
      confDir.append(File.separatorChar);
      confDir.append(CONFIG_DIR);
      StringBuilder dataDir = new StringBuilder(adept2plugin);
      dataDir.append(File.separatorChar);
      dataDir.append(EXAMPLES_DIR);
      
      // set confDir and dataDir
      System.setProperty("arflow.confdir", confDir.toString()); //$NON-NLS-1$
      System.setProperty("arflow.datadir", dataDir.toString()); //$NON-NLS-1$
    }
    TestSuiteADEPT2.initialiseRegistry();
  }

  /**
   * The classes to be checked.
   */
  public static final Class<?>[] classes = new Class<?>[]
  {
    TemplateTemporalInformation.class,
    NodeTemporalInformation.class,
    ConstraintEdgeTemporalInformation.class,
    ActivityDuration.class, //TP2
    FixedDate.class, //TP4
    ProcessDuration.class, //TP2
    ScheduleRestrictedElement.class, //TP5
    TimeLag.class, //TP1
    ValidityPeriod.class //TP7
  };
  
// LOW @Test
  public void test()
  {
    for (Class<?> clazz : classes)
    {
      CheckReport report = new CheckReport(URI.create("class:/"));
      boolean result = ViewCheck.checkClass(report, clazz);
      if (!result)
      {
        fail(report.getReportSummary());
      }
    }
  }
}
