package de.aristaflow.adept2.model.processmodel.timemodel.schedules.evaluation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.text.SimpleDateFormat;
import java.util.Calendar;

import org.junit.Ignore;
import org.junit.Test;

import de.aristaflow.adept2.core.timemanager.ScheduleResolutionException;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.ScheduleParser;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.BasicCalendar;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.BasicCalendarExpression;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.CalendarExpression;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.Dicing;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.Setops;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.expression.SingleSelection;
import de.aristaflow.adept2.model.processmodel.timemodel.schedules.lang.script.CalendarScript;

public class EvaluatorTest
{

  
  private static final SimpleDateFormat format = new SimpleDateFormat("dd.MM.yy HH:mm:ss.S");
  
  private String format(Calendar from)
  {
    return format.format(from.getTime());
  }

  @Test
  public void test()
  {
    Evaluator eval = new Evaluator(null);
    Calendar from = Calendar.getInstance();
    Calendar to = ((Calendar) from.clone());
    to.add(Calendar.MINUTE, 20);
    {
      IntervalSet result = eval.evaluate(new BasicCalendarExpression(BasicCalendar.MINUTES), from, to, true);
  
      Calendar first = result.getStart();
      Calendar last = result.getEnd();
      
      assertTrue(first.compareTo(from) <= 0);
      assertTrue(last.compareTo(to) >= 0);
    }
    
    {
      IntervalSet result = eval.evaluate(new BasicCalendarExpression(BasicCalendar.MINUTES), from, to, false);

      
      Calendar first = result.getStart();
      Calendar last = result.getEnd();
      
      assertTrue(first.compareTo(from) >= 0);
      assertTrue(last.compareTo(to) <= 0);
    }
    
  }
  
  @Test
  public void test2()
  {
    Evaluator eval = new Evaluator(null);
    Calendar from = Calendar.getInstance();
    Calendar to = ((Calendar) from.clone());
    to.add(Calendar.WEEK_OF_YEAR, 4);
    
    Dicing expression = new Dicing();
    expression.setFirstOperand(new BasicCalendarExpression(BasicCalendar.DAYS));
    expression.setSetop(Setops.during);
    expression.setStrict(true);
    expression.setSecondOperand(new BasicCalendarExpression(BasicCalendar.WEEKS));
    SingleSelection slicing = new SingleSelection();
    slicing.setSelectionFunction(1);
    slicing.setCalendarExpression(expression);
    IntervalSet result = eval.evaluate(slicing, from, to, true);

  }
  
  
  
  @Test
  public void testDuring() throws ScheduleResolutionException
  {

    // TODO implement offset
//    {
//      CalendarExpression calendarExpression = ScheduleParser.parse("2014/YEARS");
//      Evaluator eval = new Evaluator(null);
//      
//      Calendar from = Calendar.getInstance();
//      from.set(2014, 0, 1, 0, 0, 0);
//      from.set(Calendar.MILLISECOND, 0);
//     
//      
//      Calendar to = Calendar.getInstance();
//      to.set(2015, 0, 1, 0, 0, 0);
//      to.set(Calendar.MILLISECOND, 0);
//      to.add(Calendar.MILLISECOND, -1);
//
//      
//      IntervalSet result = eval.evaluate(calendarExpression, from, to, true);
//      
//      assertEquals(from, result.getStart());
//      assertEquals(to, result.getEnd());
//    }
    
    {
      CalendarExpression calendarExpression = ScheduleParser.parse("MONTHS");
      Evaluator eval = new Evaluator(null);
      
      Calendar from = Calendar.getInstance();
      from.set(2014, 0, 1, 0, 0, 0);
      from.set(Calendar.MILLISECOND, 0);
     
      
      Calendar to = Calendar.getInstance();
      to.set(2015, 0, 1, 0, 0, 0);
      to.set(Calendar.MILLISECOND, 0);
      to.add(Calendar.MILLISECOND, -1);

      
      IntervalSet result = eval.evaluate(calendarExpression, from, to, true);
      
      assertEquals(from, result.getStart());
      assertEquals(to, result.getEnd());
    }
    
    {
      CalendarExpression calendarExpression = ScheduleParser.parse("MONTHS:during:the/YEARS");
      Evaluator eval = new Evaluator(null);
      
      Calendar from = Calendar.getInstance();
      from.set(2014, 0, 1, 0, 0, 0);
      from.set(Calendar.MILLISECOND, 0);
     
      
      Calendar to = Calendar.getInstance();
      to.set(2015, 0, 1, 0, 0, 0);
      to.set(Calendar.MILLISECOND, 0);
      to.add(Calendar.MILLISECOND, -1);
      

      IntervalSet result = eval.evaluate(calendarExpression, from, to, true);

      
      assertEquals("Expected "+format(from) + " but was "+format(result.getStart()), from, result.getStart());
      assertEquals("Expected "+format(to) + " but was "+format(result.getEnd()), to, result.getEnd());
    }
    
    {
      CalendarExpression calendarExpression = ScheduleParser.parse("2/MONTHS:during:the/YEARS");
      Evaluator eval = new Evaluator(null);
      
      Calendar from = date(2014, 0, 1);
     
      
      Calendar to = date(2015, 0, 1);
      to.add(Calendar.MILLISECOND, -1);
      

      IntervalSet result = eval.evaluate(calendarExpression, from, to, true);
      
      Calendar start = date(2014, 1, 1);
      Calendar end = date(2014, 2, 1);
      end.add(Calendar.MILLISECOND, -1);

      
      assertEquals(String.format("Expected %s but was %s", format(start), format(result.getStart())), start, result.getStart());
      assertEquals(String.format("Expected %s but was %s", format(end), format(result.getEnd())), end, result.getEnd());
    }
    
    {
      CalendarExpression calendarExpression = ScheduleParser.parse("WEEKS:during:5/MONTHS:during:the/YEARS");
      Evaluator eval = new Evaluator(null);
      
      Calendar from = date(2014, 0, 1);
     
      
      Calendar to = date(2015, 0, 1);
      to.add(Calendar.MILLISECOND, -1);
      

      
      IntervalSet result = eval.evaluate(calendarExpression, from, to, true);
      
      Calendar start = date(2014, 4, 5);
      Calendar end = date(2014, 4, 26);
      end.add(Calendar.MILLISECOND, -1);

      
      assertEquals(String.format("Expected %s but was %s", format(start), format(result.getStart())), start, result.getStart());
      assertEquals(String.format("Expected %s but was %s", format(end), format(result.getEnd())), end, result.getEnd());
    }
  }
  
  @Test
  public void testSingleSelection() throws ScheduleResolutionException
  {
    {
      CalendarExpression calendarExpression = ScheduleParser.parse("1/DAYS:during:WEEKS");
      
      Evaluator eval = new Evaluator(null);
      
      Calendar from = date(2014, 4, 1);
      
      Calendar to = date(2014, 5, 1);
      to.add(Calendar.MILLISECOND, -1);
      
      IntervalSet result = eval.evaluate(calendarExpression, from, to, true);
      
      Calendar start = date(2014, 3, 28);
      
      Calendar end = date(2014, 4, 27);
      end.add(Calendar.MILLISECOND, -1);
      
      assertEquals(5, result.size());
      assertEquals(String.format("Expected %s but was %s", format(start), format(result.getStart())), start, result.getStart());
      assertEquals(String.format("Expected %s but was %s", format(end), format(result.getEnd())), end, result.getEnd());
      for (TimeFrame frame : result.getTimeFrames())
      {
        assertTrue(frame instanceof Interval);        
      }
    }
  }
  


  @Test
  public void testSetSelection() throws ScheduleResolutionException
  {
    {
      CalendarExpression calendarExpression = ScheduleParser.parse("[1-5]/DAYS:during:WEEKS");
      
      Evaluator eval = new Evaluator(null);
      
      Calendar from = date(2014, 4, 1);
      
      Calendar to = date(2014, 5, 1);
      to.add(Calendar.MILLISECOND, -1);
      
      IntervalSet result = eval.evaluate(calendarExpression, from, to, true);
      
      Calendar start = date(2014, 3, 28);
      
      Calendar end = date(2014, 4, 31);
      end.add(Calendar.MILLISECOND, -1);
      
      assertEquals(5, result.size());
      assertEquals(String.format("Expected %s but was %s", format(start), format(result.getStart())), start, result.getStart());
      assertEquals(String.format("Expected %s but was %s", format(end), format(result.getEnd())), end, result.getEnd());
      for (TimeFrame frame : result.getTimeFrames())
      {
        assertTrue(frame instanceof IntervalSet);
        IntervalSet set = (IntervalSet) frame;
        assertEquals(5, set.size());
        for (TimeFrame tf : set.getTimeFrames())
        {
          assertTrue(tf instanceof Interval);
        }
        
      }
    }
    
    {
      
      CalendarExpression calendarExpression = ScheduleParser.parse("[8-13]/HOURS:during:[1-5]/DAYS:during:WEEKS");
      
      Evaluator eval = new Evaluator(null);
      
      Calendar from = date(2014, 4, 1);
      
      Calendar to = date(2014, 5, 1);
      to.add(Calendar.MILLISECOND, -1);
      
      IntervalSet result = eval.evaluate(calendarExpression, from, to, true);
      
      Calendar start = date(2014, 3, 28, 7);
      
      Calendar end = date(2014, 4, 30, 13);
      end.add(Calendar.MILLISECOND, -1);
      
      assertEquals(5*5, result.size());
      assertEquals(String.format("Expected %s but was %s", format(start), format(result.getStart())), start, result.getStart());
      assertEquals(String.format("Expected %s but was %s", format(end), format(result.getEnd())), end, result.getEnd());
      for (TimeFrame frame : result.getTimeFrames())
      {
        assertTrue(frame instanceof IntervalSet);
        IntervalSet set = (IntervalSet) frame;
        assertEquals(6, set.size());
        for (TimeFrame tf : set.getTimeFrames())
        {
          assertTrue(tf instanceof Interval);
        }
        
      }
      
//      System.out.println(result);
    }
  }


  private Calendar date(int year, int month, int date)
  {
    return date(year, month, date, 0);
  }
  

  private Calendar date(int year, int month, int date, int hour)
  {
    Calendar cal = Calendar.getInstance();
    cal.set(year, month, date, hour, 0, 0);
    cal.set(Calendar.MILLISECOND, 0);
    return cal;
  }

  @Ignore @Test
  public void testNextInterval() throws Exception
  {
    CalendarExpression calendarExpression = ScheduleParser.parse("1/DAYS:during:WEEKS");
    
    Evaluator eval = new Evaluator(null);
    
    Calendar from = date(2014, 4, 16);
    
    TimeFrame result = eval.nextInterval(calendarExpression, from, true);
        
    Calendar expected = (Calendar) from.clone();
    expected.set(Calendar.DAY_OF_WEEK, expected.getFirstDayOfWeek());
    expected.add(Calendar.WEEK_OF_YEAR, 1);

    
    assertEquals("Expected "+format(expected) + " but was "+format(result.getStart()), expected, result.getStart());
  }
  
  @Test
  public void testBasicCalendar()
  {
    {
      
      CalendarExpression calendarExpression = new BasicCalendarExpression(BasicCalendar.DAYS);
      
      Evaluator eval = new Evaluator(null);
      
      Calendar from = date(2014, 4, 1);
      
      Calendar to = date(2014, 5, 1);
      to.add(Calendar.MILLISECOND, -1);
      
      IntervalSet result = eval.evaluate(calendarExpression, from, to, true);
      
      assertEquals(31, result.size());
      
      Calendar first = date(2014,4,1);
      Calendar last = date(2014,4,31);
      
      assertEquals(first, result.get(0).getStart());
      
      first.add(Calendar.DAY_OF_YEAR, 1);
      first.add(Calendar.MILLISECOND, -1);     
      assertEquals(first, result.get(0).getEnd());
      
      
      
      assertEquals(last, result.get(result.size()-1).getStart());
      
      last.add(Calendar.DAY_OF_YEAR, 1);
      last.add(Calendar.MILLISECOND, -1);
      assertEquals(last, result.get(result.size()-1).getEnd());
    }
    
    {
      
      CalendarExpression calendarExpression = new BasicCalendarExpression(BasicCalendar.WEEKS);
      
      Evaluator eval = new Evaluator(null);
      
      Calendar from = date(2014, 4, 1);
      
      Calendar to = date(2014, 5, 1);
      to.add(Calendar.MILLISECOND, -1);
      
      IntervalSet result = eval.evaluate(calendarExpression, from, to, true);
      
      assertEquals(5, result.size());
      
      Calendar first = date(2014,3,28);
      Calendar last = date(2014,4,26);
      
      assertEquals(first, result.get(0).getStart());
      
      first.add(Calendar.WEEK_OF_YEAR, 1);
      first.add(Calendar.MILLISECOND, -1);     
      assertEquals(first, result.get(0).getEnd());
      
      
      
      assertEquals(last, result.get(result.size()-1).getStart());
      
      last.add(Calendar.WEEK_OF_YEAR, 1);
      last.add(Calendar.MILLISECOND, -1);
      assertEquals(last, result.get(result.size()-1).getEnd());
    }
    
    {
      
      CalendarExpression calendarExpression = new BasicCalendarExpression(BasicCalendar.WEEKS);
      
      Evaluator eval = new Evaluator(null);
      
      Calendar from = date(2014, 4, 1);
      
      Calendar to = date(2014, 5, 1);
      to.add(Calendar.MILLISECOND, -1);
      
      IntervalSet result = eval.evaluate(calendarExpression, from, to, false);
      
      assertEquals(3, result.size());
      
      Calendar first = date(2014,4,5);
      Calendar last = date(2014,4,19);
      
      assertEquals(first, result.get(0).getStart());
      
      first.add(Calendar.WEEK_OF_YEAR, 1);
      first.add(Calendar.MILLISECOND, -1);     
      assertEquals(first, result.get(0).getEnd());
      
      
      
      assertEquals(last, result.get(result.size()-1).getStart());
      
      last.add(Calendar.WEEK_OF_YEAR, 1);
      last.add(Calendar.MILLISECOND, -1);
      assertEquals(last, result.get(result.size()-1).getEnd());
    }
    
    {
      
      CalendarExpression calendarExpression = new BasicCalendarExpression(BasicCalendar.MONTHS);
      
      Evaluator eval = new Evaluator(null);
      
      Calendar from = date(2014, 0, 1);
      
      Calendar to = date(2015, 0, 1);
      to.add(Calendar.MILLISECOND, -1);
      
      IntervalSet result = eval.evaluate(calendarExpression, from, to, true);
      
      assertEquals(12, result.size());
      
      Calendar first = date(2014,0,1);
      Calendar last = date(2014,11,1);
      
      assertEquals(first, result.get(0).getStart());
      
      first.add(Calendar.MONTH, 1);
      first.add(Calendar.MILLISECOND, -1);     
      assertEquals(first, result.get(0).getEnd());
      
      
      
      assertEquals(last, result.get(result.size()-1).getStart());
      
      last.add(Calendar.MONTH, 1);
      last.add(Calendar.MILLISECOND, -1);
      assertEquals(last, result.get(result.size()-1).getEnd());
    }
  }

  @Test
  public void testEvaluateCalendarScript() throws Exception
  {
    CalendarScript script = ScheduleParser.parseScript("{ workingDays = [1,2,3,4,5]/DAYS:during:WEEKS;\n"
        + "workingHours = [9,10]/HOURS:during:workingDays;\n"
        + "return(workingHours); }");
    
    Evaluator eval = new Evaluator(null);
    
    Calendar from = date(2014, 4, 1);
    
    Calendar to = date(2014, 5, 1);
    to.add(Calendar.MILLISECOND, -1);
    
    IntervalSet result = eval.evaluate(script, from, to, true);
//    System.out.println(result);
  }
  
  @Test
  public void testUnion() throws Exception
  {
    CalendarExpression expression = ScheduleParser.parse("([8-18]/HOURS:during:[1-4]/DAYS:during:WEEKS) + ([8-15]/HOURS:during:[5]/DAYS:during:WEEKS)");
    
    
    Evaluator eval = new Evaluator(null);
    
    Calendar from = date(2014, 4, 1);
    
    Calendar to = date(2014, 5, 1);
    to.add(Calendar.MILLISECOND, -1);
    
    IntervalSet result = eval.evaluate(expression, from, to, true);
  }

}
