package de.aristaflow.adept2.model.timemodel.tcn;

import de.aristaflow.adept2.model.timemodel.tools.Graph;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;
import de.aristaflow.adept2.tools.dot.AbstractDotGraph;
import de.aristaflow.adept2.tools.dot.DefaultGraphVisitor;
import de.aristaflow.adept2.tools.dot.DotEdge;
import de.aristaflow.adept2.tools.dot.DotNode;
import de.aristaflow.adept2.tools.dot.DotTools;


public class DotTCNVisitor<N extends TCNTimePoint, E extends TCNEdge<N, ?>> extends DefaultGraphVisitor<N, E> implements GraphVisitor<N,E>
{

  public DotTCNVisitor(String name)
  {
    this(name, new DotTools(), false);
  }

  public DotTCNVisitor(String name, DotTools dotTools)
  {
    this(name, dotTools, false);
  }
  
  public DotTCNVisitor(String name, boolean drawImplicitConstraints)
  {
    this(name, new DotTools(), drawImplicitConstraints);
  }
  
  public DotTCNVisitor(String name, DotTools dotTools, boolean drawImplicitConstraints)
  {
  	super(name, dotTools);
  }

  @Override
  public void discoverGraph(Graph<N, E> graph)
  {    
    super.discoverGraph(graph);
  }
  
  @Override
  public void finishGraph(Graph<N, E> graph)
  {
    //this.changeSet = null;
    super.finishGraph(graph);  
  }
  
  
  
  @Override
  public void discoverEdge(N startVertex, N endVertex, E e)
  {
    assert startVertex != null;
    assert endVertex != null;
    if (this.graphFormatter.isVisible(e))
    {
      AbstractDotGraph dotGraph = getDotGraph();
      DotNode startNode = null;
      DotNode endNode = null;
      
      startNode = dotGraph.getNode(getNodeID(startVertex));
      endNode = dotGraph.getNode(getNodeID(endVertex));
      
      assert startNode != null;
      assert endNode != null;
          
      
      DotEdge dotEdge = dotGraph
        .createEdge(startNode, endNode);
      
      dotEdge.setLabel(this.labelProvider.getEdgeLabel(e));
      dotEdge.setTailLabel(this.labelProvider.getEdgeTailLabel(e));
      this.graphFormatter.formatEdge(e, dotEdge);
    }
  }



}
