package de.aristaflow.adept2.model.timemodel.tcn.cstnImpl;

import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertEqualsIgnoreOrdering;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.junit.Before;
import org.junit.Test;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableTCN;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTransformation;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUAllMaxProjection;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTemplateTransformation;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.DefaultCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractPC2SolverTest;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledNode;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultObservationTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ObservationBinding;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.Proposition;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;
import de.aristaflow.adept2.util.xml.VersionException;
import de.aristaflow.adept2.util.xml.XMLFormatException;

public class LabelBasedPC2SolverTest extends AbstractPC2SolverTest<LabeledValueSet, LabeledEdge>
{

  private LabelBasedPC2Solver solver;

  @Before
  public void setUp()
  {
    solver = new LabelBasedPC2Solver();
  }
  
  @Override
  public LabelBasedPC2Solver getSolver()
  {
    return solver;
  }
  
  @Override
  protected TCNTransformation<LabeledTimePoint, LabeledEdge> createTransformation(TemporalTemplate template)
  {
    CSTNTemplateTransformation trafo = new CSTNTemplateTransformation(template,
                                                                      0);
    return trafo;
  }

  @Test
  public void testSimpleXOR() throws VersionException, XMLFormatException,
      IOException, TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.xorAdvancedNotPC2TestTemplate;
    
    basicTestWithTemplate(template);
  }


  @Test
  public void testSingleDiff()
  {
    IterationChain iteration = new IterationChain(0);
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint on1_2 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint on2_2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = on1_2.getObservationForEdgeCode(1l);
    ObservationBinding o1_22 = on1_2.getObservationForEdgeCode(2l);
    
    ObservationBinding o2_12 = on2_2.getObservationForEdgeCode(1l);

    
    Label label1 = new Proposition(o1_22, o2_12);
    Label label2 = new Proposition(o1_12, o2_12);
    
    assertEquals(on1_2.getObservation(), getSolver().singlediff(label1, label2));
  }
  
  @Test
  public void testObservationCaseRule() throws TCNInconsistencyException, InterruptedException
  {
    IterationChain iteration = new IterationChain(0);
    
    DefaultObservationTimePoint A = new DefaultObservationTimePoint(new NodeInstance(7, iteration, -1), "A?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding alpha = A.getObservationForEdgeCode(1l);
    DefaultObservationTimePoint B = new DefaultObservationTimePoint(new NodeInstance(8, iteration, -1), "B?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding beta = B.getObservationForEdgeCode(1l);
    DefaultObservationTimePoint G = new DefaultObservationTimePoint(new NodeInstance(9, iteration, -1), "G?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding gamma = G.getObservationForEdgeCode(1l);
    
    DefaultObservationTimePoint P = new DefaultObservationTimePoint(new NodeInstance(10, iteration, -1), "P?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding p = P.getObservationForEdgeCode(1l);
    ObservationBinding np = P.getObservationForEdgeCode(2l);
    
    LabeledTimePoint X = new DefaultLabeledNode(new NodeInstance(1, iteration, -1), "X", Proposition.EMPTY_PROPOSITION);
    LabeledTimePoint Y = new DefaultLabeledNode(new NodeInstance(2, iteration, -1), "Y", Proposition.EMPTY_PROPOSITION);


    
    
    TimeDistance u = new DefaultTimeDistance(8, 0, 0, 0, 0);
    TimeDistance v = new DefaultTimeDistance(18, 0, 0, 0, 0);
    
    LabeledEdge XY = new DefaultLabeledEdge(X, Y, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(alpha, beta, p), u)));
    LabeledEdge YX = new DefaultLabeledEdge(Y, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(beta, gamma, np), v.neg())));
    
//    Map<Label, LabeledValue> actual =  DefaultLabeledValueSet.toMap(XP.getValue());
    LabeledValue l_XY = XY.getValue().getLabeledValues().iterator().next();
    LabeledValue l_YX = YX.getValue().getLabeledValues().iterator().next();
    LabeledValue actual = getSolver().observationCaseReduction(l_XY, l_YX, P);
    
    assertNotNull(actual);
    LabeledValue expected = new DefaultLabeledValue(new Proposition(alpha, beta, gamma), DefaultTimeDistance.ZERO_TIME_DISTANCE);  
    assertEquals(expected, actual);
    

    
    List<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    nodes.add(A);
    nodes.add(B);
    nodes.add(G);
    
    nodes.add(P);
    nodes.add(X);
    nodes.add(Y);
    
    
    
    List<LabeledEdge> edges = new ArrayList<LabeledEdge>();
    edges.add(XY);
    edges.add(YX);
    
    ChangeableCSTNU cstnu = new DefaultCSTNU(nodes, edges,  
                                   new ArrayList<ChoiceVariable>());
    solve(cstnu);
    
    LabeledEdge YP = cstnu.getEdge(Y, P);
    
    assertNotNull(YP);
    assertEqualsIgnoreOrdering(new LabeledValue[]{expected}, YP.getValue().getLabeledValues());
    
  }
  
  @Test
  public void testSimplify()
  {
    IterationChain iteration = new IterationChain(0);
    
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    NodeInstance n3 = new NodeInstance(3, iteration, -1);
    
    
    DefaultLabeledValueSet labelSet;
    DefaultLabeledValueSet labelSetE;
    
    DefaultObservationTimePoint o1_2 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2_2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o3_3 = new DefaultObservationTimePoint(n3, "N3", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l});
   
    ObservationBinding nA = o1_2.getObservationForEdgeCode(1l);
    ObservationBinding A = o1_2.getObservationForEdgeCode(2l);

    ObservationBinding nB = o2_2.getObservationForEdgeCode(1l);
    ObservationBinding B = o2_2.getObservationForEdgeCode(2l);
    
    ObservationBinding o3_13 = o3_3.getObservationForEdgeCode(1l);
    ObservationBinding o3_23 = o3_3.getObservationForEdgeCode(2l);
    ObservationBinding o3_33 = o3_3.getObservationForEdgeCode(3l);
 
    
    // RLE1
    labelSet = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(A, nB), new DefaultTimeDistance(8, 0, 0, 0, 0)), // <A-B, #, 8>
      new DefaultLabeledValue(new Proposition(nB), new DefaultTimeDistance(8, 0, 0, 0, 0)) // <-B, #, 8>
      );
    
    labelSetE = new DefaultLabeledValueSet(
                                           new DefaultLabeledValue(new Proposition(nB), new DefaultTimeDistance(8, 0, 0, 0, 0)) // <-B, #, 8>
                                          );
    
    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet)); 
    
    // RLE2 Case i = j
    
    labelSet = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(A, nB), new DefaultTimeDistance(8, 0, 0, 0, 0)), // <A-B, #, 8>
      new DefaultLabeledValue(new Proposition(nA, nB), new DefaultTimeDistance(8, 0, 0, 0, 0)) // <-A-B, #, 8>
      );
    
    labelSetE = new DefaultLabeledValueSet(
                                           new DefaultLabeledValue(new Proposition(nB), new DefaultTimeDistance(8, 0, 0, 0, 0)) // <-B, #, 8>
                                          );
    
    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet)); 
    
    // RLE2 case i != j
    labelSet = new DefaultLabeledValueSet(
                                          new DefaultLabeledValue(new Proposition(nB), new DefaultTimeDistance(12, 0, 0, 0, 0)), // <-B, #, 12>
                                          new DefaultLabeledValue(new Proposition(A, nB), new DefaultTimeDistance(8, 0, 0, 0, 0)), // <A-B, #, 8>
                                          new DefaultLabeledValue(new Proposition(nA, nB), new DefaultTimeDistance(10, 0, 0, 0, 0)) // <-A-B, #, 10>
      );
    
    labelSetE = new DefaultLabeledValueSet(
                                           new DefaultLabeledValue(new Proposition(A, nB), new DefaultTimeDistance(8, 0, 0, 0, 0)), // <A-B, #, 8>
      new DefaultLabeledValue(new Proposition(nB), new DefaultTimeDistance(10, 0, 0, 0, 0)) // <-A-B, #, 10>
                                          );
    
    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet)); 
    
    // Example
    
    labelSet = new DefaultLabeledValueSet(
                                            new DefaultLabeledValue(new Proposition(), new DefaultTimeDistance(12, 0, 0, 0, 0)), // <o, #, 12> 
                                            new DefaultLabeledValue(new Proposition(A), new DefaultTimeDistance(10, 0, 0, 0, 0)), // <A, #, 10>
                                            new DefaultLabeledValue(new Proposition(nA), new DefaultTimeDistance(10, 0, 0, 0, 0)), // <-A, #, 10>
                                            new DefaultLabeledValue(new Proposition(nA, nB), new DefaultTimeDistance(8, 0, 0, 0, 0)), // <-A-B, #, 8>
                                            new DefaultLabeledValue(new Proposition(A, nB), new DefaultTimeDistance(8, 0, 0, 0, 0)), // <A-B, #, 8>
                                            new DefaultLabeledValue(new Proposition(nB), new DefaultTimeDistance(8, 0, 0, 0, 0)) // <-B, #, 8>
         );
    
    
    labelSetE = new DefaultLabeledValueSet(
                                           new DefaultLabeledValue(new Proposition(), new DefaultTimeDistance(10, 0, 0, 0, 0)), // <o, #, 10> 
                                           new DefaultLabeledValue(new Proposition(nB), new DefaultTimeDistance(8, 0, 0, 0, 0)) // <-B, #, 8>
        );

    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet));
    
    // Example
    
    labelSet = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(B), new DefaultTimeDistance(32, 0, 0, 0, 0)), // <-B , 32s>;
      new DefaultLabeledValue(new Proposition(B, nA), new DefaultTimeDistance(10, 0, 0, 0, 0)), // <-B A , 10s>;
      new DefaultLabeledValue(new Proposition(nB, nA), new DefaultTimeDistance(11, 0, 0, 0, 0)), // <B A , 11s>;
      new DefaultLabeledValue(new Proposition(nA), new DefaultTimeDistance(13, 0, 0, 0, 0)), // <A , 13s>;
      new DefaultLabeledValue(new Proposition(B, A), new DefaultTimeDistance(10, 0, 0, 0, 0)), // <-B -A , 10s>;
      new DefaultLabeledValue(new Proposition(A), new DefaultTimeDistance(14, 0, 0, 0, 0)), // <-A , 14s>;
      new DefaultLabeledValue(new Proposition(nB, A), new DefaultTimeDistance(11, 0, 0, 0, 0)) // <B -A , 11s>;
    );

    labelSetE = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(), new DefaultTimeDistance(11, 0, 0, 0, 0)), // <B , 11s>;
      new DefaultLabeledValue(new Proposition(B), new DefaultTimeDistance(10, 0, 0, 0, 0)) // <-B , 10s>;
    );
    
    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet));
    
    // Troublemaker
    //    <B , 13s> 
    //    <-B -A , 11s>
    //    <-B A , 12s>
    //    <A , 14s>
    //    <-B , 14s>
    //    <B A , 12s>
    //    <-A , 14s>
    //    <B -A , 11s>
    
    
    
    labelSet = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(B), new DefaultTimeDistance(13, 0, 0, 0, 0)), //  <B , 13s> 
      new DefaultLabeledValue(new Proposition(nB, nA), new DefaultTimeDistance(11, 0, 0, 0, 0)), // <-B -A , 11s>
      new DefaultLabeledValue(new Proposition(nB, A), new DefaultTimeDistance(12, 0, 0, 0, 0)), // <-B A , 12s>
      new DefaultLabeledValue(new Proposition(A), new DefaultTimeDistance(14, 0, 0, 0, 0)), // <A , 14s>
      new DefaultLabeledValue(new Proposition(nB), new DefaultTimeDistance(14, 0, 0, 0, 0)), // <-B , 14s>
      new DefaultLabeledValue(new Proposition(B, A), new DefaultTimeDistance(12, 0, 0, 0, 0)), // <B A , 12s>
      new DefaultLabeledValue(new Proposition(nA), new DefaultTimeDistance(14, 0, 0, 0, 0)), // <-A , 14s>
      new DefaultLabeledValue(new Proposition(B, nA), new DefaultTimeDistance(11, 0, 0, 0, 0)) // <B -A , 11s>
    );

    labelSetE = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(nA), new DefaultTimeDistance(11, 0, 0, 0, 0)), // <-A , 11s>;
      new DefaultLabeledValue(new Proposition(), new DefaultTimeDistance(12, 0, 0, 0, 0)) // <A , 12s>;
    );
    
    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet));
    
    // and another trouble maker...
    //    <-B -A , 14s>
    //    <A , 14s>
    //    <o, 14s>
    //    <B -A , 13s>
    //    <-A , 13s>

    labelSet = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(nB, nA), new DefaultTimeDistance(14, 0, 0, 0, 0)), //  <-B -A , 14s>
      new DefaultLabeledValue(new Proposition(A), new DefaultTimeDistance(14, 0, 0, 0, 0)), // <A , 14s>
      new DefaultLabeledValue(new Proposition(), new DefaultTimeDistance(14, 0, 0, 0, 0)), // <o, 14s>
      new DefaultLabeledValue(new Proposition(B, nA), new DefaultTimeDistance(13, 0, 0, 0, 0)), // <B -A , 13s>
      new DefaultLabeledValue(new Proposition(nA), new DefaultTimeDistance(13, 0, 0, 0, 0)) // <-A , 13s>
    );

    labelSetE = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(nA), new DefaultTimeDistance(13, 0, 0, 0, 0)), // <-A , 13s>;
      new DefaultLabeledValue(new Proposition(), new DefaultTimeDistance(14, 0, 0, 0, 0)) // <A , 14s>;
    );

    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet));
    
    // and another trouble maker...
    //    <(X 0_{B?};2:2) (X 0_{A?};1:2) , 0s>
    //    <(X 0_{A?};1:2) , -1s>
    //    <(X 0_{B?};1:2) (X 0_{A?};1:2) , -1s>

    labelSet = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(nB, A), new DefaultTimeDistance(0, 0, 0, 0, 0)), //  <(X 0_{B?};2:2) (X 0_{A?};1:2) , 0s>
      new DefaultLabeledValue(new Proposition(A), new DefaultTimeDistance(-1, 0, 0, 0, 0)), // <(X 0_{A?};1:2) , -1s>
      new DefaultLabeledValue(new Proposition(B, A), new DefaultTimeDistance(-1, 0, 0, 0, 0)) // <(X 0_{B?};1:2) (X 0_{A?};1:2) , -1s>
    );

    labelSetE = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(A), new DefaultTimeDistance(-1, 0, 0, 0, 0)) // <A , -1s>
    );

    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet));
    
    
    // RLE2 
    labelSet = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(o3_13), new DefaultTimeDistance(8, 0, 0, 0, 0)), 
      new DefaultLabeledValue(new Proposition(o3_23), new DefaultTimeDistance(8, 0, 0, 0, 0)) 
      );
                                        
    labelSetE = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(o3_13), new DefaultTimeDistance(8, 0, 0, 0, 0)), 
      new DefaultLabeledValue(new Proposition(o3_23), new DefaultTimeDistance(8, 0, 0, 0, 0)) 
    );
    
    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet)); 
    
    // RLE2 
    labelSet = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(o3_13), new DefaultTimeDistance(8, 0, 0, 0, 0)), 
      new DefaultLabeledValue(new Proposition(o3_23), new DefaultTimeDistance(8, 0, 0, 0, 0)), 
      new DefaultLabeledValue(new Proposition(o3_33), new DefaultTimeDistance(8, 0, 0, 0, 0)) 
      
      );
                                        
    labelSetE = new DefaultLabeledValueSet(
      new DefaultLabeledValue(Proposition.EMPTY_PROPOSITION, new DefaultTimeDistance(8, 0, 0, 0, 0))
    );
    
    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet)); 
  }
  
  @Test
  public void testIntersect()
  {
    IterationChain iteration = new IterationChain(0);
    
    LabeledValueSet labelSet1;
    LabeledValueSet labelSet2;
    LabeledValueSet labelSetR;
    
    LabeledValue label1;
    LabeledValue label2;
    LabeledValue labelR;
    

    Label proposition1;
    Label proposition2;
    Label propositionR;
    
    
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint o11_2 = new DefaultObservationTimePoint(n1, "N1_1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o11_4 = new DefaultObservationTimePoint(n1, "N1_1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l, 4l});
    DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2_4 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l, 4l});
    
    ObservationBinding o1_12 = o11_2.getObservationForEdgeCode(1l);
    ObservationBinding o1_22 = o11_2.getObservationForEdgeCode(2l);
    

    ObservationBinding o1_24 = o11_4.getObservationForEdgeCode(2l);
    ObservationBinding o1_34 = o11_4.getObservationForEdgeCode(3l);

    
    ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
    ObservationBinding o2_22 = o2.getObservationForEdgeCode(2l);
    
    ObservationBinding o2_24 = o2_4.getObservationForEdgeCode(2l);
    
    
    
    proposition1 = new Proposition(o1_22);
    proposition2 = new Proposition(o1_22);
    propositionR = new Proposition(o1_22);
    
    label1 = new DefaultLabeledValue(proposition1, new DefaultTimeDistance(3, 0, 0, 0, 0));
    label2 = new DefaultLabeledValue(proposition2, new DefaultTimeDistance(6, 0, 0, 0, 0));
    labelR = new DefaultLabeledValue(propositionR, new DefaultTimeDistance(3, 0, 0, 0, 0));
    labelSet1 = new DefaultLabeledValueSet(label1);
    labelSet2 = new DefaultLabeledValueSet(label2);
    labelSetR = new DefaultLabeledValueSet(labelR);
    assertEqualsIgnoreOrdering(labelSetR.getLabeledValues(), getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2)).values());
    
    //#########
    
    proposition1 = new Proposition(o1_22);
    proposition2 = new Proposition(o1_12);
    
    label1 = new DefaultLabeledValue(proposition1, new DefaultTimeDistance(3, 0, 0, 0, 0));
    label2 = new DefaultLabeledValue(proposition2, new DefaultTimeDistance(6, 0, 0, 0, 0));
    labelSet1 = new DefaultLabeledValueSet(label1);
    labelSet2 = new DefaultLabeledValueSet(label2);
    labelSetR = new DefaultLabeledValueSet(label1, label2);
    assertEqualsIgnoreOrdering(labelSetR.getLabeledValues(), getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2)).values());
    
    //###########
    
    proposition1 = new Proposition(o1_22);
    proposition2 = new Proposition(o2_22);
    
    label1 = new DefaultLabeledValue(proposition1, new DefaultTimeDistance(6, 0, 0, 0, 0));
    label2 = new DefaultLabeledValue(proposition2, new DefaultTimeDistance(3, 0, 0, 0, 0));
    labelSet1 = new DefaultLabeledValueSet(label1);
    labelSet2 = new DefaultLabeledValueSet(label2);
    labelSetR = new DefaultLabeledValueSet(label1, label2);
    assertEqualsIgnoreOrdering(labelSetR.getLabeledValues(), getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2)).values());
    
    //###########
    
    proposition1 = new Proposition(o1_22, o2_22);
    proposition2 = new Proposition(o1_22, o2_22);
    propositionR = new Proposition(o1_22, o2_22);
    
    label1 = new DefaultLabeledValue(proposition1, new DefaultTimeDistance(6, 0, 0, 0, 0));
    label2 = new DefaultLabeledValue(proposition2, new DefaultTimeDistance(3, 0, 0, 0, 0));
    labelR = new DefaultLabeledValue(propositionR, new DefaultTimeDistance(3, 0, 0, 0, 0));
    labelSet1 = new DefaultLabeledValueSet(label1);
    labelSet2 = new DefaultLabeledValueSet(label2);
    labelSetR = new DefaultLabeledValueSet(labelR);
    assertEqualsIgnoreOrdering(labelSetR.getLabeledValues(), getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2)).values());
    
    //###########
    
    proposition1 = new Proposition(o1_22);
    proposition2 = new Proposition(o1_22, o2_22);
    //propositionR = new Proposition(Literal.createLiteral(new NodeInstance(1, iteration), 0, 2, 2), Literal.createLiteral(new NodeInstance(2, iteration), 0, 2, 2));
    
    label1 = new DefaultLabeledValue(proposition1, new DefaultTimeDistance(6, 0, 0, 0, 0));
    label2 = new DefaultLabeledValue(proposition2, new DefaultTimeDistance(3, 0, 0, 0, 0));
    //labelR = new CSTNULabeledValue(propositionR, false, null, null, new DefaultTimeDistance(3, 0, 0, 0, 0));
    labelSet1 = new DefaultLabeledValueSet(label1);
    labelSet2 = new DefaultLabeledValueSet(label2);
    labelSetR = new DefaultLabeledValueSet(label1, label2);
    assertEqualsIgnoreOrdering(labelSetR.getLabeledValues(), getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2)).values());
    
    //###########
    
    proposition1 = new Proposition(o1_22);
    proposition2 = new Proposition(o1_22, o2_22);
    //propositionR = new Proposition(Literal.createLiteral(new NodeInstance(1, iteration), 0, 2, 2), Literal.createLiteral(new NodeInstance(2, iteration), 0, 2, 2));
    
    label1 = new DefaultLabeledValue(proposition1, new DefaultTimeDistance(3, 0, 0, 0, 0));
    label2 = new DefaultLabeledValue(proposition2, new DefaultTimeDistance(6, 0, 0, 0, 0));
    //labelR = new CSTNULabeledValue(propositionR, false, null, null, new DefaultTimeDistance(3, 0, 0, 0, 0));
    labelSet1 = new DefaultLabeledValueSet(label1);
    labelSet2 = new DefaultLabeledValueSet(label2);
    labelSetR = new DefaultLabeledValueSet(label1);
    assertEqualsIgnoreOrdering(labelSetR.getLabeledValues(), getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2)).values());
    
    //###########
    
    proposition1 = new Proposition(o1_22, o2_22);
    proposition2 = new Proposition(o1_22);
    propositionR = new Proposition(o1_22);
    
    label1 = new DefaultLabeledValue(proposition1, new DefaultTimeDistance(6, 0, 0, 0, 0));
    label2 = new DefaultLabeledValue(proposition2, new DefaultTimeDistance(3, 0, 0, 0, 0));
    labelR = new DefaultLabeledValue(propositionR, new DefaultTimeDistance(3, 0, 0, 0, 0));
    labelSet1 = new DefaultLabeledValueSet(label1);
    labelSet2 = new DefaultLabeledValueSet(label2);
    labelSetR = new DefaultLabeledValueSet(labelR);
    
    assertEqualsIgnoreOrdering(labelSetR.getLabeledValues(), getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2)).values());
    
    //###########
    
    proposition1 = new Proposition(o1_22, o2_22);
    proposition2 = new Proposition(o1_12);
    
    label1 = new DefaultLabeledValue(proposition1, new DefaultTimeDistance(6, 0, 0, 0, 0));
    label2 = new DefaultLabeledValue(proposition2, new DefaultTimeDistance(3, 0, 0, 0, 0));
    labelSet1 = new DefaultLabeledValueSet(label1);
    labelSet2 = new DefaultLabeledValueSet(label2);
    labelSetR = new DefaultLabeledValueSet(label1, label2);
    assertEqualsIgnoreOrdering(labelSetR.getLabeledValues(), getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2)).values());
    
    //###########
       
    proposition1 = new Proposition(o1_24, o2_24);
    proposition2 = new Proposition(o1_34);
    
    label1 = new DefaultLabeledValue(proposition1, new DefaultTimeDistance(6, 0, 0, 0, 0));
    label2 = new DefaultLabeledValue(proposition2, new DefaultTimeDistance(3, 0, 0, 0, 0));
    labelSet1 = new DefaultLabeledValueSet(label1);
    labelSet2 = new DefaultLabeledValueSet(label2);
    labelSetR = new DefaultLabeledValueSet(label1, label2);
    assertEqualsIgnoreOrdering(labelSetR.getLabeledValues(), getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2)).values());
    
    //###########
    
    
    proposition1 = new Proposition(o1_22, o2_12);
    proposition2 = new Proposition(o1_22, o2_22);
    
    label1 = new DefaultLabeledValue(proposition1, new DefaultTimeDistance(6, 0, 0, 0, 0));
    label2 = new DefaultLabeledValue(proposition2, new DefaultTimeDistance(3, 0, 0, 0, 0));
    labelSet1 = new DefaultLabeledValueSet(label1);
    labelSet2 = new DefaultLabeledValueSet(label2);
    labelSetR = new DefaultLabeledValueSet(label1, label2); 
    assertEqualsIgnoreOrdering(labelSetR.getLabeledValues(), getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2)).values());
    
    //##############
    
    //(<o, #, 1s>;<A -B , #, infty>;<-A B , #, 14m57s>;) u (<-A B , #, 9m57s>;) = (<o, #, 1s>;<A -B , #, infty>;<-A B , #, 9m57s>;)
        
    labelSet1 = new DefaultLabeledValueSet(
                                           new DefaultLabeledValue(new Proposition(), new DefaultTimeDistance(1, 0, 0, 0, 0)) // <o, #, 1s>
                                           );
    labelSet2 = new DefaultLabeledValueSet(
                                           new DefaultLabeledValue(new Proposition(o1_22), new DefaultTimeDistance(9, 0, 0, 0, 0)) // (<A , #, 9s>;)
                                           );
    labelSetR = new DefaultLabeledValueSet(
                                           new DefaultLabeledValue(new Proposition(), new DefaultTimeDistance(1, 0, 0, 0, 0)) // <o, #, 1s>
                                          ); 
    // <o, #, 1s> subsumes <A, #, 9s> and is also more restrictive, hence <A, #, 9s> is not necessary 
    assertEqualsIgnoreOrdering(labelSetR.getLabeledValues(), getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2)).values());
  }
  
  @Test
  public void testXOR1() throws VersionException, XMLFormatException, IOException, TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.xorTimeConstraintTestTemplate;
    
    TCNTransformation<LabeledTimePoint, LabeledEdge> trafo = createTransformation(template);
    ChangeableTCN<LabeledTimePoint, LabeledEdge> timeModel = (ChangeableTCN<LabeledTimePoint, LabeledEdge>) trafo.transform();
    Collection<NodeMapping<LabeledTimePoint>> nodeMappings = trafo.getNodeMapping();
    
    timeModel = solve(timeModel);
    
    IterationChain iterationChain = new IterationChain(template.getStartNode().getID());
    
    // Nodes 4 and 5 are on different branches
    NodeMapping<LabeledTimePoint> nodeMapping1 = getNodeMapping(nodeMappings, 4, iterationChain);
    NodeMapping<LabeledTimePoint> nodeMapping2 = getNodeMapping(nodeMappings, 5, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(timeModel.getEdge(node1, node2));
        assertNull(timeModel.getEdge(node2, node1));
      }
    }
  
  }
  


  @Test
  public void testXOR2() throws VersionException, XMLFormatException, IOException, TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.xorAdvancedNotPC2TestTemplate;
    
    TCNTransformation<LabeledTimePoint, LabeledEdge> trafo = createTransformation(template);
    ChangeableTCN<LabeledTimePoint, LabeledEdge> timeModel = (ChangeableTCN<LabeledTimePoint, LabeledEdge>) trafo.transform();
    Collection<NodeMapping<LabeledTimePoint>> nodeMappings = trafo.getNodeMapping();
    
    timeModel = solve(timeModel);
    
    IterationChain iterationChain = new IterationChain(template.getStartNode().getID());
    
    // Nodes 5 and 6 are on different branches
    NodeMapping<LabeledTimePoint> nodeMapping1 = getNodeMapping(nodeMappings, 5, iterationChain);
    NodeMapping<LabeledTimePoint> nodeMapping2 = getNodeMapping(nodeMappings, 6, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(timeModel.getEdge(node1, node2));
        assertNull(timeModel.getEdge(node2, node1));
      }
    }
  }

  protected ChangeableTCN<LabeledTimePoint, LabeledEdge> solve(ChangeableTCN<LabeledTimePoint, LabeledEdge> timeModel)
      throws TCNInconsistencyException, InterruptedException
  {
    TCNTestTools.assertConsistent(timeModel);
    ChangeableTCN<LabeledTimePoint, LabeledEdge> solution = getSolver().solve(timeModel);
    TCNTestTools.assertConsistent(solution);
//    TCNTestTools.assertComplete(solution);
    return solution;
  }
  
  
  
  @Test
  public void testXOR3() throws TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.xorAndTestTemplate;
    
    TCNTransformation<LabeledTimePoint, LabeledEdge> trafo = createTransformation(template);
    ChangeableTCN<LabeledTimePoint, LabeledEdge> tcn = (ChangeableTCN<LabeledTimePoint, LabeledEdge>) trafo.transform();
    Collection<NodeMapping<LabeledTimePoint>> nodeMappings = trafo.getNodeMapping();
    

    tcn = solve(tcn);
    
    IterationChain iterationChain = new IterationChain(template.getStartNode().getID());
    
    // Nodes are on different branches
    NodeMapping<LabeledTimePoint> nodeMapping1 = getNodeMapping(nodeMappings, 9, iterationChain);
    NodeMapping<LabeledTimePoint> nodeMapping2 = getNodeMapping(nodeMappings, 10, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(tcn.getEdge(node1, node2));
        assertNull(tcn.getEdge(node2, node1));
      }
    }
    
    // Nodes are on different branches
    nodeMapping1 = getNodeMapping(nodeMappings, 14, iterationChain);
    nodeMapping2 = getNodeMapping(nodeMappings, 15, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(tcn.getEdge(node1, node2));
        assertNull(tcn.getEdge(node2, node1));
      }
    }
  

    // Nodes are on different branches
    int[] others = new int[] { 5, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16 };
    
    for (int o : others)
    {

      nodeMapping1 = getNodeMapping(nodeMappings, 11, iterationChain);
      nodeMapping2 = getNodeMapping(nodeMappings, o, iterationChain);
      
      for (LabeledTimePoint node1 : nodeMapping1.getNodes())
      {
        for (LabeledTimePoint node2 : nodeMapping2.getNodes())
        {
          assertNull(tcn.getEdge(node1, node2));
          assertNull(tcn.getEdge(node2, node1));
        }
      }
    }
    
   
  }
  
  @Test
  public void testXOR3_2() throws TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.xorAndTestTemplate;
    
    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(template);
    ChangeableCSTNU cstnu = (ChangeableCSTNU) trafo.transform();
    TCNTestTools.assertConsistent(cstnu);
    Collection<NodeMapping<LabeledTimePoint>> nodeMappings = trafo.getNodeMapping();
//    TCNTestTools.printTCN(timeModel, this.getClass(), "XORAndTest");
    
    ChangeableTCN<LabeledTimePoint, LabeledEdge> allMaxProjection = new CSTNUAllMaxProjection(cstnu);
    TCNTestTools.assertConsistent(allMaxProjection);
    
    allMaxProjection = solve(allMaxProjection);
    
    IterationChain iterationChain = new IterationChain(template.getStartNode().getID());
    
    // Nodes are on different branches
    NodeMapping<LabeledTimePoint> nodeMapping1 = getNodeMapping(nodeMappings, 9, iterationChain);
    NodeMapping<LabeledTimePoint> nodeMapping2 = getNodeMapping(nodeMappings, 10, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(allMaxProjection.getEdge(node1, node2));
        assertNull(allMaxProjection.getEdge(node2, node1));
      }
    }
    
    // Nodes are on different branches
    nodeMapping1 = getNodeMapping(nodeMappings, 14, iterationChain);
    nodeMapping2 = getNodeMapping(nodeMappings, 15, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(allMaxProjection.getEdge(node1, node2));
        assertNull(allMaxProjection.getEdge(node2, node1));
      }
    }
  

    // Nodes are on different branches
    int[] others = new int[] { 5, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16 };
    
    for (int o : others)
    {

      nodeMapping1 = getNodeMapping(nodeMappings, 11, iterationChain);
      nodeMapping2 = getNodeMapping(nodeMappings, o, iterationChain);
      
      for (LabeledTimePoint node1 : nodeMapping1.getNodes())
      {
        for (LabeledTimePoint node2 : nodeMapping2.getNodes())
        {
          assertNull(allMaxProjection.getEdge(node1, node2));
          assertNull(allMaxProjection.getEdge(node2, node1));
        }
      }
    }
  }
  
  @Test
  public void testXOR4() throws VersionException, XMLFormatException, IOException, TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.xorEmptyBranchTestTemplate;
    
    TCNTransformation<LabeledTimePoint, LabeledEdge> trafo = createTransformation(template);
    ChangeableTCN<LabeledTimePoint, LabeledEdge> timeModel = (ChangeableTCN<LabeledTimePoint, LabeledEdge>) trafo.transform();
    Collection<NodeMapping<LabeledTimePoint>> nodeMappings = trafo.getNodeMapping();
    
    timeModel = solve(timeModel);
    
    IterationChain iterationChain = new IterationChain(template.getStartNode().getID());
    
    // Nodes 4 and 5 are on different branches
    NodeMapping<LabeledTimePoint> nodeMapping1 = getNodeMapping(nodeMappings, 4, iterationChain);
    NodeMapping<LabeledTimePoint> nodeMapping2 = getNodeMapping(nodeMappings, 5, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(timeModel.getEdge(node1, node2));
        assertNull(timeModel.getEdge(node2, node1));
      }
    }
  
  }
}
