package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertTDEquals;
import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertTPEquals;

import java.io.IOException;
import java.util.Collection;

import org.apache.commons.configuration.BaseConfiguration;
import org.apache.commons.configuration.Configuration;
import org.junit.Ignore;
import org.junit.Test;

import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.CSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableSTN;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.STN;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.CSTNTemplateTransformation;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.LabelBasedPC2Solver;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractSolverTest;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.STNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;
import de.aristaflow.adept2.util.xml.VersionException;
import de.aristaflow.adept2.util.xml.XMLFormatException;

public class CSTNU2CSTNProjectionTest extends AbstractSolverTest
{

  @Test
  public void testSolveCSTNU1() throws TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.simpleTimeTestContingentTestTemplate;
    
    CSTNUTemplateTransformation trafo = createTransformation(template);
    ChangeableCSTNU tcn = (ChangeableCSTNU) trafo.transform();
    Collection<NodeMapping<LabeledTimePoint>> nodeMappings = trafo.getNodeMapping();
    
    int startNodeID = template.getStartNode().getID();
    NodeMapping<LabeledTimePoint> startNodeMapping = getNodeMapping(nodeMappings, startNodeID, new IterationChain(startNodeID));
    
    tcn
      .updateConstraint(new DefaultLabeledEdge(
                                        tcn.getTimeBaseEvent(),
                                        startNodeMapping.getStartNode(),
                                        new DefaultLabeledValueSet(
                                                            new DefaultLabeledValue(
                                                                                  startNodeMapping
                                                                                    .getStartNode()
                                                                                    .getLabel(),
                                                                                  secToTD(2)))));
    
//    PC2Solver<LabeledEdge> updater = new PC2Solver<LabeledEdge>();
//    updater.solve(new CstnuCStnProjection(tcn, true));
    
    //tcn.visit(visitor.getInstance("C_Before"));
    
    CSTNUDCSolver solver = new CSTNUDCSolver();
    try
    {
      tcn = solver.solve(tcn);
    }
    catch (TCNInconsistencyException e)
    {
      //tcn.visit(visitor.getInstance("C_Failure"));
      throw e;
    }
    
    //TCNTestTools.printTCN(tcn, this.getClass(), "CSTNU");
    
    STN<LabeledTimePoint, STNEdge> stnProjection = new CSTNU2STNProjection(tcn);

    //TCNTestTools.printTCN(stnProjection, this.getClass(), "STN");
    
    TemplateTimeModel timeModel = new TCNTemplateTimeModel<LabeledTimePoint, STNEdge>(template, stnProjection,
                                                                                      trafo.getNodeMapping(),
                                                                                      trafo
                                                                                        .getProcessCreationTimePoint(),
                                                                                      trafo.getProcessStartTimePoint(),
                                                                                      trafo.getProcessEndTimePoint(),
                                                                                      null);
        
    assertTDEquals(m(20)+cmin(1)+stmin(1),//m(60)+cmin(1)+stmin(1), 
                   m(65), 
                   timeModel.getProcessDuration(), 1);
    assertTDEquals(sdmin(1), sdmax(1), timeModel.getActivityDuration(0), 1);
    assertTDEquals(sdmin(1), sdmax(1), timeModel.getActivityDuration(1), 1);

    assertTDEquals(m(5), m(15), timeModel.getActivityDuration(2), 1);
    assertTDEquals(m(5), m(30), timeModel.getActivityDuration(3), 1);
    assertTDEquals(m(5), m(15), timeModel.getActivityDuration(4), 1);

    assertTPEquals(0, 0, timeModel.getActivityStartTimeframe(0), 1);
    assertTPEquals(sdmin(1), 
                   sdmax(1), 
                   timeModel.getActivityCompletionTimeframe(0), 1);

    assertTPEquals(sdmin(1)+cmin(1),
                   m(5)+sdmax(1)-stmin(1),
                   timeModel.getActivityStartTimeframe(2), 1);
    assertTPEquals(m(5)+sdmin(1)+cmin(1),
                   m(20)+sdmax(1)-stmin(1),
                   timeModel.getActivityCompletionTimeframe(2), 1);

    assertTPEquals(m(5)+sdmin(1)+cmin(2),//m(15)+sdmin(1)+cmin(2),
                   m(20)+sdmax(1)-stmin(1)-cmin(1),
                   timeModel.getActivityStartTimeframe(3), 1);
    assertTPEquals(m(10)+sdmin(1)+cmin(2),//m(20)+sdmin(1)+cmin(2), 
                   m(50)+sdmax(1)-stmin(1)-cmin(1), 
                   timeModel.getActivityCompletionTimeframe(3), 1);

    assertTPEquals(m(15)+sdmin(1)+cmin(1),//m(45)+sdmin(1)+cmin(1), 
                   m(50)+sdmax(1)-stmin(1), 
                   timeModel.getActivityStartTimeframe(4), 1);
    assertTPEquals(m(20)+sdmin(1)+cmin(1),//m(50)+sdmin(1)+cmin(1), 
                   m(65)+sdmax(1)-stmin(1),  
                   timeModel.getActivityCompletionTimeframe(4), 1);

    assertTPEquals(m(20)+sdmin(1)+cmin(1)+stmin(1),//m(60)+sdmin(1)+cmin(1)+stmin(1),
                   m(65)+sdmax(1), 
                   timeModel.getActivityStartTimeframe(1), 1);
    assertTPEquals(m(20)+sdmin(2)+cmin(1)+stmin(1),//m(60)+sdmin(2)+cmin(1)+stmin(1), 
                   m(65)+sdmax(2),  
                   timeModel.getActivityCompletionTimeframe(1), 1);
  }


  @Override
  protected CSTNUTemplateTransformation createTransformation(TemporalTemplate template)
  {
    Configuration configuration = new BaseConfiguration();
    configuration.setProperty(CSTNUTemplateTransformation.CONF_ADVANCED_ACTIVITY_TRANSFORMATION, CSTNUTemplateTransformation.ACTIVITY_TRANSFORMATION_BASIC);
    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(configuration, template, 0);
    return trafo;
  }
  

  @Test
  public void testCoopIS13() throws VersionException, XMLFormatException,
      IOException, TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.coopIS13TestTemplate;

    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(
                                                                        template,
                                                                        0);
    ChangeableCSTNU cstnu = (ChangeableCSTNU) trafo.transform();

    LabelBasedPC2Solver solver = new LabelBasedPC2Solver();
    solver.solve(new CSTNU2CSTNProjection(cstnu, false, true));
  }
  
  @Test
  public void testSimpleProjection()
  {
    TemporalTemplate template = TestTemplates.coopIS13TestTemplate;

    CSTNUTemplateTransformation cstnuTrafo = createTransformation(template);
    ChangeableCSTNU cstnu = (ChangeableCSTNU) cstnuTrafo.transform();
    
    CSTNU2CSTNProjection projection = new CSTNU2CSTNProjection(cstnu, true, true);
    
    CSTNTemplateTransformation cstnTrafo = new CSTNTemplateTransformation(template);
    
    CSTN cstn = cstnTrafo.transform();
    
    // This is only equal if flexible == contingent for all activities
    // TODO fix one time... TCNTestTools.assertTCNEquals(cstn, projection);
  }
  
  @Ignore @Test // Test is no longer possible as CSTN transformation differs from CSTNU
  public void testProjection()
  {
    testProjectionWithTemplate(TestTemplates.simpleTimeTestContingentTestTemplate);
    testProjectionWithTemplate(TestTemplates.simpleTimeTestNonContingentTestTemplate);
    testProjectionWithTemplate(TestTemplates.xorAndTestTemplate);
  }
  
  @Test
  public void testSimpleTimeTestContingent() throws VersionException, XMLFormatException, IOException, TCNInconsistencyException, InterruptedException
  {  
    testPC2WithTemplate(TestTemplates.simpleTimeTestContingentTestTemplate);
    testPC2WithTemplate(TestTemplates.simpleTimeTestNonContingentTestTemplate);
  }


  private void testProjectionWithTemplate(TemporalTemplate template)
  {
    Configuration configuration = new BaseConfiguration();
    configuration.setProperty(CSTNUTemplateTransformation.CONF_ADVANCED_ACTIVITY_TRANSFORMATION, CSTNUTemplateTransformation.ACTIVITY_TRANSFORMATION_BASIC);
    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(configuration, template, 0);
    ChangeableCSTNU cstnu = (ChangeableCSTNU) trafo.transform();
    
    TCNTestTools.assertConsistent(cstnu);
    
    TCN<LabeledTimePoint,LabeledEdge> stnProjection = new CSTNU2CSTNProjection(cstnu, false, true);

    TCNTestTools.assertConsistent(stnProjection);
    
    CSTNTemplateTransformation cstnTrafo = new CSTNTemplateTransformation(template, 0);
    STN<LabeledTimePoint,LabeledEdge> expectedCSTN = cstnTrafo.transform();
    

    TCNTestTools.assertTCNEquals(expectedCSTN, stnProjection);

  }

  private void testPC2WithTemplate(TemporalTemplate template)
      throws TCNInconsistencyException, InterruptedException
  {
    CSTNUTemplateTransformation trafo = createTransformation(template);
    ChangeableCSTNU cstnu = (ChangeableCSTNU) trafo.transform();
    
    TCNTestTools.assertConsistent(cstnu);
    
    
    ChangeableSTN<LabeledTimePoint,LabeledEdge> cstn = new CSTNU2CSTNProjection(cstnu, false, true);

//    TestTools.printTCN(cstnu, this.getClass(), "CSTNU");
//    TestTools.printTCN(cstn, this.getClass(), "CSTNProjection");
    
    TCNTestTools.assertConsistent(cstn);
    
    LabelBasedPC2Solver solver = new LabelBasedPC2Solver();
    try
    {
      solver.solve(cstn);
      
      TCNTestTools.assertConsistent(cstn);
    }
    catch (TCNInconsistencyException e)
    {
//      TCNTestTools.printTCN(cstn, this.getClass(), "Failure");
      throw e;
    }

  }


}
