package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertEqualsIgnoreOrdering;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableTCN;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTransformation;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.AbstractCSTNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractSTNUDCSolverTest;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledNode;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultObservationTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ObservationBinding;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.Proposition;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;
import de.aristaflow.adept2.util.xml.VersionException;
import de.aristaflow.adept2.util.xml.XMLFormatException;

public class CSTNUDCSolverTest extends AbstractSTNUDCSolverTest
{
  private AbstractCSTNSolver solver;
  
  
  @BeforeClass
  public static void setUpBeforeCSTNUDCSolverTest() throws Exception
  {
    tmo = new DefaultTimeModelOperations();
  }
  


  @Before
  public void setUp()
  {
    solver = new CSTNUDCSolver();
  }
  
  @Override
  public AbstractCSTNSolver getSolver()
  {
    return solver;
  }
  
  /**
   * <pre>P? -- <\alpha p, -x, R, -w> --> X</pre>
   * 
   * We know <code>x <= w</code>.
   * 
   * Three cases:<ol>
   * <li> 0 <= x <= w : X may be executed before P; we cannot rely on p in the constraint; remove p
   * <li> x <= 0 <= w : X may be executed before or after P; we cannot rely on p in the constraint; remove p
   * <li> x <= w <= 0 : X can only be executed after P; no issue. </li>
   * </ol>
   * 
   * In case 1. and 2. 
   * <code>P? -- <\alpha p, -x, R, -w> --> X</code>
   * becomes
   * <code>P? -- <\alpha, -x, R, -w> --> X</code>
   */
  @Test
  public void testPropositionModificationRuleR0()
  {
    IterationChain iteration = new IterationChain(0);
    
    
    DefaultObservationTimePoint P = new DefaultObservationTimePoint(new NodeInstance(42, iteration, -1), "P?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding p = P.getObservationForEdgeCode(1l);
    
    DefaultObservationTimePoint A = new DefaultObservationTimePoint(new NodeInstance(1, iteration, -1), "A?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding alpha = A.getObservationForEdgeCode(1l);
    
    TimeDistance w = new DefaultTimeDistance(10, 0, 0, 0, 0);
    // simple case x < 0 ===============
    LabeledValue base = new DefaultLabeledValue(new Proposition(alpha, p), w.neg());
    Map<Label, LabeledValue> actual;
    actual = toMap(base);
    getSolver().propositionModificationRuleR0(P, null, actual);
    
    LabeledValue[] expected = new LabeledValue[]{new DefaultLabeledValue(new Proposition(alpha), w.neg())};
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    // ===================
    
    // simple case x > 0 ===============
    base = new DefaultLabeledValue(new Proposition(alpha, p), w);
    actual = toMap(base);
    getSolver().propositionModificationRuleR0(P, null, actual);
    
    expected = new LabeledValue[]{new DefaultLabeledValue(new Proposition(alpha, p), w)};
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    // ===================
    
    // simple case x = 0 ===============
    base = new DefaultLabeledValue(new Proposition(alpha, p), DefaultTimeDistance.ZERO_TIME_DISTANCE);
    actual = toMap(base);
    getSolver().propositionModificationRuleR0(P, null, actual);
    
    expected = new LabeledValue[]{new DefaultLabeledValue(new Proposition(alpha), DefaultTimeDistance.ZERO_TIME_DISTANCE)};
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    // ===================
    
    // x < 0 < w ==========    
    base = new DefaultContingentLabeledValue(new Proposition(alpha, p), ConstraintCase.UPPERCASE, w, P, w.neg());
    actual = toMap(base);
    getSolver().propositionModificationRuleR0(P, null, actual);
    expected = new LabeledValue[]{new DefaultContingentLabeledValue(new Proposition(alpha), ConstraintCase.UPPERCASE, w, P, w.neg())};
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    // ==================
    
    // 0 < x <= w ===============
    base = new DefaultContingentLabeledValue(new Proposition(alpha, p), ConstraintCase.UPPERCASE, w.neg(), P, w.neg());
    actual = toMap(base);
    getSolver().propositionModificationRuleR0(P, null, actual);
    expected = new LabeledValue[]{new DefaultContingentLabeledValue(new Proposition(alpha), ConstraintCase.UPPERCASE, w.neg(), P, w.neg())};
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    // ==================
    
    // x <= w < 0 ===============
    base = new DefaultContingentLabeledValue(new Proposition(alpha, p), ConstraintCase.UPPERCASE, w, P, w);
    actual = toMap(base);
    getSolver().propositionModificationRuleR0(P, null, actual);
    expected = new LabeledValue[]{new DefaultContingentLabeledValue(new Proposition(alpha, p), ConstraintCase.UPPERCASE, w, P, w)};
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    // ==================
    
    // x < w = 0 ===============
    base = new DefaultContingentLabeledValue(new Proposition(alpha, p), ConstraintCase.UPPERCASE, w, P, DefaultTimeDistance.ZERO_TIME_DISTANCE);
    actual = toMap(base);
    getSolver().propositionModificationRuleR0(P, null, actual);
    expected = new LabeledValue[]{new DefaultContingentLabeledValue(new Proposition(alpha), ConstraintCase.UPPERCASE, w, P, DefaultTimeDistance.ZERO_TIME_DISTANCE)};
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    // ==================
    
    // 0 = x < w ===============
    base = new DefaultContingentLabeledValue(new Proposition(alpha, p), ConstraintCase.UPPERCASE, DefaultTimeDistance.ZERO_TIME_DISTANCE, P, w.neg());
    actual = toMap(base);
    getSolver().propositionModificationRuleR0(P, null, actual);
    expected = new LabeledValue[]{new DefaultContingentLabeledValue(new Proposition(alpha), ConstraintCase.UPPERCASE, DefaultTimeDistance.ZERO_TIME_DISTANCE, P, w.neg())};
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    // ==================
    
    // 0 = x = w ===============
    base = new DefaultContingentLabeledValue(new Proposition(alpha, p), ConstraintCase.UPPERCASE, DefaultTimeDistance.ZERO_TIME_DISTANCE, P, DefaultTimeDistance.ZERO_TIME_DISTANCE);
    actual = toMap(base);
    getSolver().propositionModificationRuleR0(P, null, actual);
    expected = new LabeledValue[]{new DefaultContingentLabeledValue(new Proposition(alpha), ConstraintCase.UPPERCASE, DefaultTimeDistance.ZERO_TIME_DISTANCE, P, DefaultTimeDistance.ZERO_TIME_DISTANCE)};
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    // ==================
  }

  
  protected Map<Label, LabeledValue> toMap(LabeledValue base)
  {
    Map<Label, LabeledValue> map = new HashMap<Label, LabeledValue>();
    map.put(base.getLabel(), base);
    return map;
  }


  /**
   * <pre>P? - <\alpha\beta, -w*> -> X - <\beta\gamma p, x, R, v> -> Y</pre>
   * We know x <= v
   * 
   * 0 <= w* ==> if \alpha\beta is true, X may be executed at most w before P.
   * 
   * 
   * Cases<ol>
   * <li>x <= v <= w* : Y may be executed before P, thus in cases where alpha is true we cannot rely on p
   * <li>x <= w* <= v : Y may be executed before or after P, thus in cases where alpha is true we cannot rely on p
   * <li>w* <= x <= v : Y will be executed after P thus we know the value of P when we need to execute Y; ignore
   * </ol>
   * 
   * In case 1. and 2. 
   * <code>X - <\beta\gamma p, x, R, v> -> Y</code>
   * becomes
   * <code>X - <\alpha\beta\gamma, x, R, v> -> Y</code>
   * <code>X - <\not(\alpha)\beta\gamma p, x, R, v> -> Y</code>
   */
  @Test
  public void testPropositionModificationRuleR1()
  {
    IterationChain iteration = new IterationChain(0);
            
    DefaultObservationTimePoint P = new DefaultObservationTimePoint(new NodeInstance(7, iteration, -1), "P?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding p = P.getObservationForEdgeCode(1l);
//    ObservationBinding np = P.getObservationForEdgeCode(2l);
    
    DefaultObservationTimePoint A = new DefaultObservationTimePoint(new NodeInstance(1, iteration, -1), "A?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding alpha = A.getObservationForEdgeCode(1l);
    ObservationBinding notAlpha = A.getObservationForEdgeCode(2l);
    
    DefaultObservationTimePoint B = new DefaultObservationTimePoint(new NodeInstance(2, iteration, -1), "B?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding beta = B.getObservationForEdgeCode(1l);
    
    DefaultObservationTimePoint G = new DefaultObservationTimePoint(new NodeInstance(3, iteration, -1), "G?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding gamma = G.getObservationForEdgeCode(1l);
    
    LabeledTimePoint X = new DefaultLabeledNode(new NodeInstance(42, iteration, -1), "X", new Proposition(beta));
    LabeledTimePoint Y = new DefaultLabeledNode(new NodeInstance(43, iteration, -1), "Y", new Proposition(gamma));
    
    List<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    List<LabeledEdge> edges;
    nodes.add(X);
    nodes.add(Y);
    nodes.add(P);
    nodes.add(A);
    nodes.add(B);
    nodes.add(G);
    
    TimeDistance w;
    TimeDistance x;
    TimeDistance v;
    LabeledEdge PX;
    LabeledEdge XY;
    
    CSTNU cstnu;
    
    Map<Label, LabeledValue> actual;
    LabeledValue[] expected;
    
    // simple case v < w ====
    w = new DefaultTimeDistance(10, 0, 0, 0, 0);
    v = new DefaultTimeDistance(5, 0, 0, 0, 0);
    
    PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(alpha, beta), w.neg())));
    XY = new DefaultLabeledEdge(X, Y, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(beta, gamma, p), v)));
   
    edges = new ArrayList<LabeledEdge>();
    edges.add(PX);
    edges.add(XY);
    
    cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
    
    
    actual =DefaultLabeledValueSet.toMap(XY.getValue()); 
    getSolver().propositionModificationRuleR1(cstnu, X, P, actual);
    expected = new LabeledValue[]{
                                  new DefaultLabeledValue(new Proposition(alpha, beta, gamma), v),
                                  new DefaultLabeledValue(new Proposition(notAlpha, beta, gamma, p), v)
                                 };
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    
    //===========
    
    // simple case v = w ====
    w = new DefaultTimeDistance(5, 0, 0, 0, 0);
    v = new DefaultTimeDistance(5, 0, 0, 0, 0);
    
    PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(alpha, beta), w.neg())));
    XY = new DefaultLabeledEdge(X, Y, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(beta, gamma, p), v)));
   
    edges = new ArrayList<LabeledEdge>();
    edges.add(PX);
    edges.add(XY);
    
    cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
    
    
    actual =DefaultLabeledValueSet.toMap(XY.getValue()); 
    getSolver().propositionModificationRuleR1(cstnu, X, P, actual);
    expected = new LabeledValue[]{
                                  new DefaultLabeledValue(new Proposition(alpha, beta, gamma), v),
                                  new DefaultLabeledValue(new Proposition(notAlpha, beta, gamma, p), v)
                                 };
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    
    //===========
    
    // simple case v > w ====
    w = new DefaultTimeDistance(5, 0, 0, 0, 0);
    v = new DefaultTimeDistance(10, 0, 0, 0, 0);
    
    PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(alpha, beta), w.neg())));
    XY = new DefaultLabeledEdge(X, Y, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(beta, gamma, p), v)));
   
    edges = new ArrayList<LabeledEdge>();
    edges.add(PX);
    edges.add(XY);
    
    cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
    
    
    actual =DefaultLabeledValueSet.toMap(XY.getValue()); 
    getSolver().propositionModificationRuleR1(cstnu, X, P, actual);
    expected = new LabeledValue[]{
                                  new DefaultLabeledValue(new Proposition(beta, gamma, p), v)
                                 };
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    
    //===========
    
    
    // not alpha  ====
    w = new DefaultTimeDistance(10, 0, 0, 0, 0);
    v = new DefaultTimeDistance(5, 0, 0, 0, 0);
    
    PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(beta), w.neg())));
    XY = new DefaultLabeledEdge(X, Y, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(beta, gamma, p), v)));
   
    edges = new ArrayList<LabeledEdge>();
    edges.add(PX);
    edges.add(XY);
    
    cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
    
    
    actual =DefaultLabeledValueSet.toMap(XY.getValue()); 
    getSolver().propositionModificationRuleR1(cstnu, X, P, actual);
    expected = new LabeledValue[]{
                                  new DefaultLabeledValue(new Proposition(beta, gamma), v),
                                 };
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    
    //===========
    
    // x < v < w ====
    
    w = new DefaultTimeDistance(10, 0, 0, 0, 0);
    
    x = new DefaultTimeDistance(2, 0, 0, 0, 0);
    v = new DefaultTimeDistance(5, 0, 0, 0, 0);
    
    PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(alpha, beta), w.neg())));
    XY = new DefaultLabeledEdge(X, Y, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(beta, gamma, p), ConstraintCase.UPPERCASE, x, X, v.neg())));
   
    edges = new ArrayList<LabeledEdge>();
    edges.add(PX);
    edges.add(XY);
    
    cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
    
    
    actual =DefaultLabeledValueSet.toMap(XY.getValue()); 
    getSolver().propositionModificationRuleR1(cstnu, X, P, actual);
    expected = new LabeledValue[]{
                                  new DefaultContingentLabeledValue(new Proposition(alpha, beta, gamma), ConstraintCase.UPPERCASE, x, X, v.neg()),
                                  new DefaultContingentLabeledValue(new Proposition(notAlpha, beta, gamma, p), ConstraintCase.UPPERCASE, x, X, v.neg())
                                 };
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    
    //===========
    
    // x < w < v  ====
    
    w = new DefaultTimeDistance(5, 0, 0, 0, 0);
    
    x = new DefaultTimeDistance(2, 0, 0, 0, 0);
    v = new DefaultTimeDistance(10, 0, 0, 0, 0);
    
    PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(alpha, beta), w.neg())));
    XY = new DefaultLabeledEdge(X, Y, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(beta, gamma, p), ConstraintCase.UPPERCASE, x, X, v.neg())));
   
    edges = new ArrayList<LabeledEdge>();
    edges.add(PX);
    edges.add(XY);
    
    cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
    
    
    actual =DefaultLabeledValueSet.toMap(XY.getValue()); 
    getSolver().propositionModificationRuleR1(cstnu, X, P, actual);
    expected = new LabeledValue[]{
                                  new DefaultContingentLabeledValue(new Proposition(alpha, beta, gamma), ConstraintCase.UPPERCASE, x, X, v.neg()),
                                  new DefaultContingentLabeledValue(new Proposition(notAlpha, beta, gamma, p), ConstraintCase.UPPERCASE, x, X, v.neg())
                                 };
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    
    //===========
    
   // w < x < v  ====
    
    w = new DefaultTimeDistance(-6, 0, 0, 0, 0);
    
    x = new DefaultTimeDistance(5, 0, 0, 0, 0);
    v = new DefaultTimeDistance(5, 0, 0, 0, 0);
    
    PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(alpha, beta), w.neg())));
    XY = new DefaultLabeledEdge(X, Y, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(beta, gamma, p), ConstraintCase.UPPERCASE, x, X, v.neg())));
   
    edges = new ArrayList<LabeledEdge>();
    edges.add(PX);
    edges.add(XY);
    
    cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
    
    
    actual =DefaultLabeledValueSet.toMap(XY.getValue()); 
    getSolver().propositionModificationRuleR1(cstnu, X, P, actual);
    expected = new LabeledValue[]{
                                  new DefaultContingentLabeledValue(new Proposition(beta, gamma, p), ConstraintCase.UPPERCASE, x, X, v.neg()),
                                 };
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    
    //===========
    
   // no alpha  ====
    
    w = new DefaultTimeDistance(5, 0, 0, 0, 0);
    
    x = new DefaultTimeDistance(2, 0, 0, 0, 0);
    v = new DefaultTimeDistance(10, 0, 0, 0, 0);
    
    PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(beta), w.neg())));
    XY = new DefaultLabeledEdge(X, Y, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(beta, gamma, p), ConstraintCase.UPPERCASE, x, X, v.neg())));
   
    edges = new ArrayList<LabeledEdge>();
    edges.add(PX);
    edges.add(XY);
    
    cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
    
    
    actual =DefaultLabeledValueSet.toMap(XY.getValue()); 
    getSolver().propositionModificationRuleR1(cstnu, X, P, actual);
    expected = new LabeledValue[]{
                                  new DefaultContingentLabeledValue(new Proposition(beta, gamma), ConstraintCase.UPPERCASE, x, X, v.neg()),
                                 };
    
    assertEqualsIgnoreOrdering(expected, actual.values());
    
    //===========
  
  }

  @Test
  public void testPropositionModificationRuleR2()
  {
    IterationChain iteration = new IterationChain(0);
    
    DefaultObservationTimePoint P = new DefaultObservationTimePoint(new NodeInstance(7, iteration, -1), "P?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding p = P.getObservationForEdgeCode(1l);
//    ObservationBinding np = P.getObservationForEdgeCode(2l);
    
    DefaultObservationTimePoint A = new DefaultObservationTimePoint(new NodeInstance(1, iteration, -1), "A?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding alpha = A.getObservationForEdgeCode(1l);
    ObservationBinding notAlpha = A.getObservationForEdgeCode(2l);
    
    DefaultObservationTimePoint B = new DefaultObservationTimePoint(new NodeInstance(2, iteration, -1), "B?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding beta = B.getObservationForEdgeCode(1l);
    
    DefaultObservationTimePoint G = new DefaultObservationTimePoint(new NodeInstance(3, iteration, -1), "G?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding gamma = G.getObservationForEdgeCode(1l);
    
    LabeledTimePoint X = new DefaultLabeledNode(new NodeInstance(42, iteration, -1), "X", new Proposition(beta));


    
    
    TimeDistance w = new DefaultTimeDistance(10, 0, 0, 0, 0);
    TimeDistance v = new DefaultTimeDistance(15, 0, 0, 0, 0);
    
    LabeledEdge PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(alpha, beta), w.neg())));
    LabeledEdge XP = new DefaultLabeledEdge(X, P, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(beta, gamma, p), v)));
    
    List<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    nodes.add(X);
    nodes.add(P);
    nodes.add(A);
    nodes.add(B);
    nodes.add(G);
    
    List<LabeledEdge> edges = new ArrayList<LabeledEdge>();
    edges.add(PX);
    edges.add(XP);
    
    CSTNU cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
    
    Map<Label, LabeledValue> actual =  DefaultLabeledValueSet.toMap(XP.getValue());
    getSolver().propositionModificationRuleR2(cstnu, X, P, actual);
    
    LabeledValue[] expected = new LabeledValue[]{
                                                 new DefaultLabeledValue(new Proposition(alpha, beta, gamma), v),
                                                 new DefaultLabeledValue(new Proposition(notAlpha, beta, gamma, p), v)
                                                 };
    
    assertEqualsIgnoreOrdering(expected, actual.values());

  
  }

  @Test
  public void testPropositionModificationRuleR3()
  {
    IterationChain iteration = new IterationChain(0);
    
    TimeDistance w = new DefaultTimeDistance(10, 0, 0, 0, 0);
    TimeDistance v = new DefaultTimeDistance(5, 0, 0, 0, 0);
    
    {
      DefaultObservationTimePoint P = new DefaultObservationTimePoint(new NodeInstance(7, iteration, -1), "P?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
      ObservationBinding p = P.getObservationForEdgeCode(1l);
  //    ObservationBinding np = P.getObservationForEdgeCode(2l);
      
      DefaultObservationTimePoint A = new DefaultObservationTimePoint(new NodeInstance(1, iteration, -1), "A?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
      ObservationBinding alpha = A.getObservationForEdgeCode(1l);
      ObservationBinding notAlpha = A.getObservationForEdgeCode(2l);
      
      DefaultObservationTimePoint B = new DefaultObservationTimePoint(new NodeInstance(2, iteration, -1), "B?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
      ObservationBinding beta = B.getObservationForEdgeCode(1l);
      
      DefaultObservationTimePoint G = new DefaultObservationTimePoint(new NodeInstance(3, iteration, -1), "G?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
      ObservationBinding gamma = G.getObservationForEdgeCode(1l);
      
      
      LabeledTimePoint X = new DefaultLabeledNode(new NodeInstance(42, iteration, -1), "X", new Proposition(beta));
      LabeledTimePoint Y = new DefaultLabeledNode(new NodeInstance(43, iteration, -1), "Y", new Proposition(gamma));
      
      
      LabeledEdge PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(alpha, beta), w.neg())));
      LabeledEdge YX = new DefaultLabeledEdge(Y, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(beta, gamma, p), v.neg())));
      
      List<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
      nodes.add(X);
      nodes.add(Y);
      nodes.add(P);
      nodes.add(A);
      nodes.add(B);
      nodes.add(G);
      
      List<LabeledEdge> edges = new ArrayList<LabeledEdge>();
      edges.add(PX);
      edges.add(YX);
      
      CSTNU cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
      Map<Label, LabeledValue> actual = DefaultLabeledValueSet.toMap(YX.getValue());
      getSolver().propositionModificationRuleR3(cstnu, P, X, Y, actual);
      
      LabeledValue[] expected = new LabeledValue[]{
                                                   new DefaultLabeledValue(new Proposition(alpha, beta, gamma), v.neg()),
                                                   new DefaultLabeledValue(new Proposition(notAlpha, beta, gamma, p), v.neg())
                                                   };
      
      assertEqualsIgnoreOrdering(expected, actual.values());
      
    }
    
    // Test 2 #####################################
    // A X 3_{E}
    // Y: L= (A;1:2) (P;1:2)
    // X 3_{5} : L= (A;1:2) 
    // P X 12_{E} : L= (A;1:2)
    
//    {
//      ObservationNode A = new ObservationNode(new NodeInstance(1, iteration, -1), "A?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
//      ObservationBinding alpha = A.getObservationForEdgeCode(1l);
//      ObservationBinding notAlpha = A.getObservationForEdgeCode(2l);
//      
//      ObservationNode P = new ObservationNode(new NodeInstance(7, iteration, -1), "P?", new DefaultLabelSet(new Proposition(alpha)), new long[]{1l, 2l});
//      ObservationBinding p = P.getObservationForEdgeCode(1l);
//      
//      LabeledNode X = new DefaultLabeledNode(new NodeInstance(42, iteration, -1), "X", Proposition.EMPTY_PROPOSITION);
//      LabeledNode Y = new DefaultLabeledNode(new NodeInstance(43, iteration, -1), "Y", new DefaultLabelSet(new Proposition(p)));
//      
//      LabeledEdge PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(alpha), w.neg())));
//      LabeledEdge YX = new DefaultLabeledEdge(Y, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(p), v.neg())));
//      
//      List<LabeledNode> nodes = new ArrayList<LabeledNode>();
//      nodes.add(X);
//      nodes.add(Y);
//      nodes.add(P);
//      nodes.add(A);
//      
//      List<LabeledEdge> edges = new ArrayList<LabeledEdge>();
//      edges.add(PX);
//      edges.add(YX);
//      
//      CSTNU cstnu = new CSTNU(null, nodes, edges, new ArrayList<NodeMapping<LabeledNode>>(), new ArrayList<Observation>());
//      
//      Collection<LabeledValue> actual = getSolver().propositionModificationRuleR1(cstnu, X, P, toMap(YX.getValue().getLabeledValues()));
//      
//      LabeledValue[] expected = new LabeledValue[]{
//                                                   new DefaultLabeledValue(new Proposition(alpha), v.neg()),
//                                                   new DefaultLabeledValue(new Proposition(notAlpha, p), v.neg())
//                                                   };
//      
//      assertArrayEqualsIgnoreOrdering(expected, actual);
//    }
    
    // Test 3 ###########################
    {
      w = new DefaultTimeDistance(2, 0, 0, 0, 0);
      v = new DefaultTimeDistance(1, 0, 0, 0, 0);
      DefaultObservationTimePoint P = new DefaultObservationTimePoint(new NodeInstance(7, iteration, -1), "P?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
      ObservationBinding p = P.getObservationForEdgeCode(1l);
      
      LabeledTimePoint X = new DefaultLabeledNode(new NodeInstance(42, iteration, -1), "X", Proposition.EMPTY_PROPOSITION);
      LabeledTimePoint Y = new DefaultLabeledNode(new NodeInstance(43, iteration, -1), "Y", new Proposition(p));
      
      LabeledEdge PX = new DefaultLabeledEdge(P, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(), w.neg())));
      LabeledEdge YX = new DefaultLabeledEdge(Y, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(p), v.neg())));
      LabeledEdge YP = new DefaultLabeledEdge(Y, P, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(p), DefaultTimeDistance.ONE_TIME_DISTANCE.neg())));
      
      List<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
      nodes.add(X);
      nodes.add(Y);
      nodes.add(P);
      
      List<LabeledEdge> edges = new ArrayList<LabeledEdge>();
      edges.add(PX);
      edges.add(YX);
      edges.add(YP);
      
      CSTNU cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
      
      //TCNTestTools.printTCN(cstnu, getClass(), "Rule3");
//      TCNTestTools.export(cstnu, getClass(), "Rule3");
      Map<Label, LabeledValue> actual = DefaultLabeledValueSet.toMap(YX.getValue());
      getSolver().propositionModificationRuleR3(cstnu, P, X, Y, actual);
      
      
      
      LabeledValue[] expected = new LabeledValue[]{
                                                   new DefaultLabeledValue(new Proposition(p), v.neg()),
                                                   //new DefaultLabeledValue(new Proposition(p), v.neg())
                                                   };
      
      assertEqualsIgnoreOrdering(expected, actual.values());
    }
  }
  
  @Test
  public void testObservationCaseRule() throws TCNInconsistencyException, InterruptedException
  {
    IterationChain iteration = new IterationChain(0);
    
    DefaultObservationTimePoint A = new DefaultObservationTimePoint(new NodeInstance(7, iteration, -1), "A?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding alpha = A.getObservationForEdgeCode(1l);
    DefaultObservationTimePoint B = new DefaultObservationTimePoint(new NodeInstance(8, iteration, -1), "B?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding beta = B.getObservationForEdgeCode(1l);
    DefaultObservationTimePoint G = new DefaultObservationTimePoint(new NodeInstance(9, iteration, -1), "G?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding gamma = G.getObservationForEdgeCode(1l);
    
    DefaultObservationTimePoint P = new DefaultObservationTimePoint(new NodeInstance(10, iteration, -1), "P?", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    ObservationBinding p = P.getObservationForEdgeCode(1l);
    ObservationBinding np = P.getObservationForEdgeCode(2l);
    
    LabeledTimePoint X = new DefaultLabeledNode(new NodeInstance(1, iteration, -1), "X", Proposition.EMPTY_PROPOSITION);
    LabeledTimePoint Y = new DefaultLabeledNode(new NodeInstance(2, iteration, -1), "Y", Proposition.EMPTY_PROPOSITION);


    
    
    TimeDistance u = new DefaultTimeDistance(8, 0, 0, 0, 0);
    TimeDistance v = new DefaultTimeDistance(18, 0, 0, 0, 0);
    
    LabeledEdge XY = new DefaultLabeledEdge(X, Y, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(alpha, beta, p), u)));
    LabeledEdge YX = new DefaultLabeledEdge(Y, X, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(beta, gamma, np), v.neg())));
    
//    Map<Label, LabeledValue> actual =  DefaultLabeledValueSet.toMap(XP.getValue());
    LabeledValue l_XY = XY.getValue().getLabeledValues().iterator().next();
    LabeledValue l_YX = YX.getValue().getLabeledValues().iterator().next();
    LabeledValue actual = getSolver().observationCaseReduction(l_XY, l_YX, P);
    
    assertNotNull(actual);
    LabeledValue expected = new DefaultLabeledValue(new Proposition(alpha, beta, gamma), DefaultTimeDistance.ZERO_TIME_DISTANCE);  
    assertEquals(expected, actual);
    

    
    List<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    nodes.add(A);
    nodes.add(B);
    nodes.add(G);
    
    nodes.add(P);
    nodes.add(X);
    nodes.add(Y);
    
    
    
    List<LabeledEdge> edges = new ArrayList<LabeledEdge>();
    edges.add(XY);
    edges.add(YX);
    
    ChangeableCSTNU cstnu = new DefaultCSTNU(nodes, edges,  
                                   new ArrayList<ChoiceVariable>());
    solve(cstnu);
    
    LabeledEdge YP = cstnu.getEdge(Y, P);
    
    assertNotNull(YP);
    assertEqualsIgnoreOrdering(new LabeledValue[]{expected}, YP.getValue().getLabeledValues());
    
  }
  
  @Test
  public void testXOR1() throws VersionException, XMLFormatException, IOException, TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.xorTimeConstraintTestTemplate;
    
    TCNTransformation<LabeledTimePoint, LabeledEdge> trafo = createTransformation(template);
    ChangeableTCN<LabeledTimePoint, LabeledEdge> timeModel = (ChangeableTCN<LabeledTimePoint, LabeledEdge>) trafo.transform();
    Collection<NodeMapping<LabeledTimePoint>> nodeMappings = trafo.getNodeMapping();
    
    timeModel = solve(timeModel);
    
    IterationChain iterationChain = new IterationChain(template.getStartNode().getID());
    
    // Nodes 4 and 5 are on different branches
    NodeMapping<LabeledTimePoint> nodeMapping1 = getNodeMapping(nodeMappings, 4, iterationChain);
    NodeMapping<LabeledTimePoint> nodeMapping2 = getNodeMapping(nodeMappings, 5, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(timeModel.getEdge(node1, node2));
        assertNull(timeModel.getEdge(node2, node1));
      }
    }
  
  }
  
  @Test
  public void testXOR2() throws VersionException, XMLFormatException, IOException, TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.xorAdvancedNotPC2TestTemplate;
    
    TCNTransformation<LabeledTimePoint, LabeledEdge> trafo = createTransformation(template);
    ChangeableTCN<LabeledTimePoint, LabeledEdge> timeModel = (ChangeableTCN<LabeledTimePoint, LabeledEdge>) trafo.transform();
    Collection<NodeMapping<LabeledTimePoint>> nodeMappings = trafo.getNodeMapping();
    
    timeModel = solve(timeModel);
    
    IterationChain iterationChain = new IterationChain(template.getStartNode().getID());
    
    // Nodes 5 and 6 are on different branches
    NodeMapping<LabeledTimePoint> nodeMapping1 = getNodeMapping(nodeMappings, 5, iterationChain);
    NodeMapping<LabeledTimePoint> nodeMapping2 = getNodeMapping(nodeMappings, 6, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(timeModel.getEdge(node1, node2));
        assertNull(timeModel.getEdge(node2, node1));
      }
    }
  }
  
  @Test
  public void testXOR3() throws VersionException, XMLFormatException, IOException, TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.xorAndTestTemplate;
    
    CSTNUTemplateTransformation trafo = createTransformation(template);
    ChangeableCSTNU timeModel = (ChangeableCSTNU) trafo.transform();
    Collection<NodeMapping<LabeledTimePoint>> nodeMappings = trafo.getNodeMapping();
    
    TCNTestTools.assertConsistent(timeModel);
//    TCNTestTools.printTCN(timeModel, this.getClass(), "XORAndTest");
    
    TCNTestTools.assertConsistent(new CSTNUAllMaxProjection(timeModel));
    
    timeModel = solve(timeModel);
    
    IterationChain iterationChain = new IterationChain(template.getStartNode().getID());
    
    // Nodes are on different branches
    NodeMapping<LabeledTimePoint> nodeMapping1 = getNodeMapping(nodeMappings, 9, iterationChain);
    NodeMapping<LabeledTimePoint> nodeMapping2 = getNodeMapping(nodeMappings, 10, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(timeModel.getEdge(node1, node2));
        assertNull(timeModel.getEdge(node2, node1));
      }
    }
    
    // Nodes are on different branches
    nodeMapping1 = getNodeMapping(nodeMappings, 14, iterationChain);
    nodeMapping2 = getNodeMapping(nodeMappings, 15, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(timeModel.getEdge(node1, node2));
        assertNull(timeModel.getEdge(node2, node1));
      }
    }
  

    // Nodes are on different branches
    int[] others = new int[] { 5, 6, 7, 8, 9, 10, 12, 13, 14, 15, 16 };
    
    for (int o : others)
    {

      try
      {
        nodeMapping1 = getNodeMapping(nodeMappings, 11, iterationChain);
        nodeMapping2 = getNodeMapping(nodeMappings, o, iterationChain);
        
        for (LabeledTimePoint node1 : nodeMapping1.getNodes())
        {
          for (LabeledTimePoint node2 : nodeMapping2.getNodes())
          {
            assertNull(timeModel.getEdge(node1, node2));
            assertNull(timeModel.getEdge(node2, node1));
          }
        }
      }
      catch (AssertionError e)
      {
        AssertionError error = new AssertionError("Node "+o+" "+e.getMessage());
        error.initCause(e);
        error.setStackTrace(e.getStackTrace());
        throw e;
      }
    }
    
   
  }
  
  @Test
  public void testXOR4() throws VersionException, XMLFormatException, IOException, TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.xorEmptyBranchTestTemplate;
    
    TCNTransformation<LabeledTimePoint, LabeledEdge> trafo = createTransformation(template);
    ChangeableTCN<LabeledTimePoint, LabeledEdge> timeModel = (ChangeableTCN<LabeledTimePoint, LabeledEdge>) trafo.transform();
    Collection<NodeMapping<LabeledTimePoint>> nodeMappings = trafo.getNodeMapping();
    
    timeModel = solve(timeModel);
    
    IterationChain iterationChain = new IterationChain(template.getStartNode().getID());
    
    // Nodes 4 and 5 are on different branches
    NodeMapping<LabeledTimePoint> nodeMapping1 = getNodeMapping(nodeMappings, 4, iterationChain);
    NodeMapping<LabeledTimePoint> nodeMapping2 = getNodeMapping(nodeMappings, 5, iterationChain);
    
    for (LabeledTimePoint node1 : nodeMapping1.getNodes())
    {
      for (LabeledTimePoint node2 : nodeMapping2.getNodes())
      {
        assertNull(timeModel.getEdge(node1, node2));
        assertNull(timeModel.getEdge(node2, node1));
      }
    }
  }



  
  @Test
  public void testIntersection()
  {
    LabeledValueSet labelSet1;
    LabeledValueSet labelSet2;
    Map<Label, LabeledValue> actual;
    LabeledValueSet expected;
    
    LabeledTimePoint trigger = new DefaultLabeledNode(new NodeInstance(0, new IterationChain(0), -1), "E", Proposition.EMPTY_PROPOSITION);
    
    labelSet1 = new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(), 
      new DefaultTimeDistance(0, 45, TimeDistance.UNDEFINED_TIME_DISTANCE,
        TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE)));
    
    
    labelSet2 = new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(), ConstraintCase.UPPERCASE,
      new DefaultTimeDistance(0, 25, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE),
      trigger,
      new DefaultTimeDistance(0, 15, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE)));
    
    
    //(<o, 45m0s, #, NaN>;) u (<o, 25m0s, /2_{E}\, 15m0s>;) = (<o, 25m0s, #, NaN>;)
    
    expected = new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(), ConstraintCase.UPPERCASE,
      new DefaultTimeDistance(0, 25, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE),
      trigger,
      new DefaultTimeDistance(0, 15, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE, TimeDistance.UNDEFINED_TIME_DISTANCE)));
    
    
    actual = getSolver().intersect(labelSet1, DefaultLabeledValueSet.toMap(labelSet2));
    
    assertEqualsIgnoreOrdering(expected.getLabeledValues(), actual.values());
  }
  
  @Test
  public void testSimplify()
  {
    IterationChain iteration = new IterationChain(0);
    
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    
    DefaultLabeledValueSet labelSet;
    DefaultLabeledValueSet labelSetE;
    
    LabeledTimePoint trigger = new DefaultLabeledNode(new NodeInstance(3, iteration, -1), "T", Proposition.EMPTY_PROPOSITION);
    DefaultObservationTimePoint o1_2 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2_2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
   
    ObservationBinding nA = o1_2.getObservationForEdgeCode(1l);
    ObservationBinding A = o1_2.getObservationForEdgeCode(2l);

    ObservationBinding nB = o2_2.getObservationForEdgeCode(1l);
    ObservationBinding B = o2_2.getObservationForEdgeCode(2l);
    
    
    // 
    //    <B , -5s, /0_{N9}\, -10s>;
    //    <B A , -10s>;
    //    <B -A , -5s, /0_{N9}\, -10s>;
    
    
    labelSet = new DefaultLabeledValueSet(
      new DefaultContingentLabeledValue(new Proposition(B), ConstraintCase.UPPERCASE, new DefaultTimeDistance(-5, 0, 0, 0, 0), trigger, new DefaultTimeDistance(-10, 0, 0, 0, 0)), // <B , -5s, /0_{N9}\, -10s>;
      new DefaultLabeledValue(new Proposition(B, A), new DefaultTimeDistance(-10, 0, 0, 0, 0)), // <B A , -10s>;
      new DefaultContingentLabeledValue(new Proposition(B, nA), ConstraintCase.UPPERCASE, new DefaultTimeDistance(-5, 0, 0, 0, 0), trigger, new DefaultTimeDistance(-10, 0, 0, 0, 0)) // <B -A , -5s, /0_{N9}\, -10s>;
      );
    
    labelSetE = new DefaultLabeledValueSet(
      new DefaultContingentLabeledValue(new Proposition(B), ConstraintCase.UPPERCASE, new DefaultTimeDistance(-5, 0, 0, 0, 0), trigger, new DefaultTimeDistance(-10, 0, 0, 0, 0)), // <B , -5s, /0_{N9}\, -10s>;
      new DefaultLabeledValue(new Proposition(B, A), new DefaultTimeDistance(-10, 0, 0, 0, 0)) // <B A , -10s>;
    );
    
    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet)); 
    
    
    
    //    <B , 12s>;
    //    <B A, infty, /0_{N13}\, 12s>;         
    //    <B -A, 10s>;
    labelSet = new DefaultLabeledValueSet( 
      new DefaultLabeledValue(new Proposition(B), new DefaultTimeDistance(12, 0, 0, 0, 0)), // <B , 12s>;
      new DefaultContingentLabeledValue(new Proposition(B, A), ConstraintCase.UPPERCASE, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, trigger, new DefaultTimeDistance(12, 0, 0, 0, 0)), // <B A, infty, /0_{N13}\, 12s>;   
      new DefaultLabeledValue(new Proposition(B, nA), new DefaultTimeDistance(10, 0, 0, 0, 0)) //  <B -A, 10s>;
      );
    
    labelSetE = new DefaultLabeledValueSet(
      new DefaultLabeledValue(new Proposition(B), new DefaultTimeDistance(12, 0, 0, 0, 0)), //  <B , 12s>;
      new DefaultLabeledValue(new Proposition(B, nA), new DefaultTimeDistance(10, 0, 0, 0, 0)) //  <B -A, 10s>;
    );
    
    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet)); 
    
    //<A , infty, /0_{N9}\, -14s>;
    //<-A , -11s, /0_{N9}\, -16s>;
    
    labelSet = new DefaultLabeledValueSet( 
      new DefaultContingentLabeledValue(new Proposition(A), ConstraintCase.UPPERCASE, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, trigger, new DefaultTimeDistance(-14, 0, 0, 0, 0)), // <A , infty, /0_{N9}\, -14s>;   
      new DefaultContingentLabeledValue(new Proposition(nA), ConstraintCase.UPPERCASE, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, trigger, new DefaultTimeDistance(-16, 0, 0, 0, 0)) // <-A , -11s, /0_{N9}\, -16s>;
      );
    
    labelSetE = new DefaultLabeledValueSet(
      new DefaultContingentLabeledValue(new Proposition(), ConstraintCase.UPPERCASE, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, trigger, new DefaultTimeDistance(-14, 0, 0, 0, 0))
    );
    
    assertEquals(labelSetE, getSolver().applySimplificationRules(labelSet)); 
    
  }



  @Test
  public void testMerge() throws Exception
  {
    
    
    LabeledValue labeledValue1;
    LabeledValue labeledValue2;
    LabeledValue expected;
    LabeledValue actual;
    
    IterationChain iteration = new IterationChain(0);
    NodeInstance nodeInstance = new NodeInstance(0, iteration, -1);
    Label label = Proposition.EMPTY_PROPOSITION;
    LabeledTimePoint trigger = new DefaultLabeledNode(nodeInstance, "B", label);
    
    //  <o, -1s, /B\, -2s>
    labeledValue1 = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, new DefaultTimeDistance(-1, 0, 0, 0, 0), trigger, new DefaultTimeDistance(-2, 0, 0, 0, 0));
        
    //  <o, 0s, #, NaN>
    labeledValue2 = new DefaultLabeledValue(label, new DefaultTimeDistance(0, 0, 0, 0, 0));
    
    //  <o, -1s, /B\, -2s>
    expected = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, new DefaultTimeDistance(-1, 0, 0, 0, 0), trigger, new DefaultTimeDistance(-2, 0, 0, 0, 0));
    
    actual = getSolver().merge(labeledValue1, labeledValue2);
    assertEquals(expected, actual);
  
    actual = getSolver().merge(labeledValue2, labeledValue1);
    assertEquals(expected, actual);
    

    // <o, -1s, #, NaN>, <o, infty, /B\, -1s>
    labeledValue1  = new DefaultLabeledValue(label, new DefaultTimeDistance(-1, 0, 0, 0, 0));
    labeledValue2 = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, 
                                          DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, trigger, 
                                          new DefaultTimeDistance(-1, 0, 0, 0, 0));
    
    expected = new DefaultLabeledValue(label, new DefaultTimeDistance(-1, 0, 0, 0, 0));
    actual = getSolver().merge(labeledValue1, labeledValue2);
    assertEquals(expected, actual);
  }



  @Test
  public void testAddOrReplaceIfNecessary() throws Exception
  {
    LabeledValue labeledValue1;
    LabeledValue labeledValue2;
    LabeledValue expected;
    LabeledValue actual;
    
    IterationChain iteration = new IterationChain(0);
    NodeInstance nodeInstance = new NodeInstance(0, iteration, -1);
    Label label = Proposition.EMPTY_PROPOSITION;
    LabeledTimePoint trigger = new DefaultLabeledNode(nodeInstance, "B", label);
    
    
    //  <o, 0s, #, NaN>
    labeledValue1 = new DefaultLabeledValue(label, new DefaultTimeDistance(-1, 0, 0, 0, 0));
    
    //  <o, -infty, /B\, -2s>
    labeledValue2 = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, trigger, new DefaultTimeDistance(-2, 0, 0, 0, 0));
    
    //  <o, 0s, /B\, -2s>
    expected = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, new DefaultTimeDistance(-1, 0, 0, 0, 0), trigger, new DefaultTimeDistance(-2, 0, 0, 0, 0));
    
    Map<Label, LabeledValue> labels = new HashMap<Label, LabeledValue>();
    
    getSolver().addOrReplaceIfNecessary(labels, labeledValue1);
    getSolver().addOrReplaceIfNecessary(labels, labeledValue2);
    
    
    actual = labels.get(label);
    assertEquals(expected, actual);    
    
    labels.clear();
    
    getSolver().addOrReplaceIfNecessary(labels, labeledValue2);
    getSolver().addOrReplaceIfNecessary(labels, labeledValue1);
    
    
    actual = labels.get(label);
    assertEquals(expected, actual);    
    
    //##################################
    
    //  <o, -5s, #, NaN>
    labeledValue1 = new DefaultLabeledValue(label, new DefaultTimeDistance(-5, 0, 0, 0, 0));
    
    //  <o, -infty, /B\, -2s>
    labeledValue2 = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, trigger, new DefaultTimeDistance(-2, 0, 0, 0, 0));

    //  <o, -5s, #, NaN>
    expected = new DefaultLabeledValue(label, new DefaultTimeDistance(-5, 0, 0, 0, 0));
    
    labels.clear();
    
    getSolver().addOrReplaceIfNecessary(labels, labeledValue1);
    getSolver().addOrReplaceIfNecessary(labels, labeledValue2);
    
    
    actual = labels.get(label);
    assertEquals(expected, actual);    
    
    labels.clear();
    
    getSolver().addOrReplaceIfNecessary(labels, labeledValue2);
    getSolver().addOrReplaceIfNecessary(labels, labeledValue1);
    
    
    actual = labels.get(label);
    assertEquals(expected, actual);   
  }
}
