package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import java.util.UUID;

import org.apache.commons.configuration.BaseConfiguration;
import org.apache.commons.configuration.Configuration;
import org.junit.BeforeClass;
import org.junit.Test;

import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.InconsistencyException;
import de.aristaflow.adept2.model.timemodel.TimeModelSolver;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUDCSolver;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUInstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTemplateTransformation;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;

public class CSTNUInstanceTimeModelTest
{

  private static CSTNUTimeModelFactory factory;

  @BeforeClass
  public static void setUpBeforeClass() throws Exception
  {
    factory = new CSTNUTimeModelFactory();
  }

  @Test
  public void testCSTNUInstanceTimeModel() throws CloneNotSupportedException, InconsistencyException, InterruptedException
  {
    CSTNUInstanceTimeModel timeModel = createInstanceTimeModel(TestTemplates.simpleTimeTestContingentTestTemplate);

    simulateProcessInstance(timeModel);
  }

  private CSTNUInstanceTimeModel createInstanceTimeModel(TemporalTemplate template)
      throws TCNInconsistencyException, CloneNotSupportedException, InterruptedException
  {
    Configuration configuration = new BaseConfiguration();
    configuration.setProperty(CSTNUTemplateTransformation.CONF_ADVANCED_ACTIVITY_TRANSFORMATION, CSTNUTemplateTransformation.ACTIVITY_TRANSFORMATION_BASIC);
    
    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(configuration, template, 0);
    ChangeableCSTNU cstnu = (ChangeableCSTNU) trafo.transform();

    //printProcessTimes(template, tcn);
    

//    CstnuCStnProjection stnProjection = new CstnuCStnProjection(cstnu, true);
//    PC2Solver<LabeledNode, LabeledEdge> pc2Solver = new PC2Solver<LabeledNode, LabeledEdge>();
//    pc2Solver.solve(stnProjection);
//    STN<LabeledNode,LabeledEdge> stn = stnProjection.clone();
    
    
    //tcn.visit(visitor.getInstance("C_Before"));
    
    CSTNUDCSolver solver = new CSTNUDCSolver();
    try
    {
      cstnu = solver.solve(cstnu);
    }
    catch (TCNInconsistencyException e)
    {
      //tcn.visit(visitor.getInstance("C_Failure"));
      throw e;
    }
    
    //tcn.visit(visitor.getInstance("C_After"));
    
    
    CSTNUInstanceTimeModel timeModel = new CSTNUInstanceTimeModel(UUID.randomUUID(), template.getID(), cstnu,
                                                                  trafo.getNodeMapping(),
                                                                  trafo.getProcessCreationTimePoint(),
                                                                  trafo.getProcessStartTimePoint(),
                                                                  trafo.getProcessEndTimePoint(), null);
    return timeModel;
  }

  private void simulateProcessInstance(CSTNUInstanceTimeModel timeModel) throws InconsistencyException, InterruptedException
  {
    TimeModelSolver timeModelSolver = factory.getTimeModelSolver(timeModel.getClass());
    
    timeModel.setInstanceCreationTime(System.currentTimeMillis());
    
    timeModelSolver.solve(timeModel);
    
    
    //timeModel.projection.visit(visitor.getInstance("Init"));
    
    timeModel.setInstanceStartTime(System.currentTimeMillis()+5000);
    
    //timeModel.projection.visit(visitor.getInstance("InstanceStart"));
    
    //TestTools.printProcessTimes(timeModel);
    
    timeModelSolver.solve(timeModel);
    
    
    //timeModel.projection.visit(visitor.getInstance("InstanceStart_Propagation"));
    
    
  }

  

  
}
