package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertTDEquals;
import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertTPEquals;

import org.apache.commons.configuration.BaseConfiguration;
import org.apache.commons.configuration.Configuration;
import org.junit.Test;

import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.AbstractSolverTest;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;

public class CSTNUTemplateTimeModelTest extends AbstractSolverTest<LabeledValueSet, LabeledEdge>
{



  @Test
  public void testCSTNUTemplateTimeModel() throws TCNInconsistencyException, InterruptedException
  {
   
    
    TemporalTemplate template = TestTemplates.simpleTimeTestContingentTestTemplate;
//    TestTools.printTemplate(template);

    TemplateTimeModel timeModel = getCSTNUTemplateTimeModel(template);
    
    //TestTools.printProcessTimes(timeModel);
    
    assertTDEquals(m(20)+cmin(1)+stmin(1),//m(60)+cmin(1)+stmin(1), 
                   m(65), 
                   timeModel.getProcessDuration(), 1);
    assertTDEquals(sdmin(1), sdmax(1), timeModel.getActivityDuration(0), 1);
    assertTDEquals(sdmin(1), sdmax(1), timeModel.getActivityDuration(1), 1);

    assertTDEquals(m(5), m(15), timeModel.getActivityDuration(2), 1);
    assertTDEquals(m(5), m(30), timeModel.getActivityDuration(3), 1);
    assertTDEquals(m(5), m(15), timeModel.getActivityDuration(4), 1);

    assertTPEquals(0, 0, timeModel.getActivityStartTimeframe(0), 1);
    assertTPEquals(sdmin(1), 
                   sdmax(1), 
                   timeModel.getActivityCompletionTimeframe(0), 1);

    assertTPEquals(sdmin(1)+cmin(1),
                   m(5)+sdmax(1)-stmin(1),
                   timeModel.getActivityStartTimeframe(2), 1);
    assertTPEquals(m(5)+sdmin(1)+cmin(1),
                   m(20)+sdmax(1)-stmin(1),
                   timeModel.getActivityCompletionTimeframe(2), 1);

    assertTPEquals(m(5)+sdmin(1)+cmin(2),//m(15)+sdmin(1)+cmin(2),
                   m(20)+sdmax(1)-stmin(1)-cmin(1),
                   timeModel.getActivityStartTimeframe(3), 1);
    assertTPEquals(m(10)+sdmin(1)+cmin(2),//m(20)+sdmin(1)+cmin(2), 
                   m(50)+sdmax(1)-stmin(1)-cmin(1), 
                   timeModel.getActivityCompletionTimeframe(3), 1);

    assertTPEquals(m(15)+sdmin(1)+cmin(1),//m(45)+sdmin(1)+cmin(1), 
                   m(50)+sdmax(1)-stmin(1), 
                   timeModel.getActivityStartTimeframe(4), 1);
    assertTPEquals(m(20)+sdmin(1)+cmin(1),//m(50)+sdmin(1)+cmin(1), 
                   m(65)+sdmax(1)-stmin(1),  
                   timeModel.getActivityCompletionTimeframe(4), 1);

    assertTPEquals(m(20)+sdmin(1)+cmin(1)+stmin(1),//m(60)+sdmin(1)+cmin(1)+stmin(1),
                   m(65)+sdmax(1), 
                   timeModel.getActivityStartTimeframe(1), 1);
    assertTPEquals(m(20)+sdmin(2)+cmin(1)+stmin(1),//m(60)+sdmin(2)+cmin(1)+stmin(1), 
                   m(65)+sdmax(2),  
                   timeModel.getActivityCompletionTimeframe(1), 1);   
  }

  private TemplateTimeModel getCSTNUTemplateTimeModel(TemporalTemplate template)
      throws TCNInconsistencyException, InterruptedException
  {
    CSTNUTemplateTransformation trafo = createTransformation(template);
    ChangeableCSTNU cstnu = (ChangeableCSTNU) trafo.transform();
  
    CSTNUDCSolver solver = new CSTNUDCSolver();
    try
    {
      cstnu = solver.solve(cstnu);
    }
    catch (TCNInconsistencyException e)
    {
      throw e;
    }
    
    
    TemplateTimeModel timeModel = new CSTNUTemplateTimeModel(template, cstnu, trafo.getNodeMapping(),
                                                             trafo.getProcessCreationTimePoint(),
                                                             trafo.getProcessStartTimePoint(),
                                                             trafo.getProcessEndTimePoint(), 
                                                             null);
    return timeModel;
  }

  @Override
  protected CSTNUTemplateTransformation createTransformation(TemporalTemplate template)
  {
    Configuration configuration = new BaseConfiguration();
    configuration.setProperty(CSTNUTemplateTransformation.CONF_ADVANCED_ACTIVITY_TRANSFORMATION, CSTNUTemplateTransformation.ACTIVITY_TRANSFORMATION_BASIC);
    return new CSTNUTemplateTransformation(configuration, template, 0);
  }

}
