package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.fail;

import java.util.ArrayList;
import java.util.Collection;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;

import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultContingentLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledNode;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.DefaultLabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.Proposition;

public class DefaultCSTNUTest
{

  private static DefaultTimeModelOperations tmo;

  @BeforeClass
  public static void setUpBeforeClass() throws Exception
  {
    tmo = new DefaultTimeModelOperations();
  }

  @AfterClass
  public static void tearDownAfterClass() throws Exception
  {
    //
  }

  @Before
  public void setUp() throws Exception
  {
    //
  }

  @After
  public void tearDown() throws Exception
  {
    //
  }

  @Test
    public void testUpdateConstraint()
    {
      // Create some simple CSTNU
      Collection<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
      Collection<LabeledEdge> edges = new ArrayList<LabeledEdge>();
  
      int nodeID = 0;
      IterationChain iterationChain = new IterationChain(nodeID);
      Label branches = Proposition.EMPTY_PROPOSITION;
  
      LabeledTimePoint A = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "A", branches);
      LabeledTimePoint B = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "B", branches);
      LabeledTimePoint C = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "C", branches);
  
      nodes.add(A);
      nodes.add(B);
      nodes.add(C);
  
      long a = 5;
      long b = 41;
  
      LabeledEdge AB = new DefaultLabeledEdge(A, B, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(b))));
      LabeledEdge BA = new DefaultLabeledEdge(B, A, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(-a))));
      LabeledEdge AC = new DefaultLabeledEdge(A, C, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, tmo.secondsToTimeDistance(b), C, tmo.secondsToTimeDistance(a))));
      LabeledEdge CA = new DefaultLabeledEdge(C, A, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(-a))));
      LabeledEdge BC = new DefaultLabeledEdge(B, C, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(b))));
      LabeledEdge CB = new DefaultLabeledEdge(C, B, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(-a))));
  
      edges.add(AB);
      edges.add(BA);
      edges.add(AC);
      edges.add(CA);
      edges.add(BC);
      edges.add(CB);
  
      DefaultCSTNU cstnu = new DefaultCSTNU(nodes, edges, 
        new ArrayList<ChoiceVariable>());
      
      assertNotNull(cstnu.getEdge(A, B));
      assertNotNull(cstnu.getEdge(B, A));
      assertNotNull(cstnu.getEdge(A, C));
      assertNotNull(cstnu.getEdge(C, A));
      assertNotNull(cstnu.getEdge(B, C));
      assertNotNull(cstnu.getEdge(C, B));
      
      assertEquals(AB, cstnu.getEdge(A, B));
      assertEquals(BA, cstnu.getEdge(B, A));
      
      // Update the edge
      LabeledEdge AB2 = new DefaultLabeledEdge(A, B, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(17))));
      cstnu.updateConstraint(AB2);
      
      // Check the update has been successful and nothing else changed
      assertEquals(AB2, cstnu.getEdge(A, B));
      assertEquals(BA, cstnu.getEdge(B, A));
      
      try
      {
        cstnu.updateConstraint(null);
        fail("Null values are not allowed");
      }
      catch (IllegalArgumentException e)
      {
        //
      }
      
      LabeledEdge BC2 = new DefaultLabeledEdge(B, C, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, tmo.secondsToTimeDistance(17), C, tmo.secondsToTimeDistance(17))));
      
      assertEquals(BC, cstnu.getEdge(B, C));
      try
      {
        cstnu.updateConstraint(BC2);
        fail("Adding a contingent edge is not allowed");
      }
      catch (IllegalArgumentException e)
      {
        // Make sure nothing changed
        assertEquals(BC, cstnu.getEdge(B, C));
      }
      
      assertEquals(AC, cstnu.getEdge(A, C));
      LabeledEdge AC2 = new DefaultLabeledEdge(A, C, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, tmo.secondsToTimeDistance(b+5), C, tmo.secondsToTimeDistance(a))));
      cstnu.updateConstraint(AC2);
      assertEquals(AC2, cstnu.getEdge(A, C));
      
      LabeledEdge AC3 = new DefaultLabeledEdge(A, C, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, tmo.secondsToTimeDistance(b+5), C, tmo.secondsToTimeDistance(a+5))));
      try
      {
        cstnu.updateConstraint(AC3);
        fail("Updating a contingent value is not allowed");
      }
      catch (IllegalArgumentException e)
      {
        assertEquals(AC2, cstnu.getEdge(A, C));
      }
      
      AC3 = new DefaultLabeledEdge(A, C, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, tmo.secondsToTimeDistance(b), C, tmo.secondsToTimeDistance(a+5))));
      try
      {
        cstnu.updateConstraint(AC3);
        fail("Updating a contingent value is not allowed");
      }
      catch (IllegalArgumentException e)
      {
        assertEquals(AC2, cstnu.getEdge(A, C));
      }
    }

  @Test
    public void testRemoveTimePoint()
    {
      Collection<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
      Collection<LabeledEdge> edges = new ArrayList<LabeledEdge>();
  
      int nodeID = 0;
      IterationChain iterationChain = new IterationChain(nodeID);
      Label branches = Proposition.EMPTY_PROPOSITION;
  
      LabeledTimePoint A = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "A", branches);
      LabeledTimePoint B = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "B", branches);
      LabeledTimePoint C = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "C", branches);
  
      nodes.add(A);
      nodes.add(B);
      nodes.add(C);
  
      long a = 5;
      long b = 41;
  
      LabeledEdge AB = new DefaultLabeledEdge(A, B, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(b))));
  
      LabeledEdge BA = new DefaultLabeledEdge(B, A, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(-a))));
  
      LabeledEdge AC = new DefaultLabeledEdge(A, C, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(b))));
  
      LabeledEdge CA = new DefaultLabeledEdge(C, A, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(-a))));
  
      LabeledEdge BC = new DefaultLabeledEdge(B, C, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(b))));
  
      LabeledEdge CB = new DefaultLabeledEdge(C, B, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(-a))));
  
      edges.add(AB);
      edges.add(BA);
      edges.add(AC);
      edges.add(CA);
      edges.add(BC);
      edges.add(CB);
  
      DefaultCSTNU cstnu = new DefaultCSTNU(nodes, edges, 
        new ArrayList<ChoiceVariable>());
  
      assertNotNull(cstnu.getEdge(A, B));
      assertNotNull(cstnu.getEdge(B, A));
      assertNotNull(cstnu.getEdge(A, C));
      assertNotNull(cstnu.getEdge(C, A));
      assertNotNull(cstnu.getEdge(B, C));
      assertNotNull(cstnu.getEdge(C, B));
  
      cstnu.removeTimePoint(B);
  
      try
      {
        cstnu.getEdge(A, B);
        fail();
      }
      catch (IllegalArgumentException e)
      {
        // Expected
      }
      try
      {
        cstnu.getEdge(B, A);
        fail();
      }
      catch (IllegalArgumentException e)
      {
        // Expected
      }
      assertNotNull(cstnu.getEdge(A, C));
      assertNotNull(cstnu.getEdge(C, A));
      try
      {
        cstnu.getEdge(B, C);
        fail();
      }
      catch (IllegalArgumentException e)
      {
        // Expected
      }
      try
      {
        cstnu.getEdge(C, B);
        fail();
      }
      catch (IllegalArgumentException e)
      {
        // Expected
      }
    }

  @Ignore
  @Test
  public void testGetNodeMappingIntIterationChain()
  {
    fail("Not yet implemented");
  }

  @Ignore
  @Test
  public void testInsertNodeMapping()
  {
    fail("Not yet implemented");
  }

  @Test
  public void testInsertNode()
  {
    Collection<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    Collection<LabeledEdge> edges = new ArrayList<LabeledEdge>();

    int nodeID = 0;
    IterationChain iterationChain = new IterationChain(nodeID);
    Label branches = Proposition.EMPTY_PROPOSITION;

    LabeledTimePoint A = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "A", branches);
    LabeledTimePoint B = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "B", branches);

    nodes.add(A);
    nodes.add(B);

    long a = 5;
    long b = 41;

    LabeledEdge AeBs = new DefaultLabeledEdge(A, B, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(b))));

    LabeledEdge BsAe = new DefaultLabeledEdge(B, A, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo.secondsToTimeDistance(-a))));

    edges.add(AeBs);
    edges.add(BsAe);

    DefaultCSTNU cstnu = new DefaultCSTNU(nodes, edges, 
      new ArrayList<ChoiceVariable>());

    assertNotNull(cstnu.getEdge(A, B));
    assertNotNull(cstnu.getEdge(B, A));

    LabeledTimePoint C = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "C", branches);

    cstnu.insertTimePoint(C);

    assertNotNull(cstnu.getEdge(A, B));
    assertNotNull(cstnu.getEdge(B, A));

    assertNull(cstnu.getEdge(A, C));
    assertNull(cstnu.getEdge(B, C));
    assertNull(cstnu.getEdge(C, A));
    assertNull(cstnu.getEdge(C, B));
  }

}
