package de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assume.assumeTrue;

import java.util.Collection;

import org.junit.Test;

import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ExecutableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;

public class DefaultExecutableCSTNUTest
{

//  @BeforeClass
//  public static void setUpBeforeClass() throws Exception
//  {
//  }
//
//  @AfterClass
//  public static void tearDownAfterClass() throws Exception
//  {
//  }
//
//  @Before
//  public void setUp() throws Exception
//  {
//  }
//
//  @After
//  public void tearDown() throws Exception
//  {
//  }
  
  @Test
  public void testDefaultExecutableCSTNU() throws Exception
  {
    TemporalTemplate temporalTemplate = TestTemplates.simpleTimeTestContingentTestTemplate;
    
    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(temporalTemplate);
    ChangeableCSTNU cstnu = (ChangeableCSTNU) trafo.transform();
    
    LabeledTimePoint processCreationTimePoint = trafo.getProcessCreationTimePoint();
    
    CSTNUDCSolver solver = new CSTNUDCSolver();
    solver.solve(cstnu);
    
    ExecutableCSTNU executableCSTNU = new DefaultExecutableCSTNU(cstnu, solver, System.currentTimeMillis());
    
    Collection<LabeledTimePoint> executableTimePoints = executableCSTNU.getActiveTimePoints();
    assertEquals(1, executableTimePoints.size());
    assertTrue(executableTimePoints.contains(processCreationTimePoint));


  }

  @Test
  public void testExecuteTimePoint() throws Exception
  {
    TemporalTemplate temporalTemplate = TestTemplates.simpleTimeTestContingentTestTemplate;
    
    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(temporalTemplate);
    ChangeableCSTNU cstnu = (ChangeableCSTNU) trafo.transform();
    LabeledTimePoint processCreationTimePoint = trafo.getProcessCreationTimePoint();
    LabeledTimePoint processStartTimePoint = trafo.getProcessStartTimePoint();
    
    CSTNUDCSolver solver = new CSTNUDCSolver();
    solver.solve(cstnu);
    
    ExecutableCSTNU executableCSTNU = new DefaultExecutableCSTNU(cstnu, solver, System.currentTimeMillis());
    
    Collection<LabeledTimePoint> executableTimePoints = executableCSTNU.getActiveTimePoints();
    assumeTrue(executableTimePoints.contains(processCreationTimePoint));
    
    executableCSTNU.executeTimePoint(processCreationTimePoint, System.currentTimeMillis());
    
    executableTimePoints = executableCSTNU.getActiveTimePoints();
    assertEquals(1, executableTimePoints.size());
    assertTrue(executableTimePoints.contains(processStartTimePoint));
    
  }

}
