package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertTDEquals;
import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertTPEquals;

import org.junit.Test;

import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.STN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.TCNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTransformation;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;

public abstract class AbstractPC2SolverTest<V extends ConstraintValue, E extends TCNEdge<LabeledTimePoint, V>> 
  extends AbstractSolverTest<V, E>
{

  public abstract TCNSolver<LabeledTimePoint, V, E> getSolver();

  public AbstractPC2SolverTest()
  {
    super();
  }

  @Test
  public void testSolve() throws TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.simpleTimeTestContingentTestTemplate;
  
    TCNTransformation<LabeledTimePoint, E> trafo = createTransformation(template);
    ChangeableSTN<LabeledTimePoint, E> tcn = (ChangeableSTN<LabeledTimePoint, E>) trafo.transform(); 

    try
    {
      getSolver().solve(tcn);
    }
    catch (TCNInconsistencyException e)
    {
      throw e;
    }
  
    TemplateTimeModel timeModel = new TCNTemplateTimeModel<LabeledTimePoint, E>(template, tcn, trafo.getNodeMapping(), 
        trafo.getProcessCreationTimePoint(), trafo.getProcessStartTimePoint(), trafo.getProcessEndTimePoint(), null);
  
    
    assertTDEquals(m(20)+cmin(1)+stmin(1), // e(0->2) + d(2) + e(2->4) + d(4) + d(4->1) = c_i + 5min + 10min + 5min + st_i = 20min + c_i + st_i 
                   m(65), 
                   timeModel.getProcessDuration(), 1);
    assertTDEquals(sdmin(1), sdmax(1), timeModel.getActivityDuration(0), 1);
    assertTDEquals(sdmin(1), sdmax(1), timeModel.getActivityDuration(1), 1);
    
    assertTDEquals(m(5), m(15), timeModel.getActivityDuration(2), 1);
    assertTDEquals(m(5), m(30), timeModel.getActivityDuration(3), 1);
    assertTDEquals(m(5), m(15), timeModel.getActivityDuration(4), 1);
    
  
    assertTPEquals(0, 0, timeModel.getActivityStartTimeframe(0), 1);
    assertTPEquals(sdmin(1), 
                   sdmax(1), 
                   timeModel.getActivityCompletionTimeframe(0), 1);
    
    assertTPEquals(sdmin(1)+cmin(1), // d(0) + e(0->2) = sd_i + c_i
                   m(45)+sdmax(1)-stmin(1), // d(0) + e(0->1) - e(4->1) - d(4) - e(3->4) - d(4) - e(2->3) - d(2) 
                   timeModel.getActivityStartTimeframe(2), 1);
    assertTPEquals(m(5)+sdmin(1)+cmin(1),  // d(0) + e(0->2) + d(2) 
                   m(50)+sdmax(1)-stmin(1), // d(0) + e(0->1) - e(4->1) - d(4) - e(3->4) - d(4) - e(2->3) 
                   timeModel.getActivityCompletionTimeframe(2), 1);
  
    
    assertTPEquals(m(5)+sdmin(1)+cmin(2), // d(0) + e(0->2) + d(2) + e(2->3) 
                   m(55)+sdmax(1)-stmin(1)-cmin(1), // d(0) + e(0->1) - e(4->1) - d(4) - e(3->4) - d(4) 
                   timeModel.getActivityStartTimeframe(3), 1);
    assertTPEquals(m(10)+sdmin(1)+cmin(2), 
                   m(60)+sdmax(1)-stmin(1)-cmin(1), 
                   timeModel.getActivityCompletionTimeframe(3), 1);
    
    assertTPEquals(m(15)+sdmin(1)+cmin(1), 
                   m(60)+sdmax(1)-stmin(1), 
                   timeModel.getActivityStartTimeframe(4), 1);
    assertTPEquals(m(20)+sdmin(1)+cmin(1), 
                   m(65)+sdmax(1)-stmin(1),  
                   timeModel.getActivityCompletionTimeframe(4), 1);
    
    
    assertTPEquals(m(20)+sdmin(1)+cmin(1)+stmin(1), // d(0) + e(0->2) + d(2) + e(3->4) + d(4) + e(4->1)  
                   m(65)+sdmax(1), 
                   timeModel.getActivityStartTimeframe(1), 1);
    assertTPEquals(m(20)+sdmin(2)+cmin(1)+stmin(1), 
                   m(65)+sdmax(2),  
                   timeModel.getActivityCompletionTimeframe(1), 1); 
  }

  @Test
  public void testSimpleTimeTest() throws TCNInconsistencyException, InterruptedException
  {  
    basicTestWithTemplate(TestTemplates.simpleTimeTestNonContingentTestTemplate);
  }

  protected STN<LabeledTimePoint,E> basicTestWithTemplate(TemporalTemplate template)
      throws TCNInconsistencyException, InterruptedException
  {
    TCNTransformation<LabeledTimePoint, E> trafo = createTransformation(template);
    ChangeableSTN<LabeledTimePoint, E> stn = (ChangeableSTN<LabeledTimePoint, E>) trafo.transform();
  
    //TCNTestTools.printTCN(stn, getClass(), "Test");
    
    //stn.visit(visitor.getInstance("Before"));
    
    try
    {
      getSolver().solve(stn);
    }
    catch (final TCNInconsistencyException ex)
    {
      //stn.visit(visitor.getInstance("Failure"));
      throw ex;
    }
    
    //stn.visit(visitor.getInstance("After"));
  
  
    
    return stn;
  }

//  @Override
//  protected abstract TCNTransformation<LabeledNode, E> createTransformation(TemporalTemplate template);


}