package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertTDEquals;
import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertTPEquals;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.junit.Test;

import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.ChangeSet;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultChangeSet;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.Triangle;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableTCN;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.STN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.TCNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTransformation;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNU2CSTNProjection;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTemplateTransformation;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.DefaultCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;
import de.aristaflow.adept2.util.xml.VersionException;
import de.aristaflow.adept2.util.xml.XMLFormatException;

public abstract class AbstractSTNUDCSolverTest extends AbstractSolverTest<LabeledValueSet, LabeledEdge>
{

  public AbstractSTNUDCSolverTest()
  {
    super();
  }

  public abstract TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge> getSolver();

  public <T extends ChangeableTCN<LabeledTimePoint, LabeledEdge>> T solve(T tcn) throws TCNInconsistencyException, InterruptedException
  {
    TCNTestTools.assertConsistent(tcn);
    T solution = getSolver().solve(tcn);
    TCNTestTools.assertConsistent(tcn);
//    TCNTestTools.assertComplete(tcn);
    return solution;
  }

  @Test
  public void testSolveSimpleTimeTestNonContingent() throws TCNInconsistencyException, InterruptedException
  {
    try
    {
      basicTestWithTemplate(TestTemplates.simpleTimeTestNonContingentTestTemplate);
      fail("This template is not controllable");
    }
    catch (TCNInconsistencyException e)
    {
      // Expected...
    }
  }
  


  protected void basicTestWithTemplate(TemporalTemplate template) throws TCNInconsistencyException, InterruptedException
  {
    TCNTransformation<LabeledTimePoint, LabeledEdge> trafo = createTransformation(template);
    ChangeableTCN<LabeledTimePoint, LabeledEdge> timeModel = (ChangeableTCN<LabeledTimePoint, LabeledEdge>) trafo.transform();

    timeModel = solve(timeModel);

  }

  @Override
  protected CSTNUTemplateTransformation createTransformation(TemporalTemplate template)
  {
    return new CSTNUTemplateTransformation(template, 0);
  }

  @Test//(timeout = 60000)
  public void testSolveSimpleTimeTestContingent() throws TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.simpleTimeTestContingentTestTemplate;

    CSTNUTemplateTransformation trafo = createTransformation(template);
    ChangeableCSTNU tcn = (ChangeableCSTNU) trafo.transform();
    Collection<NodeMapping<LabeledTimePoint>> nodeMappings = trafo.getNodeMapping();

    int startNodeID = template.getStartNode().getID();
    NodeMapping<LabeledTimePoint> startNodeMapping = getNodeMapping(nodeMappings, startNodeID, new IterationChain(startNodeID));

    tcn
      .updateConstraint(new DefaultLabeledEdge(tcn.getTimeBaseEvent(), startNodeMapping.getStartNode(),
                                        new DefaultLabeledValueSet(new DefaultLabeledValue(startNodeMapping
                                          .getStartNode().getLabel(), DefaultTimeDistance.ONE_TIME_DISTANCE))));
    
//    TCNTestTools.printTCN(tcn, getClass(), "SimpleTimeTestContingent_Pre");

    try
    {
      tcn = solve(tcn);
    }
    catch (TCNInconsistencyException e)
    {
      throw e;
    }

    STN<LabeledTimePoint, LabeledEdge> stnProjection = new CSTNU2CSTNProjection(tcn, false, false);

    TemplateTimeModel timeModel = new TCNTemplateTimeModel<LabeledTimePoint, LabeledEdge>(template, stnProjection,
      trafo.getNodeMapping(), trafo.getProcessCreationTimePoint(), trafo.getProcessStartTimePoint(),
      trafo.getProcessEndTimePoint(), null);

//    TCNTestTools.printTCN(tcn, this.getClass(), "SimpleTimeTestContingent_Post");

    assertTDEquals(m(20) + cmin(1) + stmin(1),// m(60)+cmin(1)+stmin(1),
                   m(65), timeModel.getProcessDuration(), 1);
    assertTDEquals(sdmin(1), sdmax(1), timeModel.getActivityDuration(0), 1);
    assertTDEquals(sdmin(1), sdmax(1), timeModel.getActivityDuration(1), 1);

    assertTDEquals(m(5), m(15), timeModel.getActivityDuration(2), 1);
    assertTDEquals(m(5), m(30), timeModel.getActivityDuration(3), 1);
    assertTDEquals(m(5), m(15), timeModel.getActivityDuration(4), 1);

    assertTPEquals(0, 0, timeModel.getActivityStartTimeframe(0), 1);
    assertTPEquals(sdmin(1), sdmax(1), timeModel.getActivityCompletionTimeframe(0), 1);

    assertTPEquals(sdmin(1) + cmin(1), 
                    m(5) + sdmax(1) - stmin(1), timeModel.getActivityStartTimeframe(2), 1);
    assertTPEquals(m(5) + sdmin(1) + cmin(1), 
                    m(20) + sdmax(1) - stmin(1), timeModel.getActivityCompletionTimeframe(2), 1);

    assertTPEquals(m(5) + sdmin(1) + cmin(2),// m(15)+sdmin(1)+cmin(2),
                   m(20) + sdmax(1) - stmin(1) - cmin(1), timeModel.getActivityStartTimeframe(3), 1);
    assertTPEquals(m(10) + sdmin(1) + cmin(2),// m(20)+sdmin(1)+cmin(2),
                   m(50) + sdmax(1) - stmin(1) - cmin(1), timeModel.getActivityCompletionTimeframe(3), 1);

    assertTPEquals(m(15) + sdmin(1) + cmin(1),// m(45)+sdmin(1)+cmin(1),
                   m(50) + sdmax(1) - stmin(1), timeModel.getActivityStartTimeframe(4), 1);
    assertTPEquals(m(20) + sdmin(1) + cmin(1), // m(50)+sdmin(1)+cmin(1),
                   m(65) + sdmax(1) - stmin(1), timeModel.getActivityCompletionTimeframe(4), 1);

    assertTPEquals(m(20) + sdmin(1) + cmin(1) + stmin(1),// m(60)+sdmin(1)+cmin(1)+stmin(1),
                   m(65) + sdmax(1), timeModel.getActivityStartTimeframe(1), 1);
    assertTPEquals(m(20) + sdmin(2) + cmin(1) + stmin(1),// m(60)+sdmin(2)+cmin(1)+stmin(1),
                   m(65) + sdmax(2), timeModel.getActivityCompletionTimeframe(1), 1);
  }

  /**
   * Morris 2005, p. 5, Cutoff Algorithm
   * 
   * @throws TCNInconsistencyException
   * @throws InterruptedException 
   */
  @Test(expected = TCNInconsistencyException.class)
  public void testSolveCuttoff() throws TCNInconsistencyException, InterruptedException
  {
    Collection<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    Collection<LabeledEdge> edges = new ArrayList<LabeledEdge>();

    int nodeID = 0;
    IterationChain iterationChain = new IterationChain(nodeID);
    Label branches = Proposition.EMPTY_PROPOSITION;

    LabeledTimePoint A = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "A", branches);
    LabeledTimePoint B = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "B", branches);
    LabeledTimePoint C = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "C", branches);
    LabeledTimePoint D = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "D", branches);
    LabeledTimePoint E = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "E", branches);
    nodes.add(A);
    nodes.add(B);
    nodes.add(C);
    nodes.add(D);
    nodes.add(E);

    LabeledEdge AB = new DefaultLabeledEdge(A, B, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(-2))));
    LabeledEdge BC = new DefaultLabeledEdge(B, C,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, secToTD(1), C, secToTD(1))));

    LabeledEdge CB = new DefaultLabeledEdge(C, B,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, secToTD(-1), C, secToTD(-1))));

    LabeledEdge CD = new DefaultLabeledEdge(C, D, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(-1))));
    LabeledEdge DA = new DefaultLabeledEdge(D, A,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, secToTD(1), A, secToTD(1))));
    LabeledEdge AD = new DefaultLabeledEdge(A, D,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, secToTD(-1), A, secToTD(-1))));
    LabeledEdge AE = new DefaultLabeledEdge(A, E, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(-1))));
    edges.add(AB);
    edges.add(BC);
    edges.add(CB);
    edges.add(CD);
    edges.add(DA);
    edges.add(AD);
    edges.add(AE);

    ChangeableCSTNU cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());

    try
    {
      solve(cstnu);
      fail("Inconsistency Exception expected");
    }
    catch (TCNInconsistencyException e)
    {
      // TODO check why the cutoff bound is not reached
      // assertEquals("CutoffBound reached", e.getMessage());

      throw e;
    }

  }

  /**
   * Morris et al. 2005 p. 3 Example 1
   * 
   * @throws TCNInconsistencyException
   * @throws InterruptedException 
   * 
   */
  @Test(expected = TCNInconsistencyException.class)
  public void testSolveABCEx1() throws TCNInconsistencyException, InterruptedException
  {
    ChangeableCSTNU cstnu = getABCEx1();

//    TCNTestTools.export(cstnu, this.getClass(), "Example1");
    
    cstnu = solve(cstnu);

  }

  public static ChangeableCSTNU getABCEx1()
  {
    Collection<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    Collection<LabeledEdge> edges = new ArrayList<LabeledEdge>();

    int nodeID = 0;
    IterationChain iterationChain = new IterationChain(nodeID);
    Label branches = Proposition.EMPTY_PROPOSITION;

    LabeledTimePoint A = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "A", branches);
    LabeledTimePoint B = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "B", branches);
    LabeledTimePoint C = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "C", branches);
    nodes.add(A);
    nodes.add(B);
    nodes.add(C);

    long u = 1;
    long v = 1;

    long x = 1;
    long y = 2;

    LabeledEdge BC = new DefaultLabeledEdge(B, C, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(-u))));
    LabeledEdge CB = new DefaultLabeledEdge(C, B, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(v))));

    LabeledEdge AB = new DefaultLabeledEdge(A, B,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, secToTD(y), B, secToTD(x))));
    LabeledEdge BA = new DefaultLabeledEdge(B, A, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, secToTD(-x), B, secToTD(-y))));

    edges.add(BC);
    edges.add(CB);
    edges.add(AB);
    edges.add(BA);

    ChangeableCSTNU cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());
    return cstnu;
  }

  /**
   * Morris et al. 2005 p. 3 Example 2
   * 
   * @throws TCNInconsistencyException
   * @throws InterruptedException 
   */
  @Test
  public void testSolveABCEx2() throws TCNInconsistencyException, InterruptedException
  {
    Collection<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    Collection<LabeledEdge> edges = new ArrayList<LabeledEdge>();

    int nodeID = 0;
    IterationChain iterationChain = new IterationChain(nodeID);
    Label branches = Proposition.EMPTY_PROPOSITION;

    LabeledTimePoint A = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "A", branches);
    LabeledTimePoint B = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "B", branches);
    LabeledTimePoint C = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "C", branches);
    nodes.add(A);
    nodes.add(B);
    nodes.add(C);

    long u = 1;
    long v = 2;

    long x = 1;
    long y = 2;

    LabeledEdge BC = new DefaultLabeledEdge(B, C, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(-u))));
    LabeledEdge CB = new DefaultLabeledEdge(C, B, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(v))));

    LabeledEdge AB = new DefaultLabeledEdge(A, B,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, secToTD(y), B, secToTD(x))));
    LabeledEdge BA = new DefaultLabeledEdge(B, A, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, secToTD(-x), B, secToTD(-y))));

    edges.add(BC);
    edges.add(CB);
    edges.add(AB);
    edges.add(BA);

    ChangeableCSTNU cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());

//    TCNTestTools.export(cstnu, this.getClass(), "Example2");
    
    try
    {
      cstnu = solve(cstnu);
    }
    catch (TCNInconsistencyException e)
    {
      throw e;
    }

    // C must happen simultaneously with A
    LabeledEdge AC = cstnu.getEdge(A, C);
    assertTrue(1 == AC.getValue().getLabeledValues().size());
    LabeledValue l_AC = AC.getValue().getLabeledValues().iterator().next();
    assertTDEquals(secToTD(0), l_AC.getValue());
    assertFalse(l_AC instanceof ContingentLabeledValue);

    LabeledEdge CA = cstnu.getEdge(C, A);
    assertTrue(1 == CA.getValue().getLabeledValues().size());
    LabeledValue l_CA = CA.getValue().getLabeledValues().iterator().next();
    assertTDEquals(secToTD(0), l_CA.getValue());
    assertFalse(l_CA instanceof ContingentLabeledValue);

  }

  /**
   * Morris et al. 2005 p. 3 Example 3
   * 
   * @throws TCNInconsistencyException
   * @throws InterruptedException 
   */
  @Test
  public void testSolveABCEx3() throws TCNInconsistencyException, InterruptedException
  {
    Collection<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    Collection<LabeledEdge> edges = new ArrayList<LabeledEdge>();

    int nodeID = 0;
    IterationChain iterationChain = new IterationChain(nodeID);
    Label branches = Proposition.EMPTY_PROPOSITION;

    LabeledTimePoint A = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "A", branches);
    LabeledTimePoint B = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "B", branches);
    LabeledTimePoint C = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "C", branches);
    nodes.add(A);
    nodes.add(B);
    nodes.add(C);

    long u = -1;
    long v = 1;

    long x = 1;
    long y = 3;

    LabeledEdge BC = new DefaultLabeledEdge(B, C, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(-u))));
    LabeledEdge CB = new DefaultLabeledEdge(C, B, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(v))));

    LabeledEdge AB = new DefaultLabeledEdge(A, B,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, secToTD(y), B, secToTD(x))));
    LabeledEdge BA = new DefaultLabeledEdge(B, A, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, secToTD(-x), B, secToTD(-y))));

    edges.add(BC);
    edges.add(CB);
    edges.add(AB);
    edges.add(BA);

    ChangeableCSTNU cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());

//    TCNTestTools.export(cstnu, this.getClass(), "Example3");
    
    cstnu = solve(cstnu);

    LabeledEdge CA = cstnu.getEdge(C, A);
    assertTrue(1 == CA.getValue().getLabeledValues().size());
    LabeledValue l_CA = CA.getValue().getLabeledValues().iterator().next();
    assertTDEquals(secToTD(-2), ((ContingentLabeledValue) l_CA).getContingentValue());

    assertTrue(l_CA instanceof ContingentLabeledValue);
    assertEquals(B, ((ContingentLabeledValue) l_CA).getTrigger());

  }

  @Test(expected = TCNInconsistencyException.class)
  public void testCombi13() throws TCNInconsistencyException, InterruptedException
  {
    ChangeableCSTNU cstnu = getCombi13CSTNU();

    solve(cstnu);
  }

  public static ChangeableCSTNU getCombi13CSTNU()
  {
    IterationChain iteration = new IterationChain(0);

    NodeInstance nA = new NodeInstance(1, iteration, -1);
    NodeInstance nB = new NodeInstance(2, iteration, -1);
    NodeInstance nC = new NodeInstance(3, iteration, -1);
    NodeInstance nD = new NodeInstance(4, iteration, -1);
    NodeInstance nE = new NodeInstance(5, iteration, -1);
    NodeInstance nF = new NodeInstance(6, iteration, -1);
    NodeInstance nH = new NodeInstance(7, iteration, -1);
    NodeInstance nI = new NodeInstance(8, iteration, -1);


    long[] edgeCodes = new long[]{1l, 2l};
    DefaultObservationTimePoint A = new DefaultObservationTimePoint(nA, "A", 
                                        Proposition.EMPTY_PROPOSITION, edgeCodes);
    ObservationBinding a = A.getObservationForEdgeCode(1l);
    ObservationBinding na = A.getObservationForEdgeCode(2l);

    edgeCodes = new long[]{1l, 2l};
    DefaultObservationTimePoint B = new DefaultObservationTimePoint(nB, "B",
                                        Proposition.EMPTY_PROPOSITION, edgeCodes);
    ObservationBinding b = B.getObservationForEdgeCode(1l);
    ObservationBinding nb = B.getObservationForEdgeCode(2l);

    LabeledTimePoint C = new DefaultLabeledNode(nC, "C", new Proposition(a, b));
    LabeledTimePoint D = new DefaultLabeledNode(nD, "D", new Proposition(a, b));

    LabeledTimePoint E = new DefaultLabeledNode(nE, "E", new Proposition(na, b));
    LabeledTimePoint F = new DefaultLabeledNode(nF, "F", new Proposition(na, b));

     LabeledTimePoint H = new DefaultLabeledNode(nH, "H", new Proposition(nb));
     LabeledTimePoint I = new DefaultLabeledNode(nI, "I", new Proposition(nb));

    List<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    nodes.add(A);
    nodes.add(B);
    nodes.add(C);
    nodes.add(D);
    nodes.add(E);
    nodes.add(F);
    nodes.add(H);
    nodes.add(I);
   
    ArrayList<ChoiceVariable> observations = new ArrayList<ChoiceVariable>();
    observations.add(A.getObservation());
    observations.add(B.getObservation());

    LabeledEdge AB = new DefaultLabeledEdge(A, B, new DefaultLabeledValueSet(new DefaultLabeledValue(
      Proposition.EMPTY_PROPOSITION, new DefaultTimeDistance(-2, 0, 0, 0, 0))));
    LabeledEdge BA = new DefaultLabeledEdge(B, A, new DefaultLabeledValueSet(new DefaultLabeledValue(
      Proposition.EMPTY_PROPOSITION, new DefaultTimeDistance(11, 0, 0, 0, 0))));

    LabeledEdge AC = new DefaultLabeledEdge(A, C, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(a, b),
      new DefaultTimeDistance(5, 0, 0, 0, 0))));
    LabeledEdge CA = new DefaultLabeledEdge(C, A, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(a, b),
      new DefaultTimeDistance(-5, 0, 0, 0, 0))));

    LabeledEdge BD = new DefaultLabeledEdge(B, D, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(a, b),
      new DefaultTimeDistance(30, 0, 0, 0, 0))));
    LabeledEdge DB = new DefaultLabeledEdge(D, B, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(a, b),
      new DefaultTimeDistance(-16, 0, 0, 0, 0))));

    LabeledEdge BF = new DefaultLabeledEdge(B, F, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(na, b),
      new DefaultTimeDistance(22, 0, 0, 0, 0))));
    LabeledEdge FB = new DefaultLabeledEdge(F, B, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(na, b),
      new DefaultTimeDistance(-20, 0, 0, 0, 0))));

    LabeledEdge AE = new DefaultLabeledEdge(A, E, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(na, b),
      new DefaultTimeDistance(4, 0, 0, 0, 0))));
    LabeledEdge EA = new DefaultLabeledEdge(E, A, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(na, b),
      new DefaultTimeDistance(-2, 0, 0, 0, 0))));

    LabeledEdge AH = new DefaultLabeledEdge(A, H, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(nb), 
      new DefaultTimeDistance(10, 0, 0, 0, 0))));
    LabeledEdge HA = new DefaultLabeledEdge(H, A, new DefaultLabeledValueSet(new DefaultLabeledValue(new Proposition(nb), 
      new DefaultTimeDistance(-1, 0, 0, 0, 0))));

    LabeledEdge DC = new DefaultLabeledEdge(D, C, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(a, b),
      ConstraintCase.UPPERCASE, new DefaultTimeDistance(-10, 0, 0, 0, 0), D, new DefaultTimeDistance(-20, 0, 0, 0, 0))));
    LabeledEdge CD = new DefaultLabeledEdge(C, D, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(a, b),
      ConstraintCase.LOWERCASE, new DefaultTimeDistance(20, 0, 0, 0, 0), D, new DefaultTimeDistance(10, 0, 0, 0, 0))));

    LabeledEdge HI = new DefaultLabeledEdge(H, I, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(nb), 
      ConstraintCase.LOWERCASE, new DefaultTimeDistance(5, 0, 0, 0, 0), I, new DefaultTimeDistance(4, 0, 0, 0, 0))));
    LabeledEdge IH = new DefaultLabeledEdge(I, H, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(nb), 
      ConstraintCase.UPPERCASE, new DefaultTimeDistance(-4, 0, 0, 0, 0), I, new DefaultTimeDistance(-5, 0, 0, 0, 0))));

    LabeledEdge EF = new DefaultLabeledEdge(E, F, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(na, b),
      ConstraintCase.LOWERCASE, new DefaultTimeDistance(10, 0, 0, 0, 0), F, new DefaultTimeDistance(8, 0, 0, 0, 0))));
    LabeledEdge FE = new DefaultLabeledEdge(F, E, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(new Proposition(na, b),
      ConstraintCase.UPPERCASE, new DefaultTimeDistance(-8, 0, 0, 0, 0), F, new DefaultTimeDistance(-10, 0, 0, 0, 0))));

    List<LabeledEdge> edges = new ArrayList<LabeledEdge>();
    edges.add(AB);
    edges.add(BA);
    edges.add(AC);
    edges.add(CA);
    edges.add(BD);
    edges.add(DB);
    edges.add(BF);
    edges.add(FB);
    edges.add(AE);
    edges.add(EA);
     edges.add(AH);
     edges.add(HA);
    edges.add(DC);
    edges.add(CD);
     edges.add(HI);
     edges.add(IH);
    edges.add(EF);
    edges.add(FE);


    ChangeableCSTNU cstnu = new DefaultCSTNU(nodes, edges, observations);
    
    //TCNTestTools.export(cstnu, this.getClass(), "CombiEA13");

    LabeledEdge oB = new DefaultLabeledEdge(
                                     cstnu.getTimeBaseEvent(),
                                     B,
                                     new DefaultLabeledValueSet(
                                                                new DefaultLabeledValue(
                                                                                        Proposition.EMPTY_PROPOSITION,
                                                                                        new DefaultTimeDistance(0, 0,
                                                                                                                0, 0, 0))));
    cstnu.updateConstraint(oB);
    LabeledEdge Bo = new DefaultLabeledEdge(
                                     B,
                                     cstnu.getTimeBaseEvent(),
                                     new DefaultLabeledValueSet(
                                                                new DefaultLabeledValue(
                                                                                        Proposition.EMPTY_PROPOSITION,
                                                                                        new DefaultTimeDistance(0, 0,
                                                                                                                0, 0, 0))));
    cstnu.updateConstraint(Bo);
    return cstnu;
  }

  @Test
  public void testCoopIS13() throws VersionException, XMLFormatException, IOException, TCNInconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.coopIS13TestTemplate;

    //TCNTestTools.printTemplate(template);

    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(template, 0);
    ChangeableCSTNU cstnu = (ChangeableCSTNU) trafo.transform();

    //TCNTestTools.printTCN(cstnu, this.getClass(), "Pre");

    cstnu = solve(cstnu);

    //TCNTestTools.printTCN(cstnu, this.getClass(), "Post");

  }

  /**
   * Morris et al. 2005 p. 4
   * 
   * @throws TCNInconsistencyException
   * @throws InterruptedException 
   */
  @Test
  public void testSolveABCLabelRemoval() throws TCNInconsistencyException, InterruptedException
  {
    Collection<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    Collection<LabeledEdge> edges = new ArrayList<LabeledEdge>();

    int nodeID = 0;
    IterationChain iterationChain = new IterationChain(nodeID);
    Label branches = Proposition.EMPTY_PROPOSITION;

    LabeledTimePoint A = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "A", branches);
    LabeledTimePoint B = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "B", branches);
    LabeledTimePoint C = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "C", branches);
    nodes.add(A);
    nodes.add(B);
    nodes.add(C);

    long u = 2;
    long v = 5;

    long x = 5;
    long y = 7;

    LabeledEdge BC = new DefaultLabeledEdge(B, C, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(-u))));
    LabeledEdge CB = new DefaultLabeledEdge(C, B, new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(v))));

    LabeledEdge AB = new DefaultLabeledEdge(A, B,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, secToTD(y), B, secToTD(x))));
    LabeledEdge BA = new DefaultLabeledEdge(B, A, new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, secToTD(-x), B, secToTD(-y))));

    edges.add(BC);
    edges.add(CB);
    edges.add(AB);
    edges.add(BA);

    ChangeableCSTNU cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());

    // This setting requires one UpperCase Reduction, one LowerCase Reduction on
    // one LabelRemoval Reduction
    cstnu = solve(cstnu);

    // C must happen simultaneously with A
    LabeledEdge AC = cstnu.getEdge(A, C);
    assertNotNull(AC);
    assertTrue(1 == AC.getValue().getLabeledValues().size());
    LabeledValue l_AC = AC.getValue().getLabeledValues().iterator().next();
    assertFalse(l_AC instanceof ContingentLabeledValue);
    assertTDEquals(secToTD(x - u), l_AC.getValue());

    LabeledEdge CA = cstnu.getEdge(C, A);
    assertNotNull(CA);
    assertTrue(1 == CA.getValue().getLabeledValues().size());
    LabeledValue l_CA = CA.getValue().getLabeledValues().iterator().next();
    assertFalse(l_CA instanceof ContingentLabeledValue);
    assertTDEquals(secToTD(-(y - v)), l_CA.getValue());
  }



  @Test
  public void testNoDistance() throws TCNInconsistencyException, InterruptedException
  {
    IterationChain iteration = new IterationChain(0);

    NodeInstance nA = new NodeInstance(1, iteration, -1);
    NodeInstance nB = new NodeInstance(2, iteration, -1);
    NodeInstance nC = new NodeInstance(3, iteration, -1);

    LabeledTimePoint A = new DefaultLabeledNode(nA, "A", Proposition.EMPTY_PROPOSITION);
    LabeledTimePoint B = new DefaultLabeledNode(nB, "B", Proposition.EMPTY_PROPOSITION);
    LabeledTimePoint C = new DefaultLabeledNode(nC, "C", Proposition.EMPTY_PROPOSITION);

    List<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    nodes.add(A);
    nodes.add(B);
    nodes.add(C);

    LabeledEdge AB = new DefaultLabeledEdge(A, B,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(Proposition.EMPTY_PROPOSITION,
                                                                                      ConstraintCase.LOWERCASE,
                                                                                      new DefaultTimeDistance(11, 0, 0,
                                                                                                              0, 0), B,
                                                                                      new DefaultTimeDistance(2, 0, 0,
                                                                                                              0, 0))));
    LabeledEdge BA = new DefaultLabeledEdge(
                                     B,
                                     A,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(Proposition.EMPTY_PROPOSITION,
                                                                                      ConstraintCase.UPPERCASE,
                                                                                      new DefaultTimeDistance(-2, 0, 0,
                                                                                                              0, 0), B,
                                                                                      new DefaultTimeDistance(-11, 0,
                                                                                                              0, 0, 0))));

    LabeledEdge BC = new DefaultLabeledEdge(B, C,
                                     new DefaultLabeledValueSet(new DefaultLabeledValue(Proposition.EMPTY_PROPOSITION,
                                                                                        new DefaultTimeDistance(5 - 2,
                                                                                                                0, 0,
                                                                                                                0, 0))));
    LabeledEdge CB = new DefaultLabeledEdge(
                                     C,
                                     B,
                                     new DefaultLabeledValueSet(
                                                                new DefaultLabeledValue(
                                                                                        Proposition.EMPTY_PROPOSITION,
                                                                                        DefaultTimeDistance.ZERO_TIME_DISTANCE)));

    List<LabeledEdge> edges = new ArrayList<LabeledEdge>();
    edges.add(AB);
    edges.add(BA);
    edges.add(BC);
    edges.add(CB);

    ChangeableCSTNU cstnu = new DefaultCSTNU(nodes, edges, new ArrayList<ChoiceVariable>());

    LabeledEdge oA = new DefaultLabeledEdge(
                                     cstnu.getTimeBaseEvent(),
                                     A,
                                     new DefaultLabeledValueSet(
                                                                new DefaultLabeledValue(
                                                                                        Proposition.EMPTY_PROPOSITION,
                                                                                        DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE)));
    cstnu.updateConstraint(oA);
    LabeledEdge Ao = new DefaultLabeledEdge(
                                     A,
                                     cstnu.getTimeBaseEvent(),
                                     new DefaultLabeledValueSet(
                                                                new DefaultLabeledValue(
                                                                                        Proposition.EMPTY_PROPOSITION,
                                                                                        DefaultTimeDistance.ZERO_TIME_DISTANCE)));
    cstnu.updateConstraint(Ao);

    LabeledEdge oC = new DefaultLabeledEdge(
                                     cstnu.getTimeBaseEvent(),
                                     C,
                                     new DefaultLabeledValueSet(
                                                                new DefaultLabeledValue(
                                                                                        Proposition.EMPTY_PROPOSITION,
                                                                                        new DefaultTimeDistance(60, 0,
                                                                                                                0, 0, 0))));
    cstnu.updateConstraint(oC);

    solve(cstnu);
  }

  @Test
  public void testCompleteSolution() throws TCNInconsistencyException, InterruptedException
  {
    // Test whether a second run of the algorithm changes the result
    TemporalTemplate template = TestTemplates.simpleTimeTestContingentTestTemplate;

    CSTNUTemplateTransformation trafo = createTransformation(template);
    ChangeableCSTNU cstnu1 = (ChangeableCSTNU) trafo.transform();

    ChangeableCSTNU result1 = solve(cstnu1);
    ((DefaultCSTNU)result1).modified();

    ChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>> changeSet 
      = new DefaultChangeSet<LabeledEdge, LabeledValueSet, Triangle<LabeledTimePoint>>();
    ChangeableCSTNU result2 =  getSolver().solve((ChangeableCSTNU)result1.clone(), changeSet);

    assertFalse(changeSet.toString(), changeSet.iterator().hasNext());

    TCNTestTools.assertTCNEquals(result1, result2);
  }

  @Test
  public void testMixedCase() throws TCNInconsistencyException, InterruptedException
  {
    CSTNUGenerator generator = new CSTNUGenerator(tmo);
    IterationChain iteration = new IterationChain(1);
    LabeledTimePoint A = generator.createLabeledNode(new NodeInstance(1, iteration, -1), "A");
    LabeledTimePoint B = generator.createLabeledNode(new NodeInstance(2, iteration, -1), "B");
    LabeledTimePoint C = generator.createLabeledNode(new NodeInstance(3, iteration, -1), "C");

    generator.connectNodes(A, B, B, 2, 10);
    generator.connectNodes(A, C, 8, 20);
    generator.connectNodes(B, C, -5, 10);

    ChangeableCSTNU cstnu = generator.createCSTNU();

    solve(cstnu);
  }

}