package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.Collection;

import org.junit.AfterClass;
import org.junit.BeforeClass;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.TimeModelOperations;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;
import de.aristaflow.adept2.model.timemodel.tcn.ConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTransformation;

public abstract class AbstractSolverTest<V extends ConstraintValue, E extends TCNEdge<LabeledTimePoint, V>>
{

  protected static TimeModelOperations tmo;
  protected static long structureNodeMinDuration;
  protected static long structureNodeMaxDuration;
  protected static long controlFlowMinDistance;
  protected static long controlFlowMaxDistance;
  private static long structureNodeMinTimeLag;
  private static long structureNodeMaxTimeLag;

  @BeforeClass
  public static void setUpBeforeAbstractSolverTest() throws Exception
  {
    tmo = new DefaultTimeModelOperations();
    
    structureNodeMinDuration = TCNTransformation.DEFAULT_STRUCTURE_NODE_MIN_DURATION.toSeconds();
    structureNodeMaxDuration = TCNTransformation.DEFAULT_STRUCTURE_NODE_MAX_DURATION.toSeconds();
    structureNodeMinTimeLag = TCNTransformation.DEFAULT_STRUCTURE_NODE_MIN_TIME_LAG.toSeconds();
    structureNodeMaxTimeLag = TCNTransformation.DEFAULT_STRUCTURE_NODE_MAX_TIME_LAG.toSeconds();
    controlFlowMinDistance = TCNTransformation.DEFAULT_CONTROL_FLOW_MIN_TIME_LAG.toSeconds();
    controlFlowMaxDistance = TCNTransformation.DEFAULT_CONTROL_FLOW_MAX_TIME_LAG.toSeconds();
  }
  
  protected abstract TCNTransformation<LabeledTimePoint, E> createTransformation(TemporalTemplate template);
  
  @AfterClass
  public static void tearDownAfterAbstractSolverTest()
  {
    tmo = null;
  }

  public AbstractSolverTest()
  {
    super();
  }

  protected static TimeDistance secToTD(long seconds)
  {
    return new DefaultTimeDistance(seconds, 0, 0, 0, 0);
  }
  
  protected long sdmax(int times)
  {
    return structureNodeMaxDuration*times;
  }

  protected long sdmin(int times)
  {
    return structureNodeMinDuration*times;
  }
  
  protected long stmax(int times)
  {
    return structureNodeMaxTimeLag*times;
  }

  protected long stmin(int times)
  {
    return structureNodeMinTimeLag*times;
  }

  protected long cmin(int times)
  {
    return controlFlowMinDistance*times;
  }

  protected long cmax(int times)
  {
    return controlFlowMaxDistance*times;
  }

  protected int m(int i)
  {
    return i * 60;
  }

  protected <N extends TCNTimePoint> NodeMapping<N> getNodeMapping(Collection<NodeMapping<N>> nodeMappings, int nodeId,
      IterationChain iterationChain)
  {
    for (NodeMapping<N> nm : nodeMappings)
    {
      if (nm.getNodeInstance().getNodeID() == nodeId && nm.getNodeInstance().getIteration().equals(iterationChain))
        return nm;
    }
    return null;
  }
  
}