package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import org.apache.commons.configuration.BaseConfiguration;
import org.apache.commons.configuration.Configuration;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.Test;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNTransformation;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUAllMaxProjection;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTemplateTransformation;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;

public class AllMaxCSTNUTest
{

//  private static DotTCNVisitor<LabeledNode,LabeledEdge> visitor;


  
  @BeforeClass
  public static void setUpBeforeClass() throws Exception
  {
//    System.out.println("AllMaxCSTNUTest");
//    
//    visitor = new DotTCNVisitor<LabeledNode,LabeledEdge>("AllMaxCSTNUProjectionTest");
//    visitor.setLabelProvider(new CSTNUGraphFormatter());
//    visitor.setGraphFormatter(new CSTNUGraphFormatter());
  }

  @Test
  public void testTrafo()
  {   
    for (TemporalTemplate template : TestTemplates.getTemplates())
    {
      try
      {
        basicTestWithTemplate(template);
      }
      catch (AssertionError e)
      {
        AssertionError error = new AssertionError("AssertionError for "+template.getName());
        error.initCause(e);
        error.setStackTrace(e.getStackTrace());
        throw error;
      }
    }
    
    for (TemporalTemplate template : TestTemplates.getTemplates())
    {
      basicTestWithTemplate(template, 2, true);
    }

    
  }





  private TCN<LabeledTimePoint,LabeledEdge> basicTestWithTemplate(TemporalTemplate template)
  {   
    return basicTestWithTemplate(template, 0, true);
  }
  
  private TCN<LabeledTimePoint,LabeledEdge> basicTestWithTemplate(TemporalTemplate template, int lookAhead,
      boolean advancedActivityTransformation)
  {
    Configuration configuration = new BaseConfiguration();
    configuration.setProperty(CSTNUTemplateTransformation.CONF_ADVANCED_ACTIVITY_TRANSFORMATION, 
                              advancedActivityTransformation ? CSTNUTemplateTransformation.ACTIVITY_TRANSFORMATION_BASIC : CSTNUTemplateTransformation.ACTIVITY_TRANSFORMATION_COOPIS);
    TCNTransformation<LabeledTimePoint, LabeledEdge> trafo = new CSTNUTemplateTransformation(configuration, template, lookAhead);
    TCN<LabeledTimePoint,LabeledEdge> timeModel = new CSTNUAllMaxProjection((ChangeableCSTNU) trafo.transform());
    
//    timeModel.visit(visitor.getInstance(template.getName()));
    
    TCNTestTools.assertConsistent(timeModel);
    
    
    
    return timeModel;
  }
  
  @Test
  public void testXOREmptyBranch()
  {
    basicTestWithTemplate(TestTemplates.xorEmptyBranchTestTemplate);
        
//    TCNEvent baseEvent = timeModel.getTimeBaseEvent();
//    TCNEvent creationEvent = timeModel.getInstanceCreationEvent();
//    
//    assertNotNull(timeModel.getEdge(baseEvent, creationEvent));
//    
//    NodeMapping[] startNodes = timeModel.findNodes(xorEmptyBranchTestTemplate.getStartNode().getID());
//    assertEquals(1, startNodes.length);
//
//    NodeMapping startNode = startNodes[0];
//    assertEquals(1, startNode.startEvents.length);
//    assertEquals(0, startNode.intermediateEvents.length);
//    assertEquals(1, startNode.endEvents.length);
//    assertTrue(startNode.startEvents[0] == startNode.endEvents[0]);
//    
//    assertNotNull(timeModel.getEdge(creationEvent, startNode.startEvents[0]));
//    
//    NodeMapping[] endNodes = timeModel.findNodes(xorEmptyBranchTestTemplate.getEndNode().getID());
//    assertEquals(1, endNodes.length);
//
//    NodeMapping endNode = endNodes[0];
//    assertEquals(1, endNode.startEvents.length);
//    assertEquals(0, endNode.intermediateEvents.length);
//    assertEquals(1, endNode.endEvents.length);
//    assertTrue(endNode.startEvents[0] == endNode.endEvents[0]);
//    
//    
//    
//    NodeMapping[] xorSplitNodes = timeModel.findNodes(2); // XOR-Split
//    assertEquals(1, xorSplitNodes.length);
//    
//    NodeMapping xorSplit = xorSplitNodes[0];
//    assertEquals(1, xorSplit.startEvents.length);
//    assertEquals(EventType.XORSplit, xorSplit.startEvents[0].getEventType());
//    assertEquals(0, xorSplit.intermediateEvents.length);
//    assertEquals(3, xorSplit.endEvents.length);
//    for (TCNEvent t : xorSplit.endEvents)
//    {
//      assertEquals(EventType.XORDecision, t.getEventType());
//      assertNotNull(timeModel.getEdge(xorSplit.startEvents[0], t));
//    }
//    
//    assertNotNull(timeModel.getEdge(startNode.endEvents[0], xorSplit.startEvents[0]));
//    
//    NodeMapping[] xorJoinNodes = timeModel.findNodes(3); // XOR-Join
//    assertEquals(1, xorJoinNodes.length);
//    
//    NodeMapping xorJoin = xorJoinNodes[0];
//    assertEquals(1, xorJoin.startEvents.length);
//    assertEquals(EventType.XORJoin, xorJoin.startEvents[0].getEventType());
//    assertEquals(0, xorJoin.intermediateEvents.length);
//    assertEquals(1, xorJoin.endEvents.length);
//    assertTrue(xorJoin.startEvents[0] == xorJoin.endEvents[0]);
//    
//    int notNullCount = 0;
//    for (TCNEvent source : xorSplit.endEvents)
//    {
//      notNullCount += timeModel.getEdge(source, xorJoin.startEvents[0]) != null ? 1 : 0;
//    }
//    assertEquals(1, notNullCount);
//    
//    assertNotNull(timeModel.getEdge(xorJoin.endEvents[0], endNode.startEvents[0]));
//    
//    
//    NodeMapping[] activity1Nodes = timeModel.findNodes(4); // Activity 1
//    assertEquals(1, activity1Nodes.length);
//    
//    NodeMapping activity1 = activity1Nodes[0];
//    assertEquals(1, activity1.startEvents.length);
//    assertEquals(0, activity1.intermediateEvents.length);
//    assertEquals(1, activity1.endEvents.length);
//    assertNotNull(timeModel.getEdge(activity1.startEvents[0], activity1.endEvents[0]));
//    notNullCount = 0;
//    for (TCNEvent source : xorSplit.endEvents)
//    {
//      notNullCount += timeModel.getEdge(source,activity1.startEvents[0]) != null ? 1 : 0;
//    }
//    assertEquals(1, notNullCount);
//    assertNotNull(timeModel.getEdge(activity1.endEvents[0], xorJoin.startEvents[0]));
//    
//    NodeMapping[] activity2Nodes = timeModel.findNodes(5); // Activity 2
//    assertEquals(1, activity2Nodes.length);
//    
//    NodeMapping activity2 = activity2Nodes[0];
//    assertEquals(1, activity2.startEvents.length);
//    assertEquals(0, activity2.intermediateEvents.length);
//    assertEquals(1, activity2.endEvents.length);
//    assertNotNull(timeModel.getEdge(activity2.startEvents[0], activity2.endEvents[0]));
//    notNullCount = 0;
//    for (TCNEvent source : xorSplit.endEvents)
//    {
//      notNullCount += timeModel.getEdge(source,activity2.startEvents[0]) != null ? 1 : 0;
//    }
//    assertEquals(1, notNullCount);
//    assertNotNull(timeModel.getEdge(activity2.endEvents[0], xorJoin.startEvents[0]));
//    
    
    
    
  }
  

  
  @Test
  public void testSimpleTimeTestNonContingent()
  {      
    TCN<LabeledTimePoint, LabeledEdge> timeModel = basicTestWithTemplate(TestTemplates.simpleTimeTestNonContingentTestTemplate);
    
    for (LabeledTimePoint n1 : timeModel.getTimePoints())
    {
      for (LabeledTimePoint n2 : timeModel.getTimePoints())
      {
        LabeledEdge edge_12 = timeModel.getEdge(n1, n2);
        LabeledEdge edge_21 = timeModel.getEdge(n2, n1);
        
        if (edge_12 != null && edge_21 != null)
        {
          LabeledValue l_12 = edge_12.getValue().getLabeledValues().iterator().next();
          LabeledValue l_21 = edge_21.getValue().getLabeledValues().iterator().next();
          
          // In this setting we know that if one edge is contingent the other one is also
          if (edge_12.isContingentDuration())
          {
            ContingentLabeledValue cl_12 = (ContingentLabeledValue) l_12;
            ContingentLabeledValue cl_21 = (ContingentLabeledValue) l_21;
            Assert.assertFalse(String.format("%s <-> %s: %s : %s", n1, n2, cl_12.getContingentValue(), cl_21.getContingentValue()), add(cl_12.getContingentValue(), cl_21.getContingentValue()) > 0);
          }
          else
          {         
            // [x,y], x < y --> y+(-x) >= x+(-x) = 0
            Assert.assertFalse(String.format("%s <-> %s: %s : %s", n1, n2, l_12.getValue(), l_21.getValue()), add(l_12.getValue(), l_21.getValue()) < 0);
          }
          
        }
        
      }
    }
    
  }





  private double add(TimeDistance timeDistance, TimeDistance timeDistance2)
  {
    if (timeDistance == null)
      return Long.MAX_VALUE;
    if (timeDistance2 == null)
      return Long.MAX_VALUE;
    
    long s1 = timeDistance.toSeconds();
    long s2 = timeDistance2.toSeconds();
    
    return s1 + s2;
  }
  

}
