package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.util.ArrayList;
import java.util.List;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.TimeModelOperations;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCNEdge.FlowDirection;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.DefaultCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;

public class CSTNUGenerator
{
  private List<LabeledTimePoint> nodes;

  private List<LabeledEdge> edges;

//  private Collection<NodeMapping<LabeledTimePoint>> mapping;

  private int nodeID;

  private IterationChain iterationChain;

  private Label branches;

  private final TimeModelOperations tmo;

  public CSTNUGenerator(TimeModelOperations tmo)
  {
    this.tmo = tmo;
    nodes = new ArrayList<LabeledTimePoint>();
    edges = new ArrayList<LabeledEdge>();
//    mapping = new ArrayList<NodeMapping<LabeledTimePoint>>();

    nodeID = 0;

    iterationChain = new IterationChain(0);

    branches = Proposition.EMPTY_PROPOSITION;
  }

  public NodeMapping<LabeledTimePoint> createNodeMapping(String nodeName, long x_c, long y_C, long y_F)
  {
    NodeInstance nodeInstance = createNodeInstance();
    LabeledTimePoint A_S = createLabeledNode(nodeInstance, "S");
    LabeledTimePoint A_C = createLabeledNode(nodeInstance, "C");
    LabeledTimePoint A_E = createLabeledNode(nodeInstance, "E");
    
    NodeMapping<LabeledTimePoint> A = new NodeMapping<LabeledTimePoint>(
                                                              nodeName,
                                                              nodeInstance,
                                                              A_S,
                                                              new LabeledTimePoint[] { A_C },
                                                              A_E);
//    mapping.add(A);
    
    
    
    connectNodes(A_S, A_C, A_C, x_c, y_C);
    connectNodes(A_C, A_E, 0, y_F - x_c);
    connectNodes(A_S, A_E, 0, y_F);
    
    return A;
  }
  
  

  public NodeInstance createNodeInstance()
  {
    NodeInstance nodeInstance = new NodeInstance(nodeID++, iterationChain, -1);
    return nodeInstance;
  }
  


  public LabeledTimePoint createLabeledNode(
      NodeInstance nodeInstance, String eventID)
  {
    return createLabeledNode(nodeInstance, eventID, this.branches);
  }

  public LabeledTimePoint createLabeledNode(NodeInstance nodeInstance, String eventID, Label branches)
  {
    LabeledTimePoint A_S = new DefaultLabeledNode(nodeInstance, eventID, branches);
    nodes.add(A_S);
    return A_S;
  }
  
  public DefaultObservationTimePoint createObservationNode(NodeInstance nodeInstance, String eventID, int options)
  {
    return createObservationNode(nodeInstance, eventID, options, this.branches);
  }

  public DefaultObservationTimePoint createObservationNode(NodeInstance nodeInstance, String eventID, int options, Label branches)
  {
    long[] edgeCodes = new long[options];
    for (int i = 0; i < options; i++)
    {
      edgeCodes[i] = i+1;
    }
    
    DefaultObservationTimePoint P = new DefaultObservationTimePoint(nodeInstance, eventID, branches, edgeCodes);
    nodes.add(P);
    return P;
  }
  
  public void connectNodes(NodeMapping<LabeledTimePoint> from,
      NodeMapping<LabeledTimePoint> to, long min, long max)
  {
    connectNodes(from.getEndNode(), to.getStartNode(), min, max);
  }
  
  public void connectNodes(LabeledTimePoint fromNode, LabeledTimePoint toNode,
      long min, long max)
  {
    connectNodes(fromNode, toNode, min, max, branches);
  }

  public void connectNodes(LabeledTimePoint fromNode, LabeledTimePoint toNode, long min, long max, Label branches)
  {
    LabeledEdge AeBs  = new DefaultLabeledEdge(
                                        toNode,
                                        fromNode,
                                        new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(-min))), true, false, FlowDirection.BACKWARD);
    
    
    LabeledEdge BsAe  = new DefaultLabeledEdge(
                                        fromNode,
                                        toNode,
                                        new DefaultLabeledValueSet(new DefaultLabeledValue(branches, secToTD(max))), true, false, FlowDirection.FORWARD);
    
    edges.add(AeBs);
    edges.add(BsAe);
  }

  public void connectNodes(LabeledTimePoint fromNode, LabeledTimePoint toNode, LabeledTimePoint trigger,
      long min, long max)
  {
    connectNodes(fromNode, toNode, trigger, min, max, branches);
  }

  public void connectNodes(LabeledTimePoint fromNode, LabeledTimePoint toNode, LabeledTimePoint trigger, long min,
      long max, Label branches)
  {
    LabeledEdge AeBs = new DefaultLabeledEdge(fromNode, toNode, 
      new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, secToTD(max), 
        trigger, secToTD(min))), 
      true, false, FlowDirection.FORWARD);

    LabeledEdge BsAe = new DefaultLabeledEdge(toNode, fromNode, 
      new DefaultLabeledValueSet(new DefaultContingentLabeledValue( branches, ConstraintCase.UPPERCASE, 
        secToTD(-min), trigger, secToTD(-max))), 
      true, false, FlowDirection.BACKWARD);

    edges.add(AeBs);
    edges.add(BsAe);
  }

  public ChangeableCSTNU createCSTNU()
  {
    ChangeableCSTNU cstnu = new DefaultCSTNU(nodes, edges,
                            new ArrayList<ChoiceVariable>());
    
    edges.clear();
    connectNodes(cstnu.getTimeBaseEvent(), this.nodes.get(0), 1, 1);
    nodes.clear();
//    mapping.clear();
    
    updateCSTNU(cstnu);
    
    return cstnu;
  }

  @SuppressWarnings("unchecked")
  public void updateCSTNU(ChangeableCSTNU cstnu)
  {
    
//    for (NodeMapping<LabeledTimePoint> n : mapping)
//    {
//      ((AbstractTCN<LabeledTimePoint, LabeledValueSet, LabeledEdge>)cstnu).insertNodeMapping(n);
//    }
    
    for (LabeledTimePoint n : nodes)
    {
      ((AbstractTCN<LabeledTimePoint, LabeledValueSet, LabeledEdge>)cstnu).insertTimePoint(n);
    }
    
    for (LabeledEdge edge : edges)
    {
      cstnu.updateContingentConstraint(edge);
    }
    
    nodes.clear();
    edges.clear();
//    mapping.clear();
  }
  
  protected static TimeDistance secToTD(long seconds)
  {
    return new DefaultTimeDistance(seconds, 0, 0, 0, 0);
  }




}
