package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;

import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.InconsistencyException;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.TimeModelSolver;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.Label.ChoiceVariable;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.NodeMapping;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.AbstractCSTNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUDCSolver;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTimeModelFactory;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.DefaultCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;
import de.aristaflow.adept2.util.xml.VersionException;
import de.aristaflow.adept2.util.xml.XMLFormatException;

public class ChangeOperationsTest
{

  private static DefaultTimeModelOperations tmo;
  private CSTNUDCSolver solver;

  @BeforeClass
  public static void setUpBeforeClass() throws Exception
  {
    tmo = new DefaultTimeModelOperations();
  }

  

  @Before
  public void setUp()
  {
    solver = new CSTNUDCSolver();
  }
  
  public AbstractCSTNSolver getSolver()
  {
    return solver;
  }
  
  @Test
  public void testInsertNode() throws TCNInconsistencyException, InterruptedException
  {
    CSTNUGenerator generator = new CSTNUGenerator(tmo);
    
    NodeMapping<LabeledTimePoint> A = generator.createNodeMapping("A", 2, 3, 7);
    
    NodeMapping<LabeledTimePoint> B = generator.createNodeMapping("B", 2, 3,  7);    
    
    long a = 5;
    long b = 41;

    generator.connectNodes(A, B, a, b);
    
    ChangeableCSTNU cstnu = generator.createCSTNU();
    
    //TCNTestTools.assertConsistent(cstnu);

    
    //TCNTestTools.printTCN(cstnu, this.getClass(), "Insert_Init");
    
    solver.solve(cstnu);
    
    //TCNTestTools.printTCN(cstnu, this.getClass(), "Insert_Minimal");
    
    //TCNTestTools.printCSV(cstnu);
    
    long x_c = 13;
    long y_C = 17;
    long y_F = 21;

    NodeMapping<LabeledTimePoint> N = generator.createNodeMapping("N", x_c, y_C, y_F);
    generator.connectNodes(A, N, a, b);
    generator.connectNodes(N, B, a, b);
    
    generator.updateCSTNU(cstnu);

    
    //TCNTestTools.printTCN(cstnu, this.getClass(), "Insert_Inserted");
    
    
    
//    ChangeSetConstraintDependencyGraph<LabeledEdge, LabeledValueSet> cdg = new ChangeSetConstraintDependencyGraph<LabeledEdge, LabeledValueSet>(cstnu);
    
//    solver.solve(cstnu, cdg);
    
        
//    TCNTestTools.printTCN(cstnu, this.getClass(), "Insert_Final");
//    
//    System.out.println("===========================");
//    
//    TCNTestTools.printCSV(cstnu);
//    
//    TCNTestTools.printCDG(cdg, this.getClass(), "CDG");

  }







  
  
  @Test
  public void testANDTimeConstraint() throws VersionException, XMLFormatException, IOException, InconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.andTimeConstraintTestTemplate;
    CSTNUTimeModelFactory factory = new CSTNUTimeModelFactory();
    TemplateTimeModel timeModel = factory.createTimeModelForTemplate(template);
    
    TimeModelSolver solver = factory.getTimeModelSolver(timeModel.getClass());
    
    solver.solve(timeModel);
    
    CSTNU cstnu = ((CSTNUTemplateTimeModel)timeModel).getBaseTimeModel();
    
    //TCNTestTools.printCSV(cstnu);
    
    //TCNTestTools.printTCN(cstnu, this.getClass(), "ANDTimeConstraint");
  }
  
  @Test
  public void testXORTimeConstraint() throws VersionException, XMLFormatException, IOException, InconsistencyException, InterruptedException
  {
    TemporalTemplate template = TestTemplates.xorTimeConstraintTestTemplate;
    CSTNUTimeModelFactory factory = new CSTNUTimeModelFactory();
    TemplateTimeModel timeModel = factory.createTimeModelForTemplate(template);
    
    TimeModelSolver solver = factory.getTimeModelSolver(timeModel.getClass());
    
    solver.solve(timeModel);
    
    CSTNU cstnu = ((CSTNUTemplateTimeModel)timeModel).getBaseTimeModel();
    
    //TCNTestTools.printCSV(cstnu);
    
    //TCNTestTools.printTCN(cstnu, this.getClass(), "XORTimeConstraint");
  }


  public void testActivitySequence()
  {
    Collection<LabeledTimePoint> nodes = new ArrayList<LabeledTimePoint>();
    Collection<LabeledEdge> edges = new ArrayList<LabeledEdge>();

    int nodeID = 0;
    IterationChain iterationChain = new IterationChain(nodeID);
    Label branches = Proposition.EMPTY_PROPOSITION;

    LabeledTimePoint A_E = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "A_E", branches);
    LabeledTimePoint N_S = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "N_S", branches);
    LabeledTimePoint N_C = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "N_C", branches);
    LabeledTimePoint N_E = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "N_E", branches);
    LabeledTimePoint B_S = new DefaultLabeledNode(new NodeInstance(nodeID++, iterationChain, -1), "B_S", branches);
    
    nodes.add(A_E);
    nodes.add(N_S);
    nodes.add(N_C);
    nodes.add(N_E);
    nodes.add(B_S);

    long a = 5;
    long b = 41;

    long x_c = 13;
    long y_c = 17;
    long y_f = 21;
    
    
    LabeledEdge NsNc = new DefaultLabeledEdge(
                                     N_S,
                                     N_C,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.LOWERCASE, tmo
                                     .secondsToTimeDistance(y_c), N_C, tmo
                                     .secondsToTimeDistance(x_c))));
    LabeledEdge NcNs = new DefaultLabeledEdge(
                                     N_C,
                                     N_S,
                                     new DefaultLabeledValueSet(new DefaultContingentLabeledValue(branches, ConstraintCase.UPPERCASE, tmo
                                     .secondsToTimeDistance(-x_c), N_C, tmo
                                     .secondsToTimeDistance(-y_c))));

    LabeledEdge NcNe = new DefaultLabeledEdge(
                                     N_C,
                                     N_E,
                                     new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo
                                     .secondsToTimeDistance(y_f-x_c))));
    
    LabeledEdge NeNc = new DefaultLabeledEdge(
                                       N_E,
                                       N_C,
                                       new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo
                                       .secondsToTimeDistance(0))));
    
    LabeledEdge NsNe = new DefaultLabeledEdge(
                                       N_S,
                                       N_E,
                                       new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo
                                       .secondsToTimeDistance(y_f))));
    
    edges.add(NsNc);
    edges.add(NcNs);
    edges.add(NcNe);
    edges.add(NeNc);
    edges.add(NsNe);
    
    LabeledEdge AeNs  = new DefaultLabeledEdge(
                                        A_E,
                                        N_S,
                                        new DefaultLabeledValueSet(new DefaultLabeledValue(branches, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE)));
    

    LabeledEdge NsAe  = new DefaultLabeledEdge(
                                        N_S,
                                        A_E,
                                        new DefaultLabeledValueSet(new DefaultLabeledValue(branches, DefaultTimeDistance.ZERO_TIME_DISTANCE)));
    
    edges.add(AeNs);
    edges.add(NsAe);
    
    LabeledEdge NeBs  = new DefaultLabeledEdge(
                                        N_E,
                                        B_S,
                                        new DefaultLabeledValueSet(new DefaultLabeledValue(branches, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE)));
    

    LabeledEdge BsNe  = new DefaultLabeledEdge(
                                        B_S,
                                        N_E,
                                        new DefaultLabeledValueSet(new DefaultLabeledValue(branches, DefaultTimeDistance.ZERO_TIME_DISTANCE)));
    
    edges.add(NeBs);
    edges.add(BsNe);
    
    LabeledEdge AeBs  = new DefaultLabeledEdge(
                                        A_E,
                                        B_S,
                                        new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo
                                        .secondsToTimeDistance(b))));
    

    LabeledEdge BsAe  = new DefaultLabeledEdge(
                                        B_S,
                                        A_E,
                                        new DefaultLabeledValueSet(new DefaultLabeledValue(branches, tmo
                                        .secondsToTimeDistance(-a))));
    
    edges.add(AeBs);
    edges.add(BsAe);
    

    
    CSTNU cstnu = new DefaultCSTNU(nodes, edges,
                            new ArrayList<ChoiceVariable>());
  }
}
