package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;

import java.util.HashMap;
import java.util.Map;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;

public class DefaultLabeledValueSetTest
{

  @BeforeClass
  public static void setUpBeforeClass() throws Exception
  {
    //
  }

  @AfterClass
  public static void tearDownAfterClass() throws Exception
  {
    //
  }

  @Before
  public void setUp() throws Exception
  {
    //
  }

  @After
  public void tearDown() throws Exception
  {
    //
  }

  @Test
  public void testAddOrReplaceIfNecessary()
  {
    LabeledValue labeledValue1;
    LabeledValue labeledValue2;
    LabeledValue expected;
    LabeledValue actual;
    
    IterationChain iteration = new IterationChain(0);
    NodeInstance nodeInstance = new NodeInstance(0, iteration, -1);
    Label label = Proposition.EMPTY_PROPOSITION;
    LabeledTimePoint trigger = new DefaultLabeledNode(nodeInstance, "B", label);
    
    
    //  <o, 0s, #, NaN>
    labeledValue1 = new DefaultLabeledValue(label, new DefaultTimeDistance(-1, 0, 0, 0, 0));
    
    //  <o, -infty, /B\, -2s>
    labeledValue2 = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, trigger, new DefaultTimeDistance(-2, 0, 0, 0, 0));
    
    //  <o, 0s, /B\, -2s>
    expected = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, new DefaultTimeDistance(-1, 0, 0, 0, 0), trigger, new DefaultTimeDistance(-2, 0, 0, 0, 0));
    
    Map<Label, LabeledValue> labels = new HashMap<Label, LabeledValue>();
    
    DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, labeledValue1);
    DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, labeledValue2);
    
    
    actual = labels.get(label);
    assertEquals(expected, actual);    
    
    labels.clear();
    
    DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, labeledValue2);
    DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, labeledValue1);
    
    
    actual = labels.get(label);
    assertEquals(expected, actual);    
    
    //##################################
    
    //  <o, -5s, #, NaN>
    labeledValue1 = new DefaultLabeledValue(label, new DefaultTimeDistance(-5, 0, 0, 0, 0));
    
    //  <o, -infty, /B\, -2s>
    labeledValue2 = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, trigger, new DefaultTimeDistance(-2, 0, 0, 0, 0));

    //  <o, -5s, #, NaN>
    expected = new DefaultLabeledValue(label, new DefaultTimeDistance(-5, 0, 0, 0, 0));
    
    labels.clear();
    
    DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, labeledValue1);
    DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, labeledValue2);
    
    
    actual = labels.get(label);
    assertEquals(expected, actual);    
    
    labels.clear();
    
    DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, labeledValue2);
    DefaultLabeledValueSet.addOrReplaceIfNecessary(labels, labeledValue1);
    
    
    actual = labels.get(label);
    assertEquals(expected, actual);   
  }
  
  @Test
  public void testMerge()
  {
    LabeledValue labeledValue1;
    LabeledValue labeledValue2;
    LabeledValue expected;
    LabeledValue actual;
    
    IterationChain iteration = new IterationChain(0);
    NodeInstance nodeInstance = new NodeInstance(0, iteration, -1);
    Label label = Proposition.EMPTY_PROPOSITION;
    LabeledTimePoint trigger = new DefaultLabeledNode(nodeInstance, "B", label);
    
    //  <o, -1s, /B\, -2s>
    labeledValue1 = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, new DefaultTimeDistance(-1, 0, 0, 0, 0), trigger, new DefaultTimeDistance(-2, 0, 0, 0, 0));
        
    //  <o, 0s, #, NaN>
    labeledValue2 = new DefaultLabeledValue(label, new DefaultTimeDistance(0, 0, 0, 0, 0));
    
    //  <o, -1s, /B\, -2s>
    expected = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, new DefaultTimeDistance(-1, 0, 0, 0, 0), trigger, new DefaultTimeDistance(-2, 0, 0, 0, 0));
    
    actual = DefaultLabeledValueSet.merge(labeledValue1, labeledValue2);
    assertEquals(expected, actual);
  
    actual = DefaultLabeledValueSet.merge(labeledValue2, labeledValue1);
    assertEquals(expected, actual);
    

    // <o, -1s, #, NaN>, <o, infty, /B\, -1s>
    labeledValue1  = new DefaultLabeledValue(label, new DefaultTimeDistance(-1, 0, 0, 0, 0));
    labeledValue2 = new DefaultContingentLabeledValue(label, ConstraintCase.UPPERCASE, 
                                          DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE, trigger, 
                                          new DefaultTimeDistance(-1, 0, 0, 0, 0));
    
    expected = new DefaultLabeledValue(label, new DefaultTimeDistance(-1, 0, 0, 0, 0));
    actual = DefaultLabeledValueSet.merge(labeledValue1, labeledValue2);
    assertEquals(expected, actual);
    
    
  }

  @Test
  public void testUpdateLabeledValue() throws Exception
  {
    IterationChain iteration = new IterationChain(0);

    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);

    
    DefaultObservationTimePoint o1_2 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    LabeledTimePoint trigger = new DefaultLabeledNode(n2, "N2", Proposition.EMPTY_PROPOSITION);
    
    ObservationBinding o1_12 = o1_2.getObservationForEdgeCode(1l);
    ObservationBinding o1_22 = o1_2.getObservationForEdgeCode(2l);
    
    ContingentLabeledValue value = new DefaultContingentLabeledValue(new Proposition(o1_12), ConstraintCase.UPPERCASE, new DefaultTimeDistance(10, 0, 0, 0, 0), trigger, new DefaultTimeDistance(5, 0, 0, 0, 0));
    
    LabeledValue actual;
    TimeDistance nonContValue;
    TimeDistance contValue;
    
    nonContValue = new DefaultTimeDistance(11, 0, 0, 0, 0);
    contValue = new DefaultTimeDistance(6, 0, 0, 0, 0);
    actual = DefaultLabeledValueSet.updateLabeledValue(value, nonContValue, contValue);
    
    assertTrue(actual instanceof ContingentLabeledValue);
    assertEquals(value.getLabel(), actual.getLabel());
    assertEquals(value.getCase(), ((ContingentLabeledValue)actual).getCase());
    assertEquals(nonContValue, actual.getValue());
    assertEquals(value.getTrigger(), ((ContingentLabeledValue)actual).getTrigger());
    assertEquals(contValue, ((ContingentLabeledValue)actual).getContingentValue());
    
    nonContValue = new DefaultTimeDistance(10, 0, 0, 0, 0);
    contValue = new DefaultTimeDistance(12, 0, 0, 0, 0);
    actual = DefaultLabeledValueSet.updateLabeledValue(value, nonContValue, contValue);
    
    assertFalse(actual instanceof ContingentLabeledValue);
    assertEquals(value.getLabel(), actual.getLabel());
    assertEquals(nonContValue, actual.getValue());
    
    
    nonContValue = new DefaultTimeDistance(12, 0, 0, 0, 0);
    contValue = new DefaultTimeDistance(12, 0, 0, 0, 0);
    actual = DefaultLabeledValueSet.updateLabeledValue(value, nonContValue, contValue);
    
    assertFalse(actual instanceof ContingentLabeledValue);
    assertEquals(value.getLabel(), actual.getLabel());
    assertEquals(nonContValue, actual.getValue());
    
    value = new DefaultContingentLabeledValue(new Proposition(o1_22), ConstraintCase.LOWERCASE, new DefaultTimeDistance(11, 0, 0, 0, 0), o1_2, new DefaultTimeDistance(7, 0, 0, 0, 0));
    
    nonContValue = new DefaultTimeDistance(11, 0, 0, 0, 0);
    contValue = new DefaultTimeDistance(6, 0, 0, 0, 0);
    actual = DefaultLabeledValueSet.updateLabeledValue(value, nonContValue, contValue);
    
    assertTrue(actual instanceof ContingentLabeledValue);
    assertEquals(value.getLabel(), actual.getLabel());
    assertEquals(value.getCase(), ((ContingentLabeledValue)actual).getCase());
    assertEquals(nonContValue, actual.getValue());
    assertEquals(value.getTrigger(), ((ContingentLabeledValue)actual).getTrigger());
    assertEquals(contValue, ((ContingentLabeledValue)actual).getContingentValue());
    
    nonContValue = new DefaultTimeDistance(10, 0, 0, 0, 0);
    contValue = new DefaultTimeDistance(12, 0, 0, 0, 0);
    actual = DefaultLabeledValueSet.updateLabeledValue(value, nonContValue, contValue);
    
    assertFalse(actual instanceof ContingentLabeledValue);
    assertEquals(value.getLabel(), actual.getLabel());
    assertEquals(nonContValue, actual.getValue());
    
    
    nonContValue = new DefaultTimeDistance(12, 0, 0, 0, 0);
    contValue = new DefaultTimeDistance(12, 0, 0, 0, 0);
    actual = DefaultLabeledValueSet.updateLabeledValue(value, nonContValue, contValue);
    
    assertFalse(actual instanceof ContingentLabeledValue);
    assertEquals(value.getLabel(), actual.getLabel());
    assertEquals(nonContValue, actual.getValue());
    
  }

  @Test
  public void testSatisfiable() throws Exception
  {
    IterationChain iteration = new IterationChain(0);
    TimeDistance timeDistance = DefaultTimeDistance.ONE_TIME_DISTANCE;
    
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint o1_2 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2_2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = o1_2.getObservationForEdgeCode(1l);
    ObservationBinding o1_22 = o1_2.getObservationForEdgeCode(2l);
    
    ObservationBinding o2_12 = o2_2.getObservationForEdgeCode(1l);
    ObservationBinding o2_22 = o2_2.getObservationForEdgeCode(2l);
    
    Proposition prop1_12 = new Proposition(o1_12);
    Proposition prop1_22 = new Proposition(o1_22);
    Proposition prop2_12 = new Proposition(o2_12);
    Proposition prop2_22 = new Proposition(o2_22);
    
    Proposition prop_1_12_2_12 = new Proposition(o1_12, o2_12);
    Proposition prop_1_22_2_12 = new Proposition(o1_22, o2_12);
    Proposition prop_1_12_2_22 = new Proposition(o1_12, o2_22);
    Proposition prop_1_22_2_22 = new Proposition(o1_22, o2_22);
    LabeledValueSet set;
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop1_12, timeDistance));
    assertTrue(set.satisfiable(Proposition.EMPTY_PROPOSITION));
    assertTrue(set.satisfiable(prop1_12));
    assertFalse(set.satisfiable(prop1_22));
    assertTrue(set.satisfiable(prop2_12));
    assertTrue(set.satisfiable(prop2_22));
    assertTrue(set.satisfiable(prop_1_12_2_12));
    assertFalse(set.satisfiable(prop_1_22_2_12));
    assertTrue(set.satisfiable(prop_1_12_2_22));
    assertFalse(set.satisfiable(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop1_22, timeDistance));
    assertTrue(set.satisfiable(Proposition.EMPTY_PROPOSITION));
    assertFalse(set.satisfiable(prop1_12));
    assertTrue(set.satisfiable(prop1_22));
    assertTrue(set.satisfiable(prop2_12));
    assertTrue(set.satisfiable(prop2_22));
    assertFalse(set.satisfiable(prop_1_12_2_12));
    assertTrue(set.satisfiable(prop_1_22_2_12));
    assertFalse(set.satisfiable(prop_1_12_2_22));
    assertTrue(set.satisfiable(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop2_12, timeDistance));
    assertTrue(set.satisfiable(Proposition.EMPTY_PROPOSITION));
    assertTrue(set.satisfiable(prop1_12));
    assertTrue(set.satisfiable(prop1_22));
    assertTrue(set.satisfiable(prop2_12));
    assertFalse(set.satisfiable(prop2_22));
    assertTrue(set.satisfiable(prop_1_12_2_12));
    assertTrue(set.satisfiable(prop_1_22_2_12));
    assertFalse(set.satisfiable(prop_1_12_2_22));
    assertFalse(set.satisfiable(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop2_22, timeDistance));
    assertTrue(set.satisfiable(Proposition.EMPTY_PROPOSITION));
    assertTrue(set.satisfiable(prop1_12));
    assertTrue(set.satisfiable(prop1_22));
    assertFalse(set.satisfiable(prop2_12));
    assertTrue(set.satisfiable(prop2_22));
    assertFalse(set.satisfiable(prop_1_12_2_12));
    assertFalse(set.satisfiable(prop_1_22_2_12));
    assertTrue(set.satisfiable(prop_1_12_2_22));
    assertTrue(set.satisfiable(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop_1_12_2_12, timeDistance));
    assertTrue(set.satisfiable(Proposition.EMPTY_PROPOSITION));
    assertTrue(set.satisfiable(prop1_12));
    assertFalse(set.satisfiable(prop1_22));
    assertTrue(set.satisfiable(prop2_12));
    assertFalse(set.satisfiable(prop2_22));
    assertTrue(set.satisfiable(prop_1_12_2_12));
    assertFalse(set.satisfiable(prop_1_22_2_12));
    assertFalse(set.satisfiable(prop_1_12_2_22));
    assertFalse(set.satisfiable(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop_1_22_2_12, timeDistance));
    assertTrue(set.satisfiable(Proposition.EMPTY_PROPOSITION));
    assertFalse(set.satisfiable(prop1_12));
    assertTrue(set.satisfiable(prop1_22));
    assertTrue(set.satisfiable(prop2_12));
    assertFalse(set.satisfiable(prop2_22));
    assertFalse(set.satisfiable(prop_1_12_2_12));
    assertTrue(set.satisfiable(prop_1_22_2_12));
    assertFalse(set.satisfiable(prop_1_12_2_22));
    assertFalse(set.satisfiable(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop_1_12_2_22, timeDistance));
    assertTrue(set.satisfiable(Proposition.EMPTY_PROPOSITION));
    assertTrue(set.satisfiable(prop1_12));
    assertFalse(set.satisfiable(prop1_22));
    assertFalse(set.satisfiable(prop2_12));
    assertTrue(set.satisfiable(prop2_22));
    assertFalse(set.satisfiable(prop_1_12_2_12));
    assertFalse(set.satisfiable(prop_1_22_2_12));
    assertTrue(set.satisfiable(prop_1_12_2_22));
    assertFalse(set.satisfiable(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop_1_22_2_22, timeDistance));
    assertTrue(set.satisfiable(Proposition.EMPTY_PROPOSITION));
    assertFalse(set.satisfiable(prop1_12));
    assertTrue(set.satisfiable(prop1_22));
    assertFalse(set.satisfiable(prop2_12));
    assertTrue(set.satisfiable(prop2_22));
    assertFalse(set.satisfiable(prop_1_12_2_12));
    assertFalse(set.satisfiable(prop_1_22_2_12));
    assertFalse(set.satisfiable(prop_1_12_2_22));
    assertTrue(set.satisfiable(prop_1_22_2_22));
  }

  @Test
  public void testSubsumesLabel() throws Exception
  {
    IterationChain iteration = new IterationChain(0);
    TimeDistance timeDistance = DefaultTimeDistance.ONE_TIME_DISTANCE;
    
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint o1_2 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2_2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = o1_2.getObservationForEdgeCode(1l);
    ObservationBinding o1_22 = o1_2.getObservationForEdgeCode(2l);
    
    ObservationBinding o2_12 = o2_2.getObservationForEdgeCode(1l);
    ObservationBinding o2_22 = o2_2.getObservationForEdgeCode(2l);
    
    Proposition prop1_12 = new Proposition(o1_12);
    Proposition prop1_22 = new Proposition(o1_22);
    Proposition prop2_12 = new Proposition(o2_12);
    Proposition prop2_22 = new Proposition(o2_22);
    
    Proposition prop_1_12_2_12 = new Proposition(o1_12, o2_12);
    Proposition prop_1_22_2_12 = new Proposition(o1_22, o2_12);
    Proposition prop_1_12_2_22 = new Proposition(o1_12, o2_22);
    Proposition prop_1_22_2_22 = new Proposition(o1_22, o2_22);
    LabeledValueSet set;
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop1_12, timeDistance));
    assertTrue(set.subsumes(Proposition.EMPTY_PROPOSITION));
    assertTrue(set.subsumes(prop1_12));
    assertFalse(set.subsumes(prop1_22));
    assertFalse(set.subsumes(prop2_12));
    assertFalse(set.subsumes(prop2_22));
    assertFalse(set.subsumes(prop_1_12_2_12));
    assertFalse(set.subsumes(prop_1_22_2_12));
    assertFalse(set.subsumes(prop_1_12_2_22));
    assertFalse(set.subsumes(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop1_22, timeDistance));
    assertTrue(set.subsumes(Proposition.EMPTY_PROPOSITION));
    assertFalse(set.subsumes(prop1_12));
    assertTrue(set.subsumes(prop1_22));
    assertFalse(set.subsumes(prop2_12));
    assertFalse(set.subsumes(prop2_22));
    assertFalse(set.subsumes(prop_1_12_2_12));
    assertFalse(set.subsumes(prop_1_22_2_12));
    assertFalse(set.subsumes(prop_1_12_2_22));
    assertFalse(set.subsumes(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop2_12, timeDistance));
    assertTrue(set.subsumes(Proposition.EMPTY_PROPOSITION));
    assertFalse(set.subsumes(prop1_12));
    assertFalse(set.subsumes(prop1_22));
    assertTrue(set.subsumes(prop2_12));
    assertFalse(set.subsumes(prop2_22));
    assertFalse(set.subsumes(prop_1_12_2_12));
    assertFalse(set.subsumes(prop_1_22_2_12));
    assertFalse(set.subsumes(prop_1_12_2_22));
    assertFalse(set.subsumes(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop2_22, timeDistance));
    assertTrue(set.subsumes(Proposition.EMPTY_PROPOSITION));
    assertFalse(set.subsumes(prop1_12));
    assertFalse(set.subsumes(prop1_22));
    assertFalse(set.subsumes(prop2_12));
    assertTrue(set.subsumes(prop2_22));
    assertFalse(set.subsumes(prop_1_12_2_12));
    assertFalse(set.subsumes(prop_1_22_2_12));
    assertFalse(set.subsumes(prop_1_12_2_22));
    assertFalse(set.subsumes(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop_1_12_2_12, timeDistance));
    assertTrue(set.subsumes(Proposition.EMPTY_PROPOSITION));
    assertTrue(set.subsumes(prop1_12));
    assertFalse(set.subsumes(prop1_22));
    assertTrue(set.subsumes(prop2_12));
    assertFalse(set.subsumes(prop2_22));
    assertTrue(set.subsumes(prop_1_12_2_12));
    assertFalse(set.subsumes(prop_1_22_2_12));
    assertFalse(set.subsumes(prop_1_12_2_22));
    assertFalse(set.subsumes(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop_1_22_2_12, timeDistance));
    assertTrue(set.subsumes(Proposition.EMPTY_PROPOSITION));
    assertFalse(set.subsumes(prop1_12));
    assertTrue(set.subsumes(prop1_22));
    assertTrue(set.subsumes(prop2_12));
    assertFalse(set.subsumes(prop2_22));
    assertFalse(set.subsumes(prop_1_12_2_12));
    assertTrue(set.subsumes(prop_1_22_2_12));
    assertFalse(set.subsumes(prop_1_12_2_22));
    assertFalse(set.subsumes(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop_1_12_2_22, timeDistance));
    assertTrue(set.subsumes(Proposition.EMPTY_PROPOSITION));
    assertTrue(set.subsumes(prop1_12));
    assertFalse(set.subsumes(prop1_22));
    assertFalse(set.subsumes(prop2_12));
    assertTrue(set.subsumes(prop2_22));
    assertFalse(set.subsumes(prop_1_12_2_12));
    assertFalse(set.subsumes(prop_1_22_2_12));
    assertTrue(set.subsumes(prop_1_12_2_22));
    assertFalse(set.subsumes(prop_1_22_2_22));
    
    set = new DefaultLabeledValueSet(new DefaultLabeledValue(prop_1_22_2_22, timeDistance));
    assertTrue(set.subsumes(Proposition.EMPTY_PROPOSITION));
    assertFalse(set.subsumes(prop1_12));
    assertTrue(set.subsumes(prop1_22));
    assertFalse(set.subsumes(prop2_12));
    assertTrue(set.subsumes(prop2_22));
    assertFalse(set.subsumes(prop_1_12_2_12));
    assertFalse(set.subsumes(prop_1_22_2_12));
    assertFalse(set.subsumes(prop_1_12_2_22));
    assertTrue(set.subsumes(prop_1_22_2_22));
  }

}
