package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import static org.junit.Assert.*;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

import org.junit.Test;

import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.ContingentLabeledValue.ConstraintCase;

public class DefaultLabeledValueTest
{

  @Test
  public void testDominates() throws Exception
  {
    
    IterationChain iterationChain = new IterationChain(0);
    NodeInstance n1 = new NodeInstance(1, iterationChain, 0);
    NodeInstance n2 = new NodeInstance(2, iterationChain, 0);
    NodeInstance n3 = new NodeInstance(3, iterationChain, 0);
    DefaultObservationTimePoint observationNode1 = new DefaultObservationTimePoint(n1, "A", Proposition.EMPTY_PROPOSITION, new long[]{0l, 1l});
    DefaultObservationTimePoint observationNode2 = new DefaultObservationTimePoint(n2, "B", Proposition.EMPTY_PROPOSITION, new long[]{0l, 1l});
    LabeledTimePoint trigger = new DefaultLabeledNode(n3, "C", Proposition.EMPTY_PROPOSITION);
        
    ObservationBinding o1_12 = observationNode1.getObservationForEdgeCode(0l);
    ObservationBinding o1_22 = observationNode1.getObservationForEdgeCode(1l);
    
    ObservationBinding o2_12 = observationNode2.getObservationForEdgeCode(0l);
    ObservationBinding o2_22 = observationNode2.getObservationForEdgeCode(1l);
    
    DefaultLabeledValue l1;
    DefaultLabeledValue l2;
    
    l1 = new DefaultLabeledValue(new Proposition(), new DefaultTimeDistance(4, 0, 0, 0, 0));
    l2 = new DefaultLabeledValue(new Proposition(o1_12), new DefaultTimeDistance(4, 0, 0, 0, 0));
    
    assertTrue(l1.dominates(l2));
    assertFalse(l2.dominates(l1));
    
    l1 = new DefaultLabeledValue(new Proposition(o1_12), new DefaultTimeDistance(2, 0, 0, 0, 0));
    l2 = new DefaultLabeledValue(new Proposition(o1_12), new DefaultTimeDistance(4, 0, 0, 0, 0));
    
    assertTrue(l1.dominates(l2));
    assertFalse(l2.dominates(l1));
    
    l1 = new DefaultLabeledValue(new Proposition(o1_12, o2_12), new DefaultTimeDistance(5, 0, 0, 0, 0));
    l2 = new DefaultLabeledValue(new Proposition(o1_12), new DefaultTimeDistance(10, 0, 0, 0, 0));
    
    assertFalse(l1.dominates(l2));
    assertFalse(l2.dominates(l1));
    
    l1 = new DefaultLabeledValue(new Proposition(o1_12), new DefaultTimeDistance(5, 0, 0, 0, 0));
    l2 = new DefaultLabeledValue(new Proposition(o1_12, o2_12), new DefaultTimeDistance(10, 0, 0, 0, 0));
    
    assertTrue(l1.dominates(l2));
    assertFalse(l2.dominates(l1));
    
    l1 = new DefaultLabeledValue(new Proposition(o1_12, o2_12), new DefaultTimeDistance(5, 0, 0, 0, 0));
    l2 = new DefaultContingentLabeledValue(new Proposition(o1_12), ConstraintCase.UPPERCASE, new DefaultTimeDistance(10, 0, 0, 0, 0), trigger, new DefaultTimeDistance(8, 0, 0, 0, 0)  );
    
    assertFalse(l1.dominates(l2));
    assertFalse(l2.dominates(l1));
    
    l1 = new DefaultLabeledValue(new Proposition(o1_12, o2_12), new DefaultTimeDistance(5, 0, 0, 0, 0));
    l2 = new DefaultContingentLabeledValue(new Proposition(o1_12), ConstraintCase.UPPERCASE, new DefaultTimeDistance(10, 0, 0, 0, 0), trigger, new DefaultTimeDistance(4, 0, 0, 0, 0)  );
    
    assertFalse(l1.dominates(l2));
    assertFalse(l2.dominates(l1));
    
    l1 = new DefaultLabeledValue(new Proposition(o1_12), new DefaultTimeDistance(5, 0, 0, 0, 0));
    l2 = new DefaultContingentLabeledValue(new Proposition(o1_12, o2_12), ConstraintCase.UPPERCASE, new DefaultTimeDistance(10, 0, 0, 0, 0), trigger, new DefaultTimeDistance(8, 0, 0, 0, 0));
    
    assertTrue(l1.dominates(l2));
    assertFalse(l2.dominates(l1));
  }

}
