package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static org.mockito.Mockito.mock;

import org.junit.Test;

import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;

public class DefaultObservationTimePointTest
{

  @Test
  public void testObservationNodeNodeInstanceStringLabelLongArrayIntArray() throws Exception
  {
    NodeInstance nodeInstance = mock(NodeInstance.class);
    DefaultObservationTimePoint node = new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l}, new int[]{1, 0});
    
    assertNotNull(node.getObservation());
    assertEquals(2, node.getObservation().getOptions());
    assertEquals(node, node.getObservation().getDecisionTimePoint());
    assertNotNull(node.getChoiceVariableBindings());
    assertEquals(2, node.getChoiceVariableBindings().length);
    
    assertNull(node.getObservationForEdgeCode(0l));
    assertNotNull(node.getObservationForEdgeCode(1l));
    assertNotNull(node.getObservationForEdgeCode(2l));
    assertNull(node.getObservationForEdgeCode(3l));
    
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(1l).getVariable());
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(2l).getVariable());
    
    assertEquals(2, node.getEdgeCodeMapping().size());
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(1l)));
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(2l)));
    
    assertEquals(1, node.getObservationForEdgeCode(1l).getValue());
    assertEquals(0, node.getObservationForEdgeCode(2l).getValue());
    
    
    try
    {
      new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l}, new int[]{1, 0});
      fail();
    }
    catch (IllegalArgumentException e)
    {
      //
    }
    
    try
    {
      new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l}, new int[]{1, 0, 2});
      fail();
    }
    catch (IllegalArgumentException e)
    {
      //
    }
    
    try
    {
      new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l}, new int[]{1, 0, 3});
      fail();
    }
    catch (IllegalArgumentException e)
    {
      //
    }
  }

  @Test
  public void testObservationNodeNodeInstanceStringLabelLongArray() throws Exception
  {
    NodeInstance nodeInstance = mock(NodeInstance.class);
    DefaultObservationTimePoint node = new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    assertNotNull(node.getObservation());
    assertEquals(2, node.getObservation().getOptions());
    assertEquals(node, node.getObservation().getDecisionTimePoint());
    assertNotNull(node.getChoiceVariableBindings());
    assertEquals(2, node.getChoiceVariableBindings().length);
    
    assertNull(node.getObservationForEdgeCode(0l));
    assertNotNull(node.getObservationForEdgeCode(1l));
    assertNotNull(node.getObservationForEdgeCode(2l));
    assertNull(node.getObservationForEdgeCode(3l));
    
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(1l).getVariable());
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(2l).getVariable());
    
    assertEquals(2, node.getEdgeCodeMapping().size());
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(1l)));
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(2l)));
    
    assertEquals(0, node.getObservationForEdgeCode(1l).getValue());
    assertEquals(1, node.getObservationForEdgeCode(2l).getValue());
    
    node = new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l});
    
    assertNotNull(node.getObservation());
    assertEquals(3, node.getObservation().getOptions());
    assertEquals(node, node.getObservation().getDecisionTimePoint());
    assertNotNull(node.getChoiceVariableBindings());
    assertEquals(3, node.getChoiceVariableBindings().length);
    
    assertNull(node.getObservationForEdgeCode(0l));
    assertNotNull(node.getObservationForEdgeCode(1l));
    assertNotNull(node.getObservationForEdgeCode(2l));
    assertNotNull(node.getObservationForEdgeCode(3l));
    assertNull(node.getObservationForEdgeCode(4l));
    
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(1l).getVariable());
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(2l).getVariable());
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(3l).getVariable());
    
    assertEquals(3, node.getEdgeCodeMapping().size());
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(1l)));
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(2l)));
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(3l)));
    
    assertEquals(0, node.getObservationForEdgeCode(1l).getValue());
    assertEquals(1, node.getObservationForEdgeCode(2l).getValue());
    assertEquals(2, node.getObservationForEdgeCode(3l).getValue());
  }

  @Test
    public void testGetChoiceVariableBindings() throws Exception
    {
      NodeInstance nodeInstance = mock(NodeInstance.class);
      DefaultObservationTimePoint node = new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
      
      assertNotNull(node.getChoiceVariableBindings());
      assertEquals(2, node.getChoiceVariableBindings().length);
         
      for (ObservationBinding o : node.getChoiceVariableBindings())
      {
        assertEquals(node.getObservation(), o.getVariable());
        assertTrue(node.getEdgeCodeMapping().containsValue(o));
      }
      
      
      node = new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l});
      
      assertNotNull(node.getChoiceVariableBindings());
      assertEquals(3, node.getChoiceVariableBindings().length);
         
      for (ObservationBinding o : node.getChoiceVariableBindings())
      {
        assertEquals(node.getObservation(), o.getVariable());
        assertTrue(node.getEdgeCodeMapping().containsValue(o));
      }
    }

  @Test
  public void testGetEdgeCodeMapping() throws Exception
  {
    NodeInstance nodeInstance = mock(NodeInstance.class);
    DefaultObservationTimePoint node = new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    assertNotNull(node.getEdgeCodeMapping());
    assertEquals(2, node.getEdgeCodeMapping().size());
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(1l)));
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(2l)));
    
    
    node = new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l});
    
    assertNotNull(node.getEdgeCodeMapping());
    assertEquals(3, node.getEdgeCodeMapping().size());
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(1l)));
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(2l)));
    assertNotNull(node.getEdgeCodeMapping().get(Long.valueOf(3l)));
    
 
  }

  @Test
  public void testGetObservation() throws Exception
  {
    NodeInstance nodeInstance = mock(NodeInstance.class);
    DefaultObservationTimePoint node = new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    assertNotNull(node.getObservation());
    assertEquals(2, node.getObservation().getOptions());
    assertEquals(node, node.getObservation().getDecisionTimePoint());
    
    for (ObservationBinding o : node.getChoiceVariableBindings())
    {
      assertEquals(node.getObservation(), o.getVariable());
    }
    
    node = new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l});
    
    assertNotNull(node.getObservation());
    assertEquals(3, node.getObservation().getOptions());
    assertEquals(node, node.getObservation().getDecisionTimePoint());
    
    for (ObservationBinding o : node.getChoiceVariableBindings())
    {
      assertEquals(node.getObservation(), o.getVariable());
    }
  }

  @Test
  public void testGetObservationForEdgeCode() throws Exception
  {
    NodeInstance nodeInstance = mock(NodeInstance.class);
    DefaultObservationTimePoint node = new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    assertNull(node.getObservationForEdgeCode(0l));
    assertNotNull(node.getObservationForEdgeCode(1l));
    assertNotNull(node.getObservationForEdgeCode(2l));
    assertNull(node.getObservationForEdgeCode(3l));
    
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(1l).getVariable());
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(2l).getVariable());
    
//    assertEquals(0, node.getObservationForEdgeCode(1l).getValue());
//    assertEquals(1, node.getObservationForEdgeCode(2l).getValue());
    
    node = new DefaultObservationTimePoint(nodeInstance, "T", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l});
    
    
    assertNull(node.getObservationForEdgeCode(0l));
    assertNotNull(node.getObservationForEdgeCode(1l));
    assertNotNull(node.getObservationForEdgeCode(2l));
    assertNotNull(node.getObservationForEdgeCode(3l));
    assertNull(node.getObservationForEdgeCode(4l));
    
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(1l).getVariable());
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(2l).getVariable());
    assertEquals(node.getObservation(), node.getObservationForEdgeCode(3l).getVariable());
        
//    assertEquals(0, node.getObservationForEdgeCode(1l).getValue());
//    assertEquals(1, node.getObservationForEdgeCode(2l).getValue());
//    assertEquals(2, node.getObservationForEdgeCode(3l).getValue());
  }

}

