package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertArrayEqualsIgnoreOrdering;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;

public class PropositionTest
{

  @BeforeClass
  public static void setUpBeforeClass() throws Exception
  {
    //
  }

  @Before
  public void setUp() throws Exception
  {
    //
  }
  

  @Test
  public void testProposition() throws Exception
  {
    Observation observation = mock(Observation.class);
    when( observation.getOptions() ).thenReturn( 2 );
    
    ObservationBinding observationBinding1 = new ObservationBinding(observation, 0, 0);
    assertEquals(0, observationBinding1.getValue());
    verify(observation).getOptions();
  }


  @Test
  public void testSubsumes()
  {
    IterationChain iteration = new IterationChain(0);
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint o1 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = o1.getObservationForEdgeCode(1l);
    ObservationBinding o1_22 = o1.getObservationForEdgeCode(2l);
    ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
    ObservationBinding o2_22 = o2.getObservationForEdgeCode(2l);
    
    Label label1;
    Label label2;

    label1 = new Proposition(o1_12, o2_12);
    label2 = new Proposition(o2_12);
    
    assertTrue(label1.subsumes(label1));
    assertTrue(label2.subsumes(label2));
    
    assertTrue(label1.subsumes(label2));
    assertFalse(label2.subsumes(label1));
    
    assertTrue(label1.subsumes(Proposition.EMPTY_PROPOSITION));
    assertFalse(Proposition.EMPTY_PROPOSITION.subsumes(label1));
    assertTrue(label1.subsumes(label2));
    assertTrue(Proposition.EMPTY_PROPOSITION.subsumes(Proposition.EMPTY_PROPOSITION));
    
    
    label1 = new Proposition(o1_12);
    label2 = new Proposition(o1_22);
    
    assertFalse(label1.subsumes(label2));
    assertFalse(label2.subsumes(label1));

    label1 = new Proposition(o1_12, o2_12);
    label2 = new Proposition(o1_12, o2_22);
    
    assertFalse(label1.subsumes(label2));
    assertFalse(label2.subsumes(label1));
  }

  @Test
  public void testSatisfiableLabel()
  {
    IterationChain iteration = new IterationChain(0);
    IterationChain iteration2 = iteration.subIteration(1);
    
    NodeInstance n11 = new NodeInstance(1, iteration, -1);
    NodeInstance n12 = new NodeInstance(1, iteration2, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    Label l1 = Proposition.EMPTY_PROPOSITION;
    Label l2 = Proposition.EMPTY_PROPOSITION;
    assertTrue(l1.satisfiable(l2));
       
    l1 = Proposition.EMPTY_PROPOSITION;
    
    DefaultObservationTimePoint o11_2 = new DefaultObservationTimePoint(n11, "N1_1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o11_4 = new DefaultObservationTimePoint(n11, "N1_1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l, 4l});
    DefaultObservationTimePoint o12 = new DefaultObservationTimePoint(n12, "N1_2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o11_12 = o11_2.getObservationForEdgeCode(1l);
    ObservationBinding o11_22 = o11_2.getObservationForEdgeCode(2l);
    
    ObservationBinding o12_12 = o12.getObservationForEdgeCode(1l);
    
    ObservationBinding o11_24 = o11_4.getObservationForEdgeCode(2l);
    ObservationBinding o11_34 = o11_4.getObservationForEdgeCode(3l);
    
    ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
    ObservationBinding o2_22 = o2.getObservationForEdgeCode(2l);
    
    
    l2 = new Proposition(new ObservationBinding[]{o11_22}); // 0, true -> 2, 2
    assertTrue(l1.satisfiable(l2));
    assertTrue(l2.satisfiable(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o11_22});
    l2 = new Proposition(new ObservationBinding[]{o11_22});
    assertTrue(l1.satisfiable(l2));
    assertTrue(l2.satisfiable(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o11_22});
    l2 = new Proposition(new ObservationBinding[]{o11_12}); // 0,  false -> 1, 2
    assertFalse(l1.satisfiable(l2));
    assertFalse(l2.satisfiable(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o11_22});
    l2 = new Proposition(new ObservationBinding[]{o12_12});
    assertTrue(l1.satisfiable(l2));
    assertTrue(l2.satisfiable(l1));
    
    
    l1 = new Proposition(new ObservationBinding[]{o11_24});
    l2 = new Proposition(new ObservationBinding[]{o11_34}); // 1,  false -> 3, 4
    assertFalse(l1.satisfiable(l2));
    assertFalse(l2.satisfiable(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o11_22});
    l2 = new Proposition(new ObservationBinding[]{o2_22});
    assertTrue(l1.satisfiable(l2));
    assertTrue(l2.satisfiable(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o11_22});
    l2 = new Proposition(new ObservationBinding[]{o2_12});
    assertTrue(l1.satisfiable(l2));
    assertTrue(l2.satisfiable(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o11_22, o2_12});
    l2 = new Proposition(new ObservationBinding[]{o2_12});
    assertTrue(l1.satisfiable(l2));
    assertTrue(l2.satisfiable(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o11_22, o2_22});
    l2 = new Proposition(new ObservationBinding[]{o11_12});
    assertFalse(l1.satisfiable(l2));
    assertFalse(l2.satisfiable(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o11_22, o2_22});
    l2 = new Proposition(new ObservationBinding[]{o11_12, o2_22});
    assertFalse(l1.satisfiable(l2));
    assertFalse(l2.satisfiable(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o11_22, o2_22});
    l2 = new Proposition(new ObservationBinding[]{o11_22, o2_12});
    assertFalse(l1.satisfiable(l2));
    assertFalse(l2.satisfiable(l1));
  }

  @Test
  public void testMerge()
  {
    IterationChain iteration = new IterationChain(0);
    IterationChain iteration2 = iteration.subIteration(1);
    
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n12 = new NodeInstance(1, iteration2, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint o11_2 = new DefaultObservationTimePoint(n1, "N1_1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o11_4 = new DefaultObservationTimePoint(n1, "N1_1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l, 4l});
    DefaultObservationTimePoint o12 = new DefaultObservationTimePoint(n12, "N1_2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = o11_2.getObservationForEdgeCode(1l);
    ObservationBinding o1_22 = o11_2.getObservationForEdgeCode(2l);
    
    ObservationBinding o12_12 = o12.getObservationForEdgeCode(1l);
    
    ObservationBinding o1_24 = o11_4.getObservationForEdgeCode(2l);
    ObservationBinding o1_34 = o11_4.getObservationForEdgeCode(3l);
    
    ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
    ObservationBinding o2_22 = o2.getObservationForEdgeCode(2l);
    
    
    Label lr;
    
    Label l1 = new Proposition();
    Label l2 = new Proposition();
    assertEquals(l1, l1.merge(l2));
       
    l1 = new Proposition();    
    l2 = new Proposition(new ObservationBinding[]{o1_22});
    assertEquals(l2, l1.merge(l2));
    assertEquals(l2, l2.merge(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o1_22});
    l2 = new Proposition(new ObservationBinding[]{o1_22});
    assertEquals(l1, l1.merge(l2));
    assertEquals(l1, l2.merge(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o1_22});
    l2 = new Proposition(new ObservationBinding[]{o1_12});
    try
    {
      l1.merge(l2);
      fail("");
    }
    catch (IllegalArgumentException e)
    { //
    }
    try
    {
      l2.merge(l1);
      fail("");
    }
    catch (IllegalArgumentException e)
    { //
    }
    
    l1 = new Proposition(new ObservationBinding[]{o1_22});
    

    
    l2 = new Proposition(new ObservationBinding[]{o12_12});
    lr = new Proposition(new ObservationBinding[]{o1_22, o12_12});
    assertEquals(lr, l1.merge(l2));
    assertEquals(lr, l2.merge(l1));

    
    l1 = new Proposition(new ObservationBinding[]{o1_24});
    l2 = new Proposition(new ObservationBinding[]{o1_34});
    try
    {
      lr = new Proposition(new ObservationBinding[]{o1_24, o1_34});
      fail();
    }
    catch (AssertionError e2)
    {
      // 
    }
    try
    {
      l1.merge(l2);
      fail();
    }
    catch (IllegalArgumentException e1)
    {
      //
    }
    try
    {
      l2.merge(l1);
      fail();
    }
    catch (IllegalArgumentException e1)
    {
      // 
    }

    
    l1 = new Proposition(new ObservationBinding[]{o1_22});
    l2 = new Proposition(new ObservationBinding[]{o2_22});
    lr = new Proposition(new ObservationBinding[]{o1_22, o2_22});
    assertEquals(lr, l1.merge(l2));
    assertEquals(lr, l2.merge(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o1_22});
    l2 = new Proposition(new ObservationBinding[]{o2_12});
    lr = new Proposition(new ObservationBinding[]{o1_22, o2_12});
    assertEquals(lr, l1.merge(l2));
    assertEquals(lr, l2.merge(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o1_22, o2_12});
    l2 = new Proposition(new ObservationBinding[]{o2_12});
    lr = new Proposition(new ObservationBinding[]{o1_22, o2_12});
    assertEquals(lr, l1.merge(l2));
    assertEquals(lr, l2.merge(l1));
    
    l1 = new Proposition(new ObservationBinding[]{o1_22, o2_22});
    l2 = new Proposition(new ObservationBinding[]{o1_12});
    try
    {
      l1.merge(l2);
      fail();
    }
    catch (IllegalArgumentException e)
    {
      // 
    }
    try
    {
      l2.merge(l1);
      fail();
    }
    catch (IllegalArgumentException e)
    {//
    }
    
    l1 = new Proposition(new ObservationBinding[]{o1_22, o2_22});
    l2 = new Proposition(new ObservationBinding[]{o1_12, o2_22});
    try
    {
      l1.merge(l2);
      fail();
    }
    catch (IllegalArgumentException e)
    {
      // 
    }
    try
    {
      l2.merge(l1);
      fail();
    }
    catch (IllegalArgumentException e)
    {
      // 
    }
    
    l1 = new Proposition(new ObservationBinding[]{o1_22, o2_22});
    l2 = new Proposition(new ObservationBinding[]{o1_22, o2_12});
    try
    {
      l1.merge(l2);
      fail();
    }
    catch (IllegalArgumentException e)
    {//
    }
    try
    {
      l2.merge(l1);
      fail();
    }
    catch (IllegalArgumentException e)
    {
      // 
    }
  }

  @Test
  public void testNegate()
  {
    IterationChain iteration = new IterationChain(0);
    
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    NodeInstance n3 = new NodeInstance(3, iteration, -1);
    
    DefaultObservationTimePoint o11_2 = new DefaultObservationTimePoint(n1, "N1_1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o11_3 = new DefaultObservationTimePoint(n1, "N1_1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l});
    DefaultObservationTimePoint o11_4 = new DefaultObservationTimePoint(n1, "N1_1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l, 3l, 4l});
    DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o3 = new DefaultObservationTimePoint(n3, "N3", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = o11_2.getObservationForEdgeCode(1l);
    ObservationBinding o1_22 = o11_2.getObservationForEdgeCode(2l);
    
    ObservationBinding o1_13 = o11_3.getObservationForEdgeCode(1l);
    ObservationBinding o1_23 = o11_3.getObservationForEdgeCode(2l);
    ObservationBinding o1_33 = o11_3.getObservationForEdgeCode(3l);
    
    ObservationBinding o1_14 = o11_4.getObservationForEdgeCode(1l);
    ObservationBinding o1_24 = o11_4.getObservationForEdgeCode(2l);
    ObservationBinding o1_34 = o11_4.getObservationForEdgeCode(3l);
    ObservationBinding o1_44 = o11_4.getObservationForEdgeCode(4l);
    
    ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
    ObservationBinding o2_22 = o2.getObservationForEdgeCode(2l);
    
    ObservationBinding o3_12 = o3.getObservationForEdgeCode(1l);
    ObservationBinding o3_22 = o3.getObservationForEdgeCode(2l);
    
    Label label;
    Label[] actual;
    Label[] expected;
    
    label = new Proposition();
    try
    {
      actual = label.negate();
      fail("Empty proposition cannot be negated");
    }
    catch (IllegalStateException e)
    {
      //
    }
        
    label = new Proposition(o1_12);
    actual = label.negate();
    expected = new Label[]{new Proposition(o1_22)};
    assertArrayEqualsIgnoreOrdering(expected, actual);
    
    label = new Proposition(o1_22);
    actual = label.negate();
    expected = new Label[]{new Proposition(o1_12)};
    assertArrayEqualsIgnoreOrdering(expected, actual);
    

    
    label = new Proposition(o1_13);
    actual = label.negate();
    expected = new Label[]{
                           new Proposition(o1_23),
                           new Proposition(o1_33)
    };
    assertArrayEqualsIgnoreOrdering(expected, actual);
    
    label = new Proposition(o1_23);
    actual = label.negate();
    expected = new Label[]{
                           new Proposition(o1_13),
                           new Proposition(o1_33)
    };
    assertArrayEqualsIgnoreOrdering(expected, actual);
    
    label = new Proposition(o1_33);
    actual = label.negate();
    expected = new Label[]{
                           new Proposition(o1_13),
                           new Proposition(o1_23)
    };
    assertArrayEqualsIgnoreOrdering(expected, actual);
    

    
    label = new Proposition(o1_14);
    actual = label.negate();
    expected = new Label[]{
                           new Proposition(o1_24),
                           new Proposition(o1_34),
                           new Proposition(o1_44)
    };
    assertArrayEqualsIgnoreOrdering(expected, actual);
    
    label = new Proposition(o1_12, o2_12);
    actual = label.negate();
    
    expected = new Label[]{
                           new Proposition(o1_22, o2_12),
                           new Proposition(o1_12, o2_22),
                           new Proposition(o1_22, o2_22)
                          };
    assertArrayEqualsIgnoreOrdering(expected, actual);
    
    label = new Proposition(o1_13, o2_12);
    actual = label.negate();
    expected = new Label[]{
                           new Proposition(o1_23, o2_12),
                           new Proposition(o1_33, o2_12),
                           new Proposition(o1_13, o2_22),
                           new Proposition(o1_23, o2_22),
                           new Proposition(o1_33, o2_22)
                          };
    assertArrayEqualsIgnoreOrdering(expected, actual);

    
    label = new Proposition(o1_12, o2_12, o3_12);
    actual = label.negate();
    expected = new Label[]{
                           new Proposition(o1_22, o2_12, o3_12),
                           new Proposition(o1_12, o2_22, o3_12),
                           new Proposition(o1_22, o2_22, o3_12),
                           new Proposition(o1_22, o2_12, o3_22),
                           new Proposition(o1_12, o2_22, o3_22),
                           new Proposition(o1_12, o2_12, o3_22),
                           new Proposition(o1_22, o2_22, o3_22)
                          };
    assertArrayEqualsIgnoreOrdering(expected, actual);
  }

  @Test
  public void testGetObservations() throws Exception
  {
    IterationChain iteration = new IterationChain(0);
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint o1 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = o1.getObservationForEdgeCode(1l);
    ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
    
    Proposition label1 = new Proposition(o1_12, o2_12);
    Proposition label2 = new Proposition(o2_12);
    Proposition label3 = Proposition.EMPTY_PROPOSITION;
    
    
    assertNotNull(label1.getObservations());
    assertNotNull(label2.getObservations());
    assertNotNull(label3.getObservations());
    
    assertEquals(2, label1.getObservations().size());
    assertTrue(label1.getObservations().contains(o1_12));
    assertTrue(label1.getObservations().contains(o2_12));
    
    assertEquals(1, label2.getObservations().size());
    assertTrue(label2.getObservations().contains(o2_12));
    
    assertEquals(0, label3.getObservations().size());
    
  }

  @Test
  public void testEquals() throws Exception
  {
    IterationChain iteration = new IterationChain(0);
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint o1 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = o1.getObservationForEdgeCode(1l);
    ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
    
    Proposition label1 = new Proposition(o1_12, o2_12);
    Proposition label2 = new Proposition(o2_12);
    Proposition label3 = new Proposition();
    Proposition label4 = new Proposition(o2_12, o1_12);
    
    assertTrue(label1.equals(label1));
    assertTrue(label1.equals(label4));
    assertTrue(label4.equals(label1));
    assertFalse(label1.equals(label2));
    assertFalse(label1.equals(label3));
    
    assertTrue(label3.equals(Proposition.EMPTY_PROPOSITION));
  }

  @Test
  public void testRemoveVariable() throws Exception
  {
    IterationChain iteration = new IterationChain(0);
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint o1 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = o1.getObservationForEdgeCode(1l);
    ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
    
    Proposition label1 = new Proposition(o1_12, o2_12);
    Proposition label2 = new Proposition(o2_12);
    Proposition label3 = new Proposition();
    
    
    assertEquals(label2, label1.removeVariable(o1.getObservation()));
    assertEquals(label3, label2.removeVariable(o2.getObservation()));
    assertEquals(label3, label3.removeVariable(o1.getObservation()));
    assertEquals(label2, label2.removeVariable(o1.getObservation()));
  }

  @Test
  public void testRemoveLabel() throws Exception
  {
    IterationChain iteration = new IterationChain(0);
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint o1 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = o1.getObservationForEdgeCode(1l);
    ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
    
    Proposition label1 = new Proposition(o1_12, o2_12);
    Proposition label2 = new Proposition(o2_12);
    Proposition label3 = new Proposition();
    Proposition label4 = new Proposition(o1_12);
    
    assertEquals(label4, label1.removeLabel(label2));
    assertEquals(Proposition.EMPTY_PROPOSITION, label2.removeLabel(label1));
    assertEquals(label3, label3.removeLabel(label1));
  }

  @Test
  public void testIsEmptyLabel() throws Exception
  {
    IterationChain iteration = new IterationChain(0);
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint o1 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = o1.getObservationForEdgeCode(1l);
    ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
    
    Proposition label1 = new Proposition(o1_12, o2_12);
    Proposition label2 = new Proposition(o2_12);
    Proposition label3 = new Proposition();
    
    assertFalse(label1.isEmptyLabel());
    assertFalse(label2.isEmptyLabel());
    assertTrue(label3.isEmptyLabel());
    assertTrue(Proposition.EMPTY_PROPOSITION.isEmptyLabel());
  }

  @Test
  public void testContainsVariable() throws Exception
  {
    IterationChain iteration = new IterationChain(0);
    NodeInstance n1 = new NodeInstance(1, iteration, -1);
    NodeInstance n2 = new NodeInstance(2, iteration, -1);
    
    DefaultObservationTimePoint o1 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
    
    ObservationBinding o1_12 = o1.getObservationForEdgeCode(1l);
    ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
    
    Proposition label1 = new Proposition(o1_12, o2_12);
    Proposition label2 = new Proposition(o2_12);
    Proposition label3 = new Proposition();
    Proposition label4 = new Proposition(o1_12);
    
    assertTrue(label1.containsVariable(o1.getObservation()));
    assertTrue(label1.containsVariable(o2.getObservation()));
    assertFalse(label2.containsVariable(o1.getObservation()));
    assertTrue(label2.containsVariable(o2.getObservation()));
    assertFalse(label3.containsVariable(o1.getObservation()));
    assertFalse(label3.containsVariable(o2.getObservation()));
    assertTrue(label4.containsVariable(o1.getObservation()));
    assertFalse(label4.containsVariable(o2.getObservation()));
  }

  @Test
    public void testAddVariableBinding() throws Exception
    {
      IterationChain iteration = new IterationChain(0);
      NodeInstance n1 = new NodeInstance(1, iteration, -1);
      NodeInstance n2 = new NodeInstance(2, iteration, -1);
      
      DefaultObservationTimePoint o1 = new DefaultObservationTimePoint(n1, "N1", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
      DefaultObservationTimePoint o2 = new DefaultObservationTimePoint(n2, "N2", Proposition.EMPTY_PROPOSITION, new long[]{1l, 2l});
      
      ObservationBinding o1_12 = o1.getObservationForEdgeCode(1l);
      ObservationBinding o1_22 = o1.getObservationForEdgeCode(2l);
      ObservationBinding o2_12 = o2.getObservationForEdgeCode(1l);
      ObservationBinding o2_22 = o2.getObservationForEdgeCode(2l);
      
      Proposition label1 = new Proposition(o1_12, o2_12);
      Proposition label2 = new Proposition(o2_12);
      Proposition label3 = new Proposition();
      Proposition label4 = new Proposition(o1_12);
      
      assertEquals(label1, label1.addVariableBinding(o1_12));
      assertEquals(label1, label2.addVariableBinding(o1_12));
      assertEquals(label2, label3.addVariableBinding(o2_12));
      assertEquals(label4, Proposition.EMPTY_PROPOSITION.addVariableBinding(o1_12));
      assertEquals(label1, label4.addVariableBinding(o2_12));
      assertEquals(label1, ((Proposition)label3.addVariableBinding(o1_12)).addVariableBinding(o2_12));
      
      try
      {
        label1.addVariableBinding(o1_22);
        fail("Cannot append negation of already present ObservationBinding");
      }
      catch (IllegalArgumentException e)
      {
        // Expected
      }
      
      try
      {
        label1.addVariableBinding(o2_22);
        fail("Cannot append negation of already present ObservationBinding");
      }
      catch (IllegalArgumentException e)
      {
        // Expected
      }
      
      try
      {
        label2.addVariableBinding(o2_22);
        fail("Cannot append negation of already present ObservationBinding");
      }
      catch (IllegalArgumentException e)
      {
        // Expected
      }
      
    }




}
