package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import static org.junit.Assert.*;
import static org.hamcrest.Matchers.*;
import static de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools.assertTDEquals;

import java.io.InputStream;
import java.util.Collection;
import java.util.Set;

import org.hamcrest.Matcher;
import org.hamcrest.MatcherAssert;
import org.hamcrest.Matchers;
import org.hamcrest.core.CombinableMatcher;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Assume;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.GuardedInterval;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultGuardedInterval;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableCSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableSTN;
import de.aristaflow.adept2.model.timemodel.tcn.ChangeableSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.DotTCNVisitor;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.NodeInstance;
import de.aristaflow.adept2.model.timemodel.tcn.TCNInconsistencyException;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUDCSolver;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTemplateTransformation;
import de.aristaflow.adept2.model.timemodel.tcn.stnuImpl.STNUAllMaxProjection;
import de.aristaflow.adept2.model.timemodel.tcn.xml.TCNXMLImport;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;

public class TCNAlgorithmsTest
{

  @BeforeClass
  public static void setUpBeforeClass() throws Exception
  {
  }

  @AfterClass
  public static void tearDownAfterClass() throws Exception
  {
  }

  @Before
  public void setUp() throws Exception
  {
  }

  @After
  public void tearDown() throws Exception
  {
  }

  @Test
  @Ignore
  // Template does not fit with the test...
  public void testFindRigidComponents()
  {
    
    TemporalTemplate template = TestTemplates.simpleTimeTestNonContingentTestTemplate;
    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(template);
    CSTNU cstnu = trafo.transform();
    
//    TCNTestTools.printTCN(cstnu, getClass(), "FindRigidComponents");
    
    Collection<Set<LabeledTimePoint>> rigidComponents = TCNAlgorithms.findRigidComponents(cstnu);
    for (Set<LabeledTimePoint> rigidComponent : rigidComponents)
    {
      System.out.println(rigidComponent);
    }
    
    
    
    assertThat(rigidComponents, hasSize(2));
    Matcher<LabeledTimePoint> n2S = Matchers.<LabeledTimePoint> both(hasProperty("nodeInstance", hasProperty("nodeID", is(2)))).and(hasProperty("eventID", is("S")));
    Matcher<LabeledTimePoint> n2C = Matchers.<LabeledTimePoint> both(hasProperty("nodeInstance", hasProperty("nodeID", is(2)))).and(hasProperty("eventID", is("C")));
    Matcher<LabeledTimePoint> n4S = Matchers.<LabeledTimePoint> both(hasProperty("nodeInstance", hasProperty("nodeID", is(4)))).and(hasProperty("eventID", is("S")));
    Matcher<LabeledTimePoint> n4C = Matchers.<LabeledTimePoint> both(hasProperty("nodeInstance", hasProperty("nodeID", is(4)))).and(hasProperty("eventID", is("C")));
    
    assertThat(rigidComponents, Matchers.<Set<LabeledTimePoint>> hasItem(both(hasItem(n2S)).and(hasItem(n2C))));
    assertThat(rigidComponents, Matchers.<Set<LabeledTimePoint>> hasItem(both(hasItem(n4S)).and(hasItem(n4C))));
  }

  @Test
  public void testGetRestrictableDuration() throws Exception
  {
    
    CSTNUDCSolver solver = new CSTNUDCSolver();
    
//    CSTNUGenerator generator = new CSTNUGenerator(new DefaultTimeModelOperations());
//    NodeInstance nodeInstance = generator.createNodeInstance();
//    LabeledTimePoint n0 = generator.createLabeledNode(nodeInstance, "n0");
//    LabeledTimePoint nZ = generator.createLabeledNode(nodeInstance, "nZ");
//    
//    LabeledTimePoint S = generator.createLabeledNode(nodeInstance, "S");
//    LabeledTimePoint T = generator.createLabeledNode(nodeInstance, "T");
//    LabeledTimePoint U = generator.createLabeledNode(nodeInstance, "U");
//    generator.connectNodes(S, T, T, 5, 10);
//    generator.connectNodes(T, U, U, 5, 10);
//    generator.connectNodes(n0, S, 1, 10);
//    generator.connectNodes(U, nZ, 1, 10);
//    
//    LabeledTimePoint Q = generator.createLabeledNode(nodeInstance, "Q");
//    LabeledTimePoint R = generator.createLabeledNode(nodeInstance, "R");
//    generator.connectNodes(Q, R, R, 7, 9);
//    generator.connectNodes(n0, Q, 1, 10);
//    generator.connectNodes(R, nZ, 1, 10);
//    
//    
//    generator.connectNodes(n0, nZ, 1, 25);
//    
//    ChangeableCSTNU cstnu = generator.createCSTNU();
//    
////    cstnu.setProcessBounds(n0, n0, nZ);
//   
//    
//    solver.solve(cstnu);
//    
//    GuardedInterval<TimeDistance> duration = TCNAlgorithms.getRestrictableDuration(cstnu, n0, nZ);
//    assertEquals(new DefaultTimeDistance(13, 0, 0, 0, 0), duration.getGuardedMinimum());
//    assertEquals(new DefaultTimeDistance(22, 0, 0, 0, 0), duration.getGuardedMaximum());
//    assertEquals(new DefaultTimeDistance(25, 0, 0, 0, 0), duration.getFlexibleMaximum());
    
    TCNXMLImport importer = new TCNXMLImport();
    
    InputStream inputStream = this.getClass().getResourceAsStream("TCNAlgorithms_testGetRestrictableDuration1.xml");
    ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> tcn;
    try
    {
      tcn = (ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge>) importer.getTCNFromStream(inputStream);
    }
    finally
    {
      inputStream.close();
    }
    
    //solver.solve(tcn);
    
    
    LabeledTimePoint source = TCNAlgorithms.getSourceTimePoint(tcn);
    LabeledTimePoint destination = TCNAlgorithms.getTargetTimePoint(tcn);
    
    GuardedInterval<TimeDistance> interval = getRestrictableDurationBySTN((ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge>) tcn.clone(), source, destination);
    
    solver.solve(tcn);
    
    GuardedInterval<TimeDistance> restrictableDuration = TCNAlgorithms.getRestrictableDuration(tcn, source, destination);
    
    assertEquals(interval, restrictableDuration);
  }

  private GuardedInterval<TimeDistance> getRestrictableDurationBySTN(ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> tcn,
      LabeledTimePoint source, LabeledTimePoint destination) throws TCNInconsistencyException, InterruptedException
  {
    
    CSTNUDCSolver solver = new CSTNUDCSolver();
    solver.solve(tcn);
    
    TimeDistance flexibleMinimum;
    TimeDistance guardedMinimum;
    TimeDistance guardedMaximum;
    TimeDistance flexibleMaximum;
    
    
    LabeledEdge maxDur = tcn.getEdge(source, destination);
    
    if (maxDur != null)
    {
      flexibleMaximum = TCNAlgorithms.getTimeDistance(maxDur, true);
    }
    else
    {
      flexibleMaximum = DefaultTimeDistance.POSITIVE_INFINITE_TIME_DISTANCE;
    }
    
    
    LabeledEdge minDur = tcn.getEdge(destination, source);
    if (minDur != null)
    {
      flexibleMinimum = TCNAlgorithms.getTimeDistance(minDur, false);
      flexibleMinimum = flexibleMinimum.neg();
    }
    else
    {
      flexibleMinimum = DefaultTimeDistance.ZERO_TIME_DISTANCE;
    }
    
    STNUAllMaxProjection projection = new STNUAllMaxProjection(tcn);
    ChangeableSTN<LabeledTimePoint, STNEdge> maxSTN = (ChangeableSTN<LabeledTimePoint, STNEdge>) projection.clone();
    PC2Solver<STNEdge> pc2Solver = new PC2Solver<STNEdge>();
    DotTCNVisitor<LabeledTimePoint, STNEdge> visitor = new DotTCNVisitor<LabeledTimePoint, STNEdge>("stn");
    visitor.setName("maxSTN");
    maxSTN.visit(visitor);
    
    pc2Solver.solve(maxSTN);
    STNEdge gMaxDur = maxSTN.getEdge(destination, source);
    if (gMaxDur != null)
    {
      guardedMaximum = gMaxDur.getValue().getValue();
      guardedMaximum = guardedMaximum.neg();
    }
    else
    {
      guardedMaximum = flexibleMinimum;
    }
    
    projection = new STNUAllMaxProjection(tcn, false);
    ChangeableSTN<LabeledTimePoint, STNEdge> minSTN = (ChangeableSTN<LabeledTimePoint, STNEdge>) projection.clone();
    
    visitor.setName("minSTN");
    minSTN.visit(visitor);
    
    
    
    pc2Solver.solve(minSTN);
    
    STNEdge gMinDur = minSTN.getEdge(source, destination);
    if (gMinDur != null)
    {
      guardedMinimum = gMinDur.getValue().getValue();
    }
    else
    {
      guardedMinimum = flexibleMaximum;
    }
    
    GuardedInterval<TimeDistance> interval = new DefaultGuardedInterval<TimeDistance>(flexibleMinimum, guardedMinimum, guardedMaximum, flexibleMaximum, null); // TODO flexibility
    return interval;
  }
  

  @Test
  public void testGetSourceTimePoint() throws Exception
  {
    CSTNUGenerator generator = new CSTNUGenerator(new DefaultTimeModelOperations());
    NodeInstance nodeInstance = generator.createNodeInstance();
    LabeledTimePoint n0 = generator.createLabeledNode(nodeInstance, "n0");
    LabeledTimePoint nZ = generator.createLabeledNode(nodeInstance, "nZ");
    
    LabeledTimePoint S = generator.createLabeledNode(nodeInstance, "S");
    LabeledTimePoint T = generator.createLabeledNode(nodeInstance, "T");
    LabeledTimePoint U = generator.createLabeledNode(nodeInstance, "U");
    generator.connectNodes(S, T, T, 5, 10);
    generator.connectNodes(T, U, U, 5, 10);
    generator.connectNodes(n0, S, 1, 10);
    generator.connectNodes(U, nZ, 1, 10);
    
    LabeledTimePoint Q = generator.createLabeledNode(nodeInstance, "Q");
    LabeledTimePoint R = generator.createLabeledNode(nodeInstance, "R");
    generator.connectNodes(Q, R, R, 7, 9);
    generator.connectNodes(n0, Q, 1, 10);
    generator.connectNodes(R, nZ, 1, 10);
    
    
    generator.connectNodes(n0, nZ, 1, 25);
    
    ChangeableCSTNU cstnu = generator.createCSTNU();
    
    assertEquals(cstnu.getTimeBaseEvent(), TCNAlgorithms.getSourceTimePoint(cstnu));
    
    CSTNUDCSolver solver = new CSTNUDCSolver();
    solver.solve(cstnu);
    
    assertEquals(cstnu.getTimeBaseEvent(), TCNAlgorithms.getSourceTimePoint(cstnu));
    
    
    TCNXMLImport importer = new TCNXMLImport();
    
    InputStream inputStream = this.getClass().getResourceAsStream("TCNAlgorithms_testGetRestrictableDuration1.xml");
    ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> tcn;
    try
    {
      tcn = (ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge>) importer.getTCNFromStream(inputStream);
    }
    finally
    {
      inputStream.close();
    }
    
    solver.solve(tcn);
    assertEquals(tcn.getTimeBaseEvent(), TCNAlgorithms.getSourceTimePoint(tcn));
  }

  @Test
  public void testGetTargetTimePoint() throws Exception
  {
    CSTNUGenerator generator = new CSTNUGenerator(new DefaultTimeModelOperations());
    NodeInstance nodeInstance = generator.createNodeInstance();
    LabeledTimePoint n0 = generator.createLabeledNode(nodeInstance, "n0");
    LabeledTimePoint nZ = generator.createLabeledNode(nodeInstance, "nZ");
    
    LabeledTimePoint S = generator.createLabeledNode(nodeInstance, "S");
    LabeledTimePoint T = generator.createLabeledNode(nodeInstance, "T");
    LabeledTimePoint U = generator.createLabeledNode(nodeInstance, "U");
    generator.connectNodes(S, T, T, 5, 10);
    generator.connectNodes(T, U, U, 5, 10);
    generator.connectNodes(n0, S, 1, 10);
    generator.connectNodes(U, nZ, 1, 10);
    
    LabeledTimePoint Q = generator.createLabeledNode(nodeInstance, "Q");
    LabeledTimePoint R = generator.createLabeledNode(nodeInstance, "R");
    generator.connectNodes(Q, R, R, 7, 9);
    generator.connectNodes(n0, Q, 1, 10);
    generator.connectNodes(R, nZ, 1, 10);
    
    
    generator.connectNodes(n0, nZ, 1, 25);
    
    ChangeableCSTNU cstnu = generator.createCSTNU();
    
    assertEquals(nZ, TCNAlgorithms.getTargetTimePoint(cstnu));
    
    CSTNUDCSolver solver = new CSTNUDCSolver();
    solver.solve(cstnu);
    
    assertEquals(nZ, TCNAlgorithms.getTargetTimePoint(cstnu));
    
    TCNXMLImport importer = new TCNXMLImport();
    
    InputStream inputStream = this.getClass().getResourceAsStream("TCNAlgorithms_testGetRestrictableDuration1.xml");
    ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> tcn;
    try
    {
      tcn = (ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge>) importer.getTCNFromStream(inputStream);
    }
    finally
    {
      inputStream.close();
    }
    
    solver.solve(tcn);
    
    LabeledTimePoint destination = null;
    for (LabeledTimePoint tp : tcn.getTimePoints())
    {
      if (tp.getEventID().equals("E") && tp.getNodeInstance().getNodeID() == 0)
      {
        destination = tp;
        break;
      }
    }
    
    Assume.assumeNotNull(destination);
    
    assertEquals(destination, TCNAlgorithms.getTargetTimePoint(tcn));
  }

  @SuppressWarnings("unchecked")
  @Test
  public void testGetMinGuardDistance() throws Exception
  {
    CSTNUDCSolver solver = new CSTNUDCSolver();
    TCNXMLImport importer = new TCNXMLImport();
    
    InputStream inputStream = this.getClass().getResourceAsStream("TCNAlgorithms_minGuardTest.xml");
    ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> tcn;
    try
    {
      tcn = (ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge>) importer.getTCNFromStream(inputStream);
    }
    finally
    {
      inputStream.close();
    }
    
    solver.solve(tcn);
    
    
    LabeledTimePoint Origin = tcn.getTimeBaseEvent();
    LabeledTimePoint TP_1A = getTimePointByName(tcn, 1, "A");
    LabeledTimePoint TP_1C = getTimePointByName(tcn, 1, "C");
    LabeledTimePoint TP_2A = getTimePointByName(tcn, 2, "A");
    LabeledTimePoint TP_2C = getTimePointByName(tcn, 2, "C");
    LabeledTimePoint TP_0E = getTimePointByName(tcn, 0, "E");
    
    
    assertTDEquals(1, TCNAlgorithms.getMinGuardDistance(tcn, Origin, TP_1A));
    assertTDEquals(6, TCNAlgorithms.getMinGuardDistance(tcn, Origin, TP_1C));
    assertTDEquals(1, TCNAlgorithms.getMinGuardDistance(tcn, Origin, TP_2A));
    assertTDEquals(6, TCNAlgorithms.getMinGuardDistance(tcn, Origin, TP_2C));
    assertTDEquals(7, TCNAlgorithms.getMinGuardDistance(tcn, Origin, TP_0E)); // FIXME
   
    assertTDEquals(5, TCNAlgorithms.getMinGuardDistance(tcn, TP_1A, TP_1C));
    assertTDEquals(0, TCNAlgorithms.getMinGuardDistance(tcn, TP_1A, TP_2A));
    assertTDEquals(5, TCNAlgorithms.getMinGuardDistance(tcn, TP_1A, TP_2C));
    assertTDEquals(6, TCNAlgorithms.getMinGuardDistance(tcn, TP_1A, TP_0E));
    assertNull(TCNAlgorithms.getMinGuardDistance(tcn, TP_1A, Origin)); // Not important
    //assertTDEquals(-1, TCNAlgorithms.getMinGuardDistance(tcn, TP_1A, Origin));
    
    //assertNull(TCNAlgorithms.getMinGuardDistance(tcn, TP_1C, TP_1A)); // Not important
    assertTDEquals(-7, TCNAlgorithms.getMinGuardDistance(tcn, TP_1C, TP_1A));
    assertTDEquals(-7, TCNAlgorithms.getMinGuardDistance(tcn, TP_1C, TP_2A));
    assertTDEquals(-2, TCNAlgorithms.getMinGuardDistance(tcn, TP_1C, TP_2C));
    assertTDEquals(1, TCNAlgorithms.getMinGuardDistance(tcn, TP_1C, TP_0E));
    assertNull(TCNAlgorithms.getMinGuardDistance(tcn, TP_1C, Origin)); // Not important
    //assertTDEquals(-8, TCNAlgorithms.getMinGuardDistance(tcn, TP_1C, Origin));
    
    assertTDEquals(0, TCNAlgorithms.getMinGuardDistance(tcn, TP_2A, TP_1A));
    assertTDEquals(5, TCNAlgorithms.getMinGuardDistance(tcn, TP_2A, TP_1C));
    assertTDEquals(5, TCNAlgorithms.getMinGuardDistance(tcn, TP_2A, TP_2C));
    //assertTDEquals(6, TCNAlgorithms.getMinGuardDistance(tcn, TP_2A, TP_0E)); // FIXME
    assertNull(TCNAlgorithms.getMinGuardDistance(tcn, TP_1A, Origin)); // Not important
    //assertTDEquals(-1, TCNAlgorithms.getMinGuardDistance(tcn, TP_2A, Origin));
    
    assertTDEquals(-10, TCNAlgorithms.getMinGuardDistance(tcn, TP_2C, TP_1A));
    //assertTDEquals(-5, TCNAlgorithms.getMinGuardDistance(tcn, TP_2C, TP_1C)); // FIXME
    assertTDEquals(-10, TCNAlgorithms.getMinGuardDistance(tcn, TP_2C, TP_2A));
    //assertTDEquals(-4, TCNAlgorithms.getMinGuardDistance(tcn, TP_2C, TP_0E)); // FIXME
    assertNull(TCNAlgorithms.getMinGuardDistance(tcn, TP_1C, Origin)); // Not important
    //assertTDEquals(-11, TCNAlgorithms.getMinGuardDistance(tcn, TP_2C, Origin));
    
    assertTDEquals(-6, TCNAlgorithms.getMinGuardDistance(tcn, TP_0E, TP_1A));
    assertTDEquals(-1, TCNAlgorithms.getMinGuardDistance(tcn, TP_0E, TP_1C));
    assertTDEquals(-6, TCNAlgorithms.getMinGuardDistance(tcn, TP_0E, TP_2A));
    assertTDEquals(-1, TCNAlgorithms.getMinGuardDistance(tcn, TP_0E, TP_2C));
    assertNull(TCNAlgorithms.getMinGuardDistance(tcn, TP_0E, Origin)); // Not important
    //assertTDEquals(-7, TCNAlgorithms.getMinGuardDistance(tcn, TP_0E, Origin));
  }

  private LabeledTimePoint getTimePointByName(ChangeableSTNU<LabeledTimePoint, LabeledValueSet, LabeledEdge> tcn,
      int nodeID, String eventID)
  {
    for (LabeledTimePoint tp : tcn.getTimePoints())
      
    {
      if (tp.getNodeInstance().getNodeID() == nodeID && tp.getEventID().equals(eventID))
      {
        return tp;
      }
    }
    
    Assume.assumeTrue(false);
    return null;
  }

}
