package de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNull;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import org.xml.sax.SAXParseException;

import de.aristaflow.adept2.model.common.timedata.TimeDistance;
import de.aristaflow.adept2.model.common.timedata.TimePoint;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimeDistance;
import de.aristaflow.adept2.model.common.timedata.defaultimplementation.DefaultTimePoint;
import de.aristaflow.adept2.model.globals.TimeModelConstants.RelationType;
import de.aristaflow.adept2.model.processmodel.Node;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.timemodel.TemporalTemplateDirectedGraph;
import de.aristaflow.adept2.model.processmodel.timemodel.TemporalTemplateDotVisitor;
import de.aristaflow.adept2.model.timemodel.InstanceTimeModel;
import de.aristaflow.adept2.model.timemodel.Interval;
import de.aristaflow.adept2.model.timemodel.TemplateTimeModel;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultInterval;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.ConstraintDependencyGraph;
import de.aristaflow.adept2.model.timemodel.tcn.DotTCNVisitor;
import de.aristaflow.adept2.model.timemodel.tcn.IterationChain;
import de.aristaflow.adept2.model.timemodel.tcn.Label;
import de.aristaflow.adept2.model.timemodel.tcn.LabelSet;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValue;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledValueSet;
import de.aristaflow.adept2.model.timemodel.tcn.STN;
import de.aristaflow.adept2.model.timemodel.tcn.SimpleConstraintValue;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.TCNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.AbstractCSTNSolver;
import de.aristaflow.adept2.model.timemodel.tcn.cstnImpl.LabelBasedPC2Solver;
import de.aristaflow.adept2.model.timemodel.tcn.xml.TCNXMLExport;
import de.aristaflow.adept2.model.timemodel.tcn.xml.TCNXMLExportTest;
import de.aristaflow.adept2.model.timemodel.tools.DefaultDirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.DirectedGraph;
import de.aristaflow.adept2.tools.dot.DefaultGraphVisitor;
import de.aristaflow.adept2.tools.dot.DotTools;
import de.aristaflow.adept2.util.xml.XMLFormatException;

public class TCNTestTools
{
  


  public static final IterationChain defaultIterationChain = new IterationChain(0);

  
  private static TemporalTemplateDotVisitor templateVisitor;
  
  protected static DefaultTimeModelOperations timeModelOperations = new DefaultTimeModelOperations();

  private static DotTCNVisitor<LabeledTimePoint, LabeledEdge> tcnVisitor;

  private static DefaultGraphVisitor<Object, Object> cdgVisitor;



  
  
  static
  {
    DotTools dotTools = new DotTools();
    dotTools.cleanDir();
    
    
    if (templateVisitor == null)
      templateVisitor = new TemporalTemplateDotVisitor("", dotTools);
    
    if (tcnVisitor == null)
    {
      tcnVisitor = new DotTCNVisitor<LabeledTimePoint, LabeledEdge>("", dotTools);
      CSTNUGraphFormatter formatter = new CSTNUGraphFormatter(true, false);
      tcnVisitor.setLabelProvider(formatter);
      tcnVisitor.setGraphFormatter(formatter);
    }
    
//    if (cdgVisitor == null)
//    {
//      cdgVisitor = new DefaultGraphVisitor<Object, Object>(ChangeSetConstraintDependencyGraphTest.class.getSimpleName(), dotTools);
//      cdgVisitor.setGraphFormatter(new DefaultGraphFormatter<Object, Object>()
//      {
//
//        @Override
//        public void formatVertex(Object vertex, DotNode node)
//        {
//          if (vertex instanceof LabeledEdge)
//          {
//            LabeledEdge edge = (LabeledEdge) vertex;
//            if (!edge.isImplicit())
//              node.setStyle(NodeStyle.BOLD);
//            if (edge.isImplicit())
//              node.setStyle(NodeStyle.DASHED);
//          }
//          
//          if (vertex instanceof Tuple<?, ?>)
//          {
//            node.setShape(NodeShape.POINT);
//          }
//        }
//
//        @Override
//        public void formatGraph(Graph<Object, Object> graph,
//            AbstractDotGraph g)
//        {
//          if (g instanceof DotGraph)
//          {
//            ((DotGraph) g).setRankDir(RankDir.TopBottom);
//          }
//        }
//        
//        
//        
//
//      });
//      cdgVisitor.setLabelProvider(new DefaultLabelProvider<Object, Object>()
//      {
//        
//
//        
//        @Override
//        public String getVertexLabel(Object v)
//        {
//          if (v instanceof Tuple<?, ?>)
//            return null;
//          return super.getVertexLabel(v);
//          
//        }
//        
//      });
//    }

  }





  public static void assertCSTNUEquals(CSTNU expected, CSTNU actual)
  {
    // TODO compare observations and observation nodes.
    assertTCNEquals(expected, actual);    
  }
  
  public static void assertTCNEquals(TCN<LabeledTimePoint,LabeledEdge> expected, TCN<LabeledTimePoint,LabeledEdge> actual)
  {
    
    Collection<LabeledTimePoint> nodesExpected = expected.getTimePoints();
    Collection<LabeledTimePoint> nodesActual = actual.getTimePoints();
    
    assertEquals(nodesExpected.size(), nodesActual.size());
    
    Map<LabeledTimePoint, LabeledTimePoint> nodeEquivalenceEx = new HashMap<LabeledTimePoint, LabeledTimePoint>();
    Map<LabeledTimePoint, LabeledTimePoint> nodeEquivalenceAc = new HashMap<LabeledTimePoint, LabeledTimePoint>();
    
    for (LabeledTimePoint n1 : nodesExpected)
    {
      boolean found = false;
      
      for (LabeledTimePoint n2 : nodesActual)
      {
        if (n1.equals(n2))
        {
          if (found)
          {
            fail("Duplicate node");
          }
          found = true;
          
          if (!n1.getLabel().equals(n2.getLabel()))
            fail("Different labels!");
          
          nodeEquivalenceEx.put(n1, n2);
          nodeEquivalenceAc.put(n2, n1);
        }
      }
      
      if (!found)
        fail("No corresponding node found: "+n1);
    }
    
    assert nodeEquivalenceEx.size() == nodesExpected.size();
    assert nodeEquivalenceAc.size() == nodesExpected.size();
    
    for (LabeledTimePoint from : nodesExpected)
    {
      for (LabeledTimePoint to : nodesExpected)
      {
        LabeledEdge expectedE = expected.getEdge(from, to);
        LabeledEdge actualE = actual.getEdge(nodeEquivalenceEx.get(from), nodeEquivalenceEx.get(to));
        
        assertEdgeEquals(expected.getClass(), expectedE, actualE);
      }
    }
    
    for (LabeledTimePoint from : nodesActual)
    {
      for (LabeledTimePoint to : nodesActual)
      {
        LabeledEdge expectedE = actual.getEdge(from, to);
        LabeledEdge actualE = expected.getEdge(nodeEquivalenceAc.get(from), nodeEquivalenceAc.get(to));
        
        assertEdgeEquals(expected.getClass(), expectedE, actualE);
      }
    }
  }

  public static void assertEdgeEquals(Class<? extends TCN> clazz, LabeledEdge expectedE, LabeledEdge actualE)
  {
    if (expectedE == actualE)
      return;
    
    // Infinity is "the same" as no edge
    if (expectedE == null && actualE != null && isInfinity(actualE.getValue()))
      return;
    if (actualE == null && expectedE != null && isInfinity(expectedE.getValue()))
      return;
    
    assertFalse(String.format("Expected '%s' but got '%s'", expectedE, actualE), expectedE == null || actualE == null); // Cannot both be null (see first if) 
      
    assertEquals(expectedE.getSourceNode(), actualE.getSourceNode());
    assertEquals(expectedE.getTargetNode(), actualE.getTargetNode());

    // We don't care about Flow Direction...
    if (!equalsLabeledValueSet(clazz, expectedE.getValue(), actualE.getValue()))
      throw new AssertionError(String.format("Expected '%s' but got '%s'", expectedE, actualE));


  }
  
  public static boolean equalsLabeledValueSet(Class<? extends TCN> clazz, LabeledValueSet expected, LabeledValueSet actual)
  {
//    Map<Label, LabeledValue> mapE = DefaultLabeledValueSet.toMap(expected);
//    Map<Label, LabeledValue> mapA = DefaultLabeledValueSet.toMap(actual);
    
    Map<Label, LabeledValue> mapE = normalizeLabeledValueSet(clazz, expected);
    Map<Label, LabeledValue> mapA = normalizeLabeledValueSet(clazz, actual);
    
    assertEquals(mapE.size(), mapA.size());
    
    assertEquals(mapE.keySet(), mapA.keySet());
    
    for (Label lbl : mapE.keySet())
    {
      LabeledValue vE = mapE.get(lbl);
      LabeledValue vA = mapA.get(lbl);
      
      if (!equalsLabeledValue(vE, vA))
        return false;
    }
    
    return true;
  }

  public static boolean equalsLabeledValue(LabeledValue expected, LabeledValue actual)
  {
    if (!expected.getClass().equals(actual.getClass()))
      return false;
    
    if (expected instanceof ContingentLabeledValue)
    {
      ContingentLabeledValue cE = (ContingentLabeledValue) expected;
      ContingentLabeledValue cA = (ContingentLabeledValue) actual;
      
      TimeDistance cValueE = cE.getContingentValue();
      
      TimeDistance cValueA = cA.getContingentValue();
      
      TimeDistance valueE = expected.getValue();
      
      TimeDistance valueA = actual.getValue();
      
      return valueE.equals(valueA) && cValueE.equals(cValueA) && cE.getTrigger().equals(cA.getTrigger());
    }
    else
    {
      TimeDistance valueE = expected.getValue();
      
      TimeDistance valueA = actual.getValue();
      
      return valueE.equals(valueA);
    }
  }
  
  /**
   * Tries to represent Labeled Value Sets in a normalized way so the can be compared
   * @param clazz 
   * 
   * @param labeledValueSet
   * @return
   * 
   */
  public static Map<Label, LabeledValue> normalizeLabeledValueSet(@SuppressWarnings("rawtypes") Class<? extends TCN> clazz, LabeledValueSet labeledValueSet )
  {
    Map<Label, LabeledValue> map = DefaultLabeledValueSet.toMap(labeledValueSet);
   
    normalizeLabeledValueSet(clazz, map);
    
    return map;
  }
  
  public static Map<Label, LabeledValue> normalizeLabeledValueSet(@SuppressWarnings("rawtypes") Class<? extends TCN> clazz, Map<Label, LabeledValue> labeledValueSet)
  {
    TCNSolver<LabeledTimePoint, LabeledValueSet, LabeledEdge> tcnSolver = TCNAlgorithms.getSolver(clazz);
    
    if (!(tcnSolver instanceof AbstractCSTNSolver))
    {
      System.err.println("Trying to normalize something that is not a CSTN");
      return labeledValueSet;
    }
      
   
    AbstractCSTNSolver solver = (AbstractCSTNSolver) tcnSolver;
//    boolean changes = true;
//    int i = 10;
//    while (changes && i-- > 0)
//    {
//      changes = false;
//      changes |= solver.deriveAdditionalLabels(labeledValueSet);
//      changes |= solver.applySimplificationRules(labeledValueSet);
//    }
//    
//    if (i <= 1)
//      System.out.println("Issue case: " + labeledValueSet);
    
    solver.applySimplificationRules(labeledValueSet);
    
    return labeledValueSet;
  }

  private static boolean isInfinity(LabeledValueSet valueSet)
  {
    if (valueSet.getValues().length != 1)
      return false;
    
    SimpleConstraintValue value = valueSet.getValues()[0];
    return value.getValue().isInfinity();
  }
  



  /**
   * only for CSTN 
   * NOT for CSTNU!!!
   * @param cstn
   */
  public static void assertConsistent(TCN<LabeledTimePoint,LabeledEdge> cstn)
  {
    // TODO check time distance for consistency
    Collection<LabeledTimePoint> nodes = cstn.getTimePoints();
    for (LabeledTimePoint from : nodes)
    {
      for (LabeledTimePoint to : nodes)
      {
        if (from == to)
          continue;

          LabeledEdge dir1 = cstn.getEdge(from, to);
          LabeledEdge dir2 = cstn.getEdge(to, from);
          
          if (dir1 != null && dir2 != null)
          {
            assertTrue(dir1.isContingentDuration() == dir2.isContingentDuration());
            boolean contingentDuration = dir1.isContingentDuration();
            
            assertTrue(String.format("%s does not subsume %s", from.getLabel(), dir1.getValue().getLabels()), satisfiable(from.getLabel(), dir1.getValue()));
            assertTrue(String.format("%s does not subsume %s", to.getLabel(), dir1.getValue().getLabels()), satisfiable(to.getLabel(), dir1.getValue()));
            assertTrue(String.format("%s does not subsume %s", from.getLabel(), dir2.getValue().getLabels()), satisfiable(from.getLabel(), dir2.getValue()));
            assertTrue(String.format("%s does not subsume %s", to.getLabel(), dir2.getValue().getLabels()), satisfiable(to.getLabel(), dir2.getValue()));
            
            try
            {
              Collection<LabeledValue> lvs1 = dir1.getValue().getLabeledValues();
              Collection<LabeledValue> lvs2 = dir2.getValue().getLabeledValues();
              
              for (LabeledValue lv1 : lvs1)
              {
                assertTrue(lv1 instanceof ContingentLabeledValue || !contingentDuration);
                TimeDistance value1 = lv1.getValue();
                

                for (LabeledValue lv2 : lvs2)
                {
                  assertTrue(lv2 instanceof ContingentLabeledValue || ! contingentDuration);
                  TimeDistance value2 = lv2.getValue();
                  
                  if (lv1.getLabel().satisfiable(lv2.getLabel()))
                  {
                    if (lv1 instanceof ContingentLabeledValue && lv2 instanceof ContingentLabeledValue)
                    {
                      ContingentLabeledValue clv1 = (ContingentLabeledValue) lv1;
                      ContingentLabeledValue clv2 = (ContingentLabeledValue) lv2;
                      
                      assertFalse(clv1.getContingentValue()+" should not be > than "+clv2.getContingentValue().neg(), clv1.getContingentValue().compareTo(clv2.getContingentValue().neg()) >  0);
                    }
                    
                    // FIXME consider case if one value is contingent
                    
                    assertFalse(value1+" should not be < than -"+value2, value1.compareTo(value2.neg()) <  0);
                    
                  }
                }
              }
            } catch (AssertionError e)
            {
              throw new AssertionError(dir1+" and "+dir2+" are inconsistent: "+e.getMessage());
            }
          }

      }
    }
  }
  
  public static void assertConsistent(STN<LabeledTimePoint, STNEdge> cstn)
  {
 // TODO check time distance for consistency
    Collection<LabeledTimePoint> nodes = cstn.getTimePoints();
    for (LabeledTimePoint from : nodes)
    {
      for (LabeledTimePoint to : nodes)
      {
        if (from == to)
          continue;

          STNEdge dir1 = cstn.getEdge(from, to);
          STNEdge dir2 = cstn.getEdge(to, from);
          
          if (dir1 != null && dir2 != null)
          {
            
//            try
//            {
              
              SimpleConstraintValue lv1 = dir1.getValue();
              
              TimeDistance value1 = lv1.getValue();
              

              SimpleConstraintValue lv2 = dir2.getValue();
              
              TimeDistance value2 = lv2.getValue();
              
    
              if (lv1 instanceof ContingentLabeledValue && lv2 instanceof ContingentLabeledValue)
              {
                ContingentLabeledValue clv1 = (ContingentLabeledValue) lv1;
                ContingentLabeledValue clv2 = (ContingentLabeledValue) lv2;
                
                assertFalse(clv1.getContingentValue()+" should not be > than -"+clv2.getContingentValue(), clv1.getContingentValue().compareTo(clv2.getContingentValue().neg()) >  0);
              }
              
              // FIXME consider case if one value is contingent
              
              assertFalse(value1+" should not be < than -"+value2, value1.compareTo(value2.neg()) <  0);
                  
//            } catch (AssertionError e)
//            {
//              throw new AssertionError(dir1+" and "+dir2+" are inconsistent: "+e.getMessage());
//            }
          }

      }
    }
  }
  
  public static void assertComplete(TCN<LabeledTimePoint, LabeledEdge> tcn)
  {
    Collection<LabeledTimePoint> nodes = tcn.getTimePoints();
    for (LabeledTimePoint from : nodes)
    {
      for (LabeledTimePoint to : nodes)
      {
        if (from == to)
          continue;
        LabeledEdge edge = tcn.getEdge(from, to);
        if (from.getLabel().satisfiable(to.getLabel()))
        {
          if (edge != null) // TODO properly handle infty...
          {
            Label expected = from.getLabel().merge(to.getLabel());
            
            
            // This is not the nicest way to get the "union" of all labels, but the most simple ones (and all used functions have been tested)
            Map<Label, LabeledValue> map = new HashMap<Label, LabeledValue>();
            
            for (LabeledValue value : edge.getValue().getLabeledValues())
            {
              DefaultLabeledValueSet.addOrReplaceIfNecessary(map, new DefaultLabeledValue(value.getLabel(), DefaultTimeDistance.ONE_TIME_DISTANCE)); 
            }
            
            normalizeLabeledValueSet(tcn.getClass(), map);
            
            assertTrue(map.size() == 1);
            Label actual = map.keySet().iterator().next();
            
            assertEquals(String.format("Expected '%s'+'%s'='%s' but was: '%s'", from.getLabel(), to.getLabel(), expected, edge), expected, actual);
          }
        }
        else
        {
          assertNull(edge);
        }
      }
    }
  }
  
  private static boolean satisfiable(Label label, LabelSet labels)
  {
    for (Label l2 : labels.getLabels())
    {
      if (!label.satisfiable(l2))
        return false;
    }

    return true;
  }


  public static void printProcessTimes(TemplateTimeModel timeModel)
  {
    
    Template template = timeModel.getBaseTemplate();
    System.out.println("Process Duration: "+timeModel.getProcessDuration());

    int startNodeID = template.getStartNode().getID();
    for (Node n : template.getNodes())
    {
      System.out.println("Activity "+n.getName()+" (#"+n.getID()+")");
      System.out.println("Activity Duration: "+timeModel.getActivityDuration(n.getID()));
      System.out.println("Activity Start (absolut): "+timeModel.getActivityStartTimeframe(n.getID()));
      System.out.println("Activity Completion (absolut): "+timeModel.getActivityCompletionTimeframe(n.getID()));
      System.out.println("Activity Start (relativ): "+timeModel.getTimeLag(startNodeID, n.getID(), RelationType.START_START));
      System.out.println("Activity Completion (relativ): "+timeModel.getTimeLag(startNodeID, n.getID(), RelationType.START_END));
      System.out.println("");
    }
  }

  
  public static void printTemplate(Template template)
  {
    TemporalTemplateDirectedGraph graph = new TemporalTemplateDirectedGraph(template);
    graph.visit(templateVisitor.getInstance(template.getName()));
  }
  
  public static void assertTPEquals(long minValue, long maxValue, Interval<TimePoint> actual, long delta)
  {
    TimePoint maxTP = secondsToTimePoint(maxValue);
    TimePoint minTP = secondsToTimePoint(minValue);
    
    Interval<TimePoint> expected = new DefaultInterval<TimePoint>(minTP, maxTP);
    
    assertTPEquals(expected, actual, delta);
  }
  
  public static TimePoint secondsToTimePoint(long value)
  {
    return new DefaultTimePoint(value, 0, 0, 0, 0);
  }


  public static void assertTPEquals(Interval<TimePoint> expected,
      Interval<TimePoint> actual, long delta)
  {
    try
    {
      assertTPEquals(expected.getMinValue(), actual.getMinValue(), delta);
      assertTPEquals(expected.getMaxValue(), actual.getMaxValue(), delta);
    }
    catch (AssertionError e)
    {
      throw new AssertionError(String.format("expected:<%s> but was <%s>", expected, actual));
    }
  }

  public static void assertTPEquals(TimePoint expected, TimePoint actual, long delta)
  {
    assertTrue(Math.abs(timeModelOperations.convertToSeconds(expected) - timeModelOperations.convertToSeconds(actual)) <= delta); 
  }
  
  public static void assertTDEquals(long minValue, long maxValue, Interval<TimeDistance> actual, long delta)
  {
    TimeDistance maxTP = timeModelOperations.secondsToTimeDistance(maxValue);
    TimeDistance minTP = timeModelOperations.secondsToTimeDistance(minValue);
    
    Interval<TimeDistance> expected = new DefaultInterval<TimeDistance>(minTP, maxTP);
    
    assertTDEquals(expected, actual, delta);
  }
  
  public static void assertTDEquals(Interval<TimeDistance> expected,
      Interval<TimeDistance> actual, long delta)
  {
    try
    {
      assertTDEquals(expected.getMinValue(), actual.getMinValue(), delta);
      assertTDEquals(expected.getMaxValue(), actual.getMaxValue(), delta);
    }
    catch (AssertionError e)
    {
      throw new AssertionError(String.format("expected:<%s> but was <%s>", expected, actual));
    }
  }

  public static void assertTDEquals(TimeDistance expected, TimeDistance actual)
  {
    assertTDEquals(expected, actual, 0);
  }
  
  public static void assertTDEquals(TimeDistance expected, TimeDistance actual, long delta)
  {
    assertTrue("Expected "+expected+" but was "+actual, Math.abs(expected.toSeconds() - actual.toSeconds()) <= delta);
  }
  
  public static void assertTDEquals(long expectedSeconds, TimeDistance actual)
  {
    assertTrue("Expected "+expectedSeconds+" but was "+actual.toSeconds(), expectedSeconds == actual.toSeconds());
  }

  public static void printProcessTimes(InstanceTimeModel timeModel, Template template)
  {
//    Template template = ((CSTNUInstanceTimeModel)timeModel).getBaseTemplate();
    System.out.println("Process Duration: "+timeModel.getInstanceDuration());

    int startNodeID = template.getStartNode().getID();
    for (Node n : template.getNodes())
    {
      System.out.println("Activity "+n.getName()+" (#"+n.getID()+")");
      System.out.println("Activity Duration: "+timeModel.getActivityInstanceDuration(n.getID(), 0));
      System.out.println("Activity Start (absolut): "+timeModel.getActivityInstanceStartTimeframe(n.getID(), 0));
      System.out.println("Activity Completion (absolut): "+timeModel.getActivityInstanceCompletionTimeframe(n.getID(), 0));
      System.out.println("Activity Start (relativ): "+timeModel.getTimeLag(startNodeID, 0, n.getID(), 0, RelationType.START_START));
      System.out.println("Activity Completion (relativ): "+timeModel.getTimeLag(startNodeID, 0, n.getID(), 0, RelationType.START_END));
      System.out.println("");
    }
  }
  
  public static class DiffEdge
  {

    private LabeledValueSet expected;
    private LabeledValueSet actual;

    public DiffEdge(LabeledTimePoint sourceNode, LabeledTimePoint targetNode, LabeledValueSet expected, LabeledValueSet actual)
    {
      this.expected = expected;
      this.actual = actual;
    }
    
    @Override
    public String toString()
    {
      StringBuilder builder = new StringBuilder();
      builder.append("Expected: ");
      builder.append(this.expected);
      builder.append("\nActual: ");
      builder.append(this.actual);
      return builder.toString();
    }
    
  }
  
  public static DirectedGraph<LabeledTimePoint, DiffEdge> diff(TCN<LabeledTimePoint, LabeledEdge> expected, TCN<LabeledTimePoint, LabeledEdge> actual)
  {
    //Collection<DiffEdge> edges = new ArrayList<DiffEdge>();
    
    
    Collection<LabeledTimePoint> nodesExpected = expected.getTimePoints();
    Collection<LabeledTimePoint> nodesActual = actual.getTimePoints();
    
    assertEquals(nodesExpected.size(), nodesActual.size());
    
    Map<LabeledTimePoint, LabeledTimePoint> nodeEquivalenceEx = new HashMap<LabeledTimePoint, LabeledTimePoint>();
        
    for (LabeledTimePoint n1 : nodesExpected)
    {
      boolean found = false;
      
      for (LabeledTimePoint n2 : nodesActual)
      {
        if (n1.equals(n2))
        {
          if (found)
          {
            fail("Duplicate node");
          }
          found = true;
          
          if (!n1.getLabel().equals(n2.getLabel()))
            fail("Different labels!");
          
          nodeEquivalenceEx.put(n1, n2);
        }
      }
      
      if (!found)
        fail("No corresponding node found: "+n1);
    }
    
    
    DefaultDirectedGraph<LabeledTimePoint, DiffEdge> diffGraph = new DefaultDirectedGraph<LabeledTimePoint, DiffEdge>(expected.getTimePoints());
    
    for (LabeledTimePoint node1 : expected.getTimePoints())
    {
      for (LabeledTimePoint node2 : expected.getTimePoints())
      {
        LabeledEdge edgeE = expected.getEdge(node1, node2);
        LabeledEdge edgeA = actual.getEdge(nodeEquivalenceEx.get(node1), nodeEquivalenceEx.get(node2));
        
        if (edgeE != null)
        {
          
          if (edgeA == null || !equalsLabeledValueSet(expected.getClass(), edgeE.getValue(), edgeA.getValue()))
          {
//            if (!edgeE.isImplicit() && (edgeA == null || !edgeA.isImplicit()))
//            {
              LabeledValueSet value = new DefaultLabeledValueSet(normalizeLabeledValueSet(expected.getClass(), edgeE.getValue()).values());
              DiffEdge diff = new DiffEdge(edgeE.getSourceNode(), edgeE.getTargetNode(), value, edgeA != null ? edgeA.getValue() : null);
              diffGraph.addEdge(edgeE.getSourceNode(), edgeE.getTargetNode(), diff);
//            }
          }
        }
        else
        {
          if (edgeA != null)
          {
//            if (!edgeA.isImplicit())
//            {
              DiffEdge diff = new DiffEdge(edgeA.getSourceNode(), edgeA.getTargetNode(), null, edgeA.getValue());
              diffGraph.addEdge(edgeA.getSourceNode(), edgeA.getTargetNode(), diff);
//            }
          }
        }
        
      }
    }
    
    
    return diffGraph;
  }

  
  public static void assertEqualsIgnoreOrdering(Collection<?> expected, Collection<?> actual)
  {
    assertArrayEqualsIgnoreOrdering(expected.toArray(), actual.toArray());
  }
  
  public static void assertEqualsIgnoreOrdering(Collection<?> expected, Object[] actual)
  {
    assertArrayEqualsIgnoreOrdering(expected.toArray(), actual);
  }
  
  public static void assertEqualsIgnoreOrdering(Object[] expected, Collection<?> actual)
  {
    assertArrayEqualsIgnoreOrdering(expected, actual.toArray());
  }
  
  public static void assertArrayEqualsIgnoreOrdering(Object[] expected, Object[] actual)
  {
    if (expected == null && actual == null)
      return;
    
    if (expected == null || actual == null)
    {
      fail(String.format("Expected %s but got %s", Arrays.toString(expected), Arrays.toString(actual)));
      return; // Silence Compiler...
    }
    
    if (expected.length != actual.length)
      fail(String.format("Expected %s but got %s", Arrays.toString(expected), Arrays.toString(actual)));
    
    for (Object o : expected)
    {
      boolean found = false;
      for (Object x : actual)
      {
        if (o == x || o != null && o.equals(x))
        {
          found = true;
        }
      }
      
      if (!found)
      {
        fail(String.format("Expected %s but got %s", Arrays.toString(expected), Arrays.toString(actual)));
      }
    }
    
  }
  
  public static void assertArrayContains(Object[] array, Object obj)
  {
    for (Object o : array)
    {
      if (o == obj)
        return;
    }
    fail(String.format("%s does not contain %s", array, obj));
  }



  public static void printTCN(TCN<LabeledTimePoint, LabeledEdge> tcn, Class<?> callingClazz,
      String name)
  {   
    tcnVisitor.setName(callingClazz.getSimpleName());
    
    tcn.visit(tcnVisitor.getInstance(name));
  }



  public static void printNodeTimes(TCN<LabeledTimePoint, LabeledEdge> tcn)
  {
    LabeledTimePoint timeBaseEvent = tcn.getTimeBaseEvent();
    for (LabeledTimePoint node : tcn.getTimePoints())
    {
      LabeledEdge max = tcn.getEdge(timeBaseEvent, node);
      LabeledEdge min = tcn.getEdge(node, timeBaseEvent);
      
      System.out.println("\n"+node.toString()+": "+node.getLabel());
      
      if (min != null)
      {
        System.out.println("\tMin:");
        Collection<LabeledValue> lvs1 = min.getValue().getLabeledValues();
        
        for (LabeledValue lv1 : lvs1)
        {
          System.out.println("\t  "+lv1.getLabel()+": "+lv1.getValue().neg());
        }
      }
      if (max != null)
      {
        System.out.println("\tMax:");
        Collection<LabeledValue> lvs1 = max.getValue().getLabeledValues();
        
        for (LabeledValue lv1 : lvs1)
        {
          System.out.println("\t  "+lv1.getLabel()+": "+lv1.getValue());
        }
      }
      
      
    }
    
  }



  public static void printCDG(
      ConstraintDependencyGraph<LabeledEdge> graph, Class<?> callingClazz,
      String name)
  {
    cdgVisitor.setName(callingClazz.getSimpleName());
    
    graph.visit(cdgVisitor.getInstance(name));
  }



  public static void printCSV(TCN<LabeledTimePoint, LabeledEdge> tcn)
  {
    String format = "%1$-20s;\t%2$-50s;\t%3$-20s";
    Collection<LabeledEdge> edges = tcn.getEdges();
    
    System.out.println(String.format(format, "Source", "Constraint", "Target"));
    
    for (LabeledEdge edge : edges)
    {
      for (LabeledValue value : edge.getValue().getLabeledValues())
      {
        System.out.println(String.format(format, edge.getSourceNode(), value, edge.getTargetNode()));
      }
    }
    
  }



  public static void export(TCN<LabeledTimePoint, LabeledEdge> cstnu, Class<?> callingClazz,
      String name)
  {
    TCNXMLExport export = new TCNXMLExport();
    InputStream inputStream;
    try
    {
      inputStream = export.getInputStreamForTCN(cstnu);
      TCNXMLExportTest.writeToFile(new File("./dot/"+callingClazz.getSimpleName()+"_"+name+".xml"), inputStream);
      
      inputStream.close();
    }
    catch (XMLFormatException e)
    {
      e.printStackTrace();
      
      if (e.getCause() instanceof SAXParseException)
      {
        fail(e.getMessage()+" @ Line: "+((SAXParseException) e.getCause()).getLineNumber()+" Column: "+((SAXParseException) e.getCause()).getColumnNumber());
      }
      
      fail(e.getMessage());
    }
    catch (IOException e)
    {
      e.printStackTrace();
      fail(e.getMessage());
    }

  }




}
