package de.aristaflow.adept2.model.timemodel.tcn.xml;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.apache.commons.configuration.BaseConfiguration;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.timemodel.tcn.CSTNU;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledEdge;
import de.aristaflow.adept2.model.timemodel.tcn.LabeledTimePoint;
import de.aristaflow.adept2.model.timemodel.tcn.TCN;
import de.aristaflow.adept2.model.timemodel.tcn.cstnuImpl.CSTNUTemplateTransformation;
import de.aristaflow.adept2.model.timemodel.tcn.defaultimplementation.TCNTestTools;
import de.aristaflow.adept2.model.timemodel.templates.TestTemplates;
import de.aristaflow.adept2.util.xml.XMLFormatException;

public class TCNXMLExportTest
{

  @BeforeClass
  public static void setUpBeforeClass() throws Exception
  {
    //
  }

  @Before
  public void setUp() throws Exception
  {
    //
  }
  
  @Test
  public void testTemplates() throws XMLFormatException, IOException 
  {
    
    for (TemporalTemplate template : TestTemplates.getTemplates())
    {
      try
      {
        
        basicTestWithTemplate(template);
      }
      catch (XMLFormatException e)
      {
        throw new XMLFormatException(template.getName(), e);
      }
      catch (AssertionError e)
      {
        System.err.println(template.getName());
        throw e;
      }
      catch (IllegalStateException e)
      {
        System.err.println(template.getName());
        throw e;
      }
    }
  }

  private void basicTestWithTemplate(TemporalTemplate template) throws XMLFormatException, IOException
  {
    TCNXMLExport export = new TCNXMLExport();
    
    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(new BaseConfiguration(), template, 0);
    CSTNU cstnu = trafo.transform();
    
    InputStream stream = export.getInputStreamForTCN(cstnu);

    try
    {
      
      //writeToFile(new File("./dot/"+template.getName()+".xml"), stream);
      stream.reset();
    
      TCNXMLImport xmlImport = new TCNXMLImport();
      
      TCN<LabeledTimePoint, LabeledEdge> tcn = xmlImport.getTCNFromStream(stream);
           
      TCNTestTools.assertTCNEquals(cstnu, tcn);
    }
    finally 
    {
    
      stream.close();
    }
  }

  @Test
  public void test() throws XMLFormatException, IOException
  {
    TCNXMLExport export = new TCNXMLExport();
    
    CSTNUTemplateTransformation trafo = new CSTNUTemplateTransformation(new BaseConfiguration(), 
      TestTemplates.healthcareExampleTestTemplate, 0);
    CSTNU cstnu = trafo.transform();

    InputStream stream = export.getInputStreamForTCN(cstnu);

    try
    {
      
//      writeToFile(new File("./dot/CSTNU.xml"), stream);
//      stream.reset();
    
      TCNXMLImport xmlImport = new TCNXMLImport();
      
      TCN<LabeledTimePoint, LabeledEdge> tcn = xmlImport.getTCNFromStream(stream);
           
      TCNTestTools.assertTCNEquals(cstnu, tcn);
    }
    finally 
    {
    
      stream.close();
    }
  }
  
  public static void writeToFile(File file, InputStream inputStream)
  {
    OutputStream outputStream = null;
   
    try {
   
      // write the inputStream to a FileOutputStream
      outputStream = 
                      new FileOutputStream(file);
   
      int read = 0;
      byte[] bytes = new byte[1024];
   
      while ((read = inputStream.read(bytes)) != -1) {
        outputStream.write(bytes, 0, read);
      }
   
    } catch (IOException e) {
      e.printStackTrace();
    } finally {
      try
      {
        inputStream.reset();
      }
      catch (IOException e1)
      {
        e1.printStackTrace();
      }
      if (outputStream != null) {
        try {
          // outputStream.flush();
          outputStream.close();
        } catch (IOException e) {
          e.printStackTrace();
        }
   
      }
      
      
    }
  }

}
