package de.aristaflow.adept2.model.timemodel.templates;

import java.io.File;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

import de.aristaflow.adept2.core.checks.processmodel.timemodel.BasicTemporalProcessCheck;
import de.aristaflow.adept2.model.processmodel.Template;
import de.aristaflow.adept2.model.processmodel.TemporalTemplate;
import de.aristaflow.adept2.model.processmodel.defaultimplementation.DefaultProcessModelFactory;
import de.aristaflow.adept2.model.processmodel.defaultimplementation.DefaultTemporalTemplate;
import de.aristaflow.adept2.model.processmodel.tools.NodeRelations;
import de.aristaflow.adept2.model.processmodel.tools.ProcessElementIdentifierTools;
import de.aristaflow.adept2.model.processmodel.xml.ProcessModelXMLImport;
import de.aristaflow.adept2.model.timemodel.defaultimplementation.DefaultTimeModelOperations;
import de.aristaflow.adept2.util.CheckReport;
import de.aristaflow.adept2.util.CheckReport.ResultType;
import de.aristaflow.adept2.util.xml.VersionException;
import de.aristaflow.adept2.util.xml.XMLFormatException;

public class TestTemplates
{
  

  /**
   * <img src="doc-files/XORANDTest.jpg" />
   */
  public static TemporalTemplate xorAndTestTemplate;

  /**
   * <img src="doc-files/NestedLoops.jpg" />
   */
  public static TemporalTemplate nestedLoopTestTemplate;

  /**
   * <img src="doc-files/XOREmptyBranch.jpg" />
   */
  public static TemporalTemplate xorEmptyBranchTestTemplate;

  /**
   * <img src="doc-files/ANDEmptyBranch.jpg" />
   */
  public static TemporalTemplate andEmptyBranchTestTemplate;

  /**
   * <img src="doc-files/SyncEdgeTest.jpg" />
   */
  public static TemporalTemplate syncEdgeTestTemplate;

  /**
   * <img src="doc-files/SimpleLoop.jpg" />
   */
  public static TemporalTemplate simpleLoopTestTemplate;

  /**
   * <img src="doc-files/EmptyLoopTest.jpg" />
   */
  public static TemporalTemplate emptyLoopTestTemplate;

  /**
   * <img src="doc-files/HealthcareExample.jpg" />
   */
  public static TemporalTemplate healthcareExampleTestTemplate;

  /**
   * <img src="doc-files/SimpleTimeTestNonContingent.jpg" />
   */
  public static TemporalTemplate simpleTimeTestNonContingentTestTemplate;

  /**
   * <img src="doc-files/SimpleTimeTestContingent.jpg" />
   */
  public static TemporalTemplate simpleTimeTestContingentTestTemplate;
  
  /**
   * <img src="doc-files/XORAdvancedNotPC2.jpg" />
   */
  public static TemporalTemplate xorAdvancedNotPC2TestTemplate;
  
  public static TemporalTemplate andTimeConstraintTestTemplate;
  
  

  public static TemporalTemplate clinicalProcessTestTemplate;

  public static TemporalTemplate coopIS13TestTemplate;

  public static TemporalTemplate xorTimeConstraintTestTemplate;
  
  

  private static DefaultProcessModelFactory pmf;

  private static ProcessModelXMLImport xmlImport;

  private static TemporalTemplate[] templates;
  
  private static BasicTemporalProcessCheck templateCheck;

  static
  {
    pmf = new DefaultProcessModelFactory(null);
    pmf.setCreateChangeable(true);
    xmlImport = new ProcessModelXMLImport(pmf);
    templateCheck = new BasicTemporalProcessCheck(new DefaultTimeModelOperations());
    
    try
    {
      loadTemplates();
    }
    catch (Exception e)
    {
      // If anything goes wrong with loading templates, stop right there...
      e.printStackTrace();
      System.err.println("Initialization failed. Exiting!");
      System.exit(1);
    }
  }

  public static void loadTemplates()
  {
      try
      { 
        andEmptyBranchTestTemplate = TestTemplates.loadTemporalTemplate("ANDEmptyBranch.template");
        andTimeConstraintTestTemplate = TestTemplates.loadTemporalTemplate("ANDTimeConstraintTest.template");
        clinicalProcessTestTemplate = TestTemplates.loadTemporalTemplate("ClinicalProcess.template");
        coopIS13TestTemplate = TestTemplates.loadTemporalTemplate("CoopIS13.template");
        //cycleTestTemplate = TestTemplates.loadTemplate("Cycle.template");
        emptyLoopTestTemplate = TestTemplates.loadTemporalTemplate("EmptyLoopTest.template");
        healthcareExampleTestTemplate = TestTemplates.loadTemporalTemplate("HealthcareExample.template");
        nestedLoopTestTemplate = TestTemplates.loadTemporalTemplate("NestedLoops.template");
        simpleLoopTestTemplate = TestTemplates.loadTemporalTemplate("SimpleLoop.template");
        simpleTimeTestContingentTestTemplate = TestTemplates.loadTemporalTemplate("SimpleTimeTestContingent.template");
        simpleTimeTestNonContingentTestTemplate = TestTemplates.loadTemporalTemplate("SimpleTimeTestNonContingent.template");
        syncEdgeTestTemplate = TestTemplates.loadTemporalTemplate("SyncEdgeTest.template");
        xorAdvancedNotPC2TestTemplate = TestTemplates.loadTemporalTemplate("XORAdvancedNotPC2.template");
        xorAndTestTemplate = TestTemplates.loadTemporalTemplate("XORANDTest.template");
        xorEmptyBranchTestTemplate = TestTemplates.loadTemporalTemplate("XOREmptyBranch.template");
        xorTimeConstraintTestTemplate = TestTemplates.loadTemporalTemplate("XORTimeConstraintTest.template");
          
        TestTemplates.getTemplates();
      }
      catch (IOException e)
      {
        throw new RuntimeException(e);
      }
      catch (VersionException e)
      {
        throw new RuntimeException(e);      
      }
      catch (XMLFormatException e)
      {
        throw new RuntimeException(e);      
      }
    
  }
  
  
  public static TemporalTemplate loadTemporalTemplate(String fileName)
      throws VersionException, XMLFormatException, IOException
  {
    Template testTemplate = loadTemplate(fileName);

    return new DefaultTemporalTemplate(testTemplate);
  }


  public static Template loadTemplate(String fileName) throws VersionException, XMLFormatException, IOException
  {
    Template testTemplate;
    InputStream inputStream = TestTemplates.class
      .getResourceAsStream(fileName);
    try
    {
      testTemplate = loadTemplate(inputStream);
    }
    finally
    {
      inputStream.close();
    }
    
    if (testTemplate == null)
      throw new RuntimeException(fileName);
    return testTemplate;
  }


  public static Template loadTemplate(InputStream inputStream) throws VersionException, XMLFormatException, IOException
  {
    Template testTemplate;
   
      testTemplate = xmlImport.getTemplateFromStream(inputStream);
    

    
    
    checkTemplate(testTemplate);
    return testTemplate;
  }

  public static void checkTemplate(Template template)
  {
    NodeRelations relations = new NodeRelations(template);
    URI uri = ProcessElementIdentifierTools.getTemplateIdentifier(template, template.getName());
    CheckReport checkReport = new CheckReport(uri);
    templateCheck.performCheck(template, relations, checkReport);
    
    if (ResultType.FAILURE.compareTo(checkReport.getCheckResult()) <= 0)
    {
      throw new IllegalStateException(String.format("Template '%s' does not fullfill basic checks:\n%s", template.getName(), checkReport.getReportSummary()));
    }
    
  }

  public static TemporalTemplate[] getTemplates()
  {
    if (templates == null)
    {
      try
      {
        String[] templateFiles = listFilesInSamePackage(TestTemplates.class, ".template");
        
        
        templates = new TemporalTemplate[templateFiles.length];
        
        for (int i = 0; i < templateFiles.length; i++)
        {
          String fileName = templateFiles[i];
          templates[i] = loadTemporalTemplate(fileName);
        }
      }
      catch (VersionException e)
      {
        templates = null;
      }
      catch (XMLFormatException e)
      {
        templates = null;
      }
      catch (IOException e)
      {
        templates = null;
      }
      
    }
    return templates;
  }
  

  private static String[] listFilesInSamePackage(Class<?> locator,
      String fileExtension)
  {
    File packageFile = getPackageFile(locator);

    return listFileNames(packageFile, fileExtension.startsWith(".") ? fileExtension : "."+fileExtension);
  }

  private static File getPackageFile(Class<?> locator)
  {
    URL url = locator.getResource(locator.getSimpleName()
                                                       + ".class");
    if (url == null)
    {
      throw new RuntimeException("Cannot locate " + locator.getName());
    }

    try
    {
      return new File(url.toURI()).getParentFile();
    }
    catch (URISyntaxException e)
    {
      throw new RuntimeException(e);
    }
  }

  private static String[] listFileNames(File packageFile, final String extension)
  {
    return packageFile.list(new FilenameFilter()
    {
      @Override
      public boolean accept(File dir, String name)
      {
        return name.endsWith(extension);
      }
    });
  }

}
