package de.aristaflow.adept2.tools.dot;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import de.aristaflow.adept2.util.Tuple;

public abstract class AbstractDotGraph extends DotElement {

	protected Map<String, DotNode> nodes;
	protected Map<Tuple<DotNode, DotNode>, DotEdge> edges;
	protected final String name;
  protected Map<String, DotSubGraph> subgraphs;

	public AbstractDotGraph(String name) {
		this.name = name;
		this.nodes = new HashMap<String, DotNode>();
		this.edges = new HashMap<Tuple<DotNode, DotNode>, DotEdge>();
		this.subgraphs = new HashMap<String, DotSubGraph>();
	}
	  

	
	public DotNode createNode(String nodeID) {
		DotNode node = new DotNode(nodeID);
		nodes.put(nodeID, node);
		return node;
	}

	public DotEdge createEdge(DotNode from, DotNode to) {
		DotEdge edge = new DotEdge(from, to);
		edge.setDirected(isDirected());
		edges.put(new Tuple<DotNode, DotNode>(from, to), edge);
		return edge;
	}

	protected abstract boolean isDirected();
  
  public DotEdge createEdge(String from, String to) {
		DotNode fromNode = getNode(from);
		
		if (fromNode == null)
			throw new IllegalArgumentException(String.format("From Node %s does not exist!", from));
		
		DotNode toNode = getNode(to);
		
		if (toNode == null)
			throw new IllegalArgumentException(String.format("To Node %s does not exist!", to));
		
		return createEdge(fromNode, toNode);
	}

	public DotNode getNode(String nodeID) {
		DotNode dotNode = nodes.get(nodeID);

		return dotNode;
	}

  public DotSubGraph createSubgraph(String subGraphID)
  {
  	DotSubGraph graph = new DotSubGraph(this, subGraphID);
  	subgraphs.put(subGraphID, graph);
  	return graph;
  }
  
  public Collection<DotNode> getNodes()
  {
    return nodes.values();
  }
  
  public Collection<DotSubGraph> getSubGraphs()
  {
    return subgraphs.values();
  }
  
  public Collection<DotEdge> getEdges()
  {
    return edges.values();
  }

}
