package de.aristaflow.adept2.tools.dot;

import java.util.ArrayDeque;
import java.util.Deque;

import de.aristaflow.adept2.model.timemodel.tools.DirectedGraph;
import de.aristaflow.adept2.model.timemodel.tools.Graph;
import de.aristaflow.adept2.model.timemodel.tools.GraphVisitor;
import de.aristaflow.adept2.model.timemodel.tools.UndirectedGraph;
import de.aristaflow.adept2.tools.dot.DotGraph.GraphType;
import de.aristaflow.adept2.tools.dot.DotGraph.RankDir;

public class DefaultGraphVisitor<V, E> implements GraphVisitor<V, E>

{
  protected LabelProvider<V, E> labelProvider;
	protected GraphFormatter<V, E> graphFormatter;
	
	protected String name;
	protected String title = null;
  protected final DotTools dotTools;
  //protected DotGraph dotGraph;
  protected Deque<AbstractDotGraph> dotGraphs = new ArrayDeque<AbstractDotGraph>(4);
  protected Deque<Graph<V, E>> graphs = new ArrayDeque<Graph<V, E>>(4);
  
  protected int inc = 0;
  private GraphType graphType;
	
  
  public DefaultGraphVisitor(String name)
  {
    this(name, new DotTools());
  }
  
  
  public DefaultGraphVisitor(String name, DotTools dotTools)
  {
    this.name = name;
    this.dotTools = dotTools;
    this.labelProvider = new DefaultLabelProvider<V, E>();
    this.graphFormatter = new DefaultGraphFormatter<V, E>();
  }

  

  public void discoverGraph(Graph<V, E> graph)
  {    
    if (dotGraphs.isEmpty())
    {
      inc++;
      graphType = GraphType.GRAPH;
      if (graph instanceof UndirectedGraph<?, ?>)
      {
        graphType = GraphType.GRAPH;
      }
      else if (graph instanceof DirectedGraph<?, ?>)
      {
        graphType = GraphType.DIGRAPH;
      }
      
      DotGraph dotGraph = new DotGraph(graphType, "structs");
      if (this.name!= null)
        dotGraph.setLabel(this.name);
      dotGraph.setRankDir(RankDir.LeftRight);  
      dotGraphs.addFirst(dotGraph);
      graphs.addFirst(graph);
      graphFormatter.formatGraph(graph, dotGraph);
    }
    else
    {
      AbstractDotGraph dotGraph = getDotGraph();
      DotSubGraph subGraph = dotGraph.createSubgraph(getGraphID(graph));
      subGraph.setLabel(this.labelProvider.getGraphLabel(graph));
      dotGraphs.addFirst(subGraph);
      graphs.addFirst(graph);
      graphFormatter.formatGraph(graph, subGraph);
    }

  }


  protected String getGraphID(Graph<V, E> graph)
  {
    return "G"+Math.abs(graph.hashCode());
  }
  

  public void finishGraph(Graph<V, E> graph)
  {
    AbstractDotGraph dotGraph = dotGraphs.poll();
    graphs.remove();
    if (dotGraphs.isEmpty())
    {
      makeDotGraph((DotGraph)dotGraph);
      this.title = null;
    }
  }
  
  public void makeDotGraph(DotGraph dotGraph)
  {
    if (this.title != null)
    {
      this.dotTools.makeGraph(this.name+"_"+inc+"_"+this.title, dotGraph);
    }
    else
    {
      this.dotTools.makeGraph(this.name+"_"+inc, dotGraph);
    }
  }



  public void discoverVertex(V vertex)
  {
    AbstractDotGraph dotGraph = getDotGraph();
    
    if (dotGraph.getNode(getNodeID(vertex)) == null)
    {
      DotNode node = dotGraph.createNode(getNodeID(vertex));
      node.setLabel(this.labelProvider.getVertexLabel(vertex));
      graphFormatter.formatVertex(vertex, node);
    }
  }
  
	protected String getNodeID(V vertex)
  {
	  return "N"+Math.abs(System.identityHashCode(vertex));
  }
  

  public void finishVertex(V vertex)
  {
    // 
  }
  

  public void discoverEdge(V startVertex, V endVertex, E e)
  {
    if (this.graphFormatter.isVisible(e))
    {
      AbstractDotGraph dotGraph = getDotGraph();
      DotNode startNode = dotGraph.getNode(getNodeID(startVertex));
      DotNode endNode = dotGraph.getNode(getNodeID(endVertex));
      
      if (startNode == null || endNode == null)
      	throw new IllegalArgumentException(startVertex + " to "+ endVertex+ ": "+e);
      
      DotEdge dotEdge = dotGraph
        .createEdge(startNode, endNode);
      
      dotEdge.setDirected(GraphType.DIGRAPH.equals(graphType));
      
      dotEdge.setLabel(this.labelProvider.getEdgeLabel(e));
      graphFormatter.formatEdge(e, dotEdge);
    }
  }


  public void finishEdge(V startVertex, V endVertex, E edge)
  {
    //
  }

	public AbstractDotGraph getDotGraph()
  {
	  return dotGraphs.peek();
  }

  public Graph<V, E> getGraph()
  {
    return graphs.peek();
  }

  public Graph<V, E> getBaseGraph()
  {
    Graph<V, E> peekLast = graphs.peekLast();
    return peekLast;
  }

	public void setLabelProvider(LabelProvider<V, E> labelProvider)
	{
		if (labelProvider != null)
		{
			this.labelProvider = labelProvider;
		}
		else
		{
		  this.labelProvider = new DefaultLabelProvider<V, E>();
		}
		this.labelProvider.setGraphVisitor(this);
	}
	
	public void setGraphFormatter(GraphFormatter<V, E> graphFormatter)
	{
		if (graphFormatter != null)
		{
			this.graphFormatter = graphFormatter;
		}
		else
		{
		  this.graphFormatter = new DefaultGraphFormatter<V, E>();
		}
		this.graphFormatter.setGraphVisitor(this);
	}
	
  public void setName(String name)
  {
    this.name = name;
  }
	
	public GraphVisitor<V, E> getInstance(String title)
	{
	  this.title = title;
	  return this;
	}
}
