package  de.aristaflow.adept2.tools.dot;

import org.w3c.dom.Element;


public abstract class DotElement
{
  protected String label;
  protected DotColor fontcolor;
  private int fontSize;
  private String fontName;
  private DotColor color;
  protected boolean htmlLabel = false;

  protected abstract void createDot(StringBuilder builder);
  
  protected abstract void createGML(Element element);
  
  protected void fillElement(Element element)
  {
    if (label != null)
    {
      Element dataElement = element.getOwnerDocument().createElement("data");
      element.appendChild(dataElement);
      dataElement.setAttribute("key", "label");
      dataElement.setTextContent(this.label);
    }
  }
  
  public void setHTMLLabel(boolean htmlLabel)
  {
    this.htmlLabel = htmlLabel;
  }
  
  protected String escapeLabel(String label)
  {
    if (htmlLabel)
    {
      return String.format("<%s>", label);
    }
    else
    {
      return String.format("\"%s\"", label);
    }
  }

  public DotElement setLabel(String label)
  {
    this.label = label;
    return this;
  }
  
  protected String buildAttributes()
  {
    StringBuilder attrib = new StringBuilder();
    if (label != null)
    {
      addAttribute(attrib, String.format("label=%s", escapeLabel(label)));
    }
    
    if (color != null)
    {
      addAttribute(attrib, String.format("color=\"%s\"", color.getDotValue()));
    }
    
    if (fontcolor != null)
    {
      addAttribute(attrib, String.format("fontcolor=\"%s\"", fontcolor.getDotValue()));
    }
    
    if (fontName != null)
    {
      addAttribute(attrib, String.format("fontname=\"%s\"", fontName));
    }
    
    if (fontSize > 0)
    {
      addAttribute(attrib, String.format("fontsize=%d", Integer.valueOf(fontSize)));
    }
    
    return attrib.toString();
  }

  protected void addAttribute(StringBuilder attrib, String attribute)
  {
    if (attrib.length() != 0)
    {
      attrib.append(", ");
    }
    attrib.append(attribute);
  }

  public DotElement setFontColor(DotColor fontColor)
  {
    this.fontcolor = fontColor;
    return this;
  }
  
  public DotElement setColor(DotColor color)
  {
    this.color = color;
    return this;
  }
  
  public void setFontName(String fontName)
  {
    this.fontName = fontName;
  }
  
  public void setFontSize(int fontSize)
  {
    this.fontSize = fontSize;
  }

}
