package  de.aristaflow.adept2.tools.dot;

import org.w3c.dom.Element;

public class DotNode extends DotElement
{

  public enum NodeStyle
  {
    DASHED("dashed"), 
    DOTTED("dotted"), 
    SOLID("solid"), 
    INVISIBLE("invis"),
    BOLD("bold"),
    FILLED("filled"), 
    DIAGONALS("diagonals"),
    ROUNDED("rounded");
    
    private final String style;

    private NodeStyle(String style)
    {
      this.style = style;
    }

    public String getDotValue()
    {
      return style;
    }
  }
  
  public enum NodeShape
  {
  	BOX("box"), 
  	POLYGON("polygon"), 
  	ELLIPSE("ellipse"), 
  	OVAL("oval"), 
  	CIRCLE("circle"), 
  	POINT("point"), 
  	EGG("egg"), 
  	TRINGALE("triangle"), 
  	PAINTEXT("plaintext"),
  	DIAMOND("diamond"),
  	//trapezium parallelogram house pentagon 
  	HEXAGON("hexagon");
  	//septagon octagon doublecircle doubleoctagon tripleoctagon invtriangle invtrapezium invhouse Mdiamond Msquare Mcircle rect rectangle square none note tab folder box3d component;

  	
    private final String style;

    private NodeShape(String style)
    {
      this.style = style;
    }

    public String getDotValue()
    {
      return style;
    }
  }
  
  private final String nodeID;
  private DotColor fillcolor;
  private String group;
  private NodeStyle style;
  private NodeShape shape;


  public DotNode(String nodeID)
  {
    this.nodeID = nodeID;
  }

  @Override
  protected void createDot(StringBuilder builder)
  {
    builder.append(getNodeName());
    String attrib = buildAttributes();
    
    if (attrib.length() > 0)
    {
      builder.append(" [");
      builder.append(attrib);
      builder.append("];");
    }
  }
  
  @Override
  protected String buildAttributes()
  {
    StringBuilder attrib = new StringBuilder();
    attrib.append(super.buildAttributes());
    
    if (shape != null)
    {
    	addAttribute(attrib, String.format("shape=%s", shape.getDotValue()));
    }
    
    if (fillcolor != null)
    {
      addAttribute(attrib, String.format("fillcolor = %s",fillcolor.getDotValue()));
    }
    
    if (group != null)
    {
      addAttribute(attrib, String.format("group = \"%s\"",group));
    }
    
    if (style != null)
    {
      addAttribute(attrib, String.format("style=%s", style.getDotValue()));
    }
    
    
    
    return attrib.toString();
  }

  
  public void setFillColor(DotColor fillcolor)
  {
    this.fillcolor = fillcolor;
  }
  
  public void setGroup(String group)
  {
    this.group = group;
  }
  
  public DotNode setStyle(NodeStyle style)
  {
    this.style = style;
    return this;
  }
  
  public DotNode setShape(NodeShape shape)
  {
  	this.shape = shape;
  	return this;
  }
  
  String getNodeName()
  {
    return "node"+getNodeID();
  }

  public String getNodeID()
  {
    return nodeID;
  }
  
  
  public DotNode setLabel(String label)
  {
	super.setLabel(label);
    return this;
  }

  @Override
  protected void createGML(Element element)
  {
    Element nodeElement = element.getOwnerDocument().createElement("node");
    element.appendChild(nodeElement);
    nodeElement.setAttribute("id", getNodeName());
    fillElement(nodeElement);
  }
 
}
