package  de.aristaflow.adept2.tools.dot;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class DotSubGraph extends AbstractDotGraph {

	private ElementStyle style;
  private AbstractDotGraph parentGraph;

	public DotSubGraph(AbstractDotGraph parentGraph, String subGraphID) {
		super(subGraphID);
    this.parentGraph = parentGraph;
	}

	public enum ElementStyle
	{
		FILLED("filled"),
		BOLD("bold"),
		DOTTED("dotted");
		
		private final String style;

		private ElementStyle(String style) {
			this.style = style;
		}

		public String getDotValue() {
			return style;
		}
	}
	
	@Override
	protected void createDot(StringBuilder builder) {
		builder.append(String.format("subgraph cluster%s {\n", name));

		String attrib = buildAttributes();

		if (attrib.length() > 0) {
			builder.append(attrib);
			builder.append("\n");
		}

		for (DotElement node : nodes.values()) {
			node.createDot(builder);
			builder.append("\n");
		}

		for (DotElement edge : edges.values()) {
			edge.createDot(builder);
			builder.append("\n");
		}

		builder.append("}");
	}
	
	protected String buildAttributes() {
		StringBuilder s = new StringBuilder();
		s.append(super.buildAttributes());
		
		if (style != null)
		{
			addAttribute(s, String.format("style=%s",style.getDotValue()));
		}
		
		return s.toString();
	}

	protected void addAttribute(StringBuilder attrib, String attribute) {
		attrib.append(attribute);
		attrib.append(";\n");
	}
	
	public DotSubGraph setStyle(ElementStyle style)
	{
		this.style = style;
		return this;
	}

  @Override
  protected void createGML(Element element)
  {
    Document document = element.getOwnerDocument();
    Element nodeElement = document.createElement("node");
    element.appendChild(nodeElement);
    nodeElement.setAttribute("id", name);
    
    
    Element graphElement = document.createElement("graph");
    nodeElement.appendChild(graphElement);
    graphElement.setAttribute("id", "G_"+name);
    graphElement.setAttribute("edgedefault", "directed");
    
    for (DotElement node : nodes.values()) {
      node.createGML(graphElement);
    }
    
    for (DotElement node : subgraphs.values()) {
      node.createGML(graphElement);
    }

    for (DotElement edge : edges.values()) {
      edge.createGML(graphElement);
    }
  }

  @Override
  protected boolean isDirected()
  {
    return parentGraph.isDirected();
  }

}
