package de.aristaflow.adept2.tools.dot;

import java.awt.image.BufferedImage;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.FilenameFilter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.net.URI;
import java.net.URISyntaxException;

import javax.imageio.ImageIO;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import de.aristaflow.adept2.util.xml.XMLFormatException;
import de.aristaflow.adept2.util.xml.XMLHelperTools;

public class DotTools {

  
    /** Debug Option */
    private static final boolean WRITE_DOT_FILE = false;
  
		private static final String FILE_EXTENSION_DOT = ".dot";

		
		private final ImageType imageType;
		
		private final File directory;
		
		
		public int inc = 0;
		
		public static ResourceLocator locator = new DefaultResourceLocator();

    
		
		
		public enum ImageType {
			JPG(".jpg", "-Tjpg"),
			SVG(".svg", "-Tsvg"),
			FIG(".fig", "-Tfig"),
			PNG(".png", "-Tpng:cairo"),
			PS(".ps", "-Tps2"),
			PDF(".pdf", "-Tpdf"),
			GML(".graphml", "");
			
			public final String extension;
			public final String dotParam;

			ImageType(String extension, String param)
			{
				this.extension = extension;
				this.dotParam = param;
			}
		}
		
		public DotTools()
		{
		  this (new File("C:\\Temp\\dot\\"), ImageType.JPG);
		}
		
		public DotTools(File directory)
		{
		  this (directory, ImageType.JPG);
		}
		
		

		
		public DotTools(File directory, ImageType imageType)
    {
		  this.directory = directory;
		  this.imageType = imageType;
    }

    public void cleanDir()
    {
      if (directory.exists())
      {
        FilenameFilter filter = new FilenameFilter()
        {
          public boolean accept(File dir, String name) {
            for (ImageType type : ImageType.values())
            {
              if (name.endsWith(type.extension))
                return true;
            }
            
            
            return name.endsWith(FILE_EXTENSION_DOT);
          } 
        };
        
        for (File del : directory.listFiles(filter))
        {
          if (del.isFile())
            del.delete();
        }
      }
    }
			
		public BufferedImage getGraph(DotGraph graph)
		{
			try {
				return dotToImage(graph.createDot().getBytes());
			} catch (IOException e) {
				e.printStackTrace();
			}
			return null;
		}
		

		public void makeGraph(String file, DotGraph graph) {
			try
			{		
				if (!directory.exists())
					directory.mkdir();
				
				String fileName = new File(directory, file).getAbsolutePath();

				
        if (imageType.equals(ImageType.GML))
        {
          dotToGML(graph, fileName);
        }
        else
        {
          byte[] tempFile = graph.createDot().getBytes();
          dotToImageFile(tempFile, fileName);
        }
			}
			catch (IOException e)
			{
				e.printStackTrace();
			}
			
		}
		
		private void dotToGML(DotGraph graph, String fileName)
    {
		  Document document = XMLHelperTools.createDocument();
		  
		  Element root = document.createElement("graphml");
		  document.appendChild(root);
		  root.setAttribute("xmlns", "http://graphml.graphdrawing.org/xmlns" );
		  root.setAttribute("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance");
		  root.setAttribute("xsi:schemaLocation", "http://graphml.graphdrawing.org/xmlns http://graphml.graphdrawing.org/xmlns/1.0/graphml.xsd");
		  
		  Element key;
      key = document.createElement("key");
		  root.appendChild(key);
		  key.setAttribute("attr.name", "label");
		  key.setAttribute("attr.type", "string");
		  key.setAttribute("for", "node");
		  key.setAttribute("id", "label");
		  
		  key = document.createElement("key");
      root.appendChild(key);
      key.setAttribute("attr.name", "Edge Label");
      key.setAttribute("attr.type", "string");
      key.setAttribute("for", "node");
      key.setAttribute("id", "edgelabel");
		  
		  
		  graph.createGML(root);
		  
		  try
      {
        XMLHelperTools.writeDocumentToFile(document, new File(fileName+imageType.extension), null);
      }
      catch (IOException e)
      {
        throw new RuntimeException(e);
      }
      catch (XMLFormatException e)
      {
        throw new RuntimeException(e);
      }
	
    }

    private void dotToImageFile(byte[] dot, String fileName) throws IOException {
			String outputPath = fileName + imageType.extension;
			
			if (WRITE_DOT_FILE)
			{
        FileOutputStream out = new FileOutputStream(new File(fileName + FILE_EXTENSION_DOT));
        out.write(dot);
        out.flush();
        out.close();
			}
			
			final InputStream stream;
			try
			{
				stream = startExeApplication(getDotPath(), new String[] { imageType.dotParam }, dot);
			}
			catch (IOException e)
			{
			  if (!WRITE_DOT_FILE)
			  {
  				FileOutputStream out = new FileOutputStream(new File(fileName + FILE_EXTENSION_DOT));
  				out.write(dot);
  				out.flush();
  				out.close();
			  }
				throw e;
			}
			
			File imageFile = new File(outputPath);
			
			if (imageFile.exists())
				imageFile.delete();
			imageFile.createNewFile();
			
			FileOutputStream out = new FileOutputStream(imageFile);
			copyStream(stream, out);
			out.flush();
			out.close();
			stream.close();
		}
		
		private URI getDotPath() {
			try {
				return locator.resolve(new URI(DotTools.class.getResource("dot.exe").toExternalForm()));
			} catch (URISyntaxException e) {
			  throw new RuntimeException(e);
			}
		}

		private BufferedImage dotToImage(byte[] dot) throws IOException
		{
			ByteArrayInputStream stream = startExeApplication(getDotPath(), new String[] { imageType.dotParam }, dot);
			BufferedImage img = ImageIO.read(stream);
			stream.close();
			
			return img;
		}

		private static int copyStream(InputStream stream, OutputStream out) throws IOException {
			byte[] buff = new byte[512];
			
			int count = 0;
			int length = 0;
			while ( (count = stream.read(buff)) > -1)
			{
				out.write(buff, 0, count);
				length += count;
			}
			
			return length;
		}

  private static ByteArrayInputStream startExeApplication(URI path, String[] commandString, byte[] dot)
      throws IOException
  {
    if (path == null)
    {
      throw new IllegalArgumentException("Dot Path cannot be null");
    }

    File file = new File(path);
    if (!file.exists())
      throw new IllegalArgumentException("Dot Path does not point to a file");

    ByteArrayInputStream result = null;

    String[] command = new String[commandString.length + 1];
    command[0] = file.getPath();
    System.arraycopy(commandString, 0, command, 1, commandString.length);

    java.lang.Runtime rt = java.lang.Runtime.getRuntime();
    // platform dependent call of the executable file (see API doc
    // of the class for details)
    Process exec = rt.exec(command);

    final BufferedReader errorReader = new BufferedReader(new InputStreamReader(exec.getErrorStream()));

    ByteArrayOutputStream out = new ByteArrayOutputStream();
    final StringBuilder err = new StringBuilder();
    Thread errorStreamProcessing = new Thread(new Runnable()
    {

      public void run()
      {
        String outputLine;
        try
        {
          outputLine = errorReader.readLine();

          if (outputLine != null)
          {

            while (outputLine != null)
            {
              err.append(outputLine + "\n");
              System.err.println(outputLine);
              // System.err.println(outputLine);
              outputLine = errorReader.readLine();

            }

          }
        }
        catch (IOException e)
        {
          e.printStackTrace();
        }
        finally
        {
          try
          {
            errorReader.close();
          }
          catch (IOException e)
          {
            e.printStackTrace();
          }
        }
      }
    });
    errorStreamProcessing.start();
    
    OutputStream outputStream = new BufferedOutputStream(exec.getOutputStream());
    try
    {
      outputStream.write(dot);
      outputStream.flush();
    }
    finally
    {
      try
      {
        outputStream.close();
      }
      catch (IOException e)
      {
        e.printStackTrace();
      }
    }

    InputStream stream = exec.getInputStream();
    int outputSize;
    try
    {
      outputSize = copyStream(stream, out);
    }
    finally
    {
      try
      {
        stream.close();
      }
      catch (IOException e)
      {
        e.printStackTrace();
      }
    }

    errorStreamProcessing.interrupt();
    errorStreamProcessing = null;

    if (err.length() > 0)
    {
      if (outputSize > 0)
      {
        System.err.println("Dot Says: " + err.toString().trim());
      }
      else
      {
        throw new IOException("Dot Says: " + err.toString().trim());
      }
    }
			
    try
    {
      exec.waitFor();
    }
    catch (InterruptedException e)
    {
      e.printStackTrace();
    }

    if (exec.exitValue() != 0)
      throw new IOException("Dot Error Code: " + exec.exitValue());

    result = new ByteArrayInputStream(out.toByteArray());

    return result;
		}


	}

